#!/usr/bin/env perl

#==========================================================================
# Converts the Google n-gram tables into an SAMT language model Berkeley DB
# Do the following steps:
# cat LDC2006E88/web_5gram_1/data/1gms/total | gzip > googlelm-total.gz
# ls googlelm-total.gz LDC2006E88/*/data/*/vocab.gz LDC2006E88/*/data/*/*gm*.gz > allgooglefiles.txt
# zcat `cat allgooglefiles.txt | perl -pe 's/\n/ /g'` | convertgooglelm.pl -r /nfs/islpc7_2/ashishv/iwslt/v2/db-notight-noabstract-usenull-userefined-min0.03.targets --Vocabulary btecdb/training.en.CD.google.samtlm.vocab -o btecdb/training.en.CD.google.samtlm > & btecdb/convertgooglelm.log &
#
# If you set $BERKELEY_DB to zero, you can run the program like this:
#
# ( zcat `cat allgooglefiles.txt | replace '\n' ' '` | convertgooglelm.pl -r /nfs/islpc7_2/ashishv/iwslt/v2/db-notight-noabstract-usenull-userefined-min0.03.targets --Vocabulary btecdb/training.en.CD.google.samtlm.vocab --KeepTruecase btecdb/fullBTEC.en.clean.ibm.keeptruecase | gzip > btecdb/training.en.CD.google.samtlm.txt.gz ) > & btecdb/convertgooglelm.log &




use warnings;
use strict;

my $BERKELEY_DB = 1;		# set to zero if Berkeley DB is not available
use BerkeleyDB; # comment this out if Berkeley DB is not available

#use Encode::Arabic;

#use Encode qw/encode decode/;
#use Storable;

use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
# Note: Be careful with post-increment / decrement when tying a hash into Berkeley DB. This doesn't quite work in certain cases!
use Getopt::Long;
#use if $BERKELEY_DB, BerkeleyDB =>;

#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings

# Command line parameters


my $tmpdir;
my $outputdbfile;
my $VocabularyFile;
my $KeepTruecaseFile;
my $verbose = 0;
my $vocabRestrictionFile;
my $cachesize = 3400;		# Berkeley DB cache size in MB
my $OnlyUseBerkeleyDBForDumping = 0; # set this to 1 to do all calculations with a regular hash and only convert to Berkeley DB in the end
my $MaxMarkovOrder=4;		# max. Markov order allowed
my $MinHistoryFrequency=1;
my $MinFrequency=1;
my $UseBTree=1;
# working hash table's pack format for hash key (word index sequence): e.g. 'w*' or 'l*' 
# This is in case of --OnlyUseBerkeleyDBForDumping
my $InternalPackFormatKey = 'w*';
# working hash table's pack format for hash value: e.g. 'www' or 'ddl'
my $InternalPackFormatValue = 'www';
# output DB's pack format for hash key (word index sequence): e.g. 'w*' or 'l*'
my $PackFormatKey = 'w*';
# output DB's pack format for hash value: e.g. 'www' or 'ddl'
my $PackFormatValue = 'www';

# If this is set to nonzero, we will not truncate the existing DBFile, but rather interpolate it with the
# new ngrams from STDIN, where we multiply each STDIN ngram frequency with the value --OccurrenceCountFactor
my $InterpolateWithExistingDB = 0;
# For each n-gram read from STDIN, its occurrence frequency is multiplied with --OccurrenceCountFactor 
# (has to be a whole number!). We will also take care of smoothing issues (i.e., keep treating rare words as rare
# even when their frequencies are multiplied with large --OccurrenceCountFactor values) and not multi-counting
# ngrams in the unique-successing-ngram field.
my $OccurrenceCountFactor=1;
# If you do want to multiply the unique-ngram counts for ngrams read from STDIN by $OccurrenceCountFactor, set the 
# following to true. This has the advantage of offsetting the higher confidence in the history caused by higher 
# history frequencies, since the confidence is a function of HistoryFreq/UniqueSuccessorCount
my $MultiplyUniqueSuccessorCount = 1;

# Only count number of occurrences of the n-grams; don't do
# unique-instances counting and unused-freq-mass counting
my $OnlySimpleCounts = 0;


die if !GetOptions('tmpdir=s' => \$tmpdir, 'cachesize=s' => \$cachesize, 'OnlyUseBerkeleyDBForDumping!' => \$OnlyUseBerkeleyDBForDumping, 'KeepTruecaseFile=s' => \$KeepTruecaseFile, 'verbose=s' => \$verbose, 'DBFile=s' => \$outputdbfile, 'r=s' => \$vocabRestrictionFile, 'UseBTree!' => \$UseBTree, 'Vocabulary=s' => \$VocabularyFile, 'MaxMarkovOrder=s' => \$MaxMarkovOrder, 'MinHistoryFrequency=s' => \$MinHistoryFrequency, 'MinFrequency=s' => \$MinFrequency, 'PackFormatKey=s' => \$PackFormatKey, 'PackFormatValue=s' => \$PackFormatValue, 'InterpolateWithExistingDB!' => \$InterpolateWithExistingDB, 'OccurrenceCountFactor=s' => \$OccurrenceCountFactor, 'OnlySimpleCounts!' => \$OnlySimpleCounts  );

#my $TrainingCorpusFile = shift(@ARGV);
die "Too many arguments (expected: 0). Superfluous arguments: " . join(' ',@ARGV) if scalar(@ARGV) != 0;
# if we decide not to use a Berkeley DB, then we have to set $OnlyUseBerkeleyDBForDumping to true
$OnlyUseBerkeleyDBForDumping = 1 if !$outputdbfile;

my $UniqueSuccessorCountFactor = $MultiplyUniqueSuccessorCount ? $OccurrenceCountFactor : 1;

# $tmpdir = File::Spec->tmpdir() if (!$tmpdir);
$tmpdir = "." if (!$tmpdir);



################################## Global variables

# the vocabulary hash mapping a string to its integer vocab ID
my %vocabhash = ();
# the vocabulary list mapping element i to the string corresponding to vocab ID i
my @vocablist = ();

# create my own temp dir as subdir of $tmpdir, it and its contents will 
# automatically removed at program end
#my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 1);

my %lm = ();			# the LM: a hashtable mapping n-grams to their occurrence counts
my %preserve_case_first_words = ();
my $sentencenumber = 0;		# counts the sentences processed
my $ruleinstancecount = 0;	# counts the number of instances of ngrams encountered (i.e. multiple occurrences count multiple times)
my $writtenoutrulenumber = 0;	# number of rules written out to STDOUT / database
my $outputdb;			# file handle for Berkeley DB
my $sanity_test_total_freq = 0; # checks that Berkeley DB storage is not corrupted

# if parameter -r was passed, restrict n-grams trained to ngrams consisting of words in %vocCheck
my %vocCheck = ();


{				# main block
  print STDERR 'convertgooglelm.pl - started @ '. `date`;
  #  die "You have to specify an output DB filename via parameter --OutputDBFile (short: -o)" if !$outputdbfile;
  print STDERR ($outputdbfile ?
		"Outputting trained LM into newly created Berkeley DB file $outputdbfile\n" :
		"Outputting LM in text-form to STDOUT\n");

  die "If you're using --OutputDBFile, you have to specify a --Vocabulary to store the vocabulary in since the n-grams in the Berkeley DB are stored as sequences of word IDs, not actual word strings." if $outputdbfile && !$VocabularyFile;

  die "If you're using --InterpolateWithExistingDB, you can't use --OnlyUseBerkeleyDBForDumping" if $OnlyUseBerkeleyDBForDumping && $InterpolateWithExistingDB;
  if ($vocabRestrictionFile) {
    print STDERR "Restricting the n-grams to the words occurring in file $vocabRestrictionFile.\n";
    open(SOURCE, "< $vocabRestrictionFile") or die "Can't open $vocabRestrictionFile: $!";
    while (<SOURCE>) {
      chomp($_);
      my @sentence = ('<s>', split(' ',$_), '</s>');
      for (my $i=0; $i < scalar(@sentence); $i++) {
	$vocCheck{$sentence[$i]} = 1;
      }
    }
    close(SOURCE);
  }

  if ($KeepTruecaseFile) {
    print STDERR "Lowercasing all Google n-gram words after '.','?', '!', and '<s>' except the words occurring in $KeepTruecaseFile\n";
    open(SOURCE, "< $KeepTruecaseFile") or die "Can't open $KeepTruecaseFile: $!";
    while (<SOURCE>) {
      chomp($_);
      $preserve_case_first_words{$_} = 1;
    }
    close(SOURCE);
  }

  if ($outputdbfile) {
    if ($BERKELEY_DB) {
      no strict; no warnings;	# disable warnings for this block b/c BerkeleyDB module might not exist
      #      if ($OnlyUseBerkeleyDBForDumping) {
      #	#use this line to only use Berkeley DB at the very end:
      $outputdb = ($UseBTree ?
		   (new BerkeleyDB::Btree       
		    -Cachesize => 2**20 * $cachesize, 
		    -Filename => $outputdbfile, 
		    -Flags => DB_CREATE | ($InterpolateWithExistingDB ? 0 : DB_TRUNCATE),
		   ) :
		   (new BerkeleyDB::Hash       
		    -Cachesize => 2**20 * $cachesize, 
		    -Filename => $outputdbfile, 
		    -Flags => DB_CREATE | ($InterpolateWithExistingDB ? 0 : DB_TRUNCATE),
		   ))
	|| die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
#      } else {
#	#use this line to use Berkeley DB throughout:
#	$outputdb = tie %lm, "BerkeleyDB::Hash",
#	  -Cachesize => 2**20 * $cachesize, 
#	    -Filename => $outputdbfile, 
#	      -Flags => DB_CREATE|DB_TRUNCATE,
#		|| die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
#      }
    } else {
      die "Variable \$BERKELEY_DB is set to 0, therefore the rules cannot be output as database file but just as plain text! Bye.\n";
    }
  }

  # our vocabulary starts with index 1, not zero, so push a dummy into
  # the vocab list
  push @vocablist, '';
  if ($VocabularyFile && (-e $VocabularyFile)) {
    print STDERR "Vocabulary file $VocabularyFile exists -- reading it in.\n";
    # read in given vocabulary
    open(VOCIN, "< $VocabularyFile");
    my $wordsAddedCount = 0;
    while (<VOCIN>) {
      chomp($_);
      # add the word to our vocab
      GetWordIndex($_);
      $wordsAddedCount++;
    }
    close(VOCIN);
    if (scalar(@vocablist) != $wordsAddedCount+1) {
      die "Error: at least one word in vocab file $VocabularyFile occurred twice!";
    }
  }

  
  my $lineNumber = 0;
 LINELOOP:  while (<STDIN>) {
    $lineNumber++;
    print STDERR "At line number $lineNumber: $_" if $lineNumber % 100000 == 1;
    # immediately skip ngrams containing an <UNK> - this will assume each <UNK> entry with frequency f
    # to be f unaccounted frequency-one instances
    next LINELOOP if /\<UNK\>/ ; 

    chomp($_);
    # convert BOS / EOS into SRI style: <S> -> <s> and </S> -> </s>
    s/(\<\/?)S\>/$1s\>/g;
    my @sentence_str = split;	# split on whitespace (important: can be space but also tab)
      
    # last element is the frequency count - pop it
    my $freq = pop @sentence_str;
    # ignore it if the frequency count is too low
    next LINELOOP if $freq < $MinFrequency;

    # will be set to true if smartcasing was applied to lowercase a word in this n-gram
    my $smart_casing_occured = 0;

    # check if last word of the n-gram is the <UNK> symbol
    # my $unkEvent = scalar(@sentence_str) != 0 && $sentence_str[-1] eq '<UNK>'; 

    # convert ngram into index form, mapping words not occuring in the vocab restriction file to 0
    my @ngram = ();
    my $smartcase_next_token = 0;
    foreach my $w (@sentence_str) {
      if ($KeepTruecaseFile) {
	# lowercase first word of sentence if that word usually occurs in lowercase
	if ($smartcase_next_token && !$preserve_case_first_words{$w}) {
	  $w = lcfirst($w);
	  $smart_casing_occured = 1;
	}
	if ($w =~ /^(\.|\!|\?|\<s\>)$/) {
	  # the current token is an end-of-sentence mark or the BOS symbol
	  # smartcase the next token
	  $smartcase_next_token = 1;
	} else {
	  $smartcase_next_token = 0;
	}
      }
      if ( #$w eq '<UNK>' || 
	  ($vocabRestrictionFile && !(exists($vocCheck{$w})))) {
	# not in the restriction set or <UNK> - just convert it to 0
	# But immediately skip the rest of this loop if this $w is still part of the history
	# (b/c we don't care about histories that we'll never encounter later)
	next LINELOOP if scalar(@ngram) < scalar(@sentence_str) - 1;
	push @ngram, 0;
      } else {
	# get the word index from the vocab, possibly adding the word to the vocab if it didn't exist yet
	push @ngram, GetWordIndex($w);
      }
    }

    if (scalar(@ngram) == 0) {
      # this is the zero-gram entry
      my $ngram_packed;
      my $value; # start with undefined value
      my @stats;
      if ($OnlyUseBerkeleyDBForDumping) {
	$ngram_packed = pack($InternalPackFormatKey, @ngram);
	$value = $lm{$ngram_packed};
	@stats = defined($value)? unpack($InternalPackFormatValue, $value) : (0,0,0);
      } else {
	$ngram_packed = pack($PackFormatKey, @ngram);
	if ($outputdb->db_get($ngram_packed,$value)) {
	  # entry not found in DB - undefine $value
	  undef($value);
	}
	@stats = defined($value)? unpack($PackFormatValue, $value) : (0,0,0);
      }
      die "In line $lineNumber: '$_': second zero-gram entry encountered. There can only be one....\n" if defined($value) && !$InterpolateWithExistingDB;
      $stats[0] += $freq * $OccurrenceCountFactor;	# zerogram frequency (i.e., total number of events)
      $stats[1] += $freq * $OccurrenceCountFactor;	# unassigned frequency mass for unigram events (starts with total number of events)
      $stats[2] += $freq * $UniqueSuccessorCountFactor; # will represent  n.o. unique unigrams (i.e. n.o. unique words)
      # store the stats
      if ($OnlyUseBerkeleyDBForDumping) {
	$lm{$ngram_packed} = pack($InternalPackFormatValue, @stats);
      } else {
	$outputdb->db_put($ngram_packed,pack($PackFormatValue, @stats));
      }
      $sanity_test_total_freq += $freq * $OccurrenceCountFactor; # just a checksum to see whether the Berkeley DB works okay.
    } else {
      # a regular entry
      my @history = @ngram[0..$#ngram-1];
      # modify @history statistics only if none of the word indices in 
      # @history is zero (i.e., <UNK> or didn't occur in $vocabRestrictionFile)
      # note: this line is unnecessary now since this check is done above when iterating through @sentence_str
      # unless (grep {$_==0} @history)       
      {
	my $history_packed;
	my $value; # start with undefined value
	my @stats;
	if ($OnlyUseBerkeleyDBForDumping) {
	  $history_packed = pack($InternalPackFormatKey, @history);
	  $value = $lm{$history_packed};
	  @stats = defined($value)? unpack($InternalPackFormatValue, $value) : (0,0,0);
	} else {
	  $history_packed = pack($PackFormatKey, @history);
	  if ($outputdb->db_get($history_packed,$value)) {
	    # entry not found in DB - undefine $value
	    undef($value);
	  }
	  @stats = defined($value)? unpack($PackFormatValue, $value) : (0,0,0);
	}
	print STDERR "Warning: In line $lineNumber: '$_': History has no entry.\n" if !defined($value);
	# we should have read in the history already in form of an (n-1) gram, 
	# so don't modify the frequency counter in $stats[0]
	#	if ($unkEvent) {
	#	  # the word following this history is <UNK>, i.e., represents cumulative frequencies of events 
	#	  # that didn't reach the minimum count; thus keep frequency mass as unassigned (not decrementing
	#	  # $stats[1]). Also, assume that each of these events represented by <UNK> occurred exactly once,
	#	  # i.e., increment $stats[2] by $freq
	#	  $stats[2] += $freq;
	#	} else 
	{
	  # regular case: event that is being accounted for, so decrement unsassigned
	  # frequency mass by $freq
	  $stats[1] -= $freq * $OccurrenceCountFactor;
	  if ($stats[1] < 0) {
	    print STDERR "Warning: In line $lineNumber: '$_': History frequency $stats[0] (unaccounted freq. $stats[1], unique-event-count $stats[2]) is less than sum of frequencies of n-grams with that history (current events freq: $freq * $OccurrenceCountFactor, history length: " . 
	   scalar(@history) . ", ngram length: " . scalar(@ngram) . ").\n" if defined($value);
	    $stats[1] = 0;
	  }
	  # and adapt unique-event counter for this history
	  # (In case of $MultiplyUniqueSuccessorCount, we pretend to have seen $OccurrenceCountFactor unique words instead of 1, 
	  # such that the history confidence estimate in the SAMT language model gets corrected accordingly for the higher-frequency
	  # histories.)
	  $stats[2] -= ($freq-1) * $UniqueSuccessorCountFactor;
	  if ($stats[2] < 0) {
	    print STDERR "Warning: In line $lineNumber: '$_': Unique-successor count less than zero. History frequency $stats[0], unaccounted freq. $stats[1], unique-event-count $stats[2], current events freq: $freq * $OccurrenceCountFactor, history length: " . 
	      scalar(@history) . ", ngram length: " . scalar(@ngram) . ").\n" if defined($value);
	    $stats[2] = 0;
	  }
	}
	# store the stats
	if ($OnlyUseBerkeleyDBForDumping) {
	  $lm{$history_packed} = pack($InternalPackFormatValue, @stats); # store the stats
	} else {
	  $outputdb->db_put($history_packed, pack($PackFormatValue,@stats));
	}
      }
      # consider @ngram for statistics only if none of the word indices in 
      # @ngram is zero (i.e., <UNK> or didn't occur in $vocabRestrictionFile)
      # we know already this history indices is nonzero; still have to check the last word's index
      unless ($ngram[-1] == 0) {
	my $ngram_packed;
	my $value; # start with undefined value
	my @stats;
	if ($OnlyUseBerkeleyDBForDumping) {
	  $ngram_packed = pack($InternalPackFormatKey, @ngram);
	  $value = $lm{$ngram_packed};
	} else {
	  $ngram_packed = pack($PackFormatKey, @ngram);
	  if ($outputdb->db_get($ngram_packed,$value)) {
	    # entry not found in DB - undefine $value
	    undef($value);
	  }
	}
	if (defined($value)) {
	  print STDERR "Warning: In line $lineNumber, ngram '$_': entry occurs already the second time (this could also be a consequency of a previous history-has-no-entry error, which causes a zero-frequency entry to be created). Adding this entry's frequency to previous entry's frequency.\n" unless $smart_casing_occured || $InterpolateWithExistingDB;
	  @stats = $OnlyUseBerkeleyDBForDumping? unpack($InternalPackFormatValue, $value) : unpack($PackFormatValue, $value);
	  # increment n-gram frequency entry
	  $stats[0] += $freq * $OccurrenceCountFactor;
	  # increment unassigned frequency mass (this info will only be used if this n-gram serves as a history)
	  $stats[1] += $freq * $OccurrenceCountFactor;
	  $stats[2] += $freq * $UniqueSuccessorCountFactor;
	} else {
	  @stats = (
		    $freq * $OccurrenceCountFactor,	# @ngram frequency (i.e., total number of events)
		    $freq * $OccurrenceCountFactor,	# initialize unassigned frequency mass for (n+1) gram events whose history is @ngram
		    $freq * $UniqueSuccessorCountFactor);	# initialize n.o. unique (n+1) grams whose history is @ngram
	}
	# store the stats
	if ($OnlyUseBerkeleyDBForDumping) {
	  $lm{$ngram_packed} = pack($InternalPackFormatValue, @stats);
	} else {
	  $outputdb->db_put($ngram_packed,pack($PackFormatValue, @stats));
	}
	$sanity_test_total_freq += $freq * $OccurrenceCountFactor; # just a checksum to see whether the Berkeley DB works okay.
      }
    }
  }				# end: while (<STDIN>)
  
  
  
  # write out the Markov order info to the DB
  # key is the reserved one-tuple <0>
  if ($outputdbfile && $MaxMarkovOrder != 1000000) {
    if ($OnlyUseBerkeleyDBForDumping) {
      $lm{pack($InternalPackFormatKey, 0)} = pack($InternalPackFormatValue, ($MaxMarkovOrder, 0, 0));
    } else {
      $outputdb->db_put(pack($PackFormatKey, 0), pack($PackFormatValue, ($MaxMarkovOrder, 0, 0)));
    }
  }
  
  if ($OnlyUseBerkeleyDBForDumping && $outputdbfile) {
    # write out the n-grams in %lm
    print STDERR "\nWriting out the Berkeley DB @ `date`";
    #    my $totalrulefreqcount_sanitytest = 0;
    my $rulenumber = 0;

    my ($key, $value);
    while (($key,$value) = each %lm) { 
      $rulenumber++;
      print STDERR "Processing n-gram number $rulenumber...\n" if $rulenumber % 10000000 == 0;
      $outputdb->db_put(pack($PackFormatKey,unpack($InternalPackFormatKey,$key)), pack($PackFormatValue,unpack($InternalPackFormatValue,$value)));
    }
  }

  if (!$outputdbfile) {
    # write out the n-grams in %lm to STDOUT
    print STDERR "\nWriting out the obtained LM in text format...\n";
    #    my $totalrulefreqcount_sanitytest = 0;
    my $rulenumber = 0;

    my ($key, $value);
    while (($key,$value) = each %lm) { 
      $rulenumber++;
      print STDERR "Processing n-gram number $rulenumber...\n" if $rulenumber % 10000000 == 0;
      print join(' ',((map{$vocablist[$_]} unpack($InternalPackFormatKey,$key)), unpack($InternalPackFormatValue,$value))) . "\n";
    }
  }
    

  #    print STDERR "!!!!!!!!!!!! Something's wrong with the underlying LM database! (Maybe an network file system issue?) \$ruleinstancecount ($ruleinstancecount) != \$totalrulefreqcount_sanitytest ($totalrulefreqcount_sanitytest)!!! \n" if abs($ruleinstancecount - $totalrulefreqcount_sanitytest) > 0.001;
    

  if ($BERKELEY_DB) {
    print STDERR 'Untying Berkeley DB @ '. `date`;
    untie %lm if $outputdbfile && !$OnlyUseBerkeleyDBForDumping;
    undef $outputdb if $outputdbfile;
  }


  if ($VocabularyFile) {
    # still have to write out the vocab file
    print STDERR "Writing out the vocabulary to $VocabularyFile " . `date`;
    open(VOCOUT, "> $VocabularyFile");
    # note that we start from 1, not 0, since we don't write out the dummy word
    for (my $i=1; $i < scalar(@vocablist); $i++) {
      print VOCOUT ($vocablist[$i] . "\n");
    }
    close(VOCOUT);
  }

  print STDERR 'trainlm.pl - finished @ '. `date`;

}				# main






#################### subs


# get vocab ID of the word, adding it to the vocabulary if it wasn't in there yet
sub GetWordIndex {
  my ($wordstr) = @_;

  my $wordid = $vocabhash{$wordstr};
  if (!defined($wordid)) {
    $wordid = scalar(@vocablist);
    die "Oops.... too many unique words. Vocabulary's word IDs got too close too running out of bounds." if $wordid >= 2**31-10000;
    $vocabhash{$wordstr} = $wordid;
    push @vocablist, $wordstr;
  }
  return $wordid;
}
