#!/usr/bin/env perl

#####################################################################
# Purpose: 
# call example:
#
#################################################################

# settings
my $BERKELEY_DB = 0;		# set to zero if Berkeley DB is not available

use warnings;
use strict;
use Storable;
use Tree::R;			# obtained at http://users.tkk.fi/~jolma/pub/snapshots/Tree-R-0.05.tar.gz, extended to deal with our needs, source provided with this package
use Set::IntRange; #for set (posterior calculations)
#use Set::IntSpan; #for set (posterior calculations)
use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
use Log::Log4perl qw(:easy);
use if $BERKELEY_DB, 'BerkeleyDB';
use Getopt::Long;

#================================================================================================================================================
# Minimal Files or Stream required for the phrase extraction to function
#================================================================================================================================================
my $SourceFile = ''; #source training data that was used to extract phrases
my $TargetFile = ''; #target training data that was used to extract phrases
my $TargetParseTreeFile = ''; # enable syntactic rule extraction mode based on Penn treebank style parses of the target training sentences
my $AlignmentsFile = '';      # alignments are used to calculate lexical weights as in Koehn 2003 on initial phrases and generalized phrases
my $PhrasePairFeedFile = '';  # a feed file that includes phrases extracted on a per sentence basis
#format for --PhrasePairFeedFile: all the LOG: SRC/TGT/ALT/PARSE are optional (and we could specify them with files instead)
#LOG: NUM: 0
#LOG: SRC: s0 s2 .. sN
#LOG: TGT: t0 t2 .. tM
#LOG: ALT: s0-t1 s1-t0 s2-...
#LOG: PARSE: (S1 (S (NP man) (VP runs)))
#LOG: PHRASE:
#LOG: (0-1) (0-1)  OR homme # man
#LOG: NUM: 1
my $PhrasePairSpanInfo = 1; #by default we expect phrases in the feed with span information to properly calculate relative frequencies and generalized rules


#================================================================================================================================================
# Extraction limitations
#================================================================================================================================================
# Cap on the length of the source side. Phrase pairs whose n.o. of source words are shorter than --MaxSourceLength are not considered as initial rules
my $MaxSourceLength = 2**30;
# the following parameter restrict the kind of non-lexical rules (i.e. rules containing at least one nonterminal) allowed
# note that lexical rules (i.e., phrase pairs) are not being restricted
my $MAXSOURCEWORDCOUNT = 5; # max n.o. source words in created rules
my $MAXTARGETWORDCOUNT = 5; # same for target
my $MAXSOURCESYMBOLCOUNT = 5; # max n.o. source symbols (source words + substitutions) in created rules
my $MAXTARGETSYMBOLCOUNT = 5; # same for target
my $MAXSUBSTITUTIONCOUNT = 2; # max n.o. substitutions allowed in created rules
my $MaxSubstitutionCountAbstractRules = 0; #max n.o. substitutions allowed in abstract rules (more precisely, the max of MAXSUBSTITUTIONCOUNT and MaxSubstitutionCountAbstractRules will be the max. n.o. substitutions for abstract rules)
# note: $MAXSOURCESYMBOLCOUNT and $MAXSUBSTITUTIONCOUNT and less importantly MaxSubstitutionCountAbstractRules are the two parameters that can significantly influence execution speed, as they allow for omission of recursive subroutine calls
# for the syntax-based version: regulates what kind of extended nonterminals are used
# e.g. 'NP+V'
my $UseConcatenationCategories = 1;
# e.g. 'NP/NN' or 'DT\NP'
my $UseSubtractionCategories = 1;


#================================================================================================================================================
# Limit memory usage and size of rule DB
#================================================================================================================================================
# to save memory and limit the size of the rule database, rules are being consolidated 
# (certain one-time-occurrences removed) every ConsolidationInterval sentences
# as well as at the end of the rule extraction process
my $ConsolidationInterval = 50001;
# during consolidation, all one-time occurrence nonlexical rules are removed
# also, each one-time-occurrence lexical rule whose source side is longer than --MaxLengthSingleCountSourcePhrases are removed (unless when filtering for a dev/test set)
my $MaxLengthSingleCountSourcePhrases = 2**30;
# online pruning of the extracted rules by filtering them on the development/test source-file $SourceRestrictionCorpus (this is the parameter -r)
my $SourceRestrictionCorpus = '';
# if using -r, you have to specify a maximum source phrase length via --MaxSourceLength (see above) because we're then hashing in all n-grams of length up to $MaxSourceLength from the dev/test set
my $cachesize = 1024;		# Berkeley DB cache size in MB



#================================================================================================================================================
# Options to restart the extraction - not fully supported.
#================================================================================================================================================
my $StopSentence = 1000000000000; # the corresponding option --StopSentence allows you to process only the first N sentences
my $startsentence = 1;		# the corresponding option allows you to process only the sentence from sentence number N onwards
my $PhrasePairFeedFileSkipToStartSentence = 1; # true: read in but skip phrase pairs from --PhrasePairFeedFile before sentence number StartSentence; false: assume --PhrasePairFeedFile to start with phrase pair corresponding to StartSentence
my $TargetParseTreeFileSkipToStartSentence = 1; # read in but skip target parses from --TargetParseTreeFile before sentence number StartSentence; false: assume --TargetParseTreeFile to start with parse corresponding to StartSentence
my $resume = 0;			# set corresponding parameter to true if you want to resume rule extraction at sentence where it previously got interrupted



# not very important parameters
my $MaxSourceSymbolCountXRules = -1;
my $AbstractOverX = 0;
my $UseXIff = 1; # use _X category iff there is no other rule available!
my $TightSubRules = 1; #require that sub-rules are tight (ie dont have unaligned words at the boundaries)
#my $GluePhrasePairs = 0;
my $outputallrules = 0;
#my $exitgluemoderules = 0;
my $keepunbalancedrules = 1;
my $CalculateTargetConditionedRFs = 0;

# how to handle unary rules returned from Charniak parser syntactic parse tree (e.g.: NP--PRP)
# can be: 'all' (use all), 'top' (use only the highest-up category - NP in example above),
# 'bottom' (use only the lowest category - PRP in example above)
my $UnaryParseCategoryUtilization = 'bottom'; 

my $tmpdir;
my $outputdbfile;
my $createtmpsubdir = 1;
my $VERBOSE = 0; #change this to affect the logging level

#NOTE: this is set to 0 by default to increase efficiency. Set to 1 to
#save memory
my $UseHashcodes = 0;

#================================================================================================================================================
# Options for calculating lexical weight features
#================================================================================================================================================
# These options can each occur multiple times and lets you specify files with word-based lexical probabilities.
# Each file passed with option --LexicalWeightFile has to consist 
# of entries of probability = P(frenchword|englishword) of the form:
# frenchword englishword probability
#
# Each file passed with option --LexicalWeightFileReversed has to consist
# of entries of probability = P(englishword|frenchword) of the form:
# englishword frenchword probability
my @LexicalWeightFile = ();
my @LexicalWeightFileReversed = ();
#by default if 2 extracted rules have the same src+tgt but with different alignments then we average the resulting lexical weights 
#the alternative is to use --MaxLexicalWeights
my $MaxLexicalWeights = 0;
my $AverageLexicalWeights = 1;
#when a phrase has either only source words or only target words, one of the lexical weights is going to be 1 ie cost 0. How should we handle this?
my $HandleMismatchesByDuplicating = 0; #Copy over the prob that was not 1 (ie p(f|e) -> p(e|f) if we only produced src words)
my $HandleMismatchesByNULLCalc = 1;    #Calculate p(NULL|f) for p(e|f) if we produced only src words
#setting both options to 0 just leaves the 1 in the respective weight and we rely on additional features added in filterrules.pl
my $UniqueEvidence = 1;                #Each syntactic structure that is seen from a single sentence is seen to occur just once

my $LogLevel = "ERROR";

my $OutputProbCosts = 0; #by default we do NOT output in prob space

my $DoExpParseProbs = 1; #call exp(-val) on the scores seen in the parsing log file
my $DoNormParseProbs =1; #do normalization of the scores from the parser
my $DoNormAltProbs =1; #do normalization of the scores from the alignner

#These ranges can be used to limit the number of entries considered from the n-best list
#-1 means use the the data found
#5 means use the 5th best only
#0,5 means use the 0th -> 5th element 

#my $NBestParseRange = "-1"; 
my $NBestParseRange = "0"; 
#my $NBestAltRange = "-1";
my $NBestAltRange = "0";
my $NBestParseGeoAlpha = "1"; #When calculating the geometric mean to normalize (1) makes it arithmetic mean
my $NBestAltGeoAlpha = "1"; #When calculating the geometric mean to normalize (1) makes it arithmetic mean

my $NBestFilterFactor = "-1"; #If your posterior is < than $NBestFilterFactor*$lowestAltProb*$lowestParseProb, then PRUNE IT!
my $NBestNoNovelty = 0; #set no novelty to speed it up
my $NBestMinAltPost = 0.0;

my $SimpleNovelty = 1;

# ignore source punctuation when creating rules (i.e., remove
# punctuation words such as '.', ',', etc. before creating the rules)
my $IgnoreSourcePunctuation=0;


#remember if you set a restriction range, you might want to force renormalization

die if !GetOptions('MaxSubstitutionCountAbstractRules=s' => \$MaxSubstitutionCountAbstractRules, 'SourceFile=s' => \$SourceFile, 'TargetFile=s' => \$TargetFile, 'AlignmentsFile=s' => \$AlignmentsFile, 'TargetParseTreeFile=s' => \$TargetParseTreeFile, 'TargetParseTreeFileSkipToStartSentence!' => \$TargetParseTreeFileSkipToStartSentence, 'PhrasePairFeedFile=s' => \$PhrasePairFeedFile, 'PhrasePairSpanInfo!' => \$PhrasePairSpanInfo, 'PhrasePairFeedFileSkipToStartSentence!' => \$PhrasePairFeedFileSkipToStartSentence, 'MAXSOURCEWORDCOUNT=s' => \$MAXSOURCEWORDCOUNT, 'MAXTARGETWORDCOUNT=s' => \$MAXTARGETWORDCOUNT, 'MAXSOURCESYMBOLCOUNT=s' => \$MAXSOURCESYMBOLCOUNT, 'MAXTARGETSYMBOLCOUNT=s' => \$MAXTARGETSYMBOLCOUNT, 'MAXSUBSTITUTIONCOUNT=s' => \$MAXSUBSTITUTIONCOUNT, 'r=s' => \$SourceRestrictionCorpus, 'ConsolidationInterval=s' => \$ConsolidationInterval, 'tmpdir=s' => \$tmpdir, 'createtmpsubdir!' => \$createtmpsubdir, 'CalculateTargetConditionedRFs!' => \$CalculateTargetConditionedRFs, 'cachesize=s' => \$cachesize, 'verbose!' => \$VERBOSE, 'outputdbfile=s' => \$outputdbfile, 'UnaryParseCategoryUtilization=s' => \$UnaryParseCategoryUtilization, 'UseConcatenationCategories!' => \$UseConcatenationCategories, 'UseSubtractionCategories!' => \$UseSubtractionCategories, 'MaxSourceLength=s' => \$MaxSourceLength, 'MaxSourceSymbolCountXRules=s' => \$MaxSourceSymbolCountXRules, 'MaxLengthSingleCountSourcePhrases=s' => \$MaxLengthSingleCountSourcePhrases, 'outputallrules!' => \$outputallrules, 'AbstractOverX!' => \$AbstractOverX, 'UseXIff!' => \$UseXIff, 'TightSubRules!' => \$TightSubRules, #'GluePhrasePairs!' => \$GluePhrasePairs, 
'keepunbalancedrules!' => \$keepunbalancedrules, 'StartSentence=s' => \$startsentence, 'StopSentence=s' => \$StopSentence, 'resume!' => \$resume, 'UseHashcodes!' => $UseHashcodes, 'LexicalWeightFile=s' => \@LexicalWeightFile, 'LexicalWeightFileReversed=s' => \@LexicalWeightFileReversed, 'AverageLexicalWeights!' => \$AverageLexicalWeights, 'MaxLexicalWeights!' => \$MaxLexicalWeights, 'UniqueEvidence!' => \$UniqueEvidence, 'OutputProbCosts!'=> \$OutputProbCosts, 'DoExpParseProbs!' => \$DoExpParseProbs, 'DoNormParseProbs=s' => \$DoNormParseProbs, 'DoNormAltProbs=s' => \$DoNormAltProbs, 'NBestParseRange=s' => \$NBestParseRange, 'NBestAltRange=s' => \$NBestAltRange, 'NBestParseGeoAlpha=s' => \$NBestParseGeoAlpha, 'NBestAltGeoAlpha=s' => \$NBestAltGeoAlpha, 'NBestFilterFactor=s' => \$NBestFilterFactor, 'NBestNoNovelty!' => \$NBestNoNovelty, 'NBestMinAltPost=s' => \$NBestMinAltPost,'LogLevel=s' => \$LogLevel,
		  'IgnoreSourcePunctuation!' => $IgnoreSourcePunctuation);

if ($LogLevel eq "DEBUG"){
    Log::Log4perl->easy_init($DEBUG);
}
elsif ($LogLevel eq "INFO"){
    Log::Log4perl->easy_init($INFO);
}
elsif ($LogLevel eq "WARN"){
    Log::Log4perl->easy_init($WARN);
}
elsif ($LogLevel eq "ERROR"){
    Log::Log4perl->easy_init($ERROR);
}else{
    die "Unable to set logging level options are DEBUG, INFO, WARN, ERROR";
}

# print all messages thru the logger
my $logger = get_logger(); #for now we just use a global logger, in the future we might want to log individual functions

#================================================================================================================================================
# Sanity check some of the options and setup the tmpdir
#================================================================================================================================================
# $tmpdir = File::Spec->tmpdir() if (!$tmpdir);
$tmpdir = "." if (!$tmpdir);
$MaxLengthSingleCountSourcePhrases = 2**30 if $SourceRestrictionCorpus; # we don't need this restriction if we're filtering for a test set

if ($NBestFilterFactor>=1){
    die "NBestFilterFactor must be less than 1 because setting it to one makes 0 rules extracted if there is only n-best=1 parse/alt";
}

$MaxSourceSymbolCountXRules = $MAXSOURCESYMBOLCOUNT if $MaxSourceSymbolCountXRules == -1;
if ($MAXSOURCESYMBOLCOUNT > $MAXSUBSTITUTIONCOUNT+$MAXSOURCEWORDCOUNT) {
  print STDERR "Warning: MAXSOURCESYMBOLCOUNT ($MAXSOURCESYMBOLCOUNT) > MAXSUBSTITUTIONCOUNT ($MAXSUBSTITUTIONCOUNT) + MAXSOURCEWORDCOUNT ($MAXSOURCEWORDCOUNT). ";
  $MAXSOURCESYMBOLCOUNT = $MAXSUBSTITUTIONCOUNT + $MAXSOURCEWORDCOUNT;
  print STDERR "Setting MAXSOURCESYMBOLCOUNT=$MAXSOURCESYMBOLCOUNT\n";
}
if ($MAXTARGETSYMBOLCOUNT > $MAXSUBSTITUTIONCOUNT+$MAXTARGETWORDCOUNT) {
  print STDERR "Warning: MAXTARGETSYMBOLCOUNT ($MAXTARGETSYMBOLCOUNT) > MAXSUBSTITUTIONCOUNT ($MAXSUBSTITUTIONCOUNT) + MAXTARGETWORDCOUNT ($MAXTARGETWORDCOUNT). ";
  $MAXTARGETSYMBOLCOUNT = $MAXSUBSTITUTIONCOUNT + $MAXTARGETWORDCOUNT;
  print STDERR "Setting MAXTARGETSYMBOLCOUNT=$MAXTARGETSYMBOLCOUNT\n";
}
my $mytmpdir;
if ($BERKELEY_DB) {
  # create my own temp dir as subdir of $tmpdir, it and its contents will 
  # automatically removed at program end
  $mytmpdir = $createtmpsubdir?
    tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 1)
  :
  $tmpdir;
}



#================================================================================================================================================
# Statistics and progress tracking
#================================================================================================================================================
my $sentencenumber = 0;		# counts the sentences processed
my $phrasepairnumber = 0;	# counts the phrase pairs processed
my $ruleinstancecount = 0;	# counts the number of instances of rules extracted (sum of all %resulttypecounts values)
my $writtenoutrulenumber = 0;	# number of rules written out to STDOUT / database
my $lowest_lexical_weight_encountered = 2**30; #global variable set when loading the weight table

#phrase table usage (only still in here because the function has already been written
my %phrasetable = ();		# tied to the phrase table DB if --usephrasetable was specified
my %phrasetablebig = ();


#================================================================================================================================================
# Important variables used to accumulate data across the program. The goal is to fill %complexrules and then write it out
#================================================================================================================================================
my %complexrules = ();		# will contain the complex rules, with $key= "$source\#$target\#$result" and $value=\@costvector
my %phrasecounts = ();          # will contain occurrence counts of source side and target side of the rules and the occurrence counts of the source/target pair
my $outputdb;			# the output rule database - a B-tree
my $db1; my $db2; my $db3;	# the other DBs used

my %resulttypecounts_diluted = ();	# counts the total n.o. rules in %complexrules that have the same result type (keys are categories, values are numbers); this will enable relative-frequency calculations when writing out the rules - USES the posteriors counts 
my %resulttypecounts_undiluted = ();# counts the total n.o. rules in %complexrules that have the same result type (keys are categories, values are numbers); this will enable relative-frequency calculations when writing out the rules - USES a simple increments ++ count


my %codetable = ();
my @vocablist = ();
#restrict rules to source-side testset/devset
my %vocCheck = (); #after vocCheck is initialized it will contain all (encoded) source side n
#keep track of unique rules extracted from a particular sentence
my %g_uniquePhrasesFromSent = (); my %g_uniqueRulesFromSent = (); 
my $g_hasFailed = 0; #set to 1 if the rule has no lexical evidence for its existence in the lex.n2f/f2n (from 1st best)


my $w_abscount = 0; #warning count abstraction market
my $w_weightcount = 0; #if lexical items have weight 0
#my %srcHash = ();

#specify restrictions as to which elements to consider from the nbest list, type: Set::IntRange
my $g_nbestAltRestrictions;
my $g_nbestParseRestrictions;
my $g_minRuleFreq = -1; #The rule must have at least this freq to be stored (per sentence!)

my $NBestMinAltPost_cost = ($NBestMinAltPost>0) ? -log($NBestMinAltPost) : 9999;
#================================================================================================================================================
# Sets up all files streams that have been specified and store them in a hash. The same info in the files can also
# be passed directly thru the phrase feed by using the format described above. loadSentenceStructures(...) will look to 
# see which files streams are defined in %dataFileHandles, using the feed to override these entries if they exist
#================================================================================================================================================
#data structures where we store references to file handles based on which ones were specified
#allows us to access the file handles from anywhere just by looking up the key in this hash. 
#Keys are: SRC/TGT/ALT/PARSE
my %dataFileHandles = ();
setupFileHandles();
my $feed_ref = $dataFileHandles{"FEED"};
die "No feed file initialized" if (!defined($feed_ref));

#we cant do this anymore since we have useful info on the first line
# read and discard first line from PHRASEPAIRS pipe, which should start with "LOG: NUM: [0-9]+"
#my $firstLine = <$feed_ref>;
#die "Wrong format of log file $PhrasePairFeedFile LINE: $firstLine! Byebye.\n" if !($firstLine =~ m/^LOG: NUM:/);

  #================================================================================================================================================
# Restrict rule extraction by setting --SourceRestrictionCorpus = dev+test.fr
# Sets: $vocCheck hash to include all n-grams upto $maxsourcephraselength
#================================================================================================================================================
if ($SourceRestrictionCorpus) {
  die "When restricting rules to a dev/test set via option -r, you have to specify a maximum source phrase length via --MaxSourceLength" if $MaxSourceLength == 2**30;
  my $maxsourcephraselength = $MaxSourceLength;
  print STDERR "Loading lexical-restriction file $SourceRestrictionCorpus and hashing all contained source phrases of length <= $maxsourcephraselength.\n";
  open(SOURCE1, "< $SourceRestrictionCorpus") or die "Can't open $SourceRestrictionCorpus: $!";
  while (<SOURCE1>) {
    chomp($_);
    my @sentence = split(' ',$_);
    @sentence = encodewordlist(@sentence);
    for (my $phraselength = 1; $phraselength <= $maxsourcephraselength; $phraselength++) {
      for (my $phrasestart = 0; $phrasestart+$phraselength <= scalar(@sentence); $phrasestart++) {
	#my $wd = 
	#print STDERR "Restriction word $wd\n";
	$vocCheck{join(' ',@sentence[$phrasestart..$phrasestart+$phraselength-1])} = 1;
      }
    }
  }
  close(SOURCE1);
  # also add empty phrase
  $vocCheck{''} = 1;
}
#load lexical weight file since we have alignment information at this point
my %lexical_weight_table = ();
loadLexicalWeightTable(\%lexical_weight_table);

# store the encoded '_IGNORE' word
my $encoded_ignore_word = encodewordlist('_IGNORE');

{				# main block
  print STDERR `date`."\nRunning extractrules.pl on host $ENV{'HOST'}. Source sentence file: $SourceFile, target sentence file: $TargetFile\ntarget parse tree file: " . ($TargetParseTreeFile? $TargetParseTreeFile : "(none)") . ", " .
    ($PhrasePairFeedFile? "phrase pair feed file: $PhrasePairFeedFile, " : die "--PhrasePairFeedFile not specified specified!").       
      "temp dir: $tmpdir\nMAXSOURCEWORDCOUNT=$MAXSOURCEWORDCOUNT, MAXTARGETWORDCOUNT=$MAXTARGETWORDCOUNT, MAXSOURCESYMBOLCOUNT=$MAXSOURCESYMBOLCOUNT, MAXTARGETSYMBOLCOUNT=$MAXTARGETSYMBOLCOUNT, MAXSUBSTITUTIONCOUNT=$MAXSUBSTITUTIONCOUNT, MaxSourceLength=$MaxSourceLength, MaxSourceSymbolCountXRules=$MaxSourceSymbolCountXRules\n";
  print STDERR "Ignoring all sentences before sentence number $startsentence\n" if $startsentence > 1;
  print STDERR "Ignoring all sentences after sentence number $StopSentence\n" if $StopSentence != 1000000000000;
  print STDERR ($outputdbfile ?
		"Outputting rules into newly created Berkeley DB file $outputdbfile\n" :
		"Outputting rules in text-form to STDOUT\n");
  print STDERR ($AbstractOverX? "" : "NOT ") . "Abstracting over _X rules\n";
  #  print STDERR "Glue phrase pairs together to yield longer phrase pairs that span a syntactic category.\n" if $GluePhrasePairs;
  if ($BERKELEY_DB) {
    # old DB_File version
    #my $a = new DB_File::HASHINFO ; $a->{'cachesize'} = 2**30 ;
    #tie %complexrules, "DB_File", "$mytmpdir/complexrules", O_CREAT|O_RDWR, 0666, $a || die "Can't create temp database complexrules in $mytmpdir! Try setting \$Berkeley_DB=0 at top of extractrules.pl";
    #tie %phrasecounts, "DB_File", "$mytmpdir/phrasecounts", O_CREAT|O_RDWR, 0666, $a || die "Can't create temp database phrasecounts in $mytmpdir!";
    no strict; no warnings;
    $db1 = tie %complexrules, 
      #      "BerkeleyDB::Hash", 
      "BerkeleyDB::Btree", 
	-Cachesize => 2**19 * $cachesize, 
	  -Filename => "$mytmpdir/complexrules", 
	    -Flags => DB_CREATE # don't truncate it if it already exists# |DB_TRUNCATE
	      || die "Can't create temp database complexrules in $mytmpdir! Try setting \$Berkeley_DB=0 at top of extractrules.pl";
#    $db2 = tie %phrasecounts, 
#      #      "BerkeleyDB::Hash", 
#      "BerkeleyDB::Btree", 
#	-Cachesize => 2**19 * $cachesize, 
#	  -Filename => "$mytmpdir/phrasecounts", 
#	    -Flags => DB_CREATE # |DB_TRUNCATE
#	      || die "Can't create temp database phrasecounts in $mytmpdir!";

#    $db3 = tie %resulttypecounts, 
#      "BerkeleyDB::Hash", 
#	-Cachesize => 2**26, 
#	  -Filename => "$mytmpdir/resulttypecounts", 
#	    -Flags => DB_CREATE # |DB_TRUNCATE
#	      || die "Can't create temp database resulttypecounts in $mytmpdir!";
  } else {
    # store complex rules and phrasecounts in a real hash
    #    dbmopen(%phrasecounts, "$mytmpdir/phrasecounts", 0666) || die "Can't create temp database phrasecounts in $mytmpdir!";
    #    dbmopen(%complexrules, "$mytmpdir/complexrules", 0666) || die "Can't create temp database complexrules in $mytmpdir!";
    #    dbmopen(%resulttypecounts, "$mytmpdir/resulttypecounts", 0666) || die "Can't create temp database resulttypecounts in $mytmpdir!";
  }

  #================================================================================================================================================
  # Initialize the output DB file if its going to be used to output rules
  #================================================================================================================================================
  if ($outputdbfile) {
    no strict; no warnings;
    if ($BERKELEY_DB) {
      $outputdb = new BerkeleyDB::Btree 
	-Cachesize => 2**19 * $cachesize, 
	  -Filename => $outputdbfile, 
	    -Flags => DB_CREATE|DB_TRUNCATE,
	      -Property => DB_DUP
		|| die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
    } else {
      die "Variable \$BERKELEY_DB is set to 0, therefore the rules cannot be output as database file but just as plain text! Bye.\n";
    }
  }



 
  
  


  if ($resume) {
    die "Resume currently doesn't work - you have to make %resulttypecount_diluted and %resulttypecount_undiluted persistent.";
    # resume mode: resume from 1+sentence number that has been successfully processed
#    $onlystartwithsentence = 1 + $phrasecounts{'_SID'};
#    die "No resumption info found in database!" if !defined($onlystartwithsentence);
  }

  #================================================================================================================================================
  # Main loop: consider each src/tgt pair, along with the associated parse tree
  #================================================================================================================================================

  
  #print `cat $TargetParseTreeFile`;
  #print `pwd`;
  #print scalar <PARSES>; die;
  while (1) {

    #last if $sentencenumber >= $StopSentence;
    #================================================================================================================================================
    # Identify and create the following data structures
    # - @sourcewords
    # - @targetwords
    # - %currentparsespans
    # - $i_j_w_matrix_aref
    #================================================================================================================================================
    $sentencenumber++;
    my @sourcewords = ();
    my @targetwords = ();

    my %currentparsespans = (); #this is a ds that can only be filled AFTER you have the phrase spans, and include the occSet of spans
    my @persentencespans = (); #this is the array of hashes, each hash is the basic currentparsespans for 1 setence

    my @i_j_w_matrix = ();
    my %srcAligned = (); #track which words on the src have at least one alignment
    my %tgtAligned = (); #track which words on the tgt have at least one alignment
    #nbest 
    my @nbest_alignmentProbs = (); #the alignment probs for each of the n best alts
    my @nbest_parseProbs = (); #the parse probs for each of the n best parses

    my $hasMore = loadSentenceStructures(\@sourcewords, \@targetwords, \@persentencespans, \@i_j_w_matrix, \%srcAligned, \%tgtAligned, \@nbest_alignmentProbs, \@nbest_parseProbs);
    last if (!$hasMore);
    #print "Processing sentence\n";
    
    #The global pruning factor to throw out certain rule (at a per sentence level!)    
    $g_minRuleFreq = $NBestFilterFactor*$nbest_alignmentProbs[-1];
    if (scalar(@nbest_parseProbs)>0){
	$g_minRuleFreq *= $nbest_parseProbs[-1];
    }
    #print "g_minRuleFreq: $g_minRuleFreq, factor: $NBestFilterFactor, alt: ", $nbest_alignmentProbs[-1], " parse: ", $nbest_parseProbs[-1], "\n";

    #    my %currentphrasepairs = (); # indexes the phrase pairs of the current sentence by target spans
    my $rtree = new Tree::R;	# Create R-tree, which will store all elementary rules in such a way as to efficiently allow for sub-rule relationship testing
    # now retrieve phrase pairs for current sentence
    $logger->debug("Phrase pairs retrieved:\n");
    my $nophrasepairsfound = 1;
    {
      my %alreadyaddedphrases = (); # hash set to avoid processing multiple identical phrase pairs
      
      #if ($usephrasetable) {	# using phrase table DB
      #	if ($onlystartwithsentence > $sentencenumber) {
      #	  # ignore phrase pairs if we're not supposed to extract yet
      #	} else {	
      #fill the %alreadyaddedphrases hash structure 
      #matchPhrasesFromTable($sourceline, \@sourcewords, \@targetwords, \%phrasetable, \%alreadyaddedphrases );
      #	}
      #      } else {
      #ratio passed give the fraction that passed the per word alt prob test
      my ($passed, $considered) = matchPhrasesFromStream(\@sourcewords, \@targetwords, $feed_ref, \%alreadyaddedphrases, scalar(@nbest_alignmentProbs) );
      print STDERR "SentenceStats initial passed: $passed considered: $considered...\n" if (($sentencenumber-1) % 100 == 0 || $LogLevel eq "DEBUG");
      #      }

      
      #go thru all the phrase spans that have been identified against each of the initial spans for each sentence, building up a set
      #of extended categories that must be added back into the final %currentparsespans which include occ data
      buildPosteriorSpans(\%alreadyaddedphrases, \@persentencespans, \@sourcewords, \@targetwords,\%currentparsespans);
      #now the %currentparsespans has [span] -> hash{cat}->set_of_occ(cat)
      
      #=========================================================================================================
      # For each phrase pair that we have identified in %alreadyaddedphrases we find syntactic (or extended)
      # categories, and define regions within the R-Tree. If we have --GluePhrasePairs we save the target spans
      # to create consolidated phrases that could span the whole sentence by filling %currentphrasepairs
      #=========================================================================================================

      #POSTERIOR CHANGED
      #the key is the span, the $coststr is now the occurrence list       
      #my ($key, $coststr);
      my ($key, $occSet);

      while (($key,$occSet) = each %alreadyaddedphrases) { 
	#	  print "Phrase pair $key; $coststr\n";
	$nophrasepairsfound = 0;
	$phrasepairnumber++;

	my ($tmpminx,$tmpsizex,$tmpminy,$tmpsizey) = split(' ', $key);
	my $tmptargetspan = $tmpminy . ' ' . $tmpsizey;
	#	  # Sanity test: check for forbidden case: source phrase spans whole source sentence but target phrase doesn't
	#	  if ($tmpsizex == scalar(@sourcewords) && $tmpsizey != scalar(@targetwords)) {
	#	    die "Failed sanity test. Kill AZ. Bye. Sentence pair $sentencenumber (".
	#	      join(@sourcewords). "#".join(@targetwords)."), phrase pair ($key)\n";
	#	  }
	
	#=========================================================================================================
	# Annotation is done only on the target side, by matching a target span to a syntactic category
	# as spanned in %currentparsespans. If there is no direct match, extended categories are created
	# with preference for --UseConcatenationCategories --UseSubtractionCategories and finally _X
	#=========================================================================================================
	#my $tmpcategories = findSpanAnnotation($tmpminx, $tmpsizex, $tmpminy, $tmpsizey, \%currentparsespans, \@targetwords);

	#the findSpanAnnotation is currently used only to identify the extended categories
	my $categoryOcc_href = $currentparsespans{$tmptargetspan};

	# special case: full-sentence span
	if ($tmpsizex == scalar(@sourcewords) && $tmpsizey == scalar(@targetwords)) {
	  # add special full-sentence category _S to the list of categories - but dont if you leave it in there
	  # then it will be a category for all phrases. So make a copy of the hash!!!

	    #$categoryOcc_href{'_S'} = $set;	   
	    my %copyCat = %$categoryOcc_href;
	    #this might be a failed sentence.. so it might have @persentencespans == 0
	    my $set;
	    if (scalar(@persentencespans)>0){
		$set = Set::IntRange->new(0,scalar(@persentencespans)-1);
	    }else{
		$set = Set::IntRange->new(0,0);
	    }
	    $set->Fill();
	    $copyCat{'_S'} = $set;
	    $categoryOcc_href = \%copyCat;
	}
	
	$logger->debug("... corresponding to sourcespan (start $tmpminx length $tmpsizex), targetspan (start $tmpminy length $tmpsizey), categories: ".(defined($categoryOcc_href)? join(' ', keys %$categoryOcc_href) : "(none)") ."\n");



	#$logger->debug("... corresponding to sourcespan (start $tmpminx length $tmpsizex), targetspan (start $tmpminy length $tmpsizey), categories: ".(defined($tmpcategories)? $tmpcategories : "(none)") ."\n");


	#FORMAT OF RULE: POSTERIOR CHANGE
	# rule->[1]->[0] = the occurrence set!
	# rule->[1]->[1]= a hash of result categories + occurrences
	# rule->[2..5] = srcstart, tgtstart, srcend, tgtend
	$rtree->insert([$occSet,$categoryOcc_href],($tmpminx,$tmpminy,$tmpminx+$tmpsizex-1,$tmpminy+$tmpsizey-1));
	
	#az: commented out b/c doesn't work with ashish's modification
	#of overgoing %currentphrasepairs for regular phrase pairs
	#	if ($GluePhrasePairs) {
	#	  # also store phrase pair in hash %currentphrasepairs to allow for gluing together of phrase pairs
	#	  my $refcpp = $currentphrasepairs{$tmptargetspan};
	#	  if (
	#	      defined($refcpp) 
	#	      # oops, we have already a phrase pair with this target span
	#	      # use the one with the bigger source span
	#	      && 
	#	      $refcpp->[3] >= $tmpsizex
	#	      # the source span of the currently stored phrase pair is not smaller, so let's stick with that phrase pair
	#	     ) {
	#	    # do nothing
	#	  } else {
	#	    # store current phrase pair under the targetspan key
	#	    $currentphrasepairs{$tmptargetspan} = [\@tmpphrasecosts,$tmpcategories,$tmpminx,$tmpsizex];
	#	  }
	#	}
      }
      

      print STDERR "Warning: Found no phrase pairs for sentence $sentencenumber!\n" if $nophrasepairsfound;
      
      #      if ($GluePhrasePairs && $TargetParseTreeFile) {	  
      #	addConcatenatedPhrasesToRTree($rtree, \%currentparsespans, \%currentphrasepairs, \@sourcewords, \@targetwords);	 
      #      }# if GluePhrases && $TargetParseTreeFile
      
      #      #these rules are added directly to the complex rules structure
      #      addSentenceSpanningRules(\@sourcewords, \@targetwords, \%currentparsespans);   
    }
    
         
    # Write out the non-start-symbol elementary and complex rules
      
    #first obtain an array of all elementary rules stored 
    #(don't worry, the rules are stored by reference, so that won't take much memory)
    #my @elementaryrules = ();
    # we have to make sure that we have at least one element in the R-tree 
    # b/c of a bug in the R tree implementation that returns the root of the tree even if it is empty
    
    #reset the global evidence trackers
    %g_uniqueRulesFromSent = ();
    %g_uniquePhrasesFromSent = ();
    if (!$nophrasepairsfound) {
      #=========================================================================================================
      # Use the RTree to subtract rules from each other to create complex rules
      # Setter for: %complexrules global variable using probabilites from $i_j_w_matrix_aref (i==srcIndex+1)
      # Input: \@nbest_alignmentProbs is used to calculate the occurrence count frequency for the rule!
      #=========================================================================================================
      generalizeRulesInRTree($rtree, \@sourcewords, \@targetwords, \@i_j_w_matrix, \@nbest_alignmentProbs, \@nbest_parseProbs, \%srcAligned, \%tgtAligned);
	#for my $elemrule
    }				# if (!$nophrasepairsfound)
    
    if (($sentencenumber - $startsentence+1) % $ConsolidationInterval == 0) {
      consolidaterules();
    }

    $logger->debug("Sentence Pair: $sentencenumber |complexrules|=", scalar(keys %complexrules), "\n");
# resuming is currently deactivated    
#    if ($sentencenumber % 1000 == 0 && $BERKELEY_DB) {
#      # store number of sentence processed in case we get interrupted and want to resume later
#      $phrasecounts{'_SID'} = $sentencenumber;
#      # write the cache of the DBs to disk
#      $db1->db_sync();
#      $db2->db_sync();
#      $db3->db_sync();
#    }

  }				#while my $sourceline = <SOURCE>

  print STDERR `date`."$sentencenumber sentences, $phrasepairnumber phrase pairs. Writing out the rules (|complexrules=", scalar(keys %complexrules), ")\n";
  close $feed_ref;
  #untie %phrasetable if $usephrasetabledb;

  print STDERR `ls -lh $mytmpdir` . "\n" . `du -h $mytmpdir` . "\n" if $BERKELEY_DB;
  
  #write out the complex rules that have been gathered (doing some vary basic filtering in the process)
  #each rule is written out WITH a series of relative frequency options
  #verifySanitySrcCond(\%complexrules, \%phrasecounts);
  #this sets up the srcHash variable
  my ($totalrulefreqcount_sanitytest, $rulenumber) = writeComplexRules(\%complexrules, \%phrasecounts);

  #NOTE: $ruleinstancecount is a global
  print STDERR "!!!!!!!!!!!! Something's wrong with the underlying complexrules database! (Maybe an network file system issue?) \$ruleinstancecount ($ruleinstancecount) != \$totalrulefreqcount_sanitytest ($totalrulefreqcount_sanitytest)!!! \n" if $ruleinstancecount != $totalrulefreqcount_sanitytest;

  #write statistics about how many nonterminals there are
  writeNTCounts();

  #NOTE: $writtenoutrulenumber is a global
  print STDERR `date`.". $ruleinstancecount rule instances ($rulenumber unique rules) extracted, $writtenoutrulenumber unique rules written out.\n";

  if ($BERKELEY_DB) {
    undef $outputdb if $outputdbfile;
    undef $db1; #undef $db2; #undef $db3;
    untie %complexrules;
    #untie %phrasecounts;
  }

}				# main










#################### main subs

sub loadSentenceStructures{
  
  my ($sourcewords_aref, $targetwords_aref, $persentencespans_aref, $i_j_w_matrix_aref, $srcAligned_href, $tgtAligned_href, $nbest_alignmentProbs_aref, $nbest_parseProbs_aref) = @_;

  #read the feed until you hit the LOG: PHRASES:
  my $feed_ref = $dataFileHandles{"FEED"};
  my %foundInFeed = (); #track which data was found directly in the feed
  $foundInFeed{"FEED"} = 1; #we automatically find the FEED data stream
  # the lines that we want to initialize
  my $sourceline = '';
  my $parserline = '';
  my $alignmentline = '';
  my $targetline = '';

  #read any information that might be present on the feed
  my $feedLine;
  while ($feedLine = <$feed_ref>){
    if (!$feedLine){
      return 0;
    }

    #set up the alignment probs for each sentence
    if ($feedLine =~ m/^LOG: NUM: (.*)/){ fillAlignmentProbs($feedLine, $nbest_alignmentProbs_aref); }
    if ($feedLine =~ m/^LOG: SRC: (.*)/){ $sourceline = $1; $foundInFeed{"SRC"}=1;}
    if ($feedLine =~ m/^LOG: TGT: (.*)/){ $targetline = $1; $foundInFeed{"TGT"}=1;}
    if ($feedLine =~ m/^LOG: ALT: (.*)/){ $alignmentline = $1; $foundInFeed{"ALT"}=1;}
    if ($feedLine =~ m/^LOG: PARSE: (.*)/){ $parserline = $1; $foundInFeed{"PARSE"}=1;}
    if ($feedLine =~ m/^LOG: PHRASES/){
      #print STDERR "FOUND $feedLine";
      last;
    }
  }
  #exit;
  
  #if there was never any more to read, quit
  if (!$feedLine){
    return 0;
  }
  #post-condition: read to read some phrases
  #print "ALTPROBS : ", join(' ', @$nbest_alignmentProbs_aref), " FROM ", $g_nbestAltRestrictions->to_Enum(), "\n";
  
  #now any keys that are NOT in the %foundInFeed need to be read directly
  my ($key, $value);
  #go thru the files handles that have been registered for each type
  while (($key, $value) = each %dataFileHandles){
    #if a particular key was not already handled
    if (!defined($foundInFeed{$key})){
      #dont steal any more data from the FEED since its also 
      if ($key eq "FEED"){
	next;
      }
      #print "LOOKUP: $key\n";
      my $data = <$value>;
      if ($key =~ m/SRC/){	$sourceline = $data;       }
      if ($key =~ m/TGT/){ $targetline = $data; }
      if ($key =~ /ALT/){ $alignmentline = $data; }
      if ($key =~ /PARSE/){ $parserline = $data; }
    }
  }

  #the target side parse (or set of parses) is in $parserline
  #dont require the $parserline since we 
  if (!($alignmentline && $sourceline && $targetline )){
    $logger->warn("Possible alignment mismatch sno: $sentencenumber source $sourceline\n target $targetline\n alt $alignmentline\n parse $parserline\n");
    if ($sentencenumber==1){
	die "Possible format ERROR: Unable to extract the neccesary streams for extraction:";
    }
  }else{
    $logger->debug("Read:\n source $sourceline\n target $targetline\n alt $alignmentline\n parse $parserline\n");
  }

  normalizeinputstring($sourceline);
  my @tmp = split(' ',$sourceline);
  if ($IgnoreSourcePunctuation) {
    foreach my $x (@tmp) {
      $x =~ s/([[:punct:]])/_IGNORE/g;
    }
  }
  @$sourcewords_aref = encodewordlist(@tmp);

  normalizeinputstring($targetline);
  my @tmp2 = split(' ',$targetline);
  @$targetwords_aref = encodewordlist(@tmp2);

  $logger->debug("Processing $sentencenumber: ", join(' ',decodewordlist(@$sourcewords_aref)), " \# ", join(' ',decodewordlist(@$targetwords_aref)), "\n");
  #create a sentence level alignment which directly includes lexical weights, this alignment map ($i_j_w_matrix_aref)
  #can be used in createcomplexrules to assign lexical probabilities to phrases with elements removed. The return
  #variables ($srcAligned_href, $tgtAligned_href) allow quick lookup of un-aligned words
  createAlignmentsWithProbabilities($i_j_w_matrix_aref, $alignmentline, $sourcewords_aref, $targetwords_aref, \%lexical_weight_table, $srcAligned_href, $tgtAligned_href);

  #MODE: made a  function that builds %currentparsespans given a parserline, and the target words sequence (which is only for debuggin i think...)
  # analyze the parser outputq
  if ($parserline) {
    #buildCurrentSentenceParseSpans($currentparsespans_href, $parserline, $targetwords_aref);      
    buildCurrentSentenceParseSpans($parserline, $persentencespans_aref, $nbest_parseProbs_aref, $targetwords_aref);      
  } # end: if ($TargetParseTreeFile)
  #post-condition: %currentparsespans is filled as long as there was no parser error, otherwise its empty

  print STDERR "Processing sentence number $sentencenumber |complexrules|", scalar(keys %complexrules),"...\n" if (($sentencenumber-1) % 100 == 0 || $LogLevel eq "DEBUG");
  print STDERR  "Processing ALTPROBS : ", join(' ', @$nbest_alignmentProbs_aref), " FROM ", $g_nbestAltRestrictions->to_Enum(), "\n" if ((($sentencenumber-1) % 100 == 0) || $LogLevel eq "DEBUG");
  print STDERR "Processing PARSEPROBS : ", join(' ', @$nbest_parseProbs_aref), " FROM ", $g_nbestParseRestrictions->to_Enum(), "\n" if ((($sentencenumber-1) % 100 == 0 || $LogLevel eq "DEBUG") && (defined($g_nbestParseRestrictions)));

  

  
  return 1;
}

sub buildPosteriorSpans{
    my ($alreadyaddedphrases_href, $persentencespans_aref, $sourcewords_aref, $targetwords_aref, $currentparsespans_href) = @_;
    
    #use the $numparses to initialize the occ set for each value
    my $numparses = scalar(@$persentencespans_aref);
    foreach my $key (keys %$alreadyaddedphrases_href){
	my ($tmpminx,$tmpsizex,$tmpminy,$tmpsizey) = split(' ', $key);
	#the span that we are looking for - set up for lookup
	my $tmptargetspan = $tmpminy . ' ' . $tmpsizey;
	#we need to lookup up each span against 
	my $parseIndex = 0;
	#$logger->debug("LOOKUP KEY: $key\n");
	
	#if this is a new span, then set it up and initialize the _S, _X categories
	if (!defined($currentparsespans_href->{$tmptargetspan})){
	    my %categoriesAndOcc = ();
	    
	    #$logger->debug("SETTING RANGE: $tmpsizex $tmptargetspan\n");
	    #initialize the category _X for ALL spans, indicating existence in ALL sentences      
	    if ($UseXIff==0){
		my $setX = ($numparses>0) ? Set::IntRange->new(0,$numparses-1) : Set::IntRange->new(0,0);
		$setX->Fill();
		$categoriesAndOcc{"_X"} = $setX;
	    }
	    #initialize the sentence spanning category "_S" if this is a full sentence span	    

	    #NOTE: We dont want to assign this label right here since we dont know if the phrase that it will be applied
	    #      to is sentence spanning.
	    #if ($tmpsizex == scalar(@$sourcewords_aref) && $tmpsizey == scalar(@$targetwords_aref)) {
	    #add special full-sentence category _S to the list of categories
	    #my $set = Set::IntRange->new(0,$numparses-1);
	    #my $set = ($numparses>0) ? Set::IntRange->new(0,$numparses-1) : Set::IntRange->new(0,0);
	    #$set->Fill(); #since _S happens for each parse
	    #$categoriesAndOcc{"_S"} = $set;		
	    #}
	    $currentparsespans_href->{$tmptargetspan} = \%categoriesAndOcc;      
	}
	
	#go thru each sentence and see what annotation that sentence can provide
	foreach my $persent_href (@$persentencespans_aref){
	    #for a single sentence, we say (for now) that findSpanAnnotation can return 1 syntactic category
	    my $cat = findSpanAnnotation($tmpminx, $tmpsizex, $tmpminy, $tmpsizey, $persent_href, $targetwords_aref);
	    
	    #if there is a category, use it, otherwise just forget about it, all sentence pairs have a free _X in each span
	    #as ensured above this loop
	    if (defined($cat)){
		my $categoriesAndOcc_href = $currentparsespans_href->{$tmptargetspan};
		#go thru each category identified in THIS sentence. If the category is already defined, then add an occ, else create one
		
		if (defined $categoriesAndOcc_href->{$cat}){
		    $categoriesAndOcc_href->{$cat}->Bit_On($parseIndex); #turn on the bit for this occ count
		}else{
		    #this is the first time this category is noted for this span
		    my $set = Set::IntRange->new(0,$numparses-1);
		    $set->Bit_On($parseIndex);
		    $categoriesAndOcc_href->{$cat} = $set;
		}
	    }
	    
	    $parseIndex++;
	}#for each alternative parse    
	    if ($UseXIff==1){
		my $categoriesAndOcc_href = $currentparsespans_href->{$tmptargetspan};
		if (scalar(keys %$categoriesAndOcc_href)==0){	
		    my $setX = ($numparses>0) ? Set::IntRange->new(0,$numparses-1) : Set::IntRange->new(0,0);
		    $setX->Fill();
		    $categoriesAndOcc_href->{"_X"} = $setX;
		    $logger->debug("UseXIff: No cat for range: tmptargetspan : Using _X\n")
		    }
		$currentparsespans_href->{$tmptargetspan} = $categoriesAndOcc_href;
		
	    }
	
    }#for each span
	
	#print the posteriors learned of this block of sentences
	#while (my ($key, $value_href) = each %$currentparsespans_href){
	#  print "Span $key => (";
	#  while (my ($cat, $set) = each %$value_href){
	#    print "[$cat => ", $set->to_Enum(), "]";
	#  }
	#  print ")\n"; 
	#}
	
}

sub fillAlignmentProbs{
  my ($line, $aref) = @_;
  chomp($line);
  
  my @initialProbs = ();
  if ($line =~ m/.* ALTPROBS (.*)/){
    my $altdata = $1;
    #add in a correction, just to ensure that if any of the probs are 0, they are respected!
    $altdata =~ s/ 0 / 0.0000001 /g;
    @initialProbs = split(/\s+/, $altdata);
  }
  #If the sentence doesnt come from n-best data, then there may be not probs available
  if (scalar(@initialProbs)==0){
      push(@initialProbs, 1);
  }
  my %indexAndProb = ();
  my $index = 0;
  foreach my $val (@initialProbs){
    $indexAndProb{$index} = $val;
    $index++;
  }


  $g_nbestAltRestrictions = new Set::IntRange(0, scalar(@initialProbs)-1);
  if($NBestAltRange eq "-1"){
    $g_nbestAltRestrictions->Fill();
  }else{
    my $finalRange = ensureRangeIsCompliant($NBestAltRange, scalar(@initialProbs)-1);
    $g_nbestAltRestrictions->from_Enum($finalRange);
  }

  #these are the final indices of the values that we want to take, and we have a restriction set
  my $finalIndices_aref = restrictToRangeOfData(\%indexAndProb, $g_nbestAltRestrictions);
  for my $index (@$finalIndices_aref){
      push(@$aref, $initialProbs[$index]);
  }

  #if u need to renormalize - dont forget to set this if u restricted
  if ($DoNormAltProbs){
      NormalizeInProbSpace($aref, $NBestAltGeoAlpha);
  }
}

sub generalizeRulesInRTree{

  my ($rtree, $sourcewords_aref, $targetwords_aref, $i_j_w_matrix_aref, $nbest_alignmentProbs_aref, $nbest_parseProbs_aref, $srcAligned_href, $tgtAligned_href) = @_;
  
  my @elementaryrules = ();
  $rtree->object_and_bbox_refs(\@elementaryrules);
  
  # now iterate through these rules
  for my $elemrule (@elementaryrules) {
    #print "Elem: ", join(",",@$elemrule) ."\n";die;
    my ($sourcespanstart,$targetspanstart,$sourcespanend,$targetspanend) = @$elemrule[2..5]; # span info
    
    # r is the current rule to be written out and consists of four parts:
     
    # source part of rule (array of length equal to source span, whose elements are: 
    #  0 (write out corresponding sentence word at this position), 
    #  pos. number N (write out "#N" at this position), or 
    #  -1 (ignore this position when writing out the rule)
    my @r_source = ();
    # initialize array with zeros
    for (my $i=0; $i<=$sourcespanend-$sourcespanstart; $i++) {
      push @r_source, 0;
    }
    
    # target part of rule (analogous to source part)
    my @r_target = ();
    # initialize array with zeros
    for (my $i=0; $i<=$targetspanend-$targetspanstart; $i++) {
      push @r_target, 0;
    }
    
    # POSTERIOR CHANGES REQUIRED - once parsing stuff is ready
    #  type signature part of the rule
    #The @r_type hold the result category of the elemental rule + the categories of sub rules during the recursion
    #At each element in @r_type we need a pair of elements ($result, $occResultForRule)
    

    my @r_type = ($elemrule->[1]->[1]); # starting of with single element: the $categoryOcc_href
    my $results_href = $r_type[0];
    my @result_cats = keys %$results_href;

    # get array of (references to) elementary rules whose source and target spans are contained in $elemrule's source and target spans
    my @subrules = ();
    #POSTERIOR CHANGED: now we abstract over all spans!
    # if this rule has only _X in the result category list and $AbstractOverX is false, we won't retrieve any subrules to abstract over
    # otherwise, retrieve all subrules
    
    #either allow abstraction over X, or ensure that there are more than 1 cat, or that that one cat is NOT _X
    #OLD LINE: if ($AbstractOverX || $r_type[0] ne '_X') (asssumed that _X would come before _S (which it did in the old array version)
    if ($AbstractOverX || ($UseXIff==0) || !defined($results_href->{'_X'}) ) {
    	
	$rtree->object_and_bbox_refs_completely_within_rect(@$elemrule[2..5],\@subrules);    
	
    }else{
	$logger->debug("Not initializing subrules for: ", join(' ', @result_cats), "\n");
    }
    
    # index the array by relative source span start (rel. to source span start of $elemrule
    my @indexedsubrules = (); # array of arrays of rule references
    # initialize the array with empty arrays
    for (my $i=0; $i<=$#r_source; $i++) {
      push @indexedsubrules, [];
    }
    for my $rule (@subrules) {
      my ($subrulesourcespanstart,$subruletargetspanstart,$subrulesourcespanend,$subruletargetspanend) = @$rule[2..5]; # span info
      if ($subrulesourcespanstart < $sourcespanstart || $subruletargetspanstart < $targetspanstart || $subrulesourcespanend > $sourcespanend || $subruletargetspanend > $targetspanend) {
	print STDERR "Something's wrong with the R-tree module!\n";
      }
      my $subrule_categoryOcc_href = $rule->[1]->[1];
      my @sub_results = keys %$subrule_categoryOcc_href; #get the list of categories for the subrule as a list

      if ($rule == $elemrule) {
	# ignore $elemrule itself as a subrule
	# Note that this also causes rules such as @1#@1#PRP->NP not to be written out
	# if PRP and NP both span target word 'we'
	# we don't want to increase spurious ambiguity in the translate-parser unnecessarily
      } elsif (!$keepunbalancedrules && 
	       (
		($subrulesourcespanstart == $sourcespanstart && $subrulesourcespanend == $sourcespanend) ||
		($subruletargetspanstart == $targetspanstart && $subruletargetspanend == $targetspanend)
	       ) 
	      ) {
	# either source side of subrule is identical to source side of elemrule or target side of subrule is identical to target side
	# of elemrule
	# since we don't want unbalanced rules (such as: de @NN # @1 # @NP) , ignore this subrule
	#OLD VERSION: elsif (!$AbstractOverX && $rule->[1]->[1] eq '_X') Checked for direct equality!!
        #             now we check if the hash contains this _X rule
      }elsif(!$AbstractOverX  && defined($subrule_categoryOcc_href->{'_X'})){
	  #deals with dropping the case where the elementary rule WAS syntactic, but now we
	  #see a suggestion to abstract over X... not good
      }elsif ($TightSubRules && !checkTight($rule, $srcAligned_href, $tgtAligned_href)){
	  #ignore the sub rule if its not tight, ie the first/last src/tgt words are not-aligned
	  #print "Dropping subspan ", join(' ', @$rule[2..5]), "\n";
      }else {
	my $relsourcespanstart = $rule->[2] - $elemrule->[2];
	push @{$indexedsubrules[$relsourcespanstart]}, $rule;
      }
      
      #POSTERIOR CHANGED - there used to be one more check here
      #elsif (!$AbstractOverX && $sub_results[0] eq '_X') {
      # if this subrule has only _X in the result category list and $AbstractOverX is false
      # then don't abstract over it but just ignore this subrule
    }
    
    
    # recursively create and write out complex rules from $elemrule into %complexrules
    #createcomplexrules($elemrule, \@r_source, \@r_target, \@r_costvector, \@r_type, \@indexedsubrules, scalar(@r_source), scalar(@r_target), 0, 0, 0, 1, $sourcewords_aref, $targetwords_aref, $i_j_w_matrix_aref);
    #POSTERIOR pass thru the occ set of the elemental rule which will get recursively intersected
    my $occSet = $elemrule->[1]->[0];
    #print "OCC: $occSet\n";
    #print "BEFORE: ", $occSet->to_Enum(), "\n";
    createcomplexrules($elemrule, $occSet, \@r_source, \@r_target, \@r_type, \@indexedsubrules, scalar(@r_source), scalar(@r_target), 0, 0, 0, 1, $sourcewords_aref, $targetwords_aref, $i_j_w_matrix_aref, $nbest_alignmentProbs_aref, $nbest_parseProbs_aref);
    
  }# for each $elemrule	
}

sub checkTight{
    my $rule = shift;
    my $srcAligned_href = shift;
    my $tgtAligned_href = shift;
    my ($i1, $i2, $j1, $j2) = @$rule[2..5];
    #make sure all the src and  tgt boundaries of the phrase are properly aligned
    if (defined($srcAligned_href->{$i1}) && defined($srcAligned_href->{$i2}) && defined($tgtAligned_href->{$j1}) && defined($tgtAligned_href->{$j2})){
	return 1;
    }else{	
	return 0;
    }

}



#sub addSentenceSpanningRules{
  
#  my $sourcewords_aref = shift;
#  my $targetwords_aref = shift;
#  my $currentparsespans_href = shift;
  
#  # it follows some special handling of the complete-sentence spans:
#  # for each category that spans the whole target-sentence (e.g.: S), we add a unary abstract rule for the start symbol _S
#  # (e.g.: @1 / @1 / S->_S)
  
#  my ($minx,$sizex,$miny,$sizey) = (0, scalar @$sourcewords_aref, 0, scalar @$targetwords_aref);
#  my $tmpsourcespan = "$minx $sizex";
#  my $tmptargetspan = "$miny $sizey";	
  
#  # look whether the complete target span has at least one syntactic category
#  my $tmpcategories = $currentparsespans_href->{$tmptargetspan};
#  $tmpcategories = $tmpcategories->[0] if defined($tmpcategories);
#  if (!defined $tmpcategories) {
#    #	  print STDERR "Error: Target side sentence $sentencenumber has no syntactic category spanning the whole sentence assigned!\n" if $TargetParseTreeFile;
#    #    $tmpcategories = "_X";
#    $tmpcategories = '';
#  }
  
#  # now add unary rules for the start symbol
#  my $sourcestr = "\@1"; 
#  my $targetstr = "\@1";
#  my @startrule_type = ($tmpcategories, "_S");
#  #my @startrule_costvector = (1, 0, 0, 0, 0, 0, 0);

#  #since we log these lexical probs at the last minute, make sure they are 1s
#  my @startrule_costvector = (1);
#  my $numProbs = scalar(@LexicalWeightFile) + scalar(@LexicalWeightFileReversed);
#  for (my $nLex=0; $nLex<$numProbs; $nLex++){
#    push(@startrule_costvector, 1);
#  }
  

#  #only the first $numProbs features get logged 



#sub addConcatenatedPhrasesToRTree{
#  my $rtree = shift;
#  my $currentparsespans_href = shift;
#  my $currentphrasepairs_href = shift;

#  my $sourcewords_aref = shift;
#  my $targetwords_aref = shift;
#  # iterate through the parse spans of length > 1 bottom up (important!) and see which syntactical 
#  # span doesn't have a corresponding phrase pair yet
#  for (my $spansize = 2; $spansize <= scalar(@$targetwords_aref); $spansize++) {
#    for (my $spanstart = 0; $spanstart + $spansize <= scalar(@$targetwords_aref); $spanstart++) {
#      my $targetspan = "$spanstart $spansize";
#      #		print STDERR "\nSpan $targetspan: ". "\n";
#      my $parsespaninfo = $currentparsespans_href->{$targetspan};
#      #	      if (defined($parsespaninfo)) {
#      #		print STDERR "\nDefined: $parsespaninfo";
#      #		print STDERR "Span $targetspan: ". $parsespaninfo->[0] . "\n";
#      #	      }
#      if (defined($parsespaninfo) && !exists $currentphrasepairs_href->{$targetspan}) {
#	#		print STDERR "Considering $targetspan: ".$parsespaninfo->[0] ;
#	# this target span corresponds to a syntactic category but we don't have a phrase pair for it
#	# try to glue phrase pairs of the child nodes together to create one
#	my $failure = 0;
	
#	# iterate through the child nodes' parse spans (starting with 1 b/c 0 is this span's category field)
#	# and determine their corresponding phrase pairs if existing
#	my @childsourcespanstarts = ();
#	my @childsourcespansizes = ();
#	my @childtargetspanstarts = ();
#	my @childtargetspansizes = ();
#	my @childphrasecostvectors = ();
#	for (my $i = 1; !$failure && $i < scalar(@$parsespaninfo); $i += 2) {
#	  my $tmpspanstart = $parsespaninfo->[$i];
#	  my $tmpspansize = $parsespaninfo->[$i+1];
#	  my $phrasepairinfo = $currentphrasepairs_href->{"$tmpspanstart $tmpspansize"};
#	  if (!defined($phrasepairinfo)) {
#	    # child parse span also doesn't have a corresponding phrase pair
#	    $failure = 1;
#	  } else {
#	    push @childtargetspanstarts, $tmpspanstart;
#	    push @childtargetspansizes, $tmpspansize;
#	    push @childsourcespanstarts, $phrasepairinfo->[2]; 
#	    push @childsourcespansizes, $phrasepairinfo->[3]; 
#	    push @childphrasecostvectors, $phrasepairinfo->[0]; 
#	  }
#	}
#	$failure = 1 if scalar(@childphrasecostvectors) == 0; # there are no child nodes which we could glue together
#	if (!$failure) {
#	  #		  print STDERR " Has child nodes and they all have phrase pairs.";
#	  # syntactic category/ies of this span and thus of the to-be-created phrase pair
#	  my $resultcategories = $parsespaninfo->[0];
#	  # to-be source span of the created phrase pair
#	  my $sourcespanstart;
#	  my $sourcespansize;
#	  my @costvector; # to-be phrase cost vector
	  
#	  # first sort the child nodes by source span start; return an array of indices to the sorted children
#	  my @indices = (0..(scalar(@childtargetspanstarts)-1));
#	  my @sortedindices = sort {$childsourcespanstarts[$a] <=> $childsourcespanstarts[$b]} @indices;
#	  my $firstchild = 1;
#	  for my $i (@sortedindices) {
#	    if ($firstchild) {
#	      # first child: set our source span to child's source span
#	      $sourcespanstart = $childsourcespanstarts[$i];
#	      $sourcespansize = $childsourcespansizes[$i];
#	      # initialize also the cost vector to this child's cost vector
#	      @costvector = @{$childphrasecostvectors[$i]};
#	      $firstchild = 0;
#	    } else {
#	      # check that next child's source span either borders on or overlaps with current source span
#	      if ($childsourcespanstarts[$i] > $sourcespanstart+$sourcespansize) {
#		# there's a hole... forget it
#		$failure = 1;
#	      } else {
#		# extend our span
#		$sourcespansize = $childsourcespanstarts[$i] + $childsourcespansizes[$i] - $sourcespanstart;
#		for (my $j = 0; $j < scalar @costvector; $j++) {
#		  $costvector[$j] += $childphrasecostvectors[$i]->[$j];
#		}
#	      }
#	    }
#	    #		    print STDERR "child $i: childsourcespanstart $childsourcespanstarts[$i], childsourcespanlength $childsourcespansizes[$i]. "
#	  }
#	  if (!$failure) {
#	    # gluing succeeded - store the new phrase pair
#	    #		    print STDERR "Sucess!\n";
	    
#	    # check for special case: target span spans whole target sentence but source span doesn't span whole source sentence
#	    if ($spansize == scalar(@$targetwords_aref) && $sourcespansize != scalar(@$sourcewords_aref)) {
#		      # correct source span
#	      $sourcespanstart = 0;
#	      $sourcespansize = scalar(@$sourcewords_aref);
#	    }
	    
#	    $rtree->insert([\@costvector,$resultcategories],($sourcespanstart,$spanstart,$sourcespanstart+$sourcespansize-1,$spanstart+$spansize-1));
#	    # also store phrase pair in hash %currentphrasepairs to allow for glue-creation of super-span phrase pairs
#	    $currentphrasepairs_href->{$targetspan} = [\@costvector,$resultcategories,$sourcespanstart,$sourcespansize];
#	  }
#	}
#      }
#    }
#  }
#}

sub findSpanAnnotation{

  my ($tmpminx,$tmpsizex,$tmpminy,$tmpsizey, $currentparsespans_href, $targetwords_aref) = @_;
  my $tmptargetspan = $tmpminy . ' ' . $tmpsizey;
  

  # look whether the target span has at least one syntactic category
  my $categoriesAndOcc = $currentparsespans_href->{$tmptargetspan};
  
  #strip out all the child information
  my $tmpcategories=undef;
  if (defined($categoriesAndOcc)){
    $tmpcategories = $categoriesAndOcc->[0];
  }
  
  #this way we ensure that each category is a single NT
  
  if (!defined($tmpcategories) && $UseConcatenationCategories) {
    # no syntactic category for the target span found
    # try if we can assemble two categories leftcat and rightcat together to create a concatenated one
    for (my $leftcatsize = 1; $leftcatsize < $tmpsizey; $leftcatsize++) {
      my $leftcat; 
      my $rightcat;
      #	      print STDERR $currentparsespans{$tmpminy . ' ' . $leftcatsize};
      #	      print STDERR ($leftcat = $currentparsespans{$tmpminy . ' ' . $leftcatsize});
      #	      print STDERR ($rightcat = $currentparsespans{($tmpminy + $leftcatsize) . ' ' . ($tmpsizey - $leftcatsize)});
      if ( defined($leftcat = $currentparsespans_href->{$tmpminy . ' ' . $leftcatsize}) &&
	   defined($rightcat = $currentparsespans_href->{($tmpminy + $leftcatsize) . ' ' . ($tmpsizey - $leftcatsize)})
	 ) {
	$tmpcategories = $leftcat->[0] . '+' . $rightcat->[0];
	# in case one of the categories was a space-separated list
	# of syntactic categories: use only the first ones
	$tmpcategories =~ s/ [^+]*//g;
	last;
      }
    }
  }
  if (!defined($tmpcategories) && $UseSubtractionCategories) {
    # no concatenation category found either - try left- and right-subtracting categories
    for (my $supercatsize = $tmpsizey+1; $supercatsize <= scalar(@$targetwords_aref); $supercatsize++) {
      my $subtractionsize = $supercatsize - $tmpsizey;
      my $supercat; 
      my $subtractioncat;
      # try right minus
      if ( defined($supercat = $currentparsespans_href->{$tmpminy . ' ' . $supercatsize}) &&
	   defined($subtractioncat = $currentparsespans_href->{($tmpminy + $tmpsizey) . ' ' . $subtractionsize})
	 ) {
	$tmpcategories = $supercat->[0] . '/' . $subtractioncat->[0];
	# in case one of the categories was a space-separated list
	# of syntactic categories: use only the first ones
	$tmpcategories =~ s/ [^\/]*//g;
	last;
      }
      # try left minus
      if ( defined($supercat = $currentparsespans_href->{($tmpminy - $subtractionsize) . ' ' . $supercatsize}) &&
	   defined($subtractioncat = $currentparsespans_href->{($tmpminy - $subtractionsize) . ' ' . $subtractionsize})
	 ) {
	$tmpcategories = $subtractioncat->[0] . "\\" . $supercat->[0];
	# in case one of the categories was a space-separated list
	# of syntactic categories: use only the first ones
	$tmpcategories =~ s/ [^\\]*//g;
	last;
      }
    }
  }
  if (!defined($tmpcategories) && $UseConcatenationCategories) {
    # still no appropriate category for the target span found -- try
    # if we can assemble THREE categories leftcat, midcat and rightcat
    # together to create a concatenated one
  FOR935: for (my $leftcatsize = 1; $leftcatsize < $tmpsizey-1; $leftcatsize++) {
      for (my $midcatsize = 1; $midcatsize < $tmpsizey-$leftcatsize; $midcatsize++) {
	my $leftcat; 
	my $midcat;
	my $rightcat;
	#	      print STDERR $currentparsespans{$tmpminy . ' ' . $leftcatsize};
	#	      print STDERR ($leftcat = $currentparsespans{$tmpminy . ' ' . $leftcatsize});
	#	      print STDERR ($rightcat = $currentparsespans{($tmpminy + $leftcatsize) . ' ' . ($tmpsizey - $leftcatsize)});
	if ( defined($leftcat = $currentparsespans_href->{$tmpminy . ' ' . $leftcatsize}) &&
	     defined($midcat = $currentparsespans_href->{($tmpminy + $leftcatsize) . ' ' . $midcatsize}) &&
	     defined($rightcat = $currentparsespans_href->{($tmpminy + $leftcatsize + $midcatsize) . ' ' . ($tmpsizey - $leftcatsize - $midcatsize)})
	   ) {
	  $tmpcategories = $leftcat->[0] . '+' . $midcat->[0] . '+' . $rightcat->[0];
	  # in case one of the categories was a space-separated list
	  # of syntactic categories: use only the first ones
	  $tmpcategories =~ s/ [^+]*//g;
	  last FOR935;
	}
      }
    }
  }

  #POSTERIOR CHANGED: we are no longer going to use _X
  #if (!defined($tmpcategories)) {
    # still no category found
    # just assign it the _X category
  #  $tmpcategories = "_X";
  #}
  return $tmpcategories;
}


sub matchPhrasesFromStream{
  my $sourcewords_aref = shift;
  my $targetwords_aref = shift;
  my $stream = shift;
  my $alreadyaddedphrases_href = shift;
  my $numalternatives = shift; # this is to forcible generate the occ set for the full sentence spanning phrase

  my $numPhrasesSeen = 0; #those that have passed the AltRange restriction
  my $numPhrasesPassedCost = 0;

  #the phrase pairs are already processed and coming in a stream. but we could still limit the phrases from coming
  #from a certain rank in the nbest list by checking their occurrence set with the restrictions placed on the extraction  
  while (my $phrasepairline = <$stream>) {

    #print  "SAW $phrasepairline\n";
    # this is the clear marker to move to the next set of processing
    last if ($phrasepairline =~ /^LOG: PHRASES_END/);
    #	  next if ($onlystartwithsentence > $sentencenumber); # ignore phrase pair if we're not supposed to extract yet
    # Note: only outgoing edges are assumed to be in PHRASEPAIRS!
    
    #print STDERR $phrasepairline if $VERBOSE;
    $logger->debug($phrasepairline);
    chomp($phrasepairline);
    
    #================================================================================================================================================
    # File format for spans: each line has a i1,i2,j1,j2 => [1 2...](relative to the whole sentence)
    # The indices point to elements in the order they were found 
    #================================================================================================================================================
    if ($PhrasePairSpanInfo) {
	
	my ($srcStart, $srcEnd, $sourcePhrase, $tgtStart, $tgtEnd, $targePhrase, $enumString, $posteriorProb);
	# retrieve source and target spans into $1,$2,$3,$4
	my $oknbest = $phrasepairline =~ /\((\d+),(\d+)\)(.*)\((\d+),(\d+)\)(.*) => (.*) => (.*)/;
	if ($oknbest){
	    ($srcStart, $srcEnd, $sourcePhrase, $tgtStart, $tgtEnd, $targePhrase, $enumString, $posteriorProb) = ($1,$2,$3,$4,$5,$6,$7,$8); 
	}else{
	    my $ok = $phrasepairline =~ /\((\d+),(\d+)\)(.*)\((\d+),(\d+)\)(.*)/;
	    ($srcStart, $srcEnd, $sourcePhrase, $tgtStart, $tgtEnd, $targePhrase, $enumString, $posteriorProb) = ($1,$2,$3,$4,$5,$6); 
	    die "Syntax error in phrase pair feed line $. '$phrasepairline':" .
		" should match regexp '.*\\(\\d+,\\d+\\).*\\(\\d+,\\d+).*'" if !$ok;
	    $enumString = "0";
	    $posteriorProb = 1;
	    
	}
    
	
      
 
      #extract the span info

      $enumString =~ s/\s+/,/g;
      my $numSrcWords = $srcEnd-$srcStart;
      

      #if the instance set is just one phrase, make it a set
      #if ($enumString =~ m/^([0-9]+)$/){
#	$enumString = "$1..$1";
#      }
      
      #restric to the range provided by $numalternatives (since we might be restricting it!)
      my @restricted = ();
      my @initialOccs = split(/,/,$enumString);
      #keep ONLY the alternatives that fall within the range specified
      foreach my $val (@initialOccs){
	if ($val<$numalternatives){
	  push(@restricted, $val);
	}
      }

      if (scalar(@restricted)>0){
	  #check the per word posterior
	  $numPhrasesSeen++;
	  my $postCost = -log($posteriorProb)/$numSrcWords;
	  if ($postCost < $NBestMinAltPost_cost){
	      $enumString = join(',', @restricted);	
	      #print "STR $enumString\n";
	      my $set = Set::IntRange->new(0, $numalternatives-1);
	      $set->from_Enum($enumString); #we know that this set has occ strictly less than $numalternatives
	      # POSTERIOR CHANGE MADE HERE: THIS IS THE PERFECT PLACE TO TRACK THE LOCATION OF THE PHRASE	
	      $alreadyaddedphrases_href->{$srcStart . ' ' . ($srcEnd-$srcStart) . ' ' . $tgtStart . ' ' . ($tgtEnd-$tgtStart)} = $set;
	      $numPhrasesPassedCost++;
	  }else{
	      $logger->debug("Dropping initial against cost: $NBestMinAltPost_cost:", $phrasepairline, "\n");
	  }
      }else{
	$logger->debug("Dropping initial against NBestAltRange:", $phrasepairline, "\n");
      }
    } else  {

      #NOT SUPPORTED RIGHT NOW
      die "Option not supported: Set PhrasePairSpanInfo\n";

      # retrieve source phrase, target phrase and phrase cost vector
      my ($sourcephrasestr, $targetphrasestr, $coststr) = split(/ *\# */, $phrasepairline);
      $sourcephrasestr = encodephrase($sourcephrasestr);
      $targetphrasestr = encodephrase($targetphrasestr);
      my @sourcephrasewords = split(' ', $sourcephrasestr);
      my @targetphrasewords = split(' ', $targetphrasestr);
      
      #	  # sanity test: check if every target word occurs in the parse tree yield
      #	  for my $w (@targetphrasewords) {
      #	    if (!exists $currentparsespans{$w}) {
      #	      print STDERR "Sentence $sentencenumber: Couldn't find word '$w' in parse tree yield.\nParse tree: ".join(" ",@currentparsetokens)."\nTarget sentence: ".join(" ",@targetwords)."\nPhrase pair line: $phrasepairline\n";
      #	    }
      #	  }
      
      
      #      # check for special case: source span spans whole source sentence
      #      if (scalar(@sourcephrasewords) == scalar(@$sourcewords_aref)) {
      #	# we wanna make sure that all phrase pairs whose source side is the whole source sentence also have
      #	# as target side the whole target sentence
      #	if (scalar(@targetphrasewords) != scalar(@$targetwords_aref)) {
      #	  print STDERR "Warning: Phrase pair whose source side is the whole source sentence but whose target side ".
      #	    "is not the whole target sentence. Correcting target side to whole target sentence. Sentence pair $sentencenumber (".
      #	      join(" ",@$sourcewords_aref). "#".join(" ",@$targetwords_aref)."), phrase pair (" .
      #		join(" ",@sourcephrasewords). "#".join(" ",@targetphrasewords).")\n";
      #	  @targetphrasewords = @$targetwords_aref;
      #	}
      #      }
      
      # now determine the source/target spans of the current phrase pair
      my @tmpsourcespans = lookupspans(\@sourcephrasewords,$sourcewords_aref);
      my @tmptargetspans = lookupspans(\@targetphrasewords,$targetwords_aref);
      # iterate through all the places where the source/target pair was found 
      # and add the phrase pair to %alreadyaddedphrases
      for my $tmpsourcespan (@tmpsourcespans) {
	for my $tmptargetspan (@tmptargetspans) {
	  # make sure we haven't processed this phrase pair yet (might happen when a word occurs multiple times - in that case we hopefully have the same cost vectors for the identical phrase pairs)
	  #	    if (!exists $alreadyaddedphrases{"$tmpsourcespan $tmptargetspan"}) { 
	  #   !! this is a bug, have to avoid double-counting differently
	  # POSTERIOR CHANGE MADE HERE: THIS IS THE PERFECT PLACE TO TRACK THE LOCATION OF THE PHRASE
	  $alreadyaddedphrases_href->{"$tmpsourcespan $tmptargetspan"} = $coststr;
	}
      }
    } # if ($PhrasePairSpanInfo)
  }#while loop until we hit marker for next sentence

  # always add full-sentence span, with an all-zero cost vector (costs are bogus anyway and will never be actually used later)
  # POSTERIOR CHANGE MADE HERE: THIS IS THE PERFECT PLACE TO TRACK THE LOCATION OF THE PHRASE
  my $set = Set::IntRange->new(0, $numalternatives-1);
  $set->Fill(); #since it happens in every single alternative!
  $alreadyaddedphrases_href->{'0 ' . scalar(@$sourcewords_aref) . ' 0 ' . scalar(@$targetwords_aref)} = $set;
  #$logger->debug("PRESET: ".'0 ' . scalar(@$sourcewords_aref) . ' 0 ' . scalar(@$targetwords_aref)."\n");
  return ($numPhrasesPassedCost,$numPhrasesSeen);
}

sub matchPhrasesFromTable{
  my $sourceline = shift;
  my $sourcewords_aref = shift;
  my $targetwords_aref = shift;
  my $phrasetable_href = shift;
  my $alreadyaddedphrases_href = shift;
  
  # iterate through all possible target spans and look the corresponding target phrases up in the phrase table
  for (my $spansize = 1; $spansize <= scalar(@$targetwords_aref); $spansize++) {
    for (my $spanstart = 0; $spanstart + $spansize <= scalar(@$targetwords_aref); $spanstart++) {
      my $phrasetableentry = $phrasetable_href->{join(' ', $targetwords_aref->[$spanstart..$spanstart+$spansize-1])};
      
      #phrasetablebig is not support currently
      #if (!defined($phrasetableentry)) {
      # maybe it's in the hashtable for big phrases
      #$phrasetableentry = $phrasetablebig{join(' ', $targetwords_aref->[$spanstart..$spanstart+$spansize-1])};
      #}

      if (defined($phrasetableentry)) {
	#		print "Found phrasetable entry at $spanstart,$spansize\n"; 
	# this target phrase exists in the phrase table - see if any of the source sides matches our source sentence
	# phrasetableentry is of the form: couiffeur#0.5 0.5 0.5 0.5#couffeuse#0.5 0.5 0.5 0.5
	my @phrasedata = split('#', $phrasetableentry);
	for (my $i = 0; $i < scalar(@phrasedata); $i += 2) {
	  my @tmpsourcepositions = fastlookupspans($phrasedata[$i],$sourceline);
	  if (scalar(@tmpsourcepositions) != 0) {
	    my $tmptargetspan = "$spanstart $spansize";
	    my @tmp = split(' ',$phrasedata[$i]);
	    my $sourcespansize = scalar(@tmp);
	    # iterate through all source spans that correspond to the entry's source phrase
	    for my $sourcespanstart (@tmpsourcepositions) {
	      # store phrase pair with cost vector in %alreadyaddedphrases
	      my $tmpsourcespan = $sourcespanstart . ' ' . $sourcespansize;
		      #		      print "Found corresponding source phrase at $tmpsourcespan\n";
	      # check for special case: source span spans whole source sentence
	      if (
		  $sourcespansize == scalar(@$sourcewords_aref)
		  # we wanna make sure that all phrase pairs whose source side is the whole source sentence also have
		  # as target side the whole target sentence
		  && ($spansize != scalar(@$targetwords_aref))
		 ) {
		print STDERR "Warning: Phrase pair whose source side is the whole source sentence but whose target side ".
		  "is not the whole target sentence. Ignoring it. Sentence pair $sentencenumber (".
		    join(" ",@$sourcewords_aref). "#".join(" ",@$targetwords_aref)."), phrase pair (" .
		      $tmpsourcespan. "#". $tmptargetspan . ")\n" if $VERBOSE;
	      } else {
		
			# make sure we haven't processed this phrase pair yet (might happen when a word occurs multiple times)
		#		  if (!exists $alreadyaddedphrases{"$tmpsourcespan $tmptargetspan"}) { 
		#!! this is a bug, have to avoid double-counting differently
		
		# calculate alignment penalty (minimum between: source/target span start offset 
		# mismatch and source/target span end sentence-end-based offset mismatch
		my $stretchfactor = scalar(@$sourcewords_aref) == 0 ? 1 : scalar(@$targetwords_aref)/scalar(@$sourcewords_aref);
		my $alignmentpenalty = min(
					   abs($spanstart-$sourcespanstart*$stretchfactor), 
					   abs(scalar(@$targetwords_aref)-$spanstart-$spansize -
					       $stretchfactor*(scalar(@$sourcewords_aref)-$sourcespanstart-$sourcespansize)));
		
		# store span pair and costvector (append a zero to the cost vector to match PESA cost vector)
		$alreadyaddedphrases_href->{"$tmpsourcespan $tmptargetspan"} = $phrasedata[$i+1] . ' ' . $alignmentpenalty . ' 0';
	      }
	    }
	  }
	}
      } #if(defined($phrasetableentry))
    }#for (my $spanstart ...
  }#for (my $spansize ...
}

#handles the multiple parses per line
sub buildCurrentSentenceParseSpans{
  my ($parserline, $persentencespans_aref, $parseprobs_aref, $targetwords_aref) = @_;

  #we might see a fail!
  if ($parserline =~ m/^(_fail)/){
      #there is one dummy parse (to all the _X / _S to be used as a category)
      $g_nbestParseRestrictions = new Set::IntRange(0, 0);
      $g_nbestParseRestrictions->Fill();
      push(@$parseprobs_aref, 1.0);
      return;
  }
  #remove any trailing hashmark
  $parserline =~ s/\#\s*//g;
  #note: the one that we have to fill now is across ALL sentences
  my @multiparses = split(/ *\# */, $parserline);
  #go thru each parse - rewrite this to respect the $NBestParseRange variable
  

  my $parseIndex = 0;
  my %indexAndProb = ();
  my @parseData = ();
  #print "Numparses: ", scalar(@multiparses), "\n";
  foreach my $singleparseline (@multiparses){
    if ($singleparseline =~ m/(.*) => ([0-9\-\.e]+)/){
      my $parseline = $1;
      my $prob = $2;
      #push(@$parseprobs_aref, $prob);

      #we put ANY values in here... if we have set $DoExpParseProbs we do the sorting in reverse
      $indexAndProb{$parseIndex} = $prob;
      push(@parseData, $parseline);
      $parseIndex++;

      
      #this data structure is simply [span] => NT
      #my %currentparsespans_singleSentence = ();
      #buildCurrentSentenceParseSpansSingleSentence(\%currentparsespans_singleSentence, $singleparseline, $targetwords_aref);      
      #push (@$persentencespans_aref, \%currentparsespans_singleSentence);
    }else{
	#there is only 1 sentence, and no probabilities
	my $parseline = $singleparseline;
	my $prob = 1;
	$indexAndProb{$parseIndex} = $prob;
	push(@parseData, $parseline);
	$parseIndex++;
	last;
    }    
  }
  if ($parseIndex == 0){
      $logger->error("ERROR: No parses with probs found on: ", $parserline, "\n");
      exit;
  }
  #we know that there are at most scalar(@multiparses) options: run any restrictions now...
  $g_nbestParseRestrictions = new Set::IntRange(0, $parseIndex-1);
  if($NBestParseRange eq "-1"){
    $g_nbestParseRestrictions->Fill();
  }else{
    my $finalRange = ensureRangeIsCompliant($NBestParseRange, $parseIndex-1);
    $g_nbestParseRestrictions->from_Enum($finalRange);
  }

  #there is just 1 number in the range
  my $finalKeys_aref = restrictToRangeOfData(\%indexAndProb, $g_nbestParseRestrictions);
  
  #build the prob lists, and parse hash list for the indices that we have selected
  foreach my $key (@$finalKeys_aref){
      #print "Key $key ", $indexAndProb{$key}, "\n";
      my $prob = $indexAndProb{$key};
      my $parseline = $parseData[$key];
      push(@$parseprobs_aref, $prob);
      #this data structure is simply [span] => NT
      my %currentparsespans_singleSentence = ();
      buildCurrentSentenceParseSpansSingleSentence(\%currentparsespans_singleSentence, $parseline, $targetwords_aref);      
      push (@$persentencespans_aref, \%currentparsespans_singleSentence);
  }
  
  #if we are doing normalization AND the original probs are in regular space
  if ($DoNormParseProbs && $DoExpParseProbs==0){
      NormalizeInProbSpace($parseprobs_aref, $NBestParseGeoAlpha);
  }
  

  #if we want to normalize and we in LOG space use the log sum exp approx
  if ($DoNormParseProbs && $DoExpParseProbs==1){

      #my @backupinprob = map{ exp($_); } @$parseprobs_aref;
      #print "Basic Probs : ", join(' ', @backupinprob), "\n";

      #print "Basic Logs : ", join(' ', @$parseprobs_aref), "\n";
      NormalizeInLogToProbSpace($parseprobs_aref, $NBestParseGeoAlpha);

      #NormalizeInProbSpace(\@backupinprob);
      #print "Normed by log : ", join(' ', @$parseprobs_aref), "\n";
      #print "Normed by prob : ", join(' ', @backupinprob), "\n";
  }

  #print "PARSEPROBS : ", join(' ', @$parseprobs_aref), " FROM ", $g_nbestParseRestrictions->to_Enum(), "\n";
  VerifyDecreasingProbs($parseprobs_aref);
}

sub VerifyDecreasingProbs{
    my ($aref) = @_;
  #at the end do a sanity check to make sure the probs are in decreasing order or 0                                                              
    my $firstValue = undef;
    foreach my $value (@$aref){
	if ($value != 0){
	    if (defined($firstValue)){
		die "Current value $value > last value: $firstValue: in ", join(' ', @$aref), "\n" if ($value > $firstValue);
	    }else{
        #you have not yet hit a real value yet, so define it now                                                                                 
		$firstValue = $value;
	    }
	}else{
      #skip this value - 0 indicates that we have blanked it out                                                                                 
	}
    }

}

sub NormalizeInProbSpace{
    my ($parseprobs_aref, $alpha) = @_;
    my $sum = 0; map{ $sum+= ($_ ** $alpha); } @$parseprobs_aref;
    my @normalized = map { ($_ ** $alpha)/$sum; } @$parseprobs_aref;
    @$parseprobs_aref = @normalized;
}

sub NormalizeInLogToProbSpace{
    my ($logprobs_aref, $alpha) = @_; #sorted from largest value to lowest!
    my $lowestAbsValue  = $logprobs_aref->[-1];
    my $secondToLastIndex = scalar(@$logprobs_aref)-2;
    #uptil this value
    my @toSum = map{exp($alpha*($_ - $lowestAbsValue));} @$logprobs_aref[0 .. $secondToLastIndex];
    my $insideSum = 0;
    map{ $insideSum += $_; } @toSum;
    my $logSum = $alpha*$lowestAbsValue + log(1 + $insideSum);
    
    my @normalized = map { exp(($alpha*$_)-$logSum); } @$logprobs_aref;
    @$logprobs_aref = @normalized;
    
}

sub ensureRangeIsCompliant{
  my ($range, $maxIndex) = @_;
  #going to create a set, but ensure that none of the number are past the range
  my $finalRange;
  if ($range =~ m/^([0-9]+)$/){
    if ($1> $maxIndex){
      $finalRange = $maxIndex;
    }
    $finalRange = "$1..$1\n";
  }elsif ($range =~ m/^([0-9]+)\.\.([0-9]+)/){
    my $first = $1;
    if ($first != 0){
      die "Ranges that dont start at 0 are not supported\n";
    }
    my $second = $2;
    if ($first>$maxIndex){
      $first = $maxIndex;
    }
    if ($second>$maxIndex){
      $second = $maxIndex;
      }
    $finalRange = "$first..$second";
  }else{
    die "UNKNOWN range $range\n";
  }
  return $finalRange;
}

#select the keys that yield the best (highest values), and restrict to those speficied by the range
sub restrictToRangeOfData{
  my ($indexAndProb_href, $restrictionSet) = @_;
  my @sortedKeys = reverse sort{ $indexAndProb_href->{$a} <=> $indexAndProb_href->{$b} } keys %$indexAndProb_href;
  #@sortedKeys now has a set of indices, go thru and for each index see if the bit is set in the $restrictionSet
  my @finalKeys = ();
  for (my $index =0; $index<scalar(@sortedKeys); $index++){
    if ($restrictionSet->contains($index)){
      push(@finalKeys, $sortedKeys[$index]);
    }
  }

  #if no elements were requested, this means that the of n selected in the global restriction was greater than 
  #what was available. In this case, take the last value..
  if (scalar(@finalKeys)==0){
    push(@finalKeys, $sortedKeys[-1]);
  }

  return \@finalKeys;  
}

sub buildCurrentSentenceParseSpansSingleSentence{
  use strict;
  my $currentparsespans_href = shift;
  my $parserline = shift;
  my $targetwords_aref = shift;

  chomp($parserline);
  if ($parserline =~ /^(_fail|PARSEFAIL)/) {
    print STDERR "Parse failure at sentence number $sentencenumber: $parserline; target sentence: ".join(" ",@$targetwords_aref). "\n";
  } else {
    #expand parens to have spaces around them
    $parserline =~ s/\(/\( /g;
    $parserline =~ s/\)/ \) /g;

    my @currentparsetokens = split(' ', $parserline);
    #print join ("," , @currentparsetokens)."\n";
    #used in recursion... take care with this
    my $currentparsetokenspos = 0;
    #create the %currentparsespans entries, have span indices start from 0, not 1	
    my $sentencelengthtmp = findparsespans(0, $currentparsespans_href, \@currentparsetokens, \$currentparsetokenspos); 

    if ($sentencelengthtmp != scalar @$targetwords_aref) {
	if ($sentencelengthtmp == -1) {
	    # error analyzing parse, but we printed the error message already, so do nothing here
	} else {
	    #QUESTION: if you see this kind of mismatch, what is the plan? should you proceed to the next tuple (s,t,tree) directly?
	    #az: no, we will just annotate everything with _X in such a case
	    print STDERR "Sentence $sentencenumber: Sentence length mismatch parse tree yield vs. target sentence!\nParse tree: ".join(" ",@currentparsetokens)."\nTarget sentence: ".decodephrase(join(" ",@$targetwords_aref))."\n";	    
	}
	%$currentparsespans_href = (); #reset parse span hash for current sentence
    } else {
	print STDERR "$parserline\n" if $VERBOSE>=2;
	#(print STDERR "Target parse spans:\n"),printhash($currentparsespans_href) if $VERBOSE;
    }
  }
}

sub verifySanitySrcCond{
    my $complexrules_href = shift;
    my $phrasecounts_href = shift;

    my ($key, $value);    
    
    #global debug variable
    my %srcHash = ();
    while (($key,$value) = each %$complexrules_href) { 
	my ($sourcestr,$targetstr,$typestr) = split(/\s*\#\s*/, $key);
	my @costvector = split(" ", $value);
	my $rulefreq = $costvector[0]; #this should be the REAL rule freq
	#print STDERR "RULE: ",decodephrase($sourcestr),"\#",decodephrase($targetstr),"\#",$typestr,"\#",$value,"\n";
	$srcHash{$sourcestr} += $rulefreq;
    }
    #once you verify that you have the same probability	
    foreach my $sourcestr (keys %srcHash){
       	my $sourcefrequency = $phrasecounts_href->{hashcode($sourcestr . '#')};    # frequency of rule's source 
	if (!defined($sourcefrequency)){
	    print STDERR "ERROR: sourcefrequency undefined for:",decodephrase($sourcestr),"\n";
	}else{
	    my $summedFreq = $srcHash{$sourcestr};
	    die "Failed src: ",decodephrase($sourcestr)," summed $summedFreq - stored $sourcefrequency\n" if ($summedFreq != $sourcefrequency);
	    
	}
    }
    #output a preliminary table and see what the scores are
    while (($key,$value) = each %$complexrules_href) { 
	my ($sourcestr,$targetstr,$typestr) = split(/\s*\#\s*/, $key);
	my @costvector = split(" ", $value);
	my $rulefreq = $costvector[0];
	print "VERIFY ",decodephrase($sourcestr),"\#",decodephrase($targetstr),"\#",$typestr," => srccond: ",$rulefreq/$srcHash{$sourcestr}," freq: $rulefreq\n";

    }
}
sub writeComplexRules{
  #handle parameters
  my $complexrules_href = shift;
  my $phrasecounts_href = shift;

   # write out the rules in %complexrules
  my $totalrulefreqcount_sanitytest = 0;
  my $rulenumber = 0;
  my ($key, $value);
  while (($key,$value) = each %$complexrules_href) { 
    $rulenumber++;
    print STDERR "Processing rule $rulenumber... \n" if ($rulenumber-1) % 50000 == 0;

    #  my $neglogfrequency = - log $value->[0];
    #my $frequency = $value->[0];
    
    my ($sourcestr,$targetstr,$typestr) = split(/\s*\#\s*/, $key);
    my @costvector = split(" ", $value);

    my @ruletargetwords = split(" ", $targetstr);
    my @rulesourcewords = split(" ", $sourcestr);
    my @resultcategories = split(" ", $typestr);

    my $frequency = $costvector[0]; # frequency of rule
    my $simplefreq = $costvector[1]; # frequency of rule
    die "Assertion error: simplefreq $simplefreq frequency: $frequency for key : ",decodephrase($sourcestr), "#", decodephrase($targetstr), " => $value" if (($simplefreq==0) || ($frequency==0));
    die "Asserion error: simplefreq $simplefreq < frequency $frequency for key : ",decodephrase($sourcestr), "#", decodephrase($targetstr), " => $value" if ($simplefreq+0.1 < $frequency);
    $totalrulefreqcount_sanitytest += $frequency;

    my $sourcefrequency = $phrasecounts_href->{hashcode($sourcestr . '#')};    # frequency of rule's source side (excl. arg NTs)
    if (!defined($sourcefrequency)) {
      print STDERR "DB error: Rule $key\#$value: sourcefrequency not contained in phrasecounts DB!\n";
      exit;
      # this might have come from the rule consolidation - just give it a one-frequency
      $sourcefrequency=1; #POSTERIOR CHANGE??
    }
    #checking for pure lexical only
    #if (scalar(@resultcategories)==1){
	#my $altsrcfreq = $srcHash{$sourcestr};
    #die "Src freq does not match sourcefrequency: $sourcefrequency alt: $altsrcfreq\n" if ($altsrcfreq != $sourcefrequency);	    #}

    my $source_and_arg_frequency = $phrasecounts_href->{hashcode($sourcestr . '#' . join(' ', @resultcategories[0..$#resultcategories-1]) . '#')};    # frequency of rule's source side (incl. arg NTs)
    if (!defined($source_and_arg_frequency)) {
      print STDERR "DB error: Rule $key\# source_and_arg_frequency not contained in phrasecounts DB!\n";
      exit;
      # this might have come from the rule consolidation - just give it a one-frequency
      $source_and_arg_frequency=1; #POSTERIOR CHANGE??
    }
    my $targetfrequency = $CalculateTargetConditionedRFs? $phrasecounts_href->{hashcode('#' . $targetstr)} : -1;    # frequency of rule's target side
    if (!defined($targetfrequency)) {
      print STDERR "DB error: Rule $key\#$value: targetfrequency not contained in phrasecounts DB!\n";
      exit;
      # this might have come from the rule consolidation - just give it a one-frequency
      $targetfrequency = 1; #POSTERIOR CHANGE??
    }

    my $rulebodyfrequency = $phrasecounts{hashcode("$sourcestr\#$targetstr\#".$resultcategories[-1])};   	# frequency of rule's body (different from $frequency in that multiple types for identical rule bodies were not multiply counted)
    if (!defined($rulebodyfrequency)) {
	print STDERR "DB error: Rule ",decodephrase($sourcestr),"#",decodephrase($targetstr),"#",$resultcategories[-1]," rulebodyfrequency not contained in phrasecounts DB!\n";
	die;
    }

    # check for hash collision:
    if ($frequency > $rulebodyfrequency+0.01) {
      print STDERR "Warning: rule frequency $frequency greater than rulebodyfrequency $rulebodyfrequency; must be due to hashcode collision in phrasecounts hash. SRC: ", decodephrase($sourcestr)," TGT: ",decodephrase($targetstr),"\n";
      die;
    }
    if ($frequency > $source_and_arg_frequency+0.01) {
      print STDERR "Warning: rule frequency $frequency greater than source_and_arg_frequency $source_and_arg_frequency; must be due to hashcode collision in phrasecounts hash. Setting frequency=$source_and_arg_frequency.\n";
      $frequency = $source_and_arg_frequency;
      die;
    }
    if ($rulebodyfrequency > $sourcefrequency+0.01) {
      print STDERR "Warning: rulebodyfrequency $rulebodyfrequency greater than sourcefrequency $sourcefrequency; must be due to hashcode collision in phrasecounts hash. Setting rulebodyfrequency=$sourcefrequency.\n";
      $rulebodyfrequency = $sourcefrequency;
      die
    }
    if ($CalculateTargetConditionedRFs && $rulebodyfrequency > $targetfrequency + 0.01) {
      print STDERR "Warning: rulebodyfrequency $rulebodyfrequency greater than targetfrequency $targetfrequency; must be due to hashcode collision in phrasecounts hash. Setting rulebodyfrequency=$targetfrequency.\n";
      $rulebodyfrequency = $targetfrequency;
      die;
    }



    #Fill the @resultcount vector with 2 elements 
    #$resultcount[0] = posterior count frequency (the important count)
    #$resultcount[1] = a simple incremental count (the important count)
    my $resultcategory = $resultcategories[$#resultcategories];
    # get the count (in indiluted and in diluted interpretation) of result category
    my @resultcount = ();
    push @resultcount, $resulttypecounts_diluted{$resultcategory};
    if (!defined($resultcount[0])) {
      print STDERR "DB error: Rule $key\#$value: resultcount not contained in resulttypecounts DB!\n";
      next;
    }
    push @resultcount, $resulttypecounts_undiluted{$resultcategory};
    if (!defined($resultcount[1])) {
      print STDERR "DB error: Rule $key\#$value: resultcount not contained in resulttypecounts DB!\n";
      next;
    }

    # count source- and target-terminal symbols
    my $sourcewordcount = scalar(@rulesourcewords) + 1 - scalar(@resultcategories);
    my $targetwordcount = scalar(@ruletargetwords) + 1 - scalar(@resultcategories);

    if ( (			# either: parameter -outputallrules was passed
	  $outputallrules
	 ) || 
	 (			# or: two conditions have to hold
	  # first condition: balancedness
	  ( (($sourcewordcount == 0) == ($targetwordcount == 0))
	    # rule has no source terminals iff rule has no target terminals
	    || scalar(@resultcategories) == 1
	    # always counts as fulfilled if rule is fully lexicalized 
	    || $keepunbalancedrules
	    # always counts as fullfilled if keepunbalancedrules
	    # parameter is set
	  ) && 
	  # second condition: plausibility
	  ( $frequency > 1 
	    # seen phrase pair at least twice
	    || (scalar(@resultcategories) == 1 && scalar(@rulesourcewords) <= $MaxLengthSingleCountSourcePhrases)
	    # or: rule is fully lexicalized and source side has at most
	    # MPLSCP words
#	    || (scalar(@resultcategories) == 1
#		&& scalar(@rulesourcewords) <= $MaxLengthSingleCountSourcePhrases
#		&& $rulebodyfrequency/$sourcefrequency >= 0.15)
#	    # or: rule is fully lexicalized and source-cond. RF >= 0.15 and source side has at most
#	    # MPLSCP words
	  )
	 )
       ) {
      # write out the rule
      #print STDERR "Writing out: $key, rel. freq.: $frequency/$resultcount[0], target-cond. r.f.: $rulebodyfrequency/$targetfrequency, source-conditioned r.f.: $rulebodyfrequency/$sourcefrequency\n" if $VERBOSE;

      #convert rule format into the new style
      foreach (my $i = 0; $i < scalar @rulesourcewords; $i++) {
	if ($rulesourcewords[$i] =~ /^\@/ ) {
	  # source word starts with @ -- must be an arg marker
	  # replace this by next result category, and prepend a @ to the category name (all nonterminals will start with @ in our parsing system)
	  $rulesourcewords[$i] = "\@" . shift(@resultcategories);
	}
      }

      # @resultcategories should now contain exactly one nonterminal symbol
      print STDERR "In rule $key: Mismatch between rule's number of type categories and number of arg markers in source of the rule! Perhaps training corpus contains a word starting with @ symbol?\n" if scalar(@resultcategories) != 1;
      
      #prepend @ to the result category
      my $resultcategory = "\@" . $resultcategories[0];

      #create new rule source with modfied @NT format and decoded word indices
      my $newsourcestr = join(" ", decodewordlist(@rulesourcewords));

      #========================================================================================================
      # Produce the finalcostvector for the rule. These features here are specifically those that can only be
      # estimated during the extraction process since we know which other rules were extracted as well as the
      # alignment information used to create accurate lexical weights. One last bit of fiddling, if you have
      # specified --AverageLexicalWeights then we need to do the averaging here before converting them to log space
      #========================================================================================================
      
      #$logger->debug("InitialCosts: ",  decodephrase($sourcestr), ", ", decodephrase($targetstr), " # ", join(' ', @costvector), "\n");

      my $hasZeros = 0;
      #dont want to average over the first 2 features (which are frequencies!)
      for(my $ci = 2; $ci<scalar(@costvector); $ci++){
	if ($AverageLexicalWeights){
	    $costvector[$ci] = $costvector[$ci]/$simplefreq;
	}
	
	if ($OutputProbCosts==0){
	  if ($costvector[$ci]>0){
	    $costvector[$ci] = -log($costvector[$ci]);
	  }else{
	    $hasZeros = 1;
	  }
	}
      }


      # cost vector
      #my $finalcostvector = 
      my @final = (
		   0,		# glue feature
		   1,		# rule application counter
		   $targetwordcount, # target word counter
		   $OutputProbCosts ? $frequency/$resultcount[0] : -log($frequency/$resultcount[0]), # neglog (base e) diluted-result-conditioned RF
		   #		   -log($rulebodyfrequency/$targetfrequency), # neglog target-cond. RF
		   $OutputProbCosts ? $rulebodyfrequency/$sourcefrequency : -log($rulebodyfrequency/$sourcefrequency), # neglog source-excl.-arg-NTs-cond. RF
		   $OutputProbCosts ? $frequency/$source_and_arg_frequency : -log($frequency/$source_and_arg_frequency), # neglog source-incl.-arg-NTs-cond. RF
		   #		   -log($frequency/$ruleinstancecount) #neglog joint RF
		   #-log($frequency),	# just for info purposes: neg log of frequency of rule
		   $OutputProbCosts ? $frequency : -log($frequency),	# just for info purposes: neg log of frequency of rule
		   $OutputProbCosts ? $simplefreq : -log($simplefreq),	# just for info purposes: neg log of frequency of rule
		   @costvector[2..$#costvector]  	 # phrase lexical weights (note how we start this vector from 1!)
		  );
      

      #my @finalShort = map {sprintf("%.3g", $_);} @final;
      #$finalShort[6] = $final[6];
      #$finalShort[7] = $final[7];
      my $finalcostvector = join(' ', @final);

      #print "finalWriteComplexRules: ",decodephrase($sourcestr), "\# ",decodephrase($targetstr)," \# \n";
      

      #create rule body
      my $newtargetstr = decodephrase($targetstr);

      if ($resultcategory eq '@_S') {
	# add BOS and EOS markers
	$newsourcestr = '<s> ' . $newsourcestr . ' </s>';
	$newtargetstr = '<s> <s> ' . $newtargetstr . ' </s> </s>';
      }
      my $rulebody = $newtargetstr . "\#$resultcategory\#$finalcostvector";

      outputrule($newsourcestr,$rulebody);
    }
  }
  return ($totalrulefreqcount_sanitytest, $rulenumber);
}

#write out rules that represent the number of times each non-terminal counts has been seen
sub writeNTCounts{
  #to get a statistics about how many nonterminals there are
  my $nonterminalcount = 0; 
  #  #  print STDERR "Nonterminals X for which a rule \@X\#\@1\#\@_S has been extracted \n(for the other NTs we'll add artificial enter-glue-mode rules):\n";
  print STDERR "Nonterminals occurring as rule results and their diluted and nondiluted counts: \n";
  # create artifical _COUNT rules to store the n.o. occurrence for each nonterminal symbol
  while (my ($key,$value) = each %resulttypecounts_diluted) { 
    $nonterminalcount++;
    # get the count (in indiluted and in diluted interpretation) of result category $key
    $value .= ' ' . $resulttypecounts_undiluted{$key};
    print STDERR "$key($value) ";
    # write out a pseudo-rule storing the counts for this nonterminal for use in e.g. filterrules.pl
    outputrule("\@_COUNT", "\@1\#\@$key\#$value");
    
  }
  print STDERR "\nTotal number of nonterminal symbols (including _S): $nonterminalcount\n";
}















#################### subs

sub registerFileHandle{
  my $fileName = shift;
  my $tag = shift;
  local *IN;
  my $in = *IN{IO};
  open $in, $fileName or die "Unable to open $fileName\n";
  $dataFileHandles{$tag} = $in;
}

sub setupFileHandles{

  die "No feed file specified via --PhrasePairFeedFile" if !$PhrasePairFeedFile;
  registerFileHandle($PhrasePairFeedFile, "FEED") if $PhrasePairFeedFile;
  registerFileHandle($SourceFile, "SRC") if $SourceFile;
  registerFileHandle($TargetFile, "TGT") if $TargetFile;
  registerFileHandle($AlignmentsFile, "ALT") if $AlignmentsFile;
  registerFileHandle($TargetParseTreeFile, "PARSE") if $TargetParseTreeFile;

}


# output rule to database or standard out
sub outputrule {
  my($newsourcestr, $rulebody) = @_;
  if ($outputdbfile) {
    $outputdb->db_put($newsourcestr . "\0", $rulebody . "\0");
  } else {
    print "$newsourcestr\#$rulebody\n";
  }
  $writtenoutrulenumber++;
}

# recursively create and write out complex rules from $elemrule into %complexrules
# $indexedsubrules: array of array of subrules of $elemrule, indexed by source-span start relative to $elemrule's source span start
# $currentrelsourcewordindex: min source word position (relative to $elemrule's source span start) where subrules to be substituted are allowed to start
# recursion start: elementary rules, i.e., no words replaced yet; $currentrelsourcewordindex, $sourcesymbolcount_lowerbound and $recursiondepth are 0
sub createcomplexrules {
  my ($elemrule, $r_occSet, $r_source, $r_target, $r_type, $indexedsubrules, $countunreplacedsourcewords,$countunreplacedtargetwords, $currentrelsourcewordindex, $sourcesymbolcount_lowerbound, $recursiondepth, $abstract_rule_still_possible, $sourcewords_aref, $targetwords_aref, $i_j_w_matrix_aref, $nbest_alignmentProbs_aref, $nbest_parseProbs_aref) = @_;

  #Recurse(R, I, Lastsubrule, Recursiondepth):
  # // In Each Recursion Step, We Will Abstract Over A Sub-Rule Contained In R
  # // This Happens From Left To Right Source-Wise, Ensuring That The #I Symbols
  # // Are In The Order #1, #2, ...

  my ($r_minx,$r_miny,$r_maxx, $r_maxy) = @$elemrule[2..5]; # span of R
  my $r_source_symbolcount = $sourcesymbolcount_lowerbound + ($r_maxx+1-$r_minx)-$currentrelsourcewordindex;

  #get the category list of the most recent carve
  my $result_href = $r_type->[-1];
  my @results = keys %$result_href;
  #my @recent_carve = keys %$tmp_href; #the result category set of the current elemental rule that we are abstracting

  if (
      #check some facts about the number elemental rule to make sure we can generalize it
	(
	(
	$r_source_symbolcount == $countunreplacedsourcewords
	# rule is lexical (pure lexical on the src side)
	#	&& $r_type->[0] ne '_# and rule type is syntactic (i.e., not an X rule)
	# and it doesnt have too many symbols
	&& $r_source_symbolcount <= $MaxSourceLength
	# and rule is short enough
	)
	||
	
	(
	((
	($countunreplacedsourcewords == 0 && $countunreplacedtargetwords == 0) && ($recursiondepth <= $MaxSubstitutionCountAbstractRules)
	# rule is abstract - max sure we abstracted over at most $MaxSubstitutionCountAbstractRules
	)
	||
	(
	 ($countunreplacedsourcewords != 0 || $countunreplacedtargetwords != 0) && ($recursiondepth <= $MAXSUBSTITUTIONCOUNT)
	 # rule has lexical elements - make sure we have abstracted over at most $MAXSUBSTITUTIONCOUNT
	))
	&&
	 #check to make sure any rule is not too long (this check is only done for nonlexical rules)
	 (       
	  $countunreplacedsourcewords <= $MAXSOURCEWORDCOUNT && $countunreplacedtargetwords <= $MAXTARGETWORDCOUNT 
	  && 
	  $r_source_symbolcount <= ($results[-1] eq '_X' ? $MaxSourceSymbolCountXRules : $MAXSOURCESYMBOLCOUNT) #check the last abstraction label
	 )
	 )
	 )
      &&   # and: last lexical chunk of rule occurs as n-gram in the source-side of the dev/test data
      !($SourceRestrictionCorpus && $currentrelsourcewordindex+$r_minx != $r_maxx+1 && !exists($vocCheck{join(' ', map{ $_ eq $encoded_ignore_word ? () : ($_) } @$sourcewords_aref[$r_minx+$currentrelsourcewordindex..$r_maxx])}))
      && #rule meets balancedness criteria
      ( (($countunreplacedtargetwords!=0) ==  ($countunreplacedsourcewords!=0)) || $keepunbalancedrules)
      ){
      #   Write Out Rule R (unless $MAXTARGETSYMBOLCOUNT condition isn't fulfilled)
      
      print "About to call addrules!!\n";

    #These indices are with respect to the full source sentence
    my ($minx,$miny,$maxx, $maxy) = @$elemrule[2..5]; # span of R
    #r_source contains just the words of the phrase that we extracted

    #================================================================================================================================================
    # identifyCompactPhrase creates a $sourcestr by selecting the span from the full sentence sequence
    # we use the returned lexical index references to calculate lexical weights
    #================================================================================================================================================
    my ($srcLexicalIndices_aref, $sourcestr, $r_source_symbolcount_sanity_check) = identifyCompactPhrase($minx, $maxx, $r_source, $sourcewords_aref);

    my ($tgtLexicalIndices_aref, $targetstr, $r_target_symbolcount) = identifyCompactPhrase($miny, $maxy, $r_target, $targetwords_aref);

    die "Sanity check \$r_source_symbolcount != \$sourcesymbolcount_lowerbound + (\$r_maxx+1-\$r_minx)-\$currentrelsourcewordindex ($r_source_symbolcount == $sourcesymbolcount_lowerbound + ($r_maxx+1-$r_minx)-$currentrelsourcewordindex) failed; sourcestr=$sourcestr; targetstr=$targetstr; r=".join(" ",@$r_source)."#".join(" ",@$r_target)."#".join(" ",@$r_type)."\n" if $r_source_symbolcount != $r_source_symbolcount_sanity_check;


    #print "DEBUG: calcLex # ",decodephrase($sourcestr), " # ", decodephrase($targetstr), " srcLex: ",join(' ', @$srcLexicalIndices_aref), " tgtLex: ", join(' ', @$tgtLexicalIndices_aref),"\n";


    #================================================================================================================================================
    # Create a vector of PROBS that are associated with this rule. This is the ONLY place where costs are generated for the rule
    # We also begin tracking the number of times the rule occurred in this sentence. In the posterior model, the frequency counts are
    # summed (in addrules), but we need to set the frequency to the right value based on the occurrence counts!
    # $r_occSet is the (steadily intersected) set of alignments that support this subtraction!
    # $nbest_alignmentProbs_aref == the probabilities for the each alignment choice, we will sum over these to get the freq count
    #================================================================================================================================================
    #This calcRuleLexicalWeights is based on srcLexicalIndices_aref and tgtLexicalIndices_aref which change based on r_source and r_target
    #The first weight in the @ruleProbs is the weighted relateive frequency count taken from the alignments
    #This count does NOT consider the probability of the NT categories
    #print "Adding snum $sentencenumber rule: ",decodephrase($sourcestr), "\# ",decodephrase($targetstr), "\n";
    $g_hasFailed = 0;
    my @ruleProbs = calcRuleLexicalWeights($r_occSet, $srcLexicalIndices_aref, $tgtLexicalIndices_aref, $i_j_w_matrix_aref, $nbest_alignmentProbs_aref);


    #die "Assertion failed, too many probs in @ruleProbs " if (scalar(@ruleProbs) != 1 + scalar(@LexicalWeightFileReversed) + scalar(@LexicalWeightFile));
    # the 1+ is for the alignment posterior count that was added in calcRuleLexicalWeights

    #print "Adding rule: ",decodephrase($sourcestr), "\# ",decodephrase($targetstr)," \# ", join(' ', @ruleProbs), " => ", $r_occSet->to_Enum(), "\n";

    if (
	(
	 $r_source_symbolcount == $countunreplacedsourcewords
	 # rule is lexical 
	 #	 && $r_type->[0] ne '_X'
	 #	 # and rule type is syntactic (i.e., not an X rule)
	 && $r_source_symbolcount <= $MaxSourceLength
	 # and rule is short enough
	) ||
	# or:
	(
	 $r_target_symbolcount <= $MAXTARGETSYMBOLCOUNT
	)	
       ) {
      # iterate over the category alternatives, adding each possible resulting rule to %complexrules
      #addrules(\$sourcestr,\$targetstr,$r_costvector,$r_type,scalar(@$r_type),[]);
      if ($g_hasFailed==1 || ($ruleProbs[0]<=0)){
	$logger->debug("Rejecting rule 0 alt evidence or has no lexical evidence: ",decodephrase($sourcestr), "\# ",decodephrase($targetstr),"\n");
      }else{
	$logger->debug("Adding rule SURE: ",decodephrase($sourcestr), "\# ",decodephrase($targetstr),"\n");
	#printRType($r_type);
      
	#POSTERIOR CHANGE
	#addrules(\$sourcestr,\$targetstr,\@ruleProbs,$r_type,scalar(@$r_type),[]);
	
	#my $workingSet = new Set::IntRange(0,scalar(@$nbest_parseProbs_aref)-1);
	#$workingSet->Fill(); #initialize the working set with all entries
	#pass the miny, maxy of the elementary rule for uniqueness checking, and the length of the whole target for the bit vector set

	my $spanOccSet;
	if (!$SimpleNovelty){
	    $spanOccSet = new Set::IntRange(0, scalar(@$targetwords_aref)-1);
	    $spanOccSet->Interval_Fill($miny, $maxy);
	}
	if ($countunreplacedsourcewords == 0 && $countunreplacedtargetwords == 0){
	    print "RecursionDepth: $recursiondepth : ",decodephrase($sourcestr), "#",decodephrase($targetstr),"\n";
	}
	addrules($nbest_parseProbs_aref, \$sourcestr,\$targetstr,\@ruleProbs,$r_type,scalar(@$r_type),[], $r_occSet, $spanOccSet, scalar(@$targetwords_aref)-1, undef);

	
	#verifySanitySrcCond(\%complexrules, \%phrasecounts);
      }
    }
  }				#if $countunreplaced...

  #recursiondepth is going to increase below, make sure that it can
  if ($recursiondepth < $MAXSUBSTITUTIONCOUNT || 
      ($recursiondepth < $MaxSubstitutionCountAbstractRules && $abstract_rule_still_possible)
      ) {
    # iterate through all subrules whose rel. source span starts at or after $currentrelsourcewordindex
    # and abstract over each of those (separately)
    # in case we are only looking for abstract rules, we will only do this for loop 
    # once (for $j == $currentrelsourcewordindex); see the end of the loop for that respective 'last' command
    #Also, keep track of each sourceword we skip in $skipped_source_words in order to 
    # possibly check against the dev/test set restriction file
    my $skipped_source_words = '';
  LABEL_j_loop319:
    for (my $j = $currentrelsourcewordindex; 
	 $j < scalar @$indexedsubrules; 
	 $j++) {
    LABEL320: 
      for my $subrule (@{$indexedsubrules->[$j]}) {
	my ($sub_minx,$sub_miny,$sub_maxx, $sub_maxy) = @$subrule[2..5]; # span of $subrule
	my $new_countunreplacedsourcewords = $countunreplacedsourcewords-($sub_maxx+1-$sub_minx);
	my $new_countunreplacedtargetwords = $countunreplacedtargetwords-($sub_maxy+1-$sub_miny);
	if (!$keepunbalancedrules &&
	    ( ($new_countunreplacedsourcewords == 0) != ($new_countunreplacedtargetwords == 0) )
	   ) {
	  # unbalanced rule because source still contains words but target doesn't or vice versa
	  # ignore this rule
	  next LABEL320;
	}
	      
	my $new_sourcesymbolcount_lowerbound = $sourcesymbolcount_lowerbound + 1 + ($sub_minx-$r_minx)-$currentrelsourcewordindex ; # the new n.o. source symbols is the old one plus one (new substitution symbol @N) plus n.o. skipped (non-substituted) words
	# check that our lower bound hasn't exceeded the limit $MAXSOURCESYMBOLCOUNT yet
	if ($r_maxx == $sub_maxx) {
	  # case a) we will have processed the whole source span of r; no further substitution will be necessary
	  #POSTERIOR CHANGED
	  if ($new_sourcesymbolcount_lowerbound > ($results[-1] eq '_X' ? $MaxSourceSymbolCountXRules : $MAXSOURCESYMBOLCOUNT)) {
	  #if ($new_sourcesymbolcount_lowerbound > ($MAXSOURCESYMBOLCOUNT)) {
	    next LABEL320;	# maximum exceeded -> try next subrule
	  }
	} else {
	  # case b) we won't have processed the whole source span of r yet; at least one further substitution will be necessary
	  if ($new_sourcesymbolcount_lowerbound + 1 > ($results[-1] eq '_X' ? $MaxSourceSymbolCountXRules : $MAXSOURCESYMBOLCOUNT)) {
	  #if ($new_sourcesymbolcount_lowerbound + 1 > ($MAXSOURCESYMBOLCOUNT)) {
	    next LABEL320;	# maximum exceeded -> try next subrule
	  }
	}

	my @new_r_source = @$r_source;
	my @new_r_target = @$r_target;
	# now loop over target positions of subrule $subrule and replace positions in @new_r_target accordingly
	# still have to check if $subrule's target span is still lexicalized within R 
	# (i.e., hasn't been partially replaced yet), do that on-the-fly in the following loop
	for (my $k = $sub_miny-$r_miny; $k<=$sub_maxy-$r_miny; $k++) {
	  next LABEL320 if $r_target->[$k] != 0;
	  $new_r_target[$k] = 
	    $k==$sub_miny-$r_miny? # first position to be changed?
	      $recursiondepth+1 : -1;
	}
	
	# do similar loop over source positions of subrule $subrule and replace positions in @new_r_source accordingly
	for (my $k = $sub_minx-$r_minx; $k<=$sub_maxx-$r_minx; $k++) {
	  print STDERR "Assertion error: \$r_source->[$k] != 0\n" if $r_source->[$k] != 0;
	  $new_r_source[$k] = 
	    $k==$sub_minx-$r_minx? # first position to be changed?
	      $recursiondepth+1 : -1;
	}



	# create the new type signature - for the rule with the subrule carved out
	my $subrule_categoryOcc_href = $subrule->[1]->[1]; # result categories of $subrule
	
	#the new list is formed in reverse order, with the result category *list* of the rule always remaining at the end!
	#POSTERIOR: for each span we have a HASH_ref of categories with their associated occs. Its addrules' jobs to 
	#search thru the space of possible labelling and track occurrence intersections
	my @new_r_type = (@$r_type[0..scalar(@$r_type)-2], $subrule_categoryOcc_href, $$r_type[$#$r_type]);
	
	#	print STDERR "Recursing to next level with ".'($sub_minx,$sub_miny,$sub_maxx, $sub_maxy)='."($sub_minx,$sub_miny,$sub_maxx, $sub_maxy)\n";

	# whether it is still possible to generate an abstract rule from here
	# if j > $currentrelsourcewordindex, then we have skipped a source word, so an abstract rule is not possible anymore
	my $new_abstract_rule_still_possible = $abstract_rule_still_possible && $j == $currentrelsourcewordindex;

	#POSTERIOR CHANGED
	#we need to create a working intersected set of this new intermediate rule!
	my $new_r_occSet = $r_occSet * $subrule->[1]->[0]; #intersection operation between the current working $r_occSet and the occ set of the sub rule!
	if(!$new_r_occSet->is_empty()){
	    createcomplexrules($elemrule, $new_r_occSet, \@new_r_source, \@new_r_target, \@new_r_type, $indexedsubrules, $new_countunreplacedsourcewords,$new_countunreplacedtargetwords, $sub_maxx+1-$r_minx, $new_sourcesymbolcount_lowerbound, $recursiondepth+1, $new_abstract_rule_still_possible, $sourcewords_aref, $targetwords_aref, $i_j_w_matrix_aref, $nbest_alignmentProbs_aref,$nbest_parseProbs_aref);
	}else{
	    die "There is no reason to be here if NBestAltRange==0" if ($NBestAltRange eq "0");
	}
      }
      # about to go into the next j-iteration
      # in case that MAXSUBSTITUTIONCOUNT is exceeded (meaning we're in abstract-rule mode), we can skip the remaining
      # j's because that would mean leaving a source symbol unreplaced, i.e., no more abstract rules can come into being
      last LABEL_j_loop319 unless $recursiondepth < $MAXSUBSTITUTIONCOUNT;

      # about to go into next j-iteration: that means we are skipping
      # the current source word j
      unless (	  $$sourcewords_aref[$r_minx+$j] eq $encoded_ignore_word) {
	$skipped_source_words .= 
	  (length($skipped_source_words) == 0 ? '' : ' ') . 
	    $$sourcewords_aref[$r_minx+$j];

	# if we're in dev/test restriction mode, check that the just
	# elongated source phrase to skip over is still contained in
	# the dev/test set
	if ($SourceRestrictionCorpus && !exists($vocCheck{$skipped_source_words})) {
	  last LABEL_j_loop319;
	}	
      }

    } # for j
  }				# if $recursiondepth
}

sub printRType{
  my ($r_type) = @_;
  foreach my $posFilled_href (@$r_type){
    print "[";
    while (my ($cat, $occSet) = each %$posFilled_href){
      print "($cat => ", $occSet->to_Enum(), ") ";
    }
    print "]";
  }
  print "\n";
}
sub probForOccSet{
  my ($occSet, $nbest_alignmentProbs_aref, $dosmooth) = @_;
  
  my @supportedProbs = ();
  #iterate in sections

  my ($start_sz, $end_sz) = $occSet->Size();
  my $size = $end_sz-$start_sz+1;
  #print "SET: ", $occSet->to_Enum(), " $start_sz -> $end_sz","\n";    

  #This is a very important check. If there is no intersection, that means we are combining across
  #alignments that are NOT consistent with each other. So we need to assign a smoothing probablity
  #that is some fraction of the lowest probability alignment (and parse?) available
  my $smoothingValue = 0.000001;
  if ($occSet->is_empty()){
    #print "ZERO SET: ", $occSet->to_Enum(), "\n";        
    if ($dosmooth==1){ 
      return $smoothingValue;
    }else{
      return 0;
    }
  }
  my $start = $occSet->Min();  
  my $min; 
  my $max;

  while (($min,$max) = $occSet->Interval_Scan_inc($start)){
    #print "ARR: ",join(' ',@$nbest_alignmentProbs_aref[int($min)..int($max)]), "\n";    
    push (@supportedProbs, @$nbest_alignmentProbs_aref[int($min)..int($max)]);
    $start = $max + 1;
    if ($start>$end_sz){
      last;
    }
  }
  my $sum = 0;
  map{ $sum += $_; } @supportedProbs;
  return $sum;
}

sub calcRuleLexicalWeights{

  my $occSet = shift; #the list of alignments that SUPPORT this rule!

  my $srcLexicalIndices_aref = shift;
  my $tgtLexicalIndices_aref = shift;
  my $i_j_w_matrix_aref = shift;

  my $nbest_alignmentProbs_aref = shift; #probabilities for the alignments for this sentence

  #immediately add a counter for this rule's occurence count

  #POSTERIOR CHANGE REQUIRED!!!
  #print "SET-CALC: ", $occSet->to_Enum(), "\n";
  my $posteriorFreqCount = probForOccSet($occSet, $nbest_alignmentProbs_aref, 0);
  
  my @ruleProbs = ($posteriorFreqCount);
  my @targetConditionedProbs = calcLexicalWeightTargetCond($srcLexicalIndices_aref, $tgtLexicalIndices_aref, $i_j_w_matrix_aref);	
  my @sourceConditionedProbs = calcLexicalWeightSourceCond($srcLexicalIndices_aref, $tgtLexicalIndices_aref, $i_j_w_matrix_aref);	


  #my $HandleMismatchesWithFeature = 1; #not supported

  #if this is a pure abstract rule you are done, otherwise handle some edge cases...
  if ((scalar(@$srcLexicalIndices_aref)==0) &&  (scalar(@$tgtLexicalIndices_aref)==0)) {    
    #done
  }else{
    #You have at least one side with real words

    #Case: 0 src words, X target words -> p(f|e)==1..1 which is not quite fair, so deal with it by computing what you can compute for sure first ie p(e|f)
    if (scalar(@$srcLexicalIndices_aref)==0){
      
      if ($HandleMismatchesByDuplicating){
	@targetConditionedProbs = @sourceConditionedProbs;
      }
      if ($HandleMismatchesByNULLCalc){
	my @tmpLexical = (-1);
	@targetConditionedProbs = calcLexicalWeightTargetCond(\@tmpLexical, $tgtLexicalIndices_aref, $i_j_w_matrix_aref);
      }

    }
    
    #Case: X src words, 0 target words -> p(e|f)==1..1 which is not quite fair, so deal with it by computing what you can compute for sure first ie p(f|e)
    if (scalar(@$tgtLexicalIndices_aref)==0){
      if ($HandleMismatchesByDuplicating){
	@sourceConditionedProbs = @targetConditionedProbs;
      }
      if ($HandleMismatchesByNULLCalc){
	my @tmpLexical = (-1);
	@sourceConditionedProbs = calcLexicalWeightSourceCond($srcLexicalIndices_aref, \@tmpLexical, $i_j_w_matrix_aref);
      }

    }
  }
 
  
  push(@ruleProbs, @targetConditionedProbs);
  push(@ruleProbs, @sourceConditionedProbs);

  return @ruleProbs;
}

sub calcLexicalWeightSourceCond{
  my ($srcLexicalIndices_aref, $tgtLexicalIndices_aref, $i_j_w_matrix_aref) = @_;
  my @sourceConditionedProbs = ();

  #now handle the remaining lexical weights that calculate p(e|f)
  for (my $k = scalar(@LexicalWeightFileReversed); $k < scalar(@LexicalWeightFile) + scalar(@LexicalWeightFileReversed); $k++) {
    my $prob = 1.0;
    foreach my $tgtIndex (@$tgtLexicalIndices_aref){
      #we need to consider the NULL word now too, so push on a -1 index. since we lookup everything as $srcIndex+1
      my @sourceLexicalRange = @$srcLexicalIndices_aref;
      push(@sourceLexicalRange, -1);
      my $alignment_partner_count = 0; #track the number of alignments so that we can average over them
      my $probSum = 0.0;
      my $bestprob = 0.0;
      foreach my $srcIndex (@sourceLexicalRange){
	
	#if there is a prob (and there will ONLY be a prob IF the alignment said it was real as well)
	#then add to the probability mass of $probSum
	#print "SourceCond Looking up: srcIndex: $srcIndex tgtIndex: $tgtIndex\n";
	if (defined($i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1])) {

	  if ($i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1]->[$k] > 0) {
	    #print "Found: ", $i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1]->[$k], "\n";
	    $alignment_partner_count++;
	    $probSum += $i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1]->[$k];
	    $bestprob = max($bestprob, $i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1]->[$k]);
	  }else{
	    $logger->error("We do not support prob=0 elements in the lexical map, they should all be $lowest_lexical_weight_encountered/10 ");
	  }
	}

    }#done going thru all lexical source indices + NULL
	if ($alignment_partner_count > 0) {
	    #$cost += -log($probSum/$alignment_partner_count);
	  $prob *= $probSum/$alignment_partner_count;
      }else{
	  $logger->warn("No alignments for tgt at index: $tgtIndex - NOT EVEN to NULL!");
	  #we assign a smoother value here
	  $prob *= $lowest_lexical_weight_encountered/10;
	  $g_hasFailed = 1;
      }
    }#handled $srcIndex
    push(@sourceConditionedProbs, $prob);
  }#handled all weights in the forward direction
  return @sourceConditionedProbs;
}

sub calcLexicalWeightTargetCond{
  my ($srcLexicalIndices_aref, $tgtLexicalIndices_aref, $i_j_w_matrix_aref) = @_;
  my @targetConditionedProbs = ();
  for (my $k = 0; $k < scalar(@LexicalWeightFile); $k++) {
      my $prob = 1.0;
      foreach my $srcIndex (@$srcLexicalIndices_aref){
	  #we need to consider the NULL word now too, so push on a -1 index. since we lookup everything as $srcIndex+1
	  my @targetLexicalRange = @$tgtLexicalIndices_aref;
	  push(@targetLexicalRange, -1);
	  my $alignment_partner_count = 0; #track the number of alignments so that we can average over them
	  my $probSum = 0.0;
	  my $bestprob = 0.0;
	  foreach my $tgtIndex (@targetLexicalRange){
	      
	      #print "TargetCond Looking up: srcIndex: $srcIndex tgtIndex: $tgtIndex\n";
	      #if there is a prob (and there will ONLY be a prob IF the alignment said it was real as well)
	      #then add to the probability mass of $probSum.
	      
	      #Dont return lexical weights for NULL/NULL. The only time you will be asking for NULL | x using 
	      #is when there are only words that you condition on, ie |x|>0. 
	      if (!($srcIndex==-1 && $tgtIndex==-1)){
		  #if there in an alignment point
		  if (defined($i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1])){
		      
		      #if it WAS defined, then you should be able to trust the probability stored there, even if it is 0!
		      #but for now lets log this as potential error
		      if ($i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1]->[$k] > 0) {
			  $alignment_partner_count++;
			  $probSum += $i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1]->[$k];
			  $bestprob = max($bestprob, $i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1]->[$k]);
		      }else{
			  #there WAS an alignment point but its probability was 0 - this is a considered an error
			  $logger->error("We do not support prob=0 elements in the lexical map, they should all be $lowest_lexical_weight_encountered/10 ");
			  #use a smoother value
			  $probSum += 0.00001;
		      }
		  }
	      }
	      
	  }#done going thru all lexical target indices + NULL
	      if ($alignment_partner_count > 0) {
		  #$cost += -log($probSum/$alignment_partner_count);
		  $prob *= $probSum/$alignment_partner_count;
	      }else{
		  # this is the case where we see NO alignment points across these src/tgt sequences
		  # this could happens when we only see a source word... AND its NOT aligned to NULL or
		  # any of the words in the target. Should warn!
		  $logger->warn("No alignments for src at index: $srcIndex - NOT EVEN to NULL!");		  
		  $prob *= $lowest_lexical_weight_encountered/10;
		  $g_hasFailed = 1;
	      }
      }#handled $srcIndex
	  push(@targetConditionedProbs, $prob);
  }#handled all weights in the forward direction
      return @targetConditionedProbs;
}

sub identifyCompactPhrase{
  my ($minx,$maxx, $r_source, $sourcewords_aref) = @_;
  my $sourcestr = "";
  my $r_source_symbolcount_sanity_check = 0;

  #track the indices (relative to the full sentence) that are actual lexical entries
  my @srcLexicalIndices = ();

  for (my $i = $minx; $i <= $maxx; $i++) {
    my $w = $r_source->[$i-$minx];
    if ($w == -1) {
      # do nothing
    } elsif ($w == 0) {
      ++$r_source_symbolcount_sanity_check;

      unless ($sourcewords_aref->[$i] eq $encoded_ignore_word) {
	# write out source sentence word $i
	$sourcestr .= $sourcestr eq ""? "": " ";
	$sourcestr .= $sourcewords_aref->[$i];
	#this was a lexical index so lets track it
	push(@srcLexicalIndices, $i);      }
    } else {
      # write out "@$w"
      $sourcestr .= $sourcestr eq ""? "": " ";
      $sourcestr .= "\@$w";
      ++$r_source_symbolcount_sanity_check;
    }
  }
  return (\@srcLexicalIndices, $sourcestr, $r_source_symbolcount_sanity_check);
}

#increment counts with a posterior value. This function increments the counts that are only based on the body
sub incrementNonLabelledPhraseCounts{
  my ($refsourcestr, $reftargetstr, $posteriorFreqCount) = @_;

  #if (defined $phrasecounts{hashcode($$refsourcestr . '#')}){
  
  $phrasecounts{hashcode($$refsourcestr . '#')} += $posteriorFreqCount;
  
  #}else{
  #  $phrasecounts{hashcode($$refsourcestr . '#')} = $posteriorFreqCount;
  #}

  # increase target count for target-conditioned RF scores
  #++$phrasecounts{hashcode('#' . $$reftargetstr)} if $CalculateTargetConditionedRFs;
  #if ($CalculateTargetConditionedRFs && defined $phrasecounts{hashcode('#' . $$reftargetstr)}){

  if ($CalculateTargetConditionedRFs){
    $phrasecounts{hashcode('#' . $$reftargetstr)} += $posteriorFreqCount;
  }

  #}else{
  #   $phrasecounts{hashcode('#' . $$reftargetstr)} = $posteriorFreqCount;
  #}
  
  # increase count for source/target pair, later to be utilized by source- and target-conditioned RF scores
  #++$phrasecounts{hashcode($$refsourcestr . '#' . $$reftargetstr)};	

  
  #if (defined($phrasecounts{hashcode($$refsourcestr . '#' . $$reftargetstr)})){
  
  #moved this one over to labelled and now it include the target cat!
  #$phrasecounts{hashcode($$refsourcestr . '#' . $$reftargetstr)} += $posteriorFreqCount;
  
  #}else{
  #  $phrasecounts{hashcode($$refsourcestr . '#' . $$reftargetstr)} = $posteriorFreqCount;
  #}

}

#Checks to see if a specific span comes from a derivation that we have already seen before
#if it does, this is a duplicate, here are specific examples that we want to idenfity, and throw out
# PROBLEM:
# InitialRule1 <a b c ------->  => <a b c X> 
# InitialRule2 <a b c --->      => <a b c X> 
# this would end up being counted twice. We cant just outlaw the production of the same rule multiple
# times from one sentence, since that would underestimate very common productions like DT -> the
# that obviously happen many times within one sentence. 
# SOLUTION:
# Track the intial rule that a rule is generated from, record this span. If another rule comes in
# that is generated from a span that OVERLAPS with this span, then this is probably a duplicate rule
# and we expand the span to cover both of these regions. The duplicate matches on words generated
# and labels.
sub isruleNovel{
  my ($lookup, $src, $tgt, $labels, $parseOccSet, $altOccSet,$spanOccSet, $tgtMaxIndex, $g_uniquePhrasesFromSent, $finalProb) = @_;
  
  #my $lookupString = decodephrase($src)." # ".decodephrase($tgt)." # ".$labels;
  #print "LOOKUP : $lookupString\n";
  if (defined($g_uniquePhrasesFromSent{$lookup})){

      if ($SimpleNovelty){
	  return 0;
      }
      
    #print "EXISTING\n";
      #my $newphraseset = new Set::IntRange(0, $tgtMaxIndex);
      #$newphraseset->Interval_Fill($tgtSpanStart, $tgtSpanEnd);

    my $setList_aref = $g_uniquePhrasesFromSent{$lookup};
    my $novel = 1;

    my $tmp_spanSet = $setList_aref->[0] * $spanOccSet;
    my $tmp_parseSet = $setList_aref->[1] * $parseOccSet;
    my $tmp_altSet = $setList_aref->[2] * $altOccSet;

    #if all three sets have non-zero intersections, then this rule should be rejected
    #otherwise this rule is novel in some regard... (having an empty intersection implies novelty)
    if ( !($tmp_spanSet->is_empty()) && !($tmp_parseSet->is_empty()) && !($tmp_altSet->is_empty())){
      #print "REJECTING:  ",$lookupString," # ",$newphraseset->to_Enum(), " # ",$altOccSet->to_Enum()," # ",$parseOccSet->to_Enum(), "\n";
      #print "REJECTED AGAINST: ", $setList_aref->[0]->to_Enum(), " # ", $setList_aref->[1]->to_Enum(), " # ", $setList_aref->[2]->to_Enum(), "\n";
      return 0;
    }else{
      #do the unioning
      $setList_aref->[0] += $spanOccSet;
      $setList_aref->[1] += $parseOccSet;
      $setList_aref->[2] += $altOccSet;
    }  
    #print "RETURNING $novel\n";
    return 1;
  }else{
    #register the range that this phrase was extracted from
      #my $set = new Set::IntRange(0, $tgtMaxIndex);
      #$set->Interval_Fill($tgtSpanStart, $tgtSpanEnd);
      if ($SimpleNovelty){
	  #store the post prob of this rule choice
	  $g_uniquePhrasesFromSent{$lookup} = $finalProb;
	  return 1;
      }else{
	  #we have to maintain three sets, spans, alignments and parse occs
	  my @setList = ($finalProb, $spanOccSet, $parseOccSet, $altOccSet);
	  $g_uniquePhrasesFromSent{$lookup} = \@setList;
	  #print "ADDING NEW : $lookupString [$tgtSpanStart, $tgtSpanEnd] \n";
	  return 1; #means its novel
      }
  }
}

#fill the second vector (existing cost vector with appropriately merged values)
#sum the posterior in the first field,MAX the rest of the values
sub mergeCostVectors{
  my ($costvector_aref, $final_aref) = @_;

  my $sanityFreq = $final_aref->[1];
  $final_aref->[0] += $costvector_aref->[0];
  $final_aref->[1] += $costvector_aref->[1];
  
  for (my $i = 2; $i < scalar(@$costvector_aref); $i++) {
    
    #sum elementwise	
    #POSTERIOR CHANGE REQUIRED??
    if ($AverageLexicalWeights){
      #accumulate all the costs (scaled by the posterior freqeuency)
     die "ERROR: We expect the simple freq of the newly found (but duplicate) rule to be 1: but it was: ",$sanityFreq if ($sanityFreq!= 1);
     $final_aref->[$i] += ($costvector_aref->[$i]);
    }
    #max prob
    if ($MaxLexicalWeights){
      #if then new prob is better, then replace it
      if ($costvector_aref->[$i] > $final_aref->[$i]) {
	$final_aref->[$i] = $costvector_aref->[$i];
      }	
    }
  }
}

sub incrementLabelledCounts{
  my ($refsourcestr, $reftargetstr, $reftypeswritten, $value) = @_;
  
  my $resultcat = $reftypeswritten->[-1]; #the last element of the reftypeswritten    
  $resulttypecounts_diluted{$resultcat} += $value;  
  my $argsLabels = join(' ', @$reftypeswritten[0..scalar(@$reftypeswritten)-2]);

  #when the same arg labels are used... its safe because the second hash is added
  $phrasecounts{hashcode($$refsourcestr . '#' . $argsLabels . '#')} += $value;
  
  #the rule body is now tagged with the result cat
  $phrasecounts{hashcode($$refsourcestr . '#' . $$reftargetstr. '#'.$resultcat)} += $value;

}

#the input to add rules is actually still just span subtraction decisions, but each of these
#spans is annotated with a *set* of possible labels. We have to explore all possible combinations
#of these labels recursively! reftypeswritten is a list of the current choice of specific non
#terminals chosen during this search. parseOccSet is the contintually intersected set of occurrences
#for each of these non-terminals. When the rule is finally created, we can inspect these counts
#to assign a probability for the labelling sequence
#the r_ruleWorld determines whether a rule is being generate in the _X world or the syntax world. r_ruleWorld comes in undef
#the minute a syntax category is chosen, r_ruleWorld = 1, and subsequent category choices must remain in the same r_world 
sub addrules {
  my ($nbest_parseProbs_aref, $refsourcestr,$reftargetstr,$r_costvector,$r_type,$typestogocount,$reftypeswritten, $altOccSet, $spanOccSet, $tgtMaxIndex, $r_ruleWorld) = @_;

  #POSTERIOR CHANGE MADE - see the ++'s below
  my $posteriorFreqCount = $r_costvector->[0];
  #print "POSTERIOR $posteriorFreqCount reftypeswritten: ", scalar(@$reftypeswritten)," typestogocount: ",  scalar($typestogocount), "\n";

  #if (scalar(@$reftypeswritten)==0){
  #    incrementNonLabelledPhraseCounts($refsourcestr, $reftargetstr, $posteriorFreqCount);
  #}

  #check if we are at the recursion base level
  if ($typestogocount==0) {


    #if you are a novel phrase in the labelled world

    #key for complexrules
    my $key =  $$refsourcestr."#".$$reftargetstr."#".join(' ',@$reftypeswritten);    
    #used to determine if this is a doubly extracted rule...

    #print "CHECKING: ", $noveltyKey, "\n";
    my @costvector = @$r_costvector;
    #First calculate the final posterior value for this rule...
    my $finalOccSet = new Set::IntRange(0, scalar(@$nbest_parseProbs_aref)-1);
    $finalOccSet->Fill();
    for (my $i=0; $i<scalar(@$reftypeswritten)-1; $i++){
	my $type = $reftypeswritten->[$i];
	my $categoryOcc_href = $r_type->[$i];
        $finalOccSet *= $categoryOcc_href->{$type};
	if ($finalOccSet->is_empty()){
	    return;
	}
    }


    #my $noveltyKey = $$refsourcestr." # ".$$reftargetstr." # ".join(' ', @$reftypeswritten)." # ".$altOccSet->to_Enum()." # ".$parseOccSet->to_Enum();
    #update the cost vector here with the parse posterior
    my $resultcat = $reftypeswritten->[-1]; #the last element of the reftypeswritten
    my $parsePost = probForOccSet($finalOccSet, $nbest_parseProbs_aref, 0); #no smoothing shown by 0
    if ($parsePost==0){
      print "ERROR: parsePost==$parsePost : This means that this rule is taken across parses, which should not be allowed\n";
      exit;
    }
    #the @costvector at this point has only the posterior count in it
    #we care about the following counts: parse*alt posterior, simple +1 count (we ignore the posteriors individually)

    my $finalProb = $posteriorFreqCount*$parsePost;
    unshift (@costvector, $finalProb); #define the final posterior count (do this here)
    $costvector[1] = 1; #set the second count to 1 (matching the undiluted count)

    die "ERROR: Count issue early: simple: ",$costvector[1]+0.1," < ",$costvector[0], " at: ", decodephrase($$refsourcestr), " # ", decodephrase($$reftargetstr), "\n" if ($costvector[1]+0.1<$costvector[0]);

    #PRUNING! the rules has to have freq > $g_minRuleFreq to be considered!
    #print "CHECK: ", $costvector[0], " >= $g_minRuleFreq\n";
    if ($costvector[0] >= $g_minRuleFreq){
    if ($NBestNoNovelty || isruleNovel($key, $$refsourcestr, $$reftargetstr,  join(' ',@$reftypeswritten), $finalOccSet, $altOccSet,$spanOccSet, $tgtMaxIndex, \%g_uniquePhrasesFromSent, $finalProb)){

      incrementNonLabelledPhraseCounts($refsourcestr, $reftargetstr, $costvector[0]);
      incrementLabelledCounts($refsourcestr, $reftargetstr, $reftypeswritten, $costvector[0]);

      #this is the simple +1 count - the diluted one is the one that is the correct posterior
      $resulttypecounts_undiluted{$resultcat} += 1;


      $ruleinstancecount++; 	# update rule counter (this is just a statistics for info purposes)
      #this is truly a new rule : but there might be another rule extracted from a different place in the same sentence
      #sum their final prob*alignment posteriors over these duplicates. This is a standard frequency summing operation.
      
      #there could still be existing rules here, from other sentence for example, that we have to increment the freq for
      my $existingrule = $complexrules{$key};
      #do a merge if there is an existing rule
      if (defined($existingrule)){
	  my @existingCostsVector = split(' ', $existingrule);
	  #print STDERR "DUP:  ",decodephrase($$refsourcestr), " \# ",decodephrase($$reftargetstr), " \# ", join(' ',@$reftypeswritten), " => ", join(' ', @costvector),"  EXISTING " , join(' ', @existingCostsVector), "\n";
	  mergeCostVectors(\@existingCostsVector, \@costvector);
	  #print STDERR "DONE MERGE\n";
      }
      
      # before writing out the cost vector, round everything to 3 digits to save memory
      # but keep frequency count  field (field 0) accurate
      #my @finalCostVector = map { sprintf("%.3g", $_); } @costvector;
      #$finalCostVector[0] = $costvector[0]; #replace with the high precision one
      die "ERROR: Count issue: simple: ",$costvector[1]," < ",$costvector[0], " at: ", decodephrase($$refsourcestr), " # ", decodephrase($$reftargetstr), "\n" if ($costvector[1]+0.1<$costvector[0]);
      $complexrules{$key} = join(" ", @costvector);
      
      #$logger->debug("Adding rule complex!: ",decodephrase($$refsourcestr), "\# ",decodephrase($$reftargetstr)," # final: ", $finalCostVector[0], " # parse: ",sprintf("%.3g", $parsePost)," alt: ",sprintf("%.3g",$posteriorFreqCount)," =>", join(' ', @$reftypeswritten[0..scalar(@$reftypeswritten)-2]), " => $resultcat \n");
      
     
      
    }else{
      #$logger->debug("Dropping duplicate phrase in labelled phrase count: ", decodephrase($$refsourcestr." ".$$reftargetstr), "\n");
      #print "Dropping duplicate phrase in labelled phrase count: ", decodephrase($$refsourcestr." ".$$reftargetstr), "\n";
      #THIS DUPLICATE MIGHT HAVE A BETTER FREQ COUNT!!!
      #print "CHECK IF BETTER\n";
      my $existingProb = $g_uniquePhrasesFromSent{$key};
      if (!defined($existingProb)){die "Assertion error: If a rule was a duplicate, there must be a rule in the g_uniquePhrasesFromSent\n";}

      #frequency field...
      if ($costvector[0]>$existingProb){
	  #print STDERR "BETTER THO:  ",decodephrase($$refsourcestr), " # ",decodephrase($$reftargetstr), " # ", join(' ',@$reftypeswritten), " => ", join(' ', @costvector)," > " , $existingCostsVector[0], "\n";
	  my $existingaccrule = $complexrules{$key};
	  if (!defined($existingaccrule)){die "Assertion error: If a rule was ruled a duplicate, you must be able to find the duplicate!!!\n";}

	  my @existingAcc = split(' ', $existingaccrule); #THIS IS THE ACCUMULATED VECTOR
	  
	  $existingAcc[0] += $costvector[0]-$existingProb;
	  $complexrules{$key} = join (' ',@existingAcc);	 
	  
	  incrementNonLabelledPhraseCounts($refsourcestr, $reftargetstr, $costvector[0]-$existingProb);	
	  incrementLabelledCounts($refsourcestr, $reftargetstr, $reftypeswritten, $costvector[0]-$existingProb);

	  #keep this highest prob value in there!
	  $g_uniquePhrasesFromSent{$key} = $costvector[0];
       }
  }
    
  }else{#g_minRuleFreq check
	    #$logger->debug("Dropping: ",decodephrase($$refsourcestr), "\# ",decodephrase($$reftargetstr)," # final: ", $costvector[0], " # parse: ",sprintf("%.3g", $parsePost)," alt: ",sprintf("%.3g",$posteriorFreqCount)," =>", join(' ', @$reftypeswritten[0..scalar(@$reftypeswritten)-2]), " => $resultcat \n");
	    #print STDERR "Dropping: ",decodephrase($$refsourcestr), "\# ",decodephrase($$reftargetstr)," # final: ", $costvector[0], " # parse: ",sprintf("%.3g", $parsePost)," alt: ",sprintf("%.3g",$posteriorFreqCount)," =>", join(' ', @$reftypeswritten[0..scalar(@$reftypeswritten)-2]), " => $resultcat \n";
	}
  } else {
    #more unrolling to do...

    my $categoryOcc_href =  $r_type->[--$typestogocount]; #the categories alternatives for the NT at positions --$typestogocount
    unshift @$reftypeswritten, ''; # prepend dummy to the front of the typeswritten array, allows the [0] value to be reused in the loop
    my @keys = keys %$categoryOcc_href;
    #pre-condition: the _X category type is available for ALL spans
    foreach my $type (@keys){
      my $occSet = $categoryOcc_href->{$type};
      #this allowed lexical rules of category X to come thru. but now, we add an _X to each category type if its not there.
      #if ($AbstractOverX || $type ne '_X' || ((scalar(@$reftypeswritten) == 1) && ($typestogocount == 0)) ) {


      my $new_ruleWorld = undef;
      #if $r_ruleWorld has not been defined yet, lets define it based on the type
      if (!defined($r_ruleWorld)){
	# this is where you are coming into the rule for the very first time! you can check the result cat  
	if ($type eq '_X'){ 
	    #if you can only use X if you have to, then ensure that there is only one category option
	    #if (!$UseXIff || (scalar(@keys)==1)){
		#$new_ruleWorld = 0;
	    #}else{
		#$logger->debug("DONT PURSUE (UseXIff): ", decodephrase($$refsourcestr), "\# ",decodephrase($$reftargetstr), "\#", $type, "\n");
		#next;
	    #}
	    #die "Assertion error: UseXIff was set, but this span has multiple categories: ", join(' ', @keys), "\n" if ($UseXIff && scalar(@keys)!=1);
	    $new_ruleWorld = 0; #could be pure lexical so it might still get thru
	}else{
	  $new_ruleWorld = 1;
	}
      }else{
	  #NOTE: once you are in here, you are NOT a pure lexical rule. Pure lexical rules start with $r_ruleWorld=undef
	  #      and since there are no more cats, it will have already been output
	  #$r_world is already defined (by the result), if we are in-compliance we its value, then we proceed, else, reject this iteration
	  if (($r_ruleWorld==0 && $type ne '_X')  || ($r_ruleWorld==1 && $type eq '_X') || ($type eq '_X' && !$AbstractOverX)){
	      $logger->debug("DONT PURSUE: type $type in r_ruleWorld $r_ruleWorld based on: ", join(' ', @$reftypeswritten), "\n");
	      next;
	  }
	$new_ruleWorld = $r_ruleWorld;
      }

      # replace dummy with $type
      $reftypeswritten->[0] = $type;
      
      #print "IN MERGE: ", $parseOccSet->to_Enum(), " with $type =>", $occSet->to_Enum(), "\n";
      #my $new_parseOccSet = $parseOccSet * $occSet;
      # call addrules recursively

      #if (!$new_parseOccSet->is_empty()){
      #my $parsePost = probForOccSet($new_parseOccSet, $nbest_parseProbs_aref, 0); #no smoothing shown by 0
	#  if ($posteriorFreqCount*$parsePost >= $g_minRuleFreq){
      #addrules($nbest_parseProbs_aref, $refsourcestr,$reftargetstr,$r_costvector,$r_type,$typestogocount, $reftypeswritten, $new_parseOccSet, $altOccSet, $spanOccSet, $tgtMaxIndex, $new_ruleWorld);
      addrules($nbest_parseProbs_aref, $refsourcestr,$reftargetstr,$r_costvector,$r_type,$typestogocount, $reftypeswritten, $altOccSet, $spanOccSet, $tgtMaxIndex, $new_ruleWorld);
	 # }else{
	  #    $logger->debug("Dropping early: ",decodephrase($$refsourcestr), "\# ",decodephrase($$reftargetstr), " # parse: ",sprintf("%.3g", $parsePost)," alt: ",sprintf("%.3g",$posteriorFreqCount)," =>", join(' ', @$reftypeswritten[0..scalar(@$reftypeswritten)-1]),"\n");
	  #}
      #}else{
	#$logger->debug("No evidence for!: ",decodephrase($$refsourcestr), "\# ",decodephrase($$reftargetstr)," => ",join(' ', @$reftypeswritten[0..scalar(@$reftypeswritten)-1]), " => ", $new_parseOccSet->to_Enum(), "\n");
      #}
      #print "RETURNED FROM REC $typestogocount finishing type: $type\n";
    #}
    }

    shift @$reftypeswritten; 
  }
  #print "RETURNED typestogocount $typestogocount\n";
}

# normalizes the string by removing leading and trailing spaces and
# multiple-spaces and the newline symbol
sub normalizeinputstring {
  die "Target/source/parsetree mismatch (nonequal number of sentences) at sentence number $sentencenumber. Byebye." if !defined($_[0]);
  chomp($_[0]);
  $_[0] =~ s/^\s+|\s+$//g;		# delete leading and trailing spaces
  $_[0] =~ s/\s+/ /g;
}


#returns a hash with TARGET indexed spans in the form key: [$spanstart $spanlength]
sub findparsespans {
  #my ($spanstart) = @_;
  use strict;
  my $spanstart = shift;
  my $currentparsespans_href = shift; #named to indicate that this is a hash ref
  my $currentparsetokens_aref = shift;#named to indicate that this is a arr ref
  my $currentparsetokenspos_ref = shift; #used to track in the recursion, simple variable reference

  #currentparsetokenspos is globally updated
  if ($$currentparsetokenspos_ref < scalar(@$currentparsetokens_aref)) {        	# only if we are not beyond the end of the array, we can succeed
    my $token = $currentparsetokens_aref->[$$currentparsetokenspos_ref++];
    if ($token eq "(") {
      #new nonterminal root
      my $root = $currentparsetokens_aref->[$$currentparsetokenspos_ref++];
      #print "root $root\n";
      my $spanpos = $spanstart;
      my @backpointerinfo = (); # list of spanstart/spanlength info pairs of the child nodes
      while ($currentparsetokens_aref->[$$currentparsetokenspos_ref] ne ")") { # iterate over all child nodes
	#my $childspanlength = findparsespans($spanpos);
	my $childspanlength = findparsespans($spanpos, $currentparsespans_href, $currentparsetokens_aref, $currentparsetokenspos_ref);
	if ($childspanlength == -1) {
	  return -1;
	}
	push @backpointerinfo, ($spanpos, $childspanlength); # store the child's span
	$spanpos += $childspanlength;
      }
      # arrived at the ")" telling us that this span is completed
      $$currentparsetokenspos_ref++;
      my $spanlength = $spanpos-$spanstart;
      if ($spanlength == 0) {
	# parsing error - do nothing, below we will fail
      } else {
	my $erroroccurred = 0;
	if (($root ne "S1") && ($root ne 'ROOT')) {	# exclude the case of artifical root symbols such as S1 (Charniak Parser) or ROOT (Stanford Parser)
	  my $hashindex = "$spanstart $spanlength";
	  #print "$spanstart $spanlength\n";
	  
	  my $hashvalue = $currentparsespans_href->{$hashindex};
	  if (defined $hashvalue) {
	    # this span has already been indexed - this must be a unary rule node (one child only)
	    $erroroccurred = 1 if (scalar(@backpointerinfo) != 2); # assert that we only have one child
	    # add nonterminal $root to the list indexed by our span
	    #	    $hashvalue .= " $root";
	    if ($UnaryParseCategoryUtilization eq 'all') {
	      $hashvalue->[0] .= " $root"; # add our category to category field
	    } elsif ($UnaryParseCategoryUtilization eq 'top') {
	      $hashvalue->[0] = $root; # replace category with this category (b/c it's higher in the hierarchy)
	    } elsif ($UnaryParseCategoryUtilization eq 'bottom') {
	      # keep existing category b/c it's lower in the hierarchy)
	    } else {
	      die "Wrong argument specified for option --UnaryParseCategoryUtilization";
	    }
	  } else {
	    if (scalar(@backpointerinfo) == 2) {
	      # unary rule (only one child node) - don't want backpointer info
	      # this case occurs if and only if this node is a POS tag since it's child node is not in the $currentparsespans hash
	      @backpointerinfo = (); # remove backpointer to the terminal symbol's span, which is identical to this span
	    }
	    unshift @backpointerinfo, $root; # insert category as first element
	    $hashvalue = \@backpointerinfo; # store reference to array whose first element is the category and whose other elements are children's span info pairs
	  }
	  $currentparsespans_href->{$hashindex} = $hashvalue;
	}
	return $spanlength if !$erroroccurred;
      }
    } else {			#if ($token eq " ") {
      # we've arrived at the word level
      #print "word ".$token."\n";
      if ($currentparsetokens_aref->[$$currentparsetokenspos_ref] eq ")") {
	return 1;		# length of span is one (the word we just encountered)
      } else {
	# parsing error - do nothing, below we will fail
      }
    }
  }
  print STDERR "Parsing error in parsing log file, sentence number $sentencenumber around token number " . ($currentparsetokenspos_ref+1) . ": ".join(" ",@$currentparsetokens_aref)."\n";
  return -1;
}



# look up array of words referenced by $phrasewords as subarray of array referenced 
# by $sentencewords and return the matching spans in form of a list each of whose
# entries is of the form "$spanstart $spanlength" ($spanstart starts with 0, $spanlength==1 means spanning one word)
sub lookupspans {
  my ($phrasewords,$sentencewords) = @_;
  my @matchingspans = ();
  my $spanlength = scalar @$phrasewords;
  # efficiency is not an issue in this sub, so do stupid search
 LABEL438:
  for (my $i=0;$i<=scalar(@$sentencewords)-scalar(@$phrasewords);$i++) {
    for (my $j=0;$j<scalar(@$phrasewords);$j++) {
      next LABEL438 if $phrasewords->[$j] ne $sentencewords->[$i+$j];
    }
    # haven gotten so far means we found a match
    push @matchingspans,"$i $spanlength";
  }
  return @matchingspans;
}

# look up $phrasewords in $sentencewords
# return the list of word positions in $sentencewords where matches occurred (starting with 0)
sub fastlookupspans {
  #  my ($phrasewords,$sentencewords) = @_;
  my $phrasewords = ' ' . $_[0] . ' ';
  my $sentencewords = ' ' . $_[1] . ' ';
  my @matchingspans = ();

  my $strpos = index $sentencewords, $phrasewords;
  while ($strpos != -1) {
    # found something at position $strpos

    #split everything before the found occurrence into words (offset 1 to account for the leading ' ')
    my @tmp = split(' ',substr($sentencewords, 1, $strpos));
    # record the match's span
    push @matchingspans, scalar(@tmp);
    #    print "Found match of $phrasewords in $sentencewords at position " . scalar(@tmp) . "\n"; 
    # find next match
    $strpos = index $sentencewords, $phrasewords, $strpos+1;
  }

  return @matchingspans;
}


sub printhash {
  my ($hashref) = @_;

  my ($key, $value);
  while (($key,$value) = each %$hashref) { 
    #  my $neglogfrequency = - log $value->[0];
    print STDERR "$key --> $value\n";
  }
}



sub num2code {
  # use ANSI characters from 65 ('A') up to 126 (~), i.e., 62-character alphabet
  # don't use 64 because that's the @ arg marker
  # output a number in the number system with base 62
  my $x = $_[0];
  my $resultstr = '';
  while ($x != 0) {
    my $remainder = $x % 62;
    $x = ($x - $remainder) / 62;
    $resultstr .= chr($remainder+65);
  }
  if (length($resultstr) == 0) {
    # we don't want 0 to map to the empty string, so output 0 in base-62, which is equivalent
    $resultstr = chr(0+65);
  }
  return $resultstr;
}

sub code2num {
  my $x = $_[0]; # the code string
  my $resultnum = 0;
  while (length($x) != 0) {
    my $char = chop($x);
    $resultnum = $resultnum * 62 + ord($char)-65;
  }
  return $resultnum;
}

sub encodewordlist {
  my @resultlist = ();
  foreach my $x (@_) {
    if ($x =~ /^\@/) {
	print STDERR "Warning! Abstraction symbol $x occurring in data. This might mess up the rules.\n" if ($w_abscount==0);
	$w_abscount = 1;
      # ignore this token
    } else {
      my $code = $codetable{$x};
      if (!defined($code)) {
	push @vocablist, $x;
	$code = num2code($#vocablist);
	$codetable{$x} = $code;
      }
      push @resultlist, $code;
    }
  }
  return @resultlist;
}

sub decodewordlist {
  my @resultlist = ();
  foreach my $x (@_) {
    if ($x =~ /^\@/) {
      # nonterminal or abstraction marker - simply output this token itself
      push @resultlist, $x;
    }elsif ($x =~ /\#/){
      die "What is this??? should never happen. az.";
	# has sign for rules
	push @resultlist, $x;
    }else {
      push @resultlist, $vocablist[code2num($x)];
    }
  }
  return @resultlist;
}

sub encodephrase {
  my @wordlist = split(' ', $_[0]);
  return join(' ', encodewordlist(@wordlist));
}

sub decodephrase {
    my @wordlist = split(' ', $_[0]);
    return join(' ', decodewordlist(@wordlist));
}

  

sub hashcode {
  if (!$UseHashcodes) {
    return $_[0]; # just return input, don't compute hash code
  } else {
  use bytes;
  # we need to calculate two independent hash codes because with a 32bit hash code the collision 
  # probability gets too high
  # 48 bit is okay - if we have only one or two collisions then that's an insignificant distortion
  my $hc = 1;
#  my $hc2 = 1;
  {
    use integer;
    my ($inputstr) = @_;
    for (my $i=0; $i < length($inputstr); $i++) {
      my $currentcharcode = ord(substr($inputstr, $i, 1));
      $hc = 31*$hc + $currentcharcode;
#      $hc2 = 53*$hc2 + $currentcharcode;
    }
  }

  no warnings;
  # don't use integer anymore because we want to do an unsigned shift >>
  return chr($hc & 0xFF) . chr(($hc & 0xFF00) >> 8) . chr(($hc & 0xFF0000) >> 16) . chr(($hc & 0xFF000000) >> 24)
    . chr(($hc & 0xFF00000000) >> 32) . chr(($hc & 0xFF0000000000) >> 40);
#    . chr($hc2 & 0xFF) . chr($hc2 & 0xFF00 >> 8);
  }
}

sub consolidaterules {

  print STDERR "\n" . `ls -lh $mytmpdir` . "\n" . `du -h $mytmpdir` . "\n" if $BERKELEY_DB;
  print STDERR `date` . "Consolidating rule DBs, throwing out non-lexical one-time-occurrence rules to save disk space...\n";
  
  die "ERROR: Consolidation is not supported right now!!!\n";
  my $totalrulefreqcount_sanitytest = 0;
  my $rulenumber = 0;
  my $rulesdeleted = 0;
  my $sourcephrasesdeleted = 0;
  my $source_and_arg_phrases_deleted = 0;
  my $targetphrasesdeleted = 0;
  my $rulebodiesdeleted = 0;
  my ($key, $value);
  while (($key,$value) = each %complexrules) { 
    $rulenumber++;
    print STDERR "Processing rule $rulenumber...\r" if ($rulenumber-1) % 50000 == 0;

    #  my $neglogfrequency = - log $value->[0];
    #my $frequency = $value->[0];
    
    my ($sourcestr,$targetstr,$typestr) = split("#", $key);
    #print "Handling : ", decodephrase($sourcestr), " \# ", decodephrase($targetstr), " # ", "\n";
    my @costvector = split(" ", $value);
    my @resultcategories = split(" ", $typestr);
    my $frequency = $costvector[0]; # frequency of rule
    my $simplefreq = $costvector[1]; # frequency of rule

    my $sourcefrequency = $phrasecounts{hashcode($sourcestr . '#')};    # frequency of rule's source side (excl. arg NTs)
    if (!defined($sourcefrequency)) {
      #print STDERR "DB error: Rule $key\#$value: sourcefrequency not contained in phrasecounts DB!\n";
      #next;
      # this might have come from the rule consolidation - just give it a one-frequency
      $sourcefrequency=1;
    } elsif ($sourcefrequency == 1) {
      # remove this entry to save memory
      delete $phrasecounts{hashcode($sourcestr . '#')};
      $sourcephrasesdeleted++;
    }

    my $tmpkey = hashcode($sourcestr . '#' . join(' ', @resultcategories[0..$#resultcategories-1]) . '#');
    my $source_and_arg_frequency = $phrasecounts{$tmpkey};    # frequency of rule's source side (incl. arg NTs)
    if (!defined($source_and_arg_frequency)) {
      #print STDERR "DB error: Rule $key\#$value: sourcefrequency not contained in phrasecounts DB!\n";
      #next;
      # this might have come from the rule consolidation - just give it a one-frequency
      $source_and_arg_frequency=1;
    } elsif ($source_and_arg_frequency == 1) {
      # remove this entry to save memory
      delete $phrasecounts{$tmpkey};
      $source_and_arg_phrases_deleted++;
    }

    my $targetfrequency = $CalculateTargetConditionedRFs? $phrasecounts{hashcode('#' . $targetstr)} : -1;    # frequency of rule's target side
    if (!defined($targetfrequency)) {
#      print STDERR "DB error: Rule $key\#$value: targetfrequency not contained in phrasecounts DB!\n";
#      next;
      # this might have come from the rule consolidation - just give it a one-frequency
      $targetfrequency = 1;
    } elsif ($targetfrequency == 1) {
      # remove this entry to save disk space
      delete $phrasecounts{hashcode('#' . $targetstr)};
      $targetphrasesdeleted++;
    }

    my $rulebodyfrequency = $phrasecounts{hashcode("$sourcestr\#$targetstr\#".$resultcategories[-1])};   	# frequency of rule's body (different from $frequency in that multiple types for identical rule bodies were not multiply counted)
    if (!defined($rulebodyfrequency)) {
	print STDERR "DB error: Rule ",decodephrase($sourcestr),"#",decodephrase($targetstr),"#",$resultcategories[-1]," rulebodyfrequency not contained in phrasecounts DB!\n";
	die;
      #      next;
      # this might have come from the rule consolidation - just give it a one-frequency
      $rulebodyfrequency = 1;
    } elsif ($rulebodyfrequency == 1) {
      # remove this entry to save disk space
      delete $phrasecounts{hashcode("$sourcestr\#$targetstr\#".$resultcategories[-1])};
      $rulebodiesdeleted++;
    }

#    # find the result category in the rule
#    $key =~ /[ \#]([^ \#]*)$/;
#    #my $resultcategory = $1;
#    # get the count (in indiluted and in diluted interpretation) of result category $1
#    my @resultcount = ();
#    push @resultcount, $resulttypecounts_diluted{$1};
#    if (!defined($resultcount[0])) {
#      print STDERR "DB error: Rule $key\#$value: resultcount not contained in resulttypecounts DB!\n";
#      next;
#    }
#    push @resultcount, $resulttypecounts_undiluted{$1};
#    if (!defined($resultcount[1])) {
#      print STDERR "DB error: Rule $key\#$value: resultcount not contained in resulttypecounts DB!\n";
#      next;
#    }

    my @ruletargetwords = split(" ", $targetstr);
    my @rulesourcewords = split(" ", $sourcestr);

    # count source- and target-terminal symbols
    my $sourcewordcount = scalar(@rulesourcewords) + 1 - scalar(@resultcategories);
    my $targetwordcount = scalar(@ruletargetwords) + 1 - scalar(@resultcategories);

    if ( (
	  # first condition: balancedness
	  ( (($sourcewordcount == 0) == ($targetwordcount == 0))
	    # rule has no source terminals iff rule has no target terminals
	    || scalar(@resultcategories) == 1
	    # always counts as fulfilled if rule is fully lexicalized 
	    || $keepunbalancedrules
	    # always counts as fullfilled if keepunbalancedrules
	    # parameter is set
	  ) && 
	  # second condition: plausibility
	  ( $simplefreq > 1 
	    # seen phrase pair at least twice
	    || (scalar(@resultcategories) == 1
		&& scalar(@rulesourcewords) <= $MaxLengthSingleCountSourcePhrases
		# && $rulebodyfrequency/$sourcefrequency >= 0.15
	       )
	    # or: rule is fully lexicalized and source-cond. RF >= 0.15 and source side has at most
	    # MPLSCP words
	  )
	 )
       ) {
      # don't do anything
      $totalrulefreqcount_sanitytest += $frequency;
    } else {
      # delete rule to save disk space
      delete $complexrules{$key};
      $rulesdeleted++;
      # decrease rule instance count to account for lost rule
      $ruleinstancecount -= $frequency;
    }
  }
  #NOTE: $ruleinstancecount is a global
  print STDERR "\n!!!!!!!!!!!! Something's wrong with the underlying complexrules database! (Maybe an network file system issue?) \$ruleinstancecount ($ruleinstancecount) != \$totalrulefreqcount_sanitytest ($totalrulefreqcount_sanitytest)!!! \n" if $ruleinstancecount != $totalrulefreqcount_sanitytest;

  print STDERR "\nConsolidation finished going through $rulenumber rules, deleting $rulesdeleted rules, $sourcephrasesdeleted source-excl.-argNTs-phrases, $source_and_arg_phrases_deleted source-incl.-argNTs-phrases, $targetphrasesdeleted targetphrases and $rulebodiesdeleted rulebodies \@" . `date`;
} # end of sub consolidaterules

# Lexical weight table format: key = srcWord tgtWord ( file entries in the reversed set are reversed to comply)
# Value = array reference to a sequence of weights. [w1..wN]
sub loadLexicalWeightTable{
  my ($lexical_weight_table_href) = @_;
  # load in Lexical weight files
  my $lexical_weight_entry_count = 0;
  foreach my $filename ((@LexicalWeightFile, @LexicalWeightFileReversed)) {
    # n.o. lexical weight files so far
    $lexical_weight_entry_count++;
    # each weight file causes two features (lexical cost and refined lexical cost), so increase $additionalfeaturecount by 2
    #$additionalfeaturecount += 2;
    # if we're beyond the LexicalWeightFile entries, set reversemode to true to switch french and english
    my $reversemode = ($lexical_weight_entry_count > scalar(@LexicalWeightFile));
    print STDERR "Reading in lexical weight file $filename...\n";
    open(WTS, "< $filename") or die "Can't open lexical weight file $filename: $!";
    while (<WTS>) {
      chomp($_);
      my ($french, $english, $weight);
      if (!$reversemode) {
	($french, $english, $weight) = split(' ',$_);
      } else {
	($english, $french, $weight) = split(' ',$_);
      }
      my $frenchEncoded = join(' ',encodewordlist(split(' ',$french)));
      my $englishEncoded = join(' ',encodewordlist(split(' ',$english)));

      #don't add entries that arent in the vocCheck IFF you have turned on restrictions
      #NOTE: we can't compare the $frenchEncoded to NULL, only the $french
      if ($SourceRestrictionCorpus && !$vocCheck{$frenchEncoded} && $french ne 'NULL') {
	# source side doesn't occur in the restriction dev/test set -- ignore this entry
	  $logger->debug("Dropping entry encoded $frenchEncoded - real $french - $english - $weight\n");
      } elsif ($weight <= 0) {
	  print STDERR "Warning: Dropping weight $weight <= 0 which occured in entries like '$_'\n" if ($w_weightcount==0);
	  $w_weightcount = 1;
      } else {
	$lowest_lexical_weight_encountered = $weight if $weight < $lowest_lexical_weight_encountered;
	#my $key = $french . ' ' . $english;
	my $key = $frenchEncoded . ' ' . $englishEncoded;


	my $weights_aref =  $lexical_weight_table_href->{$key};
	if (!defined($weights_aref)){
	  my $numWeights =  scalar(@LexicalWeightFile)+scalar(@LexicalWeightFileReversed);
	  my @weights = (0) x $numWeights;
	  $weights_aref = \@weights;
	  $lexical_weight_table_href->{$key} = $weights_aref;
	}
	#assign the weight thru the $weights_aref thus changing the values in the hash
	$weights_aref->[$lexical_weight_entry_count-1] = $weight;

	#we guarantee that is scalar(@LexicalWeightFile)+scalar(@LexicalWeightFileReversed) elements 	
	#my $value =  $lexical_weight_table_href->{$key};
	#my @weights = ();	
	#if (defined($value)) {
	#  @weights = split(' ', $value);
	#} else {
	#  for (my $i = 0; $i<scalar(@LexicalWeightFile)+scalar(@LexicalWeightFileReversed); $i++) {
	#    push @weights,0;
	#  }
	#}
        #	$weights[$lexical_weight_entry_count-1] = $weight;
	#	  $weights[$i] = 0
	#print "DEBUG: adding wt:", $key, "=>",  join(' ', @weights), "\n";	
	#$lexical_weight_table_href->{$key} = join(' ', @weights);
      }
    }    
  }
  
  #now go thru and update all 0 entries with $lowest_lexical_weight_encountered/10. This is to ensure
  #that if weightFile1 has the entry, but weightFile2 does not, we still have a valid probability when
  #we lookup the value from weightfile2.
  #while (my ($key, $weights_aref)=each %$lexical_weight_table_href){
  #  for (my $wi=0; $wi=scalar(@$weights_aref); $wi++){
  #    if ($weights_aref->[$wi]==0){
#	$weights_aref->[$wi]= $lowest_lexical_weight_encountered/10;
#      }
#    }
#  }

}

sub displayIJWMatrix{
  my ($sourcewords_aref, $targetwords_aref, $i_j_w_matrix_aref) = @_;
  #iterate across real indices in the src, tgt vectors, but lookup weights in index+1
  #go down to -1 to print NULL
  my $retString = "";
  for (my $j=scalar(@$targetwords_aref)-1; $j>=-1; $j--){    
    if ($j==-1){ 
      #print "NULL\t| ";
      $retString  = $retString."NULL\t| ";
    }else{ 
      #print decodephrase($targetwords_aref->[$j]), "\t| "; 
      $retString = $retString.decodephrase($targetwords_aref->[$j])."\t| "; 
    }
    for (my $i=-1; $i<scalar(@$sourcewords_aref); $i++){
      my $weights_aref = $i_j_w_matrix_aref->[$i+1]->[$j+1];
      #if (scalar(@$weights_aref)==0){ 
      if (defined($weights_aref)){ 
	#print join('/', @$weights_aref), "\t\t";
	$retString = $retString.join('/', @$weights_aref). "\t\t";
      }else{
	#print "-\t\t";
	$retString = $retString."-\t\t";
      }
    }
    #print "\n";
    $retString = $retString."\n";
  }
  #print "\t| ";
  $retString = $retString."\t| ";
  for (my $i=-1; $i<scalar(@$sourcewords_aref); $i++){
    if ($i==-1){
      #print "NULL\t\t";
      $retString = $retString."NULL\t\t";
    } else {
      #print decodephrase($sourcewords_aref->[$i]), "\t\t";
      $retString = $retString.decodephrase($sourcewords_aref->[$i])."\t\t";
    }
  }
  #print "\n";
  $retString = $retString."\n";
  return $retString;
}

#Start with alignments in the form srcIndex-tgtIndex (ie the pharaoh alignment table)
#and intersect this alignment map with lexical weights. The $lexical_weight_table_href
#is looked up by src-tgt pair which returns an array reference to a series of weights.
#
#This function fills an alignment map, the %i_j_w_matrix_aref with the weights 
#corresponding to source i, target j. undef if there is no alignment, $lowest_lexical_weight_encountered/10 x |wts|
#if there is an alignment but no weights.
sub createAlignmentsWithProbabilities{
  my ($i_j_w_matrix_aref, $alignmentString, $sourcewords_aref, $targetwords_aref, $lexical_weight_table_href, $srcAligned_href, $tgtAligned_href) = @_;

  #first build the space for the matrix with all blank elements
  #my @i_j_w_matrix = ();
  for (my $i = -1; $i < scalar(@$sourcewords_aref); $i++) {
    my @j_w_matrix = ();

    #lets make sure there is always a source vector that can be queried, but you can see
    #if the target vector is actually defined (at which point you are guaranteed that probs exist)
    for (my $j = -1; $j < scalar(@$targetwords_aref); $j++) {
      #initialize the weight vector so that we can always ask for any model's weight
      #my @weights = ();
      #for (my $w =0; $w<scalar(@LexicalWeightFile)+scalar(@LexicalWeightFileReversed); $w++){
	#push(@weights, 0.0);
      #}
      #my $refweights = \@weights;

      #push an undefined matrix
      my $refweights = undef;
      push @j_w_matrix, $refweights;
    }
    my $ref_j_w_matrix = \@j_w_matrix;
    #push @i_j_w_matrix, $ref_j_w_matrix;
    push @$i_j_w_matrix_aref, $ref_j_w_matrix;
  }

  #now look at all the alignments that we know exist. If an alignment
  #exists, we lookup the corresponding lexical weight from the $lexical_weight_table_href
  #and set the approriate @i_j cell. We track which src positions have at least one aligment
  #and when we are done we can find any that are aligned to NULL and assign the NULL prob
  #my %srcPositionsAligned = ();
  #my %tgtPositionsAligned = ();

  #create a backup set of null weights in case we find any alignments but no lexical evidence
  #in ANY of the lexical structures. Creating it once here allows them to share
  my $numWts = scalar(@LexicalWeightFile) + scalar(@LexicalWeightFileReversed);
  my @weightsDummy = $lowest_lexical_weight_encountered x $numWts;
  

  #these indices start at 0 < @$sourcewords_aref
  foreach my $pair (split(' ', $alignmentString)){
    my ($srcIndex, $tgtIndex) = split('-',$pair);

    #print "DEBUG: Alignment $srcIndex-$tgtIndex\n";
    $srcAligned_href->{$srcIndex} = 1;
    $tgtAligned_href->{$tgtIndex} = 1;

    #see what the actual words are, and lookup their lexical probabilities
    my $key = $sourcewords_aref->[$srcIndex] . ' ' . $targetwords_aref->[$tgtIndex];
    #print "DEBUG: lookup wts for => $key\n";
    my $value = $lexical_weight_table_href->{$key};
    if (defined($value)) {
      #my @weights = split(' ', $value);
      #print "DEBUG: Saving with: ", join(' ', @$value), "\n";
      #remember to file it in the +1 position since both sides have a NULL aligment
      #$i_j_w_matrix[$srcIndex+1]->[$tgtIndex+1] = \@weights;
      $i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1] = $value;
    }else{
      #You did see the alignment, but there is no lexical evidence for it

      if ($SourceRestrictionCorpus){
	#there might be no lexical evidence because you removed it (since it is outside of the vocab restriction), in this case don't warn
	if ($vocCheck{$sourcewords_aref->[$srcIndex]}){	
	    #$logger->warn("No lex evidence for: ", decodephrase($sourcewords_aref->[$srcIndex]), "-", decodephrase($targetwords_aref->[$tgtIndex]), " setting: ",  $lowest_lexical_weight_encountered/10 );
	}
      }else{
	  #$logger->warn("No lex evidence for: ", decodephrase($sourcewords_aref->[$srcIndex]), "-", decodephrase($targetwords_aref->[$tgtIndex]), " setting: ",  $lowest_lexical_weight_encountered/10 );
      }

      #you need to create a vector or elements on the fly here

       $i_j_w_matrix_aref->[$srcIndex+1]->[$tgtIndex+1] = \@weightsDummy;
    }
  }
  
  #now the @i_j_w_matrix has weight entries for all the aligned words, but we still need to 
  #add the non-aligned words by looking at which src, tgt indices where not aligned
  for (my $srcIndex = 0; $srcIndex < scalar(@$sourcewords_aref); $srcIndex++) {
    
    #only for unaligned src words
    if (!(defined($srcAligned_href->{$srcIndex}))){
      #add a link from $srcIndex to -1 by looking up the alignment to NULL
      my $value = $lexical_weight_table_href->{$sourcewords_aref->[$srcIndex] . ' ' .encodephrase('NULL')};
      if (defined($value)){
	#my @weights = split(' ', $value);
	#fill the NULL spot
	#$i_j_w_matrix[$srcIndex+1]->[-1+1] = \@weights;
	$i_j_w_matrix_aref->[$srcIndex+1]->[-1+1] = $value;
      }else{
	#You did see the alignment but there is no lexical evidence for it
	if ($SourceRestrictionCorpus){
	  #there might be no lexical evidence because you removed it (since it is outside of the vocab restriction), in this case don't warn
	  if ($vocCheck{$sourcewords_aref->[$srcIndex]}){
	      #$logger->warn("No lex evidence for: ", decodephrase($sourcewords_aref->[$srcIndex]), "-NULL", " setting: ",  $lowest_lexical_weight_encountered/10 );
	  }
	}else{
	    #$logger->warn("No lex evidence for: ", decodephrase($sourcewords_aref->[$srcIndex]), "-NULL", " setting: ",  $lowest_lexical_weight_encountered/10 );
	}
	$i_j_w_matrix_aref->[$srcIndex+1]->[-1+1] = \@weightsDummy;
      }
    }
  }
  #repeat for the target side
  for (my $tgtIndex = 0; $tgtIndex < scalar(@$targetwords_aref); $tgtIndex++) {

    #only for unaligned target words
    if (!(defined($tgtAligned_href->{$tgtIndex}))){
      #add a link from $srcIndex to -1 by looking up the alignment to NULL
      my $tgtWord = $targetwords_aref->[$tgtIndex];
      my $value = $lexical_weight_table_href->{encodephrase('NULL').' '. $tgtWord}; 
      if (defined($value)){
	#my @weights = split(' ', $value);
	#fill the NULL spot
	#$i_j_w_matrix[-1+1]->[$tgtIndex+1] = \@weights;
	$i_j_w_matrix_aref->[-1+1]->[$tgtIndex+1] = $value;
      }else{
	#You did see the alignment but there is no lexical evidence for it - there is no question of dropping the lexical entry because of src side restriction
	  #$logger->warn("No lex evidence for: ", "NULL-", decodephrase($targetwords_aref->[$tgtIndex]), " setting: ",  $lowest_lexical_weight_encountered/10 );
	$i_j_w_matrix_aref->[-1+1]->[$tgtIndex+1] = \@weightsDummy;
      }
    }
  }
  #print "ABOUT TO DISPLAY\n";
  #display matrix that we have generated
  $logger->debug("Lexical WeightMatrix \n\n",displayIJWMatrix($sourcewords_aref, $targetwords_aref, $i_j_w_matrix_aref));
}



