#!/usr/bin/env perl

use warnings;
use strict;
#use Encode::Arabic;

#use Encode qw/encode decode/;
#use Storable;

use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
use BerkeleyDB;			# uncomment this line if you don't have Berkeley DB installed on your unix system
use Getopt::Long;


#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings

my $BERKELEY_DB = 1;		# set to zero if Berkeley DB is not available

# options and input params

# puts everything into probability space - use for STTK
#my $neglogprobs2probs
#use: HtdToTD.sh | ConvertLogTDtoProbs.pl for that



# Command line parameters

# if set to 1 (option --UsePerlHash), uses regular perl hash for the temporary rule DB and phrasal counts / lexical weights (fast, but memory-intensive)
# if set to 0 (--noUsePerlHash), uses temporary Berkeley DB instead (slow, but keeps memory usage under control)
my $UsePerlHashForRules = 1;
my $UsePerlHashForLexicalWeights = 1;

my $ConsiderSourceSpanImpact = 0;

# n.o. phrasal features (everything after the frequency feature AND any simply freq feature that u added
# passed from extractrules.pl
# obligatory parameter (unless --UnknownInputRuleFile is specified)
my $PhrasalFeatureCount;

# if true, then we'll make as few assumptions about the rule file's features as possible (everything must be in neg-log space though)
my $UnknownInputRuleFile = 0;

# offset (starting with index 0) of the neg log source cond. relative frequency feature (used for pruning)
# ie frequency/source_and_arg
my $InputRuleFileSourceCondRFFeatureOffset = 5;

my $InputRuleFileIgnoreFirstNFeatures = 0;

# tells us that we have to take exp(-x) to get occurrence count from 7th (and now 8th for the simple) feature
# if $UnknownInputRuleFile is true, the occurrence count feature will be ignored, so this option has no effect
my $NegLogOccurrenceCountFeature = 1;

# by default, we will recompute the result-conditioned RF feature based on the cumulative result-category counts of the rules.
# This way, rule bases can be created independently from training corpora split into several parts and then joined
# via filterrules.pl
my $RecomputeResultCondRF = 1;

# set this less than inf if you want to restrict the number of
# abstraction sites in the rules
my $maxabstractioncount = 2**30;

my $MaxSourceSymbolsNonlexicalRules = 2**30;
my $MaxTargetSymbolsNonlexicalRules = 2**30;
my $AllowAdjacentSourceNonterminals = 1;
#my $AllowAdjacentTargetNonterminals = 1;
my $AllowAbstractRules = 0;
my $AllowRulesWithOnlyTargetTerminals = 0;

my $RestrictResultCategoriesPerRuleBody = 0;

my $keepunbalancedpunctuationrules = 1; #keep rules that have unbalanced source-target punctuation / word behavior 

# set this to true to keep natural _S rules (e.g. @NP @VP#@1 @2#@_S#0.5 ....) in the rule DB
my $KeepNaturalSentenceRules = 1;

# set this to false if you want to filter out unbalanced rules
# unbalanced means having source words but no target words and vice
# versa or ratio betwen source word count and target word count (or
# vice versa) is bigger than $bal_limit
my $keepunbalancedrules = 1;

#modified to 1000 to prevent removal of totally unbalanced rules
my $bal_limit = 1000;		# max. ratio between source word count and target word count and vice versa
# set this to "n.o. target words in training corpus" / "n.o. source words in training corpus"
my $MeanTargetSourceRatio = 1;
# if set to 0, removes all rules containing _X in the rule body, but keeps lexical rules with result category _X
my $AbstractOverX = 1;

my $crazyclusterrules = 0;	# 1 for crazy (only keep first two letters from categories), 2 for very crazy (only keep first letter from categories), 3 for insane (everything becomes _X - this together with $maxabstractioncount=0 plus some reordering rule simulates STTK mode )

# set to true what you want to be written out
my $gluerules = 1;
#my $exitgluemoderules = 0;
#my $entergluemoderules = 1;
#my $joininggluerule = 1;
my $unknownrules = 1;

# If --NullRules is passed, we create for each source word a rule translating that word to _NULL
# For these rules, feature $nullrulefeatureoffset with be set to 1
# If you make use of this, make sure your post processing normalization script deletes the _NULL word
my $NullRules = 0;

my $tmpdir;
my $warned_too_many_features=0;
my $outputdbfile;
my $verbose = 0;
my $vocabFile = '';

# These options can each occur multiple times and lets you specify files with word-based lexical probabilities.
# Each file passed with option --LexicalWeightFile has to consist 
# of entries of probability = P(frenchword|englishword) of the form:
# frenchword englishword probability
#
# Each file passed with option --LexicalWeightFileReversed has to consist
# of entries of probability = P(englishword|frenchword) of the form:
# englishword frenchword probability
#
# The resulting lexical features (two for each weight file) for the rules will be 
# appended to the end of the rules' feature vectors,
# first the ones corresponding to the --LexicalWeightFile files specified,
# then the ones corresponding to the --LexicalWeightFileReversed files specified.
# The first lexical weight feature is the neg-log IBM1 prob for the phrase pair (not considering nonterminals).
# The second lexical weight feature neg-log best-possible joint lexical probability:
#   For phrase pair f_1,...,f_m/e_1,...,e_n, in case of --LexicalWeightFile that is:
#     -log(prod_i max_j P(f_i|e_j)
#   For phrase pair f_1,...,f_m/e_1,...,e_n, in case of --LexicalWeightFileReversed that is:
#     -log(prod_j max_i P(e_j|f_i)
my @LexicalWeightFile = ();
my @LexicalWeightFileReversed = ();
# The following parameter concerns the exact way the IBM1 phrasal prob is estimated.
# Set the following option to 0 if you want the traditional IBM1 estimate
# By default (1), a refined IBM1 estimate is used, which performs better in practice (paper on that coming soon)
my $UseRefinedIBM1ProbEstimate = 1;
# make use of null alignment probability in lexical weight files
my $UseNULL = 1;
my $AlsoComputeWrongDirections = 0;

# binary vector B1_B2_..._Bn indicating by Bi=1 that i-th feature should be written out, and by Bi=0 that it shouldn't
my $RestrictToFeatures = '';
my $sttk = 0;			# set crazyclusterrules = 3, maxabstractioncount = 0, add reordering rule, change glue feature into reordering feature
my $chiang = 0;
#my $KeepOccurrenceCountFeature = 0;
#my $LostOccurrenceCountFeature = 0;
my $maxsourcephraselength = 100;	# upper bound for test-sentence filtering

# removes a NONLEXICAL rule source/target/type if its occurrence count is less than $minoccurrencecount
my $minoccurrencecountnonlexicalrules = 0;
# removes a LEXICAL rule source/target/type if its occurrence count is less than $minoccurrencecount and there is another rule source/target2/type2 with higher occurrence count
my $minoccurrencecountlexicalrules = 0;
# in case such a low-occurrence-count lexical rule DOES have the highest occurrence count amongst all rules with same source,
# it will be added if its source length is less or equal $MaxSourceLengthLowOccurrenceRules
my $MaxSourceLengthLowOccurrenceRules = 2**30;

# set this greater than zero if you want to remove a rule if it has a smaller
# lexical-source conditioned relative frequency than $minlexsourcecondrf and
# there is another rule with the same source with higher occurrence count
my $minlexsourcecondrflexicalrules = 0.000001;
my $minlexsourcecondrfnonlexicalrules = 0.000001;
# set this to greater than zero to only output rules whose frequency is at least $beamfactor * 'frequency of highest-frequency rule with same source side'
my $beamfactorlexicalrules = 0; #by default no pruning away
my $beamfactornonlexicalrules = 0; #by default no pruning away


# set this to a file name if you want that phrase table to be merged into the rule file
# format of phrase table: @NEGLOG#sourcephrase#targetphrase#costvector
my $MergePhrasetable = '';
# the LHS category assigned to merged phrasepairs that don't exist as rules
my $merge_phrasetable_result_nonterminal = '@_X';

my $cachesize = 512;		# Berkeley DB cache size in MB

# write out rules assuming source sentences start with <s> and end
# with </s> and target sentences start with <s> <s> and end with </s> </s>
my $UseBOSandEOS = 1;
my $UseCountMerge = 0;		#when merging multiple phrase tables you should update several accumulate counts
my $PrintPhraseCounts = 0;	#print out phrase counts, allows you to refilter based on the db itself
my $HasSimpleCount=1;
my $LengthBoostAlpha = 0;	#extra cost given to rule that 

my $RemoveSourcePunctuation = 0;
my $MaxCategories = 10000;	#max number of concatenations in the result cat like NP+VP = 2
my $MergeWithPosterior = 1;	# merge with the posterior freq instead of the simple one
my $PosteriorBasedRFPruning = 1; # do the relative frequency - and beam-based pruning based on the posterior count, not based on the simplecount

print STDERR 'filterrules.pl - started @ '. `date`;

die if !GetOptions('tmpdir=s' => \$tmpdir, 'cachesize=s' => \$cachesize, 'verbose=s' => \$verbose, 
		   'UsePerlHashForRules!' => \$UsePerlHashForRules, 'UsePerlHashForLexicalWeights!' => \$UsePerlHashForLexicalWeights,
		   'outputdbfile=s' => \$outputdbfile, 'r=s' => \$vocabFile, 
		   'KeepNaturalSentenceRules!' => \$KeepNaturalSentenceRules, 'AbstractOverX!' => \$AbstractOverX, 
		   'maxsourcephraselength=s' => \$maxsourcephraselength, 
		   'MaxSourceLength=s' => \$maxsourcephraselength, 
'RestrictResultCategoriesPerRuleBody=s' => \$RestrictResultCategoriesPerRuleBody, 'minoccurrencecountlexicalrules=s' => \$minoccurrencecountlexicalrules, 'minoccurrencecountnonlexicalrules=s' => \$minoccurrencecountnonlexicalrules, 'MaxSourceLengthLowOccurrenceRules=s' => \$MaxSourceLengthLowOccurrenceRules, 
		   'MeanTargetSourceRatio=s' => \$MeanTargetSourceRatio, 
		   #======= rule pruning ==============
		   'minlexsourcecondrflexicalrules=s' => \$minlexsourcecondrflexicalrules, 
		   'MaxSourceSymbolsNonlexicalRules=s' => \$MaxSourceSymbolsNonlexicalRules,
		   'MaxTargetSymbolsNonlexicalRules=s' => \$MaxTargetSymbolsNonlexicalRules,
		   'PosteriorBasedRFPruning!' => \$PosteriorBasedRFPruning,
		   'minlexsourcecondrfnonlexicalrules=s' => \$minlexsourcecondrfnonlexicalrules, 
		   'beamfactorlexicalrules=s' => \$beamfactorlexicalrules,
		   'beamfactornonlexicalrules=s' => \$beamfactornonlexicalrules,
		   'LengthBoostAlpha=s' => \$LengthBoostAlpha,
		   'RecomputeResultCondRF!' => \$RecomputeResultCondRF, 
		   'AllowAbstractRules!' => \$AllowAbstractRules, 'AllowAdjacentSourceNonterminals!' => \$AllowAdjacentSourceNonterminals, 
		   #'AllowAdjacentTargetNonterminals!' => \$AllowAdjacentTargetNonterminals, 
		   'UseRefinedIBM1ProbEstimate!' => \$UseRefinedIBM1ProbEstimate, 'NegLogOccurrenceCountFeature!' => \$NegLogOccurrenceCountFeature, 'STTK' => \$sttk, 'Chiang' => \$chiang, 'RestrictToFeatures=s' => \$RestrictToFeatures, 'CrazyClusterRules=s' => \$crazyclusterrules, 'GlueRules!' => \$gluerules, 'UnknownRules!' => \$unknownrules, 
		   'LexicalWeightFile=s' => \@LexicalWeightFile, 'LexicalWeightFileReversed=s' => \@LexicalWeightFileReversed, 
		   'InputRuleFileSourceCondRFFeatureOffset=s' => \$InputRuleFileSourceCondRFFeatureOffset, 
		   'UnknownInputRuleFile!' => \$UnknownInputRuleFile, 'InputRuleFileIgnoreFirstNFeatures=s' => \$InputRuleFileIgnoreFirstNFeatures, 'MergePhrasetable=s' => \$MergePhrasetable, 
		   'KeepUnbalancedRules!' => \$keepunbalancedrules, 
		   'keepunbalancedpunctuationrules!' => \$keepunbalancedpunctuationrules, 
		   'UseBOSandEOS!' => \$UseBOSandEOS, 'UseCountMerge!' => \$UseCountMerge, 'PrintPhraseCounts!' => \$PrintPhraseCounts, 'UseNULL!' => \$UseNULL, 'ConsiderSourceSpanImpact!' => \$ConsiderSourceSpanImpact, 
		   'RemoveSourcePunctuation!' => \$RemoveSourcePunctuation, 
		   'NullRules!' => \$NullRules,
		   'PhrasalFeatureCount=s' => \$PhrasalFeatureCount, 'HasSimpleCount!' => \$HasSimpleCount, 'maxabstractioncount=s' => \$maxabstractioncount, 'MergeWithPosterior!' => \$MergeWithPosterior, "MaxCategories=s" => \$MaxCategories);

die "filterrules.pl: You need to specify either --UnknownInputRuleFile or --PhrasalFeatureCount=N where N denotes the n.o. phrasal (lexical weight etc.) features extractrules.pl created (all features after the frequency count feature)" if !$UnknownInputRuleFile && !defined($PhrasalFeatureCount);
$PhrasalFeatureCount = 0 if !defined($PhrasalFeatureCount);

die "If you want to use --RestrictResultCategoriesPerRuleBody, you have to specify --noUsePerlHashForRules" if $UsePerlHashForRules && $RestrictResultCategoriesPerRuleBody;
die "If you don't have a Berkeley DB, you have to specify option --UsePerlHashForRules and --UsePerlHashForLexicalWeights" if !$BERKELEY_DB && (!$UsePerlHashForRules || !$UsePerlHashForLexicalWeights);

# $originalfeaturecount denotes the n.o. relevant features from input
# rule files (all features beyond this count are ignored)
my $originalfeaturecount = 7 + ($HasSimpleCount?1:0) + $PhrasalFeatureCount;
#0		   0,		# glue feature
#1		   1,		# rule application counter
#2		   $targetwordcount, # target word counter
#3		   -log($frequency/$undilutedresultcount), # neglog (base e) result-conditioned RF
#######4		   -log($rulebodyfrequency/$targetfrequency), # neglog target-cond. RF
#4		   -log($rulebodyfrequency/$sourcefrequency), # neglog source-cond. RF (not considering nontermial symbols)
#5		   -log($frequency/$source_and_arg_frequency), # neglog source-cond. RF (considering nonterminal symbols)
#6   		   -log($frequency) #neglog frequency of rule ($NegLogOccurrenceCountFeature == true)
#	or:	   $frequency,	# frequency of rule (NegLogOccurrenceCountFeature == false)
#     (filterrules.pl will always put this feature into neg-log space)
#7                 -log($frequencySimple) # a simple frequency count
#8..8+PhrasalFeatureCount-1	:extractrules' @costvector[1..$#costvector]  	 # phrasal costs (5 values)

# n.o. additional features being created by filterrules.pl
# will be increased  if --LexicalWeightFile or --LexicalWeightFileReversed or --MergePhrasetable parameters are passed
my $additionalfeaturecount = 11;
#0  purely-lexical feature (rule body consists only of terminals)
my $lexicalfeatureoffset = $originalfeaturecount+0;
##1  purely-abstract feature (rule body consists only of nonterminals)
#my $abstractfeatureoffset = $originalfeaturecount+1;
#1 adjacent-NT feature
my $adjacentsourceNTfeatureoffset = $originalfeaturecount+1;
#2  _X rule feature
my $xrulefeatureoffset = $originalfeaturecount+2;
#3  unbalancedness penalty 1
my $unbalancednesspenalty1offset = $originalfeaturecount+3;
#4  unbalancedness penalty 2
my $unbalancednesspenalty2offset = $originalfeaturecount+4;
#4 source terminals but no target terminals feature
my $onlysourceterminalsfeatureoffset = $originalfeaturecount+5;
# # target terminals but no source terminals feature
# my $onlytargetterminalsfeatureoffset = $originalfeaturecount+6;
# source-punctuation-was-removed feature
#my $sourcepunctuationremovedfeatureoffset = $originalfeaturecount+6;
my $unbalancedpunctuationfeatureoffset = $originalfeaturecount+6;
#  non-monotonicity penalty
my $nonmonotonicitypenaltyoffset = $originalfeaturecount+7;
#  rareness penalty
my $rarenesspenaltyoffset = $originalfeaturecount+8;
# -log($simpleFreq/$undilutedresultcount), # neglog (base e) result-conditioned RF
my $undilutedresultcountoffset = $originalfeaturecount+9; #the diluted count is the important ONE
my $nullrulefeatureoffset = $originalfeaturecount+10; # 1 if this is a null rule; 0 if it's a regular or glue rule

# in addition, we have the lexical-weight features, three for each file passed in via 
# --LexicalWeightFile or --LexicalWeightFileReversed
my $lexical_costs_offset = $additionalfeaturecount + $originalfeaturecount;

# if option --MergePhrasetable is specified, we will have even more features after that:
# is_rule(0/1), is_phrase(0/1), phrase_feature_1, ..., phrase_feature_n
my $merge_phrasetable_feature_offset;


my $minlexsourcecondrflexicalrules_cost = -log($minlexsourcecondrflexicalrules);
my $minlexsourcecondrfnonlexicalrules_cost = -log($minlexsourcecondrfnonlexicalrules);



# $tmpdir = File::Spec->tmpdir() if (!$tmpdir);
$tmpdir = "." if (!$tmpdir);

if ($sttk) {
  $crazyclusterrules = 3;
  $maxabstractioncount = 0;
}
if ($chiang) {
  $crazyclusterrules = 3;
  $maxabstractioncount = 2;
  $AllowAbstractRules = 0;
  $AllowAdjacentSourceNonterminals = 0;
  #  $AllowAdjacentTargetNonterminals = 0;
  #  $MaxSourceSymbolsNonlexicalRules = 5;
  #  $MaxTargetSymbolsNonlexicalRules = 5;
}

my @featureRestrictionVector;

my %vocCheck = ();
if ($vocabFile) {
  die "Parameter --MaxSourceLength must be specified!\n" if !$maxsourcephraselength;
  open(SOURCE, "< $vocabFile") or die "Can't open $vocabFile: $!";
  while (<SOURCE>) {
    chomp($_);
    my @sentence = ('<s>', split(' ',$_), '</s>');
    for (my $phraselength = 1; $phraselength <= $maxsourcephraselength; $phraselength++) {
      for (my $phrasestart = 0; $phrasestart+$phraselength <= scalar(@sentence); $phrasestart++) {
	$vocCheck{join(' ',@sentence[$phrasestart..$phrasestart+$phraselength-1])} = 1;
      }
    }
  }
  close(SOURCE);
}



# global variables

# create my own temp dir as subdir of $tmpdir, it and its contents will 
# automatically removed at program end
my $mytmpdir;
$mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 1) if !$UsePerlHashForRules || !$UsePerlHashForLexicalWeights;
#my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 0);

my $sentencenumber = 0;		# counts the sentences processed
my $phrasepairnumber = 0;	# counts the phrase pairs processed
my $ruleinstancecount = 0;	# counts the number of instances of rules extracted
my $writtenoutrulenumber = 0;	# number of rules written out to STDOUT / database

# will contain the complex rules, with $key= "$source\#$target\#$result" and $value=\@costvector
# also stores for each rule source side the maximum occurrence count achieved by a rule with that source side
my %complexrules = ();		
#my %phrasecounts = ();          # will map each rule source side to the maximum occurrence count achieved by a rule with that source side
my $outputdb = ();		# the output rule database - a B-tree

#THIS DILUTED COUNT IS THE CORRECT ONE - that uses the posterior frequency
my %resulttypecounts_diluted = (); # counts the total n.o. rules in %complexrules that have the same result type (keys are categories, values are numbers); this will enable relative-frequency calculations when writing out the rules
#THIS UNDILUTED COUNT IS A SMOOTHER! - it uses a simple frequency count
my %resulttypecounts_undiluted = (); # counts the total n.o. rules in %complexrules that have the same result type (keys are categories, values are numbers); this will enable relative-frequency calculations when writing out the rules
my %resulttypecounts_diluted_logsourcespanlength_meanstat = (); # see extractrules.pl
my %resulttypecounts_diluted_logsourcespanlength_variancestat = (); # see extractrules.pl

my %resulttypes_confirmed = (); # result categories that actually occur as results in the filtered rules

my %source_counts = ();		# phrase count structures for proper merging
my %sourceandarg_counts = ();	#count structures for proper merging
my %rulebody_counts = ();	#count structures for proper merging
my %target_counts = ();		#count structures for proper merging

# n.o. lexical weights
my $lexical_weight_entry_count;
my $lexical_features_per_entry_count;
# lowest weight we will encounter
my $lowest_lexical_weight_encountered;
my %lexical_weight_table = ();

# when we are merging a phrasetable into the rule database, we 
# have to remember the rule bodies of lexical rules
my %lexical_rule_bodies = ();
my $w_weightZero = 0;

# hashes all terminals occurring as source words in the rules -- used for creating the Null rules
my %source_vocab = ();
# avoid adding Null rules for <s> and </s>
$source_vocab{'<s>'} = 1; $source_vocab{'</s>'} = 1; 

# n.o. phrase features in case of --MergePhrasetable
my $merge_phrasetable_feature_count = 0;

{				# main block
  print STDERR ($outputdbfile ?
		"Outputting rules into newly created Berkeley DB file $outputdbfile\n" :
		"Outputting rules in text-form to STDOUT\n");

  if ($outputdbfile) {
    if ($BERKELEY_DB) {
      $outputdb = new BerkeleyDB::Btree 
	-Cachesize => 2**20 * $cachesize/4, 
	  -Filename => $outputdbfile, 
	    -Flags => DB_CREATE|DB_TRUNCATE,
	      -Property => DB_DUP
		|| die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
      die "weird database creation error" if !defined($outputdb);
    } else {
      die "Variable \$BERKELEY_DB is set to 0, therefore the rules cannot be output as database file but just as plain text! Bye.\n";
    }
  }

  my $db1;
  if (!$UsePerlHashForRules) {
    $db1 = tie %complexrules,
      #    "BerkeleyDB::Hash", 
      "BerkeleyDB::Btree", 
	-Cachesize => 2**20 * $cachesize * 1/4, 
	  -Filename => "$mytmpdir/complexrules", 
	    -Flags => DB_CREATE | DB_TRUNCATE
	      || die "Can't create temp database complexrules in $mytmpdir!";
    my  $db2 = tie %sourceandarg_counts, 
      #      "BerkeleyDB::Hash", 
      "BerkeleyDB::Btree", 
	-Cachesize => 2**20 * $cachesize * 1/4, 
    	  -Filename => "$mytmpdir/sourceandarg_counts", 
    	    -Flags => DB_CREATE | DB_TRUNCATE
	      || die "Can't create temp database phrasecounts in $mytmpdir!";
    my  $db3 = tie %source_counts, 
      #      "BerkeleyDB::Hash", 
      "BerkeleyDB::Btree", 
	-Cachesize => 2**20 * $cachesize * 1/4, 
    	  -Filename => "$mytmpdir/source_counts", 
    	    -Flags => DB_CREATE | DB_TRUNCATE
	      || die "Can't create temp database phrasecounts in $mytmpdir!";
    my  $db4 = tie %rulebody_counts, 
      #      "BerkeleyDB::Hash", 
      "BerkeleyDB::Btree", 
	-Cachesize => 2**20 * $cachesize * 1/4, 
    	  -Filename => "$mytmpdir/rulebody_counts", 
    	    -Flags => DB_CREATE | DB_TRUNCATE
	      || die "Can't create temp database phrasecounts in $mytmpdir!";
  }  
  if (!$UsePerlHashForLexicalWeights) {
    my $db5 =  tie %lexical_weight_table, 
      #      "BerkeleyDB::Hash", 
      "BerkeleyDB::Btree", 
	-Cachesize => 2**20 * $cachesize * 1/4, 
    	  -Filename => "$mytmpdir/lexical_weight_table", 
    	    -Flags => DB_CREATE | DB_TRUNCATE
	      || die "Can't create temp database lexical_weight_table in $mytmpdir!";
  }  


  if ($MergePhrasetable) {
    print STDERR "Merging rules with phrase table $MergePhrasetable\n";
    # push back '-' (meaning STDIN) and the phrase table file into @ARGV in order to read in the phrases after all rules have been
    # processed
    if (scalar(@ARGV) == 0) {
      @ARGV = ('-', $MergePhrasetable);
    } else {
      push @ARGV, $MergePhrasetable;
    }
  }

  # load in Lexical weight files
  $lowest_lexical_weight_encountered = 2**30;
  $lexical_weight_entry_count = 0;
  foreach my $filename ((@LexicalWeightFile, @LexicalWeightFileReversed)) {
    # n.o. lexical weight files so far
    $lexical_weight_entry_count++;
    # if we're beyond the LexicalWeightFile entries, set reversemode to true to switch french and english
    my $reversemode = ($lexical_weight_entry_count > scalar(@LexicalWeightFile));
    print STDERR "Reading in lexical weight file $filename...\n";
    open(SOURCE, "< $filename") or die "Can't open lexical weight file $filename: $!";
    while (<SOURCE>) {
      chomp($_);
      my ($french, $english, $weight);
      if (!$reversemode) {
	($french, $english, $weight) = split(' ',$_);
      } else {
	($english, $french, $weight) = split(' ',$_);
      }
      if ($vocabFile && !$vocCheck{$french} && $french ne 'NULL') {
	# source side doesn't occur in the restriction dev/test set -- ignore this entry
      } elsif ($weight <= 0) {
	print STDERR "Warning: weight $weight <= 0 occurred in entry '$_'\n" if ($w_weightZero==0);
	$w_weightZero = 1;
      } else {
	$lowest_lexical_weight_encountered = $weight if $weight < $lowest_lexical_weight_encountered;
	my $key = $french . ' ' . $english;
	my $value =  $lexical_weight_table{$key};
	my @weights = ();
	if (defined($value)) {
	  @weights =unpack('f*', $value);
	} else {
	  for (my $i = 0; $i<scalar(@LexicalWeightFile)+scalar(@LexicalWeightFileReversed); $i++) {
	    push @weights,0;
	  }
	}
	$weights[$lexical_weight_entry_count-1] = $weight;
	#	  $weights[$i] = 0
	$lexical_weight_table{$key} = pack('f*', @weights);
      }
    }
    close(SOURCE);
  }
  # each weight file causes 2 features
  $lexical_features_per_entry_count = 2 * ($AlsoComputeWrongDirections?2:1);
  $additionalfeaturecount += $lexical_features_per_entry_count * $lexical_weight_entry_count;
  # adjust offset for phrase features
  $merge_phrasetable_feature_offset = $additionalfeaturecount + $originalfeaturecount;


  my $adjusted_sourcecondrf_offset = 0;
 FIRSTLOOP:
  while (<>) {
    $sentencenumber++;
    print STDERR "Processing rule number $sentencenumber...\r" if $sentencenumber % 10000 == 0;

    chomp;
    my $orgrule = $_;
    my @rule = split(/ *\# */,$_);
    
    
    # determine whether it's a phrase pair from --MergePhrasetable, not an actual rule
    my $is_phrasepair = 0;
    my @phrasepair_features;
    if ($rule[0] eq '@NEGLOG') {
      $is_phrasepair = 1;
      # save the phrase features
      @phrasepair_features = split(' ',$rule[3]);
      $merge_phrasetable_feature_count = scalar(@phrasepair_features);
      
      # rearrange rule, and prepend glue feature, rule app count feature, and target word count feature
      # target word count feature will be computed below
      @rule = ($rule[1],$rule[2],$merge_phrasetable_result_nonterminal,'0 1 -1' . concatenatenumber(0, $originalfeaturecount-3));

      # remove possible source or target words starting with @
      $rule[0] =~ s/\B\@/___at___/g; # all @ that are not preceded by a letter or number
      $rule[1] =~ s/\B\@/___at___/g; # all @ that are not preceded by a letter or number
    }

    if ($crazyclusterrules == 1) {
      $rule[0] =~ s/(\@[A-Za-z][A-Za-z])\S+/$1/g; # substitute e.g. @NNS to @NN in sourceside
      $rule[2] =~ s/(\@[A-Za-z][A-Za-z])\S+/$1/g; # substitute e.g. @NNS to @NN in resultcategory
    } elsif ($crazyclusterrules == 2) {
      $rule[0] =~ s/(\@[A-Za-z])\S+/$1/g; # substitute e.g. @NP to @N in sourceside
      $rule[2] =~ s/(\@[A-Za-z][A-Za-z])\S+/$1/g; # substitute e.g. @NP to @N in resultcategory
    } elsif ($crazyclusterrules == 3) {
      $rule[0] =~ s/\@[^ _][^ ]*/\@_X/g; # substitute all @something (except underscore categories) to @_X
      $rule[2] =~ s/\@[^ _][^ ]*/\@_X/g; # same in resultcategory
      $rule[0] =~ s/\@_X[^ ]*/\@_X/g; # also substitute all @_Xsomething to @_X
      $rule[2] =~ s/\@_X[^ ]*/\@_X/g; # same in resultcategory
    }

    if ($rule[0] eq '@_COUNT') {
      # pseudo-rule storing the counts of a result nonterminal
      my @rc = split(' ',$rule[3]);
      print STDERR "Warning: mean / variance statistics not specified for NT $rule[2] - old rule file format\n" if !defined($rc[3]);
      $rc[2] = 0 if !defined($rc[2]); # for compatibility
      $rc[3] = 0 if !defined($rc[3]); # for compatibility
      # update diluted result category occurrence frequency using += magic
      $resulttypecounts_diluted{$rule[2]} += $rc[0];
      # update undiluted result category occurrence frequency using += magic
      $resulttypecounts_undiluted{$rule[2]} += $rc[1];
      # update mean log source span length statistics using += magic
      $resulttypecounts_diluted_logsourcespanlength_meanstat{$rule[2]} += $rc[2];
      # update variance of log source span length statistics using += magic
      $resulttypecounts_diluted_logsourcespanlength_variancestat{$rule[2]} += $rc[3];

    } elsif ($rule[0] =~ m/^\@_PHRASE.*/) {
      #handle the phrase count structures in a objects that handle phrasecounts
      #source_counts, sourceandarg_counts, rulebody_counts - target_counts
      #dont need to read in (or write out all this stuff if you are not using it
      if ($UseCountMerge) {
	handlePhraseCount(\@rule);
      }
    } elsif ($rule[2] =~ /^\@_FORCE(.*)/) {
      # it's a forced rule (e.g. special number handling
      # make sure that it gets wonderful scores

      # first remove '@_FORCE' from the result category to give it its real name
      $rule[2] = $1;

      my @src = split(' ',$rule[0]);
      my @tgt = split(' ',$rule[1]);
      my @costs = (0,		# glue rule feature
		   1,		# rule-app-count
		   scalar(@tgt), # target-word count
		   (0) x ($lexical_costs_offset - 3), # everything up to the lexical weights
		   (-10*scalar(@src)) x ($originalfeaturecount+$additionalfeaturecount-$lexical_costs_offset) # the lexical weights
		  );
      # output the rule right away, so that it can't get filtered out in the second loop
      $rule[3] = join(' ',@costs);
      outputrule($rule[0],join("#",@rule[1..3]));
      
   } else {

      if (scalar(@rule) != 4) {
	die "Unknown rule format snum=$sentencenumber: $_\n";
      }	

      if (!$AbstractOverX && ($rule[0] =~ /\@_X/)) {
	# _X occurs in rule's body - skip rule
      } elsif ($rule[2] eq '@_G' || $rule[0] =~ /^\@_G/ || $rule[0] eq '_UNKNOWN' || 
	       (!$KeepNaturalSentenceRules && $rule[2] eq '@_S')) {
	# skip enter / exit glue rules as well as unknown-rules
	# in case $KeepOccurrenceCountFeature is false, skip also natural _S result rules
      } else {
	if ($rule[2] eq '@_S' && $UseBOSandEOS) {
	  # add BOS and EOS markers if not there yet
	  $rule[0] = '<s> ' . $rule[0] . ' </s>' if !($rule[0] =~ /^<s>/);
	  $rule[1] = '<s> <s> ' . $rule[1] . ' </s> </s>' if !($rule[1] =~ /^<s>/);
	}

	my @rulesourcewords = split(" ", $rule[0]);
	my @rulesourcewords_after_punct_removal = @rulesourcewords;
	if ($RemoveSourcePunctuation) {
	  # For certain feature calculations, already remove the source punctuation.
	  # For others (like lexical weight, keep it in for now). That's why we use both
	  # @rulesourcewords and @rulesourcewords_after_punct_removal
	  #  @rulesourcewords_after_punct_removal = grep{!/^([[:punct:]]|\xef\xb9\x90)$/} @rulesourcewords;
	  @rulesourcewords_after_punct_removal = grep{!/^([:,.]|\xef\xb9\x90)$/} @rulesourcewords;

	  # the source-side might have become empty by this action - in that case move on to the next rule
	  next FIRSTLOOP if scalar(@rulesourcewords_after_punct_removal) == 0;
	}

	#restrict rules to those that have sources that occur in the source data + abstract elememnts
	my $vocMatched = 1;	#by default it passes
	my $phraseToCheck = '';	#used for the voc test
	#checking for mixed punctuation and words
	my $hasPunct_src =0;
	my $hasWords_src =0;
	#compute the abstraction count
	my $abstractioncount = 0;
	#check if two adjacent nonterminals occur
	my $foundadjacentsourcenonterminals = 0;
	#	my $foundadjacenttargetnonterminals = 0;
	my $lastwordwasnonterminal = 0;
	foreach (my $i = 0; $i < scalar @rulesourcewords_after_punct_removal; $i++) {
	  if ($rulesourcewords_after_punct_removal[$i] =~ /^\@/ ) {
	    # source word starts with @ -- must be an arg marker
	    $abstractioncount++;
	    $foundadjacentsourcenonterminals = 1 if $lastwordwasnonterminal;
	    $lastwordwasnonterminal = 1;
	    # this abstraction stops a consecutive phrase, so check the current phrase
	    unless ($phraseToCheck eq '' || $vocCheck{$phraseToCheck}) {
	      $vocMatched = 0;
	      my $bequiet = 1;
	      if (!$bequiet) {
		print STDERR "Reject rule '$_' due to failed testset restriction for phrase '$phraseToCheck'\n" unless $phraseToCheck =~ /^<s>/;
	      }
	    }
	    $phraseToCheck = ''; # reset phraseToCheck
	  } else {
	    $lastwordwasnonterminal = 0;
	    if ($vocabFile) {
	      # append the current word to the phrase-to-check
	      $phraseToCheck .= ($phraseToCheck eq '' ? '' : ' ') . $rulesourcewords_after_punct_removal[$i];
	    }
	    if (!$keepunbalancedpunctuationrules) {
	      if ($rulesourcewords_after_punct_removal[$i] =~ /^[,\!\"\'\`\%\(\)\+\-\/\:\;]+$/) {
		$hasPunct_src++; #modify to have an actual number
	      } else {
		$hasWords_src=1;
	      }
	    }
	    # possibly add a Null rule for this source word
	    if ( $NullRules &&
		 (!$vocabFile || $vocCheck{$rulesourcewords_after_punct_removal[$i]}) &&
		 !($source_vocab{$rulesourcewords_after_punct_removal[$i]}++)
	       ) {
	      # the source word is relevant and new -- create a Null rule for it
	      my @costs = (0,		# glue rule feature
			   1,		# rule-app-count
			   1, # target-word count
			   (0) x ($nullrulefeatureoffset - 3), # everything up to $nullrulefeatureoffset
			   1, # Null rule feature
			   (0) x ($originalfeaturecount+$additionalfeaturecount-$nullrulefeatureoffset-1) # all the rest
			  );
	      # output the rule right away, so that it can't get filtered out in the second loop
	      outputrule($rulesourcewords_after_punct_removal[$i],'_NULL#@_X#' . join(' ',@costs));
	      # make sure that @_X really is a confirmed NT, so that it definitely gets a glue rule later
	      ++$resulttypes_confirmed{'@_X'};      
	    }
	  }
	}
	# have to check one more time, for last phrase of the rule
	unless ($phraseToCheck eq '' || $vocCheck{$phraseToCheck}) {
	  $vocMatched = 0;
	  my $bequiet = 1;
	  if (!$bequiet) {
	    print STDERR "Reject rule '$_' due to failed testset restriction for phrase '$phraseToCheck'\n" unless $phraseToCheck =~ /^<s>/;
	  }
	}

	if ($vocMatched) {
	  my @ruletargetwords = split(" ", $rule[1]);

	  # target side monotonicity and punctuation check
	  my $hasPunct_tgt =0;
	  my $hasWords_tgt =0;
	  my $is_monotonic = 1;
	  my $last_arg_number = 0;
	  #if (!$keepunbalancedpunctuationrules) 
	  for (my $i = 0; $i < scalar @ruletargetwords; $i++) {
	    if ($ruletargetwords[$i] =~ /^\@(.*)$/ ) {
	      # abstraction marker
	      if ($1 < $last_arg_number) {
		$is_monotonic = 0;
	      }
	      $last_arg_number = $1;
	    } else {
	      if ($ruletargetwords[$i] =~ /^[,\!\"\'\`\%\(\)\+\-\/\:\;]+$/) {
		$hasPunct_tgt++;
	      } else {
		$hasWords_tgt=1;
	      }
	    }
	  }			# for
	  

	  my $punctPassed = $keepunbalancedpunctuationrules || 
	    ($hasWords_src == $hasWords_tgt && $hasPunct_src == $hasPunct_tgt);
	  if (!$punctPassed) {
	    print STDERR "Reject Unbalanced by number punctuation for rule: $_ (hasPunct_tgt=$hasPunct_tgt, hasPunct_src=$hasPunct_src, hasWords_tgt=$hasWords_tgt, hasWords_src=$hasWords_src)\n"       if ($hasPunct_src != $hasPunct_tgt && $hasPunct_src >0 && $hasPunct_tgt > 0);
	  }
	
	  # count source- and target-terminal symbols
	  my $sourcewordcount = scalar(@rulesourcewords) - $abstractioncount;
	  my $sourcewordcount_after_punct_removal = scalar(@rulesourcewords_after_punct_removal) - $abstractioncount;
	  my $targetwordcount = scalar(@ruletargetwords) - $abstractioncount;
	  die "In rule $_: sourcewordcount ($sourcewordcount) longer than maxsourcephraselength ($maxsourcephraselength)!" if $sourcewordcount > $maxsourcephraselength && $maxsourcephraselength;
	
	  # check whether rule is balanced
	  my $isbalanced = 
	    ( (($sourcewordcount == 0) == ($targetwordcount == 0))
	      # rule has no source terminals iff rule has no target terminals
	    );

	  
	  if ($targetwordcount>0 && $sourcewordcount > 0) {
	    # check if source / target ratio is okay
	    my $ratioSrcTgt = $sourcewordcount / $targetwordcount;
	    if ( $ratioSrcTgt>$bal_limit || $ratioSrcTgt<1.0/$bal_limit) {
	      $isbalanced = 0;
	    }
	  }
	  #      print STDERR "Bad source/target length ratio for rule: $_\n" if !$isbalanced && ($vocMatched && $rulefreq >= $minoccurrencecount);


	  my @rulecostvector  = split(" ", $rule[3]);

	  #THE REAL RULE FREQ - BUT ALSO FOLLOWED BY A SIMPLE (+1) style count
	  my $rulefreq;
	  my $simplerulefreq;

	  if ($is_phrasepair) {
	    # we don't know the rule's occurrence count - just write 1
	    $rulefreq = 1;
	  } elsif ($UnknownInputRuleFile) {
	    # we don't know the rule's occurrence count - just write 1
	    $rulefreq = 1;
	    # prepend glue feature, rule application counter and target word count feature to the rule cost vector
	    # and then zeros so that we get the regular n.o. features $originalfeaturecount
	    my @old_rulecostvector = @rulecostvector[$InputRuleFileIgnoreFirstNFeatures..$#rulecostvector];
	    my $number_of_zeros_to_pad = $originalfeaturecount - 3 - scalar(@old_rulecostvector);
	    die(
		(-$number_of_zeros_to_pad) . 
		" too many features in input rule file's rule feature vector $rule[3]. " .
		"Simply increase your --PhrasalFeatureCount parameter by that amount."
	       ) if $number_of_zeros_to_pad < 0;
	    if (!$adjusted_sourcecondrf_offset) {
	      # adjust the offset for the source cond. RF feature
	      $InputRuleFileSourceCondRFFeatureOffset += 3 + $number_of_zeros_to_pad;
	      $adjusted_sourcecondrf_offset = 1;
	      print STDERR "--UnknownInputRuleFile option was passed. Prepended glue feature, rule app counter, target word counter, and $number_of_zeros_to_pad zeros. Adjusted InputRuleFileSourceCondRFFeatureOffset to $InputRuleFileSourceCondRFFeatureOffset\n";
	    }
	    @rulecostvector = (0, 1, 0);
	    while ($number_of_zeros_to_pad--) {
	      push @rulecostvector, 0;
	    }
	    push @rulecostvector, @old_rulecostvector;
	    die 'assertion error' if scalar(@rulecostvector) != $originalfeaturecount;
	  } else {
	    #we know this is a proper SMT style rule! - dont round the posterior frequency down
	    $rulefreq = $NegLogOccurrenceCountFeature? exp(-$rulecostvector[6]) : $rulecostvector[6]; # rule frequency count
	    if ($HasSimpleCount) {
	      #round down the simple count, otherwise you end up with all these 2.0000001 freq (which we keep as strings)
	      $simplerulefreq = $NegLogOccurrenceCountFeature? sprintf("%.2f", exp(-$rulecostvector[7])) : $rulecostvector[7]; # rule frequency count
	    }
	    #die "assertion error: rulecostvector: ", scalar(@rulecostvector), " orig $originalfeaturecount\n" if scalar(@rulecostvector) != $originalfeaturecount;
	    
	  }
	  #print "Processing rule: ", join('#', @rule), " rulefreq = $rulefreq simplerulefreq $simplerulefreq\n";
	  if ($rulefreq==0.0) {
	    #if you dont have the post vs simple distinction, then this is really bad
	    if (!$HasSimpleCount) {
	      die "Dropping rulefreq=0 (logspace=",$rulecostvector[6], ") Raw rule: ", join('#', @rule), "\n";
	    } else {
	      #This means the rule freq was WAY low - it really should happen like this since we started out in prob space!
	      print STDERR "WARNING (dropping) rulefreq=0 (logspace=",$rulecostvector[6], ") Raw rule: ", join('#', @rule), "\n";
	      next;
	    }
	  }

	  if ( 
	      # first condition: balancedness
	      ( $isbalanced || $keepunbalancedrules
	      ) &&
	      # second condition: rule needs to be either abstract or
	      # have at least one source terminal in it b/c otherwise
	      # the decoding process is immensely slowed down
	      # (also abstract rules slow down decoding and can be removed by
	      # passing --noAllowAbstractRules)
	      ( $AllowRulesWithOnlyTargetTerminals || !($sourcewordcount_after_punct_removal == 0 && $targetwordcount > 0)
	      ) &&
	      (
	       # third condition: should be no glue rule
	       $rulecostvector[0] == 0
	      ) &&
	      (
	       # condition: abstraction count
	       $abstractioncount <= $maxabstractioncount
	      ) &&
	      (
	       # condition: n.o. source and target side symbols for non-lexical rules
	       $abstractioncount == 0 || 
	       (scalar(@rulesourcewords) <= $MaxSourceSymbolsNonlexicalRules &&
		scalar(@ruletargetwords) <= $MaxTargetSymbolsNonlexicalRules
	       )
	      ) &&
	      (
	       # condition: non-abstract rule
	       $AllowAbstractRules || !($sourcewordcount_after_punct_removal == 0 && $targetwordcount == 0)
	      ) &&
	      (
	       # condition: no two adjacent source nonterminals
	       $AllowAdjacentSourceNonterminals || !($foundadjacentsourcenonterminals)
	      ) &&
	      #	      (
	      #	       # condition: no two adjacent target nonterminals
	      #	       $AllowAdjacentTargetNonterminals || !($foundadjacenttargetnonterminals)
	      #	      ) &&
	      (
	       # condition: punctuation balanced (or $keepunbalancedpunctuationrules == 1)
	       $punctPassed
	      )
	     ) {
	      
	    #debuggin 
	    #if ($keepunbalancedrules && !$isbalanced){
	    #  if ($abstractioncount>2){
	    #      print "Lets start: ", $orgrule, " ABS $abstractioncount MAX $maxabstractioncount\n";
	    #      exit;
	    #  }
	    #}

	    # update total rule instance counter, use the simple one here (since we call -- on it later)
	    #if there is no simple count, then use the real count
	    $ruleinstancecount += (!$HasSimpleCount) ? $rulefreq : $simplerulefreq; 
	      
	    # add result category to our confirmed-nonterminals set using ++ magic
	    ++$resulttypes_confirmed{$rule[2]};

	    if (scalar(@rulecostvector) > $originalfeaturecount) {
	      # cut off all features beyond the original feature count in case there are any
	      print STDERR "Warning: some rules have more than $originalfeaturecount features. Cutting off all these additional features. If that's not intended, correct your parameter --PhrasalFeatureCount\n" if !($warned_too_many_features++);
	      $#rulecostvector = $originalfeaturecount - 1;
	    } else {
	      #	      print STDERR "Rulecostvector size $#rulecostvector \n"; #!!!!
	      # append as many zeros as necessary
	      #	      for (my $i = 0; $i < $originalfeaturecount-scalar(@rulecostvector); $i++) {
	      while (scalar(@rulecostvector) < $originalfeaturecount) {
		push @rulecostvector, 0;
	      }
	    }

	    if (!$UnknownInputRuleFile) {
	      # put rule occurrence count feature into frequency space if still in neg-log space
	      $rulecostvector[6] = $rulefreq;
	      if ($HasSimpleCount) {
		$rulecostvector[7] = $simplerulefreq;
	      }
	    }
	    #now rule cost vector has 6,7 in prob space as well!

	    # if in ConsiderSourceSpanImpact mode, we will multiply the concerned feature values by
	    # the number of source symbols (terminals and NTs) of this rule
	    my $ruleimpact = ($ConsiderSourceSpanImpact? scalar(@rulesourcewords) : 1);

	    # set n.o. target words feature
	    if ($targetwordcount != $rulecostvector[2]) {
	      #only output the error msg if EOS BOS adding doesnt explain it...
	      my $tgtCopy = $rule[1];
	      my $eosboscount = 0;	      
	      $eosboscount++ while $tgtCopy =~ /\<\/?s\>/g;
	      
	      print STDERR "Wrong target word counter in feature 2 (index starting with 0) of rule: ", join('#',@rule), " in $sentencenumber, rule cost vector $rule[3]: target word count feature is $rulecostvector[2], but actual target word count is $targetwordcount. Correcting feature to $targetwordcount (EOSBOS=$eosboscount).\n" if !$UnknownInputRuleFile && !$is_phrasepair && $targetwordcount != $rulecostvector[2]+$eosboscount;

	      $rulecostvector[2] = $targetwordcount; #set it to the correct value
	    }

	    # set lexical-feature ($lexicalfeatureoffset == $originalfeaturecount)
	    $rulecostvector[$lexicalfeatureoffset] = $ruleimpact * ($abstractioncount == 0? 1 : 0);

	    # set abstract-feature
	    #$rulecostvector[$abstractfeatureoffset] = $ruleimpact * ($sourcewordcount == 0 && $targetwordcount == 0? 1 : 0);
	    # set adjacent-source-nonterminals feature
	    $rulecostvector[$adjacentsourceNTfeatureoffset] = $ruleimpact * ($foundadjacentsourcenonterminals? 1 : 0);

	    # set _X rule feature
	    $rulecostvector[$xrulefeatureoffset] = $ruleimpact * ($rule[2] eq '@_X'? 1 : 0);

	    # set unbalancedness penalty 1
	    $rulecostvector[$unbalancednesspenalty1offset] = abs($MeanTargetSourceRatio * $sourcewordcount - $targetwordcount) ;

	    # set unbalancedness penalty 2
	    $rulecostvector[$unbalancednesspenalty2offset] = 
	      ($sourcewordcount == 0 || $targetwordcount == 0 ?
	       0 : 
	       scalar(@rulesourcewords) * (log($targetwordcount / $MeanTargetSourceRatio / $sourcewordcount))**2
	      );

	    $rulecostvector[$onlysourceterminalsfeatureoffset] = $ruleimpact * ((($sourcewordcount > 0) && ($targetwordcount == 0))? 1 : 0);

	    #	    $rulecostvector[$onlytargetterminalsfeatureoffset] = $ruleimpact * ((($sourcewordcount == 0) && ($targetwordcount > 0))? 1 : 0);

#	    $rulecostvector[$sourcepunctuationremovedfeatureoffset] = $ruleimpact * (($sourcewordcount != $sourcewordcount_after_punct_removal)? 1 : 0);
	    $rulecostvector[$unbalancedpunctuationfeatureoffset] = $ruleimpact * 
	      (	( $rule[0] =~ /(^| )[,\!\"\'\`\%\(\)\+\-\/\:\;]/ ) != ( $rule[1] =~ /(^| )[,\!\"\'\`\%\(\)\+\-\/\:\;]/ ) ? 1 : 0);

	    # set non-monotonicity penalty
	    $rulecostvector[$nonmonotonicitypenaltyoffset] = $ruleimpact * ($is_monotonic? 0 : 1);
	    

	    #NOTE: by this point, the freqs are in prob space!!!
	    die "Assertion error (1): freqs is in prob space but is: rulefreq $rulefreq in rule",join("\#",@rule), "\n" if ($rulecostvector[6]==0);
	    die "Assertion error (1): simplefreq is in prob space but is: $simplerulefreq\n" if ($HasSimpleCount && $rulecostvector[7]==0);
	    #exit;
	    
	    # are we dealing with a phrasepair to be merged in?
	    if ($is_phrasepair) {
	      # check if there's any lexical rule whose body is identical to the phrasepair
	      my $tmpresultstr = $lexical_rule_bodies{join('#', @rule[0..1])};
	      if (defined($tmpresultstr)) {
		# yes there is - retrieve all of them
		# but first undo the rule instance counter increase done above
		$ruleinstancecount--;
		my @tmpresultsarray = split(' ',$tmpresultstr);
		foreach my $tmpresult (@tmpresultsarray) {
		  # look up rule in rule database %complexrules
		  my $keytmp = join('#', (@rule[0..1], $tmpresult));
		  my $valuetmp = $complexrules{$keytmp};
		  if (!defined($valuetmp)) {
		    die "Assertion error."
		  } else {
		    #  my @tmpcostvector = split(' ', $valuetmp);
		    my @tmpcostvector = unpack('f*', $valuetmp);
		    if (scalar(@tmpcostvector) > $merge_phrasetable_feature_offset) {
		      print STDERR "Warning: phrase table to be merged in doesn't appear to be uniqued; concerning phrase pair: $keytmp\n";
		      $#tmpcostvector = $merge_phrasetable_feature_offset-1;
		    }
		    #		    $valuetmp = join(' ', @tmpcostvector) . 
		    #		      concatenatenumber(0, $merge_phrasetable_feature_offset-scalar(@tmpcostvector)) .
		    #			" $ruleimpact" . # SAMT feature
		    #			  " $ruleimpact" . # phrasetable feature .
		    #			    ' ' . join(' ',@phrasepair_features);
		    #		    $complexrules{$keytmp} = $valuetmp;
		    $complexrules{$keytmp} = pack('f*', (@tmpcostvector, 
							 (0) x ($merge_phrasetable_feature_offset-scalar(@tmpcostvector)),
							 $ruleimpact, # SAMT feature
							 $ruleimpact, #phrasetable feature
							 @phrasepair_features));
		  }
		}
	      } else {
		# there's no lexical rule matching this phrasepair --
		# add the phrasepair on its own to the rule database
		$complexrules{join('#', @rule[0..2])} = 
		  #		  join(' ', @rulecostvector) . 
		  #		    concatenatenumber(0, $merge_phrasetable_feature_offset-scalar(@rulecostvector)) .
		  #		      ' 0' .	# SAMT 0/1 feature
		  #			" $ruleimpact" . # phrasetable 0/1 feature .
		  #			  ' ' . join(' ',@phrasepair_features);
		  pack('f*', (@rulecostvector,
			      (0) x ($merge_phrasetable_feature_offset-scalar(@rulecostvector)),
			      0, # SAMT feature
			      $ruleimpact, #phrasetable feature
			      @phrasepair_features));
	      }
	    } else {		# else branch of: if ($is_phrasepair)
	    
	      # look up rule in so-far created rule database %complexrules
	      my $keytmp = join('#', @rule[0..2]);
	      my $valuetmp = $complexrules{$keytmp};
	      #valuetmp has 6,7 fields as probs already
	      if (!defined($valuetmp)) {
		# it's a new rule
		
		# dealing with a regular and lexical rule?
		if (!$is_phrasepair && $abstractioncount == 0) {
		  # yes; store its body and result category
		  # this is in addition to the regular storage of the rule in %complexrules
		  my $tmpresultstr = $lexical_rule_bodies{join('#', @rule[0..1])};
		  if (defined($tmpresultstr)) {
		    $tmpresultstr .= ' ' . $rule[2];
		  } else {
		    $tmpresultstr = $rule[2];
		  }
		  $lexical_rule_bodies{join('#', @rule[0..1])} = $tmpresultstr;
		}

		#		print STDERR join(' ',@rulecostvector) . "\n";
		$valuetmp = pack('f*', @rulecostvector);
		#	    print "Adding $valuetmp...\n";
	      } else {
		# this case occurs when we're combining rule files - we have to merge the cost vectors now
		my @tmpcostvector = unpack('f*', $valuetmp);
		#tmpcostvector has 6,7 as probs
		if ($UnknownInputRuleFile) {
		  # max all costs to minimize impact of inconsistency problems
		  # the first 3 fields are indicators about the rule, dont worry about these
		  for (my $i = 3; $i < $originalfeaturecount; $i++) {
		    if ($rulecostvector[$i] > $tmpcostvector[$i]) {
		      $tmpcostvector[$i] = $rulecostvector[$i];
		    }
		  }
		} else {
		  # calculate weighted average of the costs (not the freq terms) in prob space
		  #rulefreq, simplerulefreq are in probspace right now
		  #tmpcostvector: the rule vector stored in complexrules (with 6,7 as probs)
		  #we want to simulate what would have happenned in extractrules when an item is foundin complexrules, ie
		  #sum the frequencies in prob space, and weight the rest of the score with SIMPLE counts ONLY
		  #complex counts should never be used - if you have set HasSimpleCount ensure that you use the simple count
		  #which is in spot 7. If you have HasSimpleCount, you can use the simplefreq (since u have set it)
		  #otherwise your rulefreq is only (simple) freq you have


		  my $existingFrequency;
		  my $newruleFrequency;
		  if ($HasSimpleCount && $MergeWithPosterior) {
		    $existingFrequency =  $tmpcostvector[6];
		    $newruleFrequency = $rulefreq;
		  } else {
			
		    $existingFrequency = ($HasSimpleCount) ? $tmpcostvector[7] : $tmpcostvector[6];
		    $newruleFrequency = ($HasSimpleCount) ? $simplerulefreq : $rulefreq;
		  }

		  my $normalization_constant = log($existingFrequency + $newruleFrequency);
		  #go thru each field that is NOT an indicator function: ie 3 and above, but NOT the freq field (6 and maybe 7)
		  for (my $i = 3; $i < $originalfeaturecount; $i++) {
		    if ($i != 6 && ($i!=6+($HasSimpleCount?1:0)) # leave out the frequency field
			&& $rulecostvector[$i] != $tmpcostvector[$i] # calculation is only necessary if the costs differ
		       ) {
		      #take the weighted average
		      #update the existing cost vector - the one that will be finally used
		      $tmpcostvector[$i] = -log(
						exp(-$tmpcostvector[$i]) * $existingFrequency + exp(-$rulecostvector[$i]) * $newruleFrequency
					       ) + $normalization_constant;
		    }
		  }
		  $rulefreq += $tmpcostvector[6]; # revise this rule's frequency
		  $simplerulefreq += $tmpcostvector[7] if ($HasSimpleCount); #revise this rule's simple freq too!
		  #update the existing cost vector - the one that will be finally used
		  $tmpcostvector[6] = $rulefreq; # store revised rule freq
		  $tmpcostvector[7] = $simplerulefreq if ($HasSimpleCount);
		}
		#valuetmp contains the updated cost after merging
		#	$valuetmp = join(' ', @tmpcostvector);
		$valuetmp = pack('f*', @tmpcostvector);
	      }
	      $complexrules{$keytmp} = $valuetmp;
	    }			# end of: if ($is_phrasepair)

	    # check if we have set a new record with the occurrence count of this rule amongst all rules with the same source
	    # Use posterior frequency since we are checking to see which is the rule with the highest freq among same source, 
	    # rather than simply checking if we have seen it enough
	    my $usethisfreq = $PosteriorBasedRFPruning || !$HasSimpleCount? $rulefreq:$simplerulefreq;
	    my $MaxOccCount = $complexrules{$rule[0]};
	    if (!defined($MaxOccCount) || $usethisfreq > $MaxOccCount) {
	      # new record - store occurrence count
	      $complexrules{$rule[0]} = $usethisfreq;
	    }
	  }
	} else {
	  #	  print STDERR "Kicking out PT rule $_\n" if $rule[0] =~ /PT/;
	}			# if vocMatched
      }
    }
  }

  # adjust $additionalfeaturecount
  $additionalfeaturecount += $merge_phrasetable_feature_count + 2 if $MergePhrasetable;
  # check if feature restriction vector is correct - we can do this here only b/c 
  # only here we know the final number of features
  if ($RestrictToFeatures) {
    # first remove comments from Restriction vector
    $RestrictToFeatures =~ s/\#[^_]*//g;
    @featureRestrictionVector = split('_', $RestrictToFeatures);
    die "Wrong length of feature restriction vector $RestrictToFeatures" if scalar(@featureRestrictionVector) != $originalfeaturecount+$additionalfeaturecount;
  }

  {
    # write out the rules in %complexrules
    print STDERR "\nLeft with " . scalar(keys %complexrules) . " rules. Writing out the rules...\n";
    my $totalrulefreqcount_sanitytest = 0;
    my $rulenumber = 0;
    # note: the following works only if using Berkeley DB:
    # we'll iterate through the B-Tree in sorted order
    # if $RestrictResultCategoriesPerRuleBody==N where N>0, we'll accumulate the results and costvectors of rules
    # with identical rule bodies and only write out rules for the N most frequent rules
    my $lastrulesource = ''; my $lastruletarget = ''; # remembers the rule body we're currently at
    my @lastruleinfos = ();	# frequency/result/costvector triples of all accumulated rules with same rule body
    my ($key, $value);
  LABEL_2ND_PASS:
    while (($key,$value) = each %complexrules) { 
      my ($sourcestr,$targetstr,$result) = split("#", $key);
      if (!defined($targetstr)) {
	# this is just the max-occurrence count statistics - ignore this entry
      } else {
	$rulenumber++;
	print STDERR "Processing rule number $rulenumber...\r" if $rulenumber % 10000 == 0;
    
	my @rulecostvector = 
	  #split(" ", $value);
	  unpack('f*', $value);
	#print "Retrieved for $key => $value\n"; 

	# determine whether we're talking about a phrase pair rule that is not an SAMT rule
	my $is_phrasepair_but_not_samt_rule = $rulecostvector[$merge_phrasetable_feature_offset+1] && !$rulecostvector[$merge_phrasetable_feature_offset];

	#The rulecostvector has the freqs in prob space
	my $rulefreq = $rulecostvector[6]; # rule frequency count
	my $simplerulefreq = ($HasSimpleCount? $rulecostvector[7] : $rulefreq); # rule frequency count

	#print "READ (probspace!!) freq: $rulefreq simple $simplerulefreq\n";
	die "Assertion error (2): freqs is in prob space but is: rulefreq $rulefreq: key $key => $value\n" if ($rulefreq==0);
	die "Assertion error (2): simplefreq is in prob space but is: $simplerulefreq\n" if ($HasSimpleCount && (!defined($simplerulefreq) || $simplerulefreq==0));
	#exit;
	
	$totalrulefreqcount_sanitytest += (!$HasSimpleCount) ? $rulefreq :$simplerulefreq;
	#$totalrulefreqcount_sanitytest += $rulefreq;

	print STDERR 'Sanity check failed for rule ' . "$key\#$value (rulefreq=$rulefreq > $complexrules{$sourcestr}=bestrulefreq!" if ($rulefreq > $complexrules{$sourcestr} + 0.1);

	my $usethisfreq = $PosteriorBasedRFPruning || !$HasSimpleCount? $rulefreq:$simplerulefreq;
	my @rulesourcewords = split(' ',$sourcestr);
	my @ruletargetwords = split(' ',$targetstr);
	my @rulesourcewords_lexical = grep {!/^@/} @rulesourcewords;
	my @ruletargetwords_lexical = grep {!/^@/} @ruletargetwords;

	my $costDelta = 0;
	#there must be some lexical component, otherwise this is an abstract rule
	if (scalar(@ruletargetwords_lexical) >0 || scalar(@rulesourcewords_lexical) > 0) {
	  if (scalar(@rulesourcewords_lexical)==0) {
	    #dealing with the special case here where is an empty src going to multiple target, we assume one extra word is ok
	    $costDelta = abs(log(scalar(@ruletargetwords_lexical)/1));
	  }

	  if (scalar(@ruletargetwords_lexical)==0) {
	    #dealing with the special case here where is an empty src going to multiple target, we assume one extra word is ok
	    $costDelta = abs(log(scalar(@rulesourcewords_lexical)/1));
	  }

	  if (scalar(@ruletargetwords_lexical)>0 && scalar(@rulesourcewords_lexical)>0) {
	    my $realTargetSourceRatio = scalar(@ruletargetwords_lexical)/scalar(@rulesourcewords_lexical);
	    $costDelta = abs(log($realTargetSourceRatio/$MeanTargetSourceRatio));
	  }
	}
	
	if ($UseCountMerge) { 
	  #update all the relative frequencies based on updated merged counts - note we are doing everything
	  #in the space of smoothed counts here
	  updateMergedRelativeFreqs(\@rulecostvector, \@rulesourcewords, \@ruletargetwords, $result, $rulefreq, $simplerulefreq);
	}


	my $srcCondRelFreqCosPruningCost = $rulecostvector[$InputRuleFileSourceCondRFFeatureOffset] + $LengthBoostAlpha*$costDelta;
	#die "UH: unequal values: " if  ($srcCondRelFreqCosPruningCost != $rulecostvector[$InputRuleFileSourceCondRFFeatureOffset]);
	$_ = $result;
	my $numPlusses = 0;
	while (/\+/g) {
	  ++$numPlusses;
	}
	my $numCats = $numPlusses +1;
	
	my $beamfactor = ($rulecostvector[$lexicalfeatureoffset]) ? $beamfactorlexicalrules : $beamfactornonlexicalrules;
	my $bestfreq = $complexrules{$sourcestr};
	
	my $debug = 0;
	if ($debug) {
	  my $beamtest =  ($usethisfreq >= $bestfreq*$beamfactor);
	  my $probtest = ($srcCondRelFreqCosPruningCost <= ($rulecostvector[$lexicalfeatureoffset]? $minlexsourcecondrflexicalrules_cost : $minlexsourcecondrfnonlexicalrules_cost));
	  my $freqtest = ($simplerulefreq + 0.1 >= ($rulecostvector[$lexicalfeatureoffset]? $minoccurrencecountlexicalrules : $minoccurrencecountnonlexicalrules));
	  my $besttest = ($usethisfreq + 0.0001 >= $bestfreq);

	  if (($beamtest && $probtest && $freqtest) || ($besttest)) {
	    print STDERR "PASSED b($beamtest) p($probtest) f($freqtest) or best($besttest) freq $usethisfreq best: $bestfreq : ", $sourcestr, "#",$targetstr, " costs: ", join(' ', @rulecostvector), "\n"; 
	  } else {
	    print STDERR "FAILED b($beamtest) p($probtest) f($freqtest) or best($besttest) : ", $sourcestr, "#",$targetstr, " costs: ", join(' ', @rulecostvector), "\n"; 
	  }
	}

	if (
	    #check conditions on the rule
	    ( 
	     #check that the rule is good enough against its alernatives
	     (
	      ($usethisfreq >= $bestfreq*$beamfactor)
	      &&
	      ($srcCondRelFreqCosPruningCost <= ($rulecostvector[$lexicalfeatureoffset]? $minlexsourcecondrflexicalrules_cost : $minlexsourcecondrfnonlexicalrules_cost))
	      &&
	      ($simplerulefreq + 0.1 >= ($rulecostvector[$lexicalfeatureoffset]? $minoccurrencecountlexicalrules : $minoccurrencecountnonlexicalrules))
	     )
	     #or you are the best rule for your src side AND lexical, then you are also thru
	     ||
	     ($rulecostvector[$lexicalfeatureoffset] && $usethisfreq + 0.0001 >= $bestfreq)
	    )
	    && ($numCats <= $MaxCategories) #limits the number of categories per result
	   ) {
	  
	    

	  
	  # if in ConsiderSourceSpanImpact mode, we will multiply the concerned feature values by
	  # the number of source symbols (terminals and NTs) of this rule
	  my $ruleimpact = ($ConsiderSourceSpanImpact? scalar(@rulesourcewords) : 1);
	  
	  # re-compute result-conditioned neg log prob
	  # ARV CHANGED
	  # The diluted count is the important (smooth prob) one!
	  
	  if (!$UnknownInputRuleFile && !$is_phrasepair_but_not_samt_rule) {
	    if (defined($resulttypecounts_diluted{$result})) {
	      $rulecostvector[3] = 
		-log($rulefreq/$resulttypecounts_diluted{$result}) * $ruleimpact;
	    } else {
	      print STDERR "Warning: no _COUNT statistics for '$result'. Concerning rule: $key\#$value.\n";
	      $rulecostvector[3] = 0;
	    }
	  }



	  # add some more features

	  # rareness penalty
	  # 1 for freq-one rules, then levelling off with slope determined 
	  # by feature weight, with limit 0 for $rulefreq --> infinity
	  $rulecostvector[$rarenesspenaltyoffset] = $ruleimpact * exp(1-$rulefreq);

	  #	  # squared neg log result-cond. RF
	  #	  $rulecostvector[$originalfeaturecount+5] = $rulecostvector[3] ** 2;

	  # compute diluted-result-conditioned neg log prob
	  # ARV CHANGED to use the simple count to calculate this!
	  if ((defined($resulttypecounts_undiluted{$result}) && $resulttypecounts_undiluted{$result}==0)) {
	    print STDERR "WARNING: result = $result undiluted = 0\n";
	  }
	  if ($HasSimpleCount && !defined($simplerulefreq)) {
	    print STDERR "ERROR simplerulefreq not defined $key => $value: \n";
	  }
	  if (!defined($result)) {
	    print STDERR "ERROR result not defined $key => $value: \n";
	  }
	  #if you only have a single count option, then we set a blank for this backup field
	  #which is the simple count based result conditioned relative freq
	  if ($UnknownInputRuleFile || $is_phrasepair_but_not_samt_rule || !defined($resulttypecounts_undiluted{$result}) || !$HasSimpleCount) {
	    $rulecostvector[$undilutedresultcountoffset] = 0;
	  } else {
	      
	    #print "RULE: $key => $value\n";
	    $rulecostvector[$undilutedresultcountoffset] = -log($simplerulefreq/$resulttypecounts_undiluted{$result}) * $ruleimpact;
	  }

	  $rulecostvector[$nullrulefeatureoffset] = 0;
	  

	  #	  # repetition of phrase costs, zeroed out for non-lexical rules
	  #	  $rulecostvector[$originalfeaturecount+7] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[7] : 0;
	  #	  $rulecostvector[$originalfeaturecount+8] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[8] : 0;
	  #	  $rulecostvector[$originalfeaturecount+9] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[9] : 0;
	  #	  $rulecostvector[$originalfeaturecount+10] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[10] : 0;
	  #	  $rulecostvector[$originalfeaturecount+11] = $rulecostvector[$lexicalfeatureoffset]? $rulecostvector[11] : 0;

	  # change occurrence count feature back into neg-log space
	  $rulecostvector[6] = -log($rulefreq) if !$UnknownInputRuleFile;
	  #ARVCHANGED, do this for both types of frequency!
	  $rulecostvector[7] = -log($simplerulefreq) if (!$UnknownInputRuleFile && $HasSimpleCount);
	  

	  # the lexical costs
	  @rulecostvector[$lexical_costs_offset..$lexical_costs_offset + $lexical_features_per_entry_count*$lexical_weight_entry_count -1] =
	    compute_lexical_costs(\@rulesourcewords, \@ruletargetwords, $ruleimpact);

	  if ($MergePhrasetable) {
	    # possibly have to pad the rule cost vector
	    if (scalar(@rulecostvector) < $merge_phrasetable_feature_offset) {
	      die "Assertion error: rule cost vector " . join(' ',@rulecostvector) . " smaller than merge_phrasetable_feature_offset ($merge_phrasetable_feature_offset)";
	    } elsif (scalar(@rulecostvector) == $merge_phrasetable_feature_offset) {
	      # this is a pure SAMT rule
	      $rulecostvector[$merge_phrasetable_feature_offset] = 1 * $ruleimpact; # SAMT feature
	      $rulecostvector[$merge_phrasetable_feature_offset+1] = 0; # phrasetable feature
	      for (my $i=0; $i < $merge_phrasetable_feature_count; $i++) {
		push @rulecostvector, 0;
	      }
	    }
	  }
	  
	  # make sure that the rule cost vector is full
	  die "Assertion error: rule cost vector " . join(' ',@rulecostvector) . " of wrong size (should be size $originalfeaturecount+$additionalfeaturecount)" if scalar(@rulecostvector) != $additionalfeaturecount+$originalfeaturecount;

	  # output the rule
	  #	  for (my $i = 0; $i<=scalar @rulecostvector; $i++) 
	  #	    {
	  #	      print $rulecostvector[$i] . ' ';
	  #	    } 
	  #	  print "\n";
	  
	  # round the costs in the costvector, except for the frequency count feature
	  #for (m
	  my $coststr = join(' ', @rulecostvector);

	  #we are in the second pass right now, we are no longer doing any merging of rules
	  if ($RemoveSourcePunctuation) {
	    # Now that all the cost calculations have been done, we can remove the source punctuation.
	    # Note that this means that
	    # a) features based on the source word count remain based on the version with source-punctuation
	    # b) multiple identical rules (but with different cost values) might be output
	    # c) re-running filterrules.pl will result in different cost vectors (and merged rules) [this is
	    #    not recommended since the lexical weight calculation will compute large costs for target-
	    #    punctuation due to missing source-punctuation counterparts]
	    #@rulesourcewords = grep{!/^([[:punct:]]|\xef\xb9\x90)$/} @rulesourcewords;
	    @rulesourcewords = grep{!/^([:,.]|\xef\xb9\x90)$/} @rulesourcewords;

	    # the source-side might have become empty by this action - in that case move on to the next rule
	    # this should already have been done above in the first loop
	    # next LABEL_2ND_PASS if scalar(@rulesourcewords) == 0;
	    die if scalar(@rulesourcewords) == 0;

	    # re-create sourcestr
	    $sourcestr = join(' ',@rulesourcewords);
	  }
	      
	  
	  if (!$RestrictResultCategoriesPerRuleBody) {
	    outputrule($sourcestr,join("#",($targetstr,$result,$coststr)));
	  } else {
	    if ($lastrulesource eq $sourcestr && $lastruletarget eq $targetstr) {
	      # we're still treating the same rule body
	    } else {
	      if ($lastrulesource ne '') {
		# write out the previously stored stuff
		# first sort the rule infos by frequency
		my @sortedinfos = sort {$b->[0] <=> $a->[0]} @lastruleinfos;
		if (scalar(@sortedinfos) > $RestrictResultCategoriesPerRuleBody) {
		  # truncate the array
		  $#sortedinfos = $RestrictResultCategoriesPerRuleBody-1;
		}
		# write out the $RestrictResultCategoriesPerRuleBody top rules
		for my $ruleinfo (@sortedinfos) {
		  outputrule($lastrulesource, join('#',($lastruletarget,$ruleinfo->[1],$ruleinfo->[2])));
		}
	      }
	      # update $lastrulesource / $lastruletarget and reset @lastruleinfos
	      $lastrulesource = $sourcestr;
	      $lastruletarget = $targetstr;
	      @lastruleinfos = ();
	    }
	    # add result and costvector to our array
	    push @lastruleinfos, [$rulefreq,$result,$coststr];
	  }
	} else {
	  #	  print STDERR "!!!!Warning!!!! Kicking out PT rule $key\n" if $sourcestr =~ /PT/;
	}
      }
    }
      
    if ($RestrictResultCategoriesPerRuleBody && $lastrulesource ne '') {
      # write out the previously stored stuff
      # first sort the rule infos by frequency
      my @sortedinfos = sort {$b->[0] <=> $a->[0]} @lastruleinfos;
      if (scalar(@sortedinfos) > $RestrictResultCategoriesPerRuleBody) {
	# truncate the array
	$#sortedinfos = $RestrictResultCategoriesPerRuleBody-1;
      }
      # write out the $RestrictResultCategoriesPerRuleBody top rules
      for my $ruleinfo (@sortedinfos) {
	outputrule($lastrulesource, join('#',($lastruletarget,$ruleinfo->[1],$ruleinfo->[2])));
      }
    }

    print STDERR "!!!!!!!!!!!! Something's wrong with the underlying complexrules database! (Maybe an network file system issue?) \$ruleinstancecount ($ruleinstancecount) != \$totalrulefreqcount_sanitytest ($totalrulefreqcount_sanitytest)!!! \n" if abs($ruleinstancecount - $totalrulefreqcount_sanitytest) > 0.1;
  }


  
  #to get a statistics about how many nonterminals there are
  my $nonterminalcount = 0;

  #add some manual rules now
  my $manualrulecount = 0;

  # glue rules: theory is: any result can be transformed into _X
  # (enter glue mode: e.g. @NP#@1#@_X); _X rules can be joined (@_X
  # @_X#@1 @2#@_X); _X can be transformed into any nonterminal (exit
  # glue mode: e.g. @_X#@1#@NP) in practise, avoid doubling the
  # n.o. hyps in the chart tables by using _S instead of _X 
  # also, the exit-glue mode rules will be inserted only when needed
  # directly by the parser (the parser only applies the self-rule to
  # _S only when the whole chart cell contains no hypergraph nodes
  # with non-"_S" result nonterminals)

  #phrase count data in case we need to re-filter this rule table
  if ($UseCountMerge && $PrintPhraseCounts) {
    printPhraseCounts();
  }

  # enter and exit glue-mode rules, one for each nonterminal symbol
  print STDERR "Nonterminals occurring as result categories in the filtered ruleset with posterior count, simple-frequency count, mean source span length: \n";
  my $key;
  my $tmpvalue;
  while (($key,$tmpvalue) = each %resulttypes_confirmed) { 
    my $nt_count_simplefreq = $resulttypecounts_undiluted{$key}; # consider the undiluted counts for the glue rules below
    my $nt_count_posteriorcounting = $resulttypecounts_diluted{$key}; # and also diluted counts
    my $meanstat = $resulttypecounts_diluted_logsourcespanlength_meanstat{$key};
    my $variancestat = $resulttypecounts_diluted_logsourcespanlength_variancestat{$key};
    if (!defined($nt_count_posteriorcounting)) {
      print STDERR "Warning: no COUNT information for nonterminal $key. Using 1 as counts.\n";
      $nt_count_posteriorcounting = 1;
      $nt_count_simplefreq = 1;
      $meanstat = 0;
      $variancestat = 0;
    }
    print STDERR "$key\t$nt_count_posteriorcounting\t$nt_count_simplefreq\t" . exp($meanstat/$nt_count_posteriorcounting) . "\n";
    # write out a pseudo-rule storing the counts for this nonterminal for use in e.g. filterrules.pl
    outputrule("\@_COUNT", "\@1\#$key\#$nt_count_posteriorcounting $nt_count_simplefreq $meanstat $variancestat");
    $manualrulecount++;

    if ($key eq '@_S' || $key eq '@_G') {
      # ignore _S and _G
    } else {
      $nonterminalcount++;
      #      if ($exitgluemoderules) {
      #	# add exit-glue-mode rule for $key
      #	outputrule('@_S', "\@1\#$key\#1 0 0 0 0 0 0 0 0 0 0 0");
      #	$manualrulecount++;
      #      }
      # unary glue rules are next 
      if ($gluerules) {
	# let each category directly finish up as a full sentence _S)
	#	# if rule has already been extracted in a natural way - don't write
	#	# out extra unary glue rule
	#	if (!exists($complexrules{"$key\#\@1\#\@_S"})) {
	if (!$UseBOSandEOS) {
	  outputrule($key, 
		     "\@1\#\@_S\#" . ($sttk? 0 : 1) . " 1 0 " . 
		     -log($nt_count_posteriorcounting) . #ARV SWAPPED (the diluted is the important one)
		     concatenatenumber(0, $undilutedresultcountoffset-4) .
		     ' ' . -log($nt_count_simplefreq) .  
		     concatenatenumber(0, $originalfeaturecount+$additionalfeaturecount-$undilutedresultcountoffset-1)
		    );
	  $manualrulecount++;
	}

	#	# now add enter-glue-mode rule for $key
	#	outputrule($key, "\@1\#\@_G\#0 0 0 " . 
	#		   -log($nt_count_simplefreq) . 
	#		   concatenatenumber(0, $undilutedresultcountoffset-4) .
	#		   ' ' . -log($nt_count_diluted) . 
	#		   concatenatenumber(0, $originalfeaturecount+$additionalfeaturecount-$undilutedresultcountoffset-1)
	#		  );
	#	$manualrulecount++;
	
	#the joining glue rule
	if (!$sttk) {
	  outputrule("\@_S $key", "\@1 \@2\#\@_S\#1 1 0 " . 
		     -log($nt_count_posteriorcounting) . #ARV SWAPPED (the diluted is the important one)
		     concatenatenumber(0, $undilutedresultcountoffset-4) .
		     ' ' . -log($nt_count_simplefreq) . 
		     concatenatenumber(0, $originalfeaturecount+$additionalfeaturecount-$undilutedresultcountoffset-1)
		    );
	  $manualrulecount++;
	} else {
	  die "STTK sim has to be repaired.";
	  # phrase decoder simulation: add a regular glue rule without glue feature...
	  outputrule("\@_S \@_G", "\@1 \@2\#\@_S\#0 1 0" . concatenatenumber(0, $originalfeaturecount+$additionalfeaturecount-3));
	  # and a binary reodering rule (reordering two _S rules) with glue (now meaning: reordering) feature
	  outputrule("\@_S \@_S", "\@2 \@1\#\@_S\#1 1 0" . concatenatenumber(0, $originalfeaturecount+$additionalfeaturecount-3));
	  $manualrulecount += 2;
	}

      }
      if ($unknownrules && $UnknownInputRuleFile) {
	# if not UnknownInputRuleFile, we will ouput a few manual unknown rules below, otherwise we do it here
	# for all nonterminals in the rule file (except _S and _G)

	outputrule('_UNKNOWN', '_UNKNOWN#' . $key . '#0 1 1' . 
		   concatenatenumber(1000-log(1/100), $originalfeaturecount-3) .
		   concatenatenumber(0, $additionalfeaturecount)
		  );
	$manualrulecount += 1;
      }
    }
  }
  print STDERR "\nTotal number of nonterminal symbols (excluding _S and _G): $nonterminalcount\n";

  if ($UseBOSandEOS) {
    # BOS / EOS handling
    outputrule('<s>', '<s> <s>#' . '@_S' . '#1 1 0' . 
	       concatenatenumber(0, $originalfeaturecount-3) .
	       concatenatenumber(0, $additionalfeaturecount)
	      );
    outputrule('@_S </s>', '@1 </s> </s>#' . '@_S' . '#1 1 0' . 
	       concatenatenumber(0, $originalfeaturecount-3) .
	       concatenatenumber(0, $additionalfeaturecount)
	      );
    $manualrulecount += 2;
  }

  if ($unknownrules & !$UnknownInputRuleFile) {
    #unknown rules: here only the relative weights of unknown rules w.r.t. each other matter
    #to make sure that their feature weights are positive (i.e., they are costs), we add the same value to all neglog prob features
    if (!$crazyclusterrules) {
      outputrule('_UNKNOWN', '_UNKNOWN#@NNP#0 1 1' . 
		 concatenatenumber(1000-log(9/1000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);		# proper name, weight 9/10; -ln(9/10)=0.1
      outputrule('_UNKNOWN', '_UNKNOWN#@NP#0 1 1' . 
		 concatenatenumber(1000-log(9/1000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);		# az 06/03/09: allow also directly NP to cope with the new unary-rule handling of extractrules.pl
      outputrule('_UNKNOWN', '_UNKNOWN#@JJ#0 1 1' . 
		 concatenatenumber(1000-log(1/3000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);		# adjective, weight 1/30
      outputrule('_UNKNOWN', '_UNKNOWN#@RB#0 1 1' . 
		 concatenatenumber(1000-log(1/3000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);		# adverb, weight 1/30
      outputrule('_UNKNOWN', '_UNKNOWN#@NN#0 1 1' . 
		 concatenatenumber(1000-log(1/3000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);		# noun singular, weight 1/30
      $manualrulecount += 5;
    } elsif ($crazyclusterrules == 1) {
      # substitute e.g. @NNS to @NN in resultcategory
      outputrule('_UNKNOWN', '_UNKNOWN#@NN#0 1 1' . 
		 concatenatenumber(1000-log(9/1000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);
      outputrule('_UNKNOWN', '_UNKNOWN#@NP#0 1 1' . 
		 concatenatenumber(1000-log(9/1000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);		# az 06/03/09: allow also directly NP to cope with the new unary-rule handling of extractrules.pl
      outputrule('_UNKNOWN', '_UNKNOWN#@JJ#0 1 1' . 
		 concatenatenumber(1000-log(1/3000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);
      outputrule('_UNKNOWN', '_UNKNOWN#@RB#0 1 1' . 
		 concatenatenumber(1000-log(1/3000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);
      $manualrulecount += 4;
    } elsif ($crazyclusterrules == 2) {
      # substitute e.g. @NP to @N in resultcategory
      outputrule('_UNKNOWN', '_UNKNOWN#@N#0 1 1' . 
		 concatenatenumber(1000-log(9/1000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);
      outputrule('_UNKNOWN', '_UNKNOWN#@J#0 1 1' . 
		 concatenatenumber(1000-log(1/3000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);
      outputrule('_UNKNOWN', '_UNKNOWN#@R#0 1 1' . 
		 concatenatenumber(1000-log(1/3000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);
      $manualrulecount += 3;
    } elsif ($crazyclusterrules == 3) {
      # substitute @something to @_X in resultcategory
      outputrule('_UNKNOWN', '_UNKNOWN#@_X#0 1 1' . 
		 concatenatenumber(1000-log(9/1000), $originalfeaturecount-3) .
		 concatenatenumber(0, $additionalfeaturecount)
		);
      $manualrulecount += 1;
    }

    #    if ($gluerules) {
    #      # also output a copy-glue-rule for '.' for the case that '.' was removed during training
    #      outputrule('.', '.#@_G#0 1 1' . 
    #		 concatenatenumber(1000-log(1/10), $originalfeaturecount-3) .
    #		 concatenatenumber(0, $additionalfeaturecount)
    #		);
    #      $manualrulecount += 1;
    #      # and a full-sentence rule in the very unlikely case that '.' was removed during 
    #      # training AND the whole sentence consists only of a single '.'
    #      outputrule('.', '.#@_S#1 1 1' . 
    #		 concatenatenumber(1000-log(1/10), $originalfeaturecount-3) .
    #		 concatenatenumber(0, $additionalfeaturecount)
    #		);
    #      $manualrulecount += 1;
    #    }
  }

  print STDERR "$writtenoutrulenumber rules written out. ";
  print STDERR 'filterrules.pl - finished @ '. `date`;

  if ($BERKELEY_DB) {
    undef $outputdb if $outputdbfile;
  }

}				# main





#################### subs

#register count and phrase count values in hash structures for lookup
sub handlePhraseCount{
  my ($rule_aref) = @_;
  #we want to put the rule into the appropriate category
  #source_counts, sourceandarg_counts, rulebody_counts - target_counts
  if ($rule_aref->[0] eq "\@_PHRASE-SRC") {
    my $tmp = $source_counts{$rule_aref->[1]};
    $tmp = 0 if !defined($tmp);
    $source_counts{$rule_aref->[1]} = $tmp + $rule_aref->[2];
  } elsif ($rule_aref->[0] eq "\@_PHRASE-SRCANDARG") {
    my $val = $rule_aref->[-1];
    my $last = scalar(@$rule_aref)-2; 
    my $key = join('#', @$rule_aref[1..$last]);
    my $tmp = $sourceandarg_counts{$key};
    $tmp = 0 if !defined($tmp);
    $sourceandarg_counts{$key} = $tmp + $val;
  } elsif ($rule_aref->[0] eq "\@_PHRASE-RULEBODY") {
    #include all the args that are left
    my $last = scalar(@$rule_aref)-2;
    my $key = join('#', @$rule_aref[1..$last]);
    #my $value = $rule_aref->[-1];
    my $tmp = $rulebody_counts{$key};
    $tmp = 0 if !defined($tmp);
    $rulebody_counts{$key} = $tmp + $rule_aref->[-1]; #the last element is the value 
  } elsif ($rule_aref->[0] eq "\@_PHRASE-TGT") {
    my $tmp = $target_counts{$rule_aref->[1]};
    $tmp = 0 if !defined($tmp);
    $target_counts{$rule_aref->[1]} = $tmp + $rule_aref->[2];
  }
}

#output all the phrase counts thru the rule mechanism
sub printPhraseCounts{
  while (my ($key,$value) = each %source_counts) { 
    outputrule("\@_PHRASE-SRC", $key."#".$value);
  }
  while (my ($key,$value) = each %sourceandarg_counts) { 
    outputrule("\@_PHRASE-SRCANDARG", $key."#".$value);
  }
  while (my ($key,$value) = each %rulebody_counts) { 
    outputrule("\@_PHRASE-RULEBODY", $key."#".$value);
  }
  while (my ($key,$value) = each %target_counts) { 
    outputrule("\@_PHRASE-TGT", $key."#".$value);
  }
}


#update the relative frequency features for a given rule based on merged values accumulated thru the 
#phrasecount and count special variables
sub updateMergedRelativeFreqs{
  my ($rulecostvector_aref, $sourcewords_aref, $targetwords_aref, $result, $frequency, $simplerulefreq) = @_;
  #build keys for each of the lookups
  my  $count = 0;
  my @labels = ();
  #replace the @s with the counters and drop any <s>'s
  my @srcnolabelsorg = map { if (m/^\@(.*)/){push(@labels, $1); $count++; "@".$count;}else{ $_; }} @$sourcewords_aref;
  my @srcnolabels = grep {!/s\>/} @srcnolabelsorg;
  #print "LOOKUP SRC: ", join(" ", @$sourcewords_aref), " STRIPPED: ", join(" ", @srcnolabels), " LAB: ", join(" ", @labels), "\n";

  my $srcbody = join(" ",@srcnolabels);
  my $sourcefrequency_key = join(" ",@srcnolabels);

  #case:
  #@_PHRASE-SRCANDARG # otros miembros de #  #89
  #note the blank result at the end for the pure lexical rules
  my $source_and_arg_frequency_key = (scalar(@labels)>0) ? join(" ",@srcnolabels)."#".join(" ",@labels) : join(" ",@srcnolabels)."#";
  my $RuleBodyNoResult = 1 ;
  #my $rulebodyfreq_key  = ($RuleBodyNoResult) ? join(" ",@srcnolabels)."#".join(" ",@$targetwords_aref) : join(" ",@srcnolabels)."#".join(" ",@$targetwords_aref)."#$result";

  my $tgtbody = join(" ",@$targetwords_aref);
  $tgtbody =~ s/\s*\<s\>\s*//g;
  $tgtbody =~ s/\s*\<\/s\>\s*//g;
  #print "RBG:$srcbody#$tgtbody numsrcnolab: ", scalar(@srcnolabels)," srcorg: ", scalar(@$sourcewords_aref), "\n";
  my $rulebodyfreq_key  = $srcbody."#".$tgtbody;


  #lookup the values
  my $rulebodyfrequency = $rulebody_counts{$rulebodyfreq_key};
  my $sourcefrequency = $source_counts{$sourcefrequency_key};
  my $source_and_arg_frequency = $sourceandarg_counts{$source_and_arg_frequency_key};
    
  #Sometimes you might look up a phrase count and not find its entry in the db. Here are the possible causes
  #a) the phrase count src was purely lexical - in which case it is dropped it it occurred less than 2 times, or less than minlexfreq
  #   this is the only case where the phrase count is removed, but the actual rule could still be in there (if minlexfreq==1)
  #   such entries are dropped since we can look them up, see that they dont exist, and assign simplecount==1 and posteriorcount==rule's posterior count to them safely
  #b) the phrase count had NTs in it - in which case it was dropped if it didnt occur less the minnonlexfreq times
  #   this is safe because it must occur each time the rule occured


  # if one of them was not defined and the rule's simplerulefreq is different from 1, then something is wrong
  if (abs($simplerulefreq-1) > 0.1 && (!defined($sourcefrequency) || !defined($source_and_arg_frequency) || !defined($rulebodyfrequency))) {

    print STDERR "Error: rbf: rulebodyfrequency not defined for src: ", join(' ', @$sourcewords_aref), " tgt: ", join(' ', @$targetwords_aref),"\n" if (!defined($rulebodyfrequency));
    print STDERR "Error: srcf: sourcefrequency not defined for ",join(' ', @$sourcewords_aref),"\n" if (!defined($sourcefrequency));
    print STDERR "Error: srcand: source_and_arg_frequency not defined for ", join(' ', @$sourcewords_aref), "\n" if (!defined($source_and_arg_frequency));
    print STDERR "(even though simplerulefreq=$simplerulefreq)\n";
      
    #die "ERROR: Some phrase count information was not identified: yet rule frequency = $frequency RULE: src: ", join(' ', @$sourcewords_aref), " tgt: ", join(' ', @$targetwords_aref), "\n" if ($frequency!=1); #and this is not an n-best game
    
  }

  #note: this frequency might have been a posterior frequency - so we have to set counts to that value instead of just giving them frequency 1
  $sourcefrequency = $frequency if (!defined($sourcefrequency));
  $source_and_arg_frequency = $frequency if (!defined($source_and_arg_frequency));
  $rulebodyfrequency = $frequency if (!defined($rulebodyfrequency));
      
    
  $rulecostvector_aref->[4] = -log($rulebodyfrequency/$sourcefrequency);
  $rulecostvector_aref->[5] = -log($frequency/$source_and_arg_frequency);
}

# compute lexical phrase costs
sub compute_lexical_costs {
  my ($refsourcewords, $reftargetwords, $ruleimpact) = @_;

  # our return value
  my @lexical_costs = ();

  # first set up the source/target/weightvector matrix @i_j_w_matrix
  my $i_count = 0;		# counts the n.o. actual source words (excl. nonterminals)
  my $j_count = 0;		# counts the n.o. actual target words (excl. nonterminals)
  my @i_j_w_matrix = ();
  my %unique_src_words = ();
  for (my $i = -1; 
       $i < scalar(@$refsourcewords); 
       $i++) {
    my $src = $i==-1 ? 'NULL' : $$refsourcewords[$i];
    if (!(
	  $src =~ /^\@/		# no terminal symbol - ignore
	  or $src eq '<s>' or $src eq '</s>' # BOS / EOS marker - ignore
	  or $unique_src_words{$src}++ # word occurred already earlier in the phrase
	 )) {
      # process target
      print STDERR "$i: $src\n" if $verbose > 4;
      my @j_w_matrix = ();
      my %unique_tgt_words = ();
      for (my $j = -1; $j < scalar(@$reftargetwords); $j++) {
	my $tgt = $j==-1 ? 'NULL' : $$reftargetwords[$j];
	if (!($tgt =~ /^\@/ or $tgt eq '<s>' or $tgt eq '</s>' 
	      or $unique_tgt_words{$tgt}++ # word occurred already earlier in the phrase
	     )) {
	  # process weight vector
	  my @weights = ();
	  my $value = $lexical_weight_table{$src . ' ' . $tgt};
	  if (defined($value)) {
	    @weights = unpack('f*', $value);
	  }
	  # what do we do with source-target pairs for which a certain weight entry doesn't exist in the lexical weight file?
	  # set weight 0, meaning these words are unaligned
	  for (my $k = 0; $k < $lexical_weight_entry_count; $k++) {
	    # $weights[$k] = $lowest_lexical_weight_encountered / 10 if !defined($weights[$k]) || $weights[$k] == 0;
	    $weights[$k] = 0 if !defined($weights[$k]);
	  }
	  print STDERR "i$i j$j: " . join(' ', @weights) . "\n" if $verbose > 4;
	  my $refweights = \@weights;
	  push @j_w_matrix, $refweights;
	}
      }				# for $j
      my $ref_j_w_matrix = \@j_w_matrix;
      push @i_j_w_matrix, $ref_j_w_matrix;
      $i_count++;		# to determine n.o. actual source words
      $j_count = scalar(@j_w_matrix); # to determine n.o. actual target words
    }
  }				# for $i

  die if $i_count<1 || $j_count<1;
  my $abstractrule = $i_count==1 && $j_count==1; # only NULL --> must be an abstract rule

  # now compute the target-conditioned lexical costs
  # iterate through target-conditioned lexical-weight files
  for (my $k = 0; $k < ($AlsoComputeWrongDirections? $lexical_weight_entry_count : scalar(@LexicalWeightFile)); $k++) {

    # first feature: IBM1-prob-based lexical cost (either traditional or refined, we'll calculate both here and will decide which one to use according to $UseRefinedIBM1ProbEstimate parameter)
    # traditional IBM1 cost estimate:
    my $cost = 0;
    # refined IBM1 estimate - publication forthcoming:
    my $cost_refined = 0;

    # second feature: neg-log best-possible joint lexical prob f|e
    my $best_possible_cost = 0;
    my $best_possible_cost_refined = 0;

    if (!$abstractrule) {
      # iterate through source words
      my $i_start = $i_count==1?
	0 :			# no source words, so use NULL instead
	    1;			# start with 1 since we are not using NULL source words if we have actual words in the source phrase
      for (my $i = $i_start; $i < $i_count; $i++) {
	my $sum = 0;
	my $sum_refined = 0;
	my $bestprob = 0;
	my $bestprob_refined = 0;
	print STDERR "i$i: $i_j_w_matrix[$i]\n" if $verbose > 4;
	my $alignment_partner_count = 0;
	# iterate through target words - the potential alignment partners, including NULL
	for (my $j = 0; $j < $j_count; $j++) {
	  if ($i_j_w_matrix[$i]->[$j]->[$k] > 0) {
	    $alignment_partner_count++;
	    print STDERR "i$i j$j: " . join(';',@{$i_j_w_matrix[$i]->[$j]}) . "\n" if $verbose > 4;
	    $sum += $i_j_w_matrix[$i]->[$j]->[$k];
	    $bestprob = max($bestprob, $i_j_w_matrix[$i]->[$j]->[$k]);
	
	    # for refined estimate, need to compute discounted weight mass:
	    my $discount_mass = 0;
	    for (my $i2 = $i_start; $i2 < $i_count; $i2++) {
	      $discount_mass += $i_j_w_matrix[$i2]->[$j]->[$k] if $i2 != $i;
	    }
	    my $margin_after_discount = 1 - $discount_mass;
	    # don't let the margin get lower than the lexical prob f_i|e_j (this can happen due to two identical source words - no, not anymore, words are now unique)
	    if ($margin_after_discount < $i_j_w_matrix[$i]->[$j]->[$k] - 0.0001) {
	      print STDERR "Warning: margin_after_discount= $margin_after_discount < $i_j_w_matrix[$i]->[$j]->[$k] = i_j_w_matrix[$i]->[$j]->[$k] for phrase pair " . join(' ',@$refsourcewords) . "#" . join(' ',@$reftargetwords) . "\n";
	      $sum_refined += 1;
	      $bestprob_refined = 1;
	    } else {
	      $sum_refined +=  
		$i_j_w_matrix[$i]->[$j]->[$k] / $margin_after_discount;
	      $bestprob_refined = max($bestprob_refined, $i_j_w_matrix[$i]->[$j]->[$k] / $margin_after_discount);
	    }
	  }
	}
	if ($alignment_partner_count > 0) {
	  $cost += -log($sum/$alignment_partner_count);
	  $cost_refined += -log($sum_refined/$alignment_partner_count);
	  $best_possible_cost += -log($bestprob);
	  $best_possible_cost_refined += -log($bestprob_refined);
	} else {
	  # wasn't aligned to anything
	  $cost += -log($lowest_lexical_weight_encountered / 10);
	  $cost_refined += -log($lowest_lexical_weight_encountered / 10);
	  $best_possible_cost += -log($lowest_lexical_weight_encountered / 10);
	  $best_possible_cost_refined += -log($lowest_lexical_weight_encountered / 10);
	}
      }
    }
    push @lexical_costs, $ruleimpact * ($UseRefinedIBM1ProbEstimate? $cost_refined : $cost);
    push @lexical_costs, $ruleimpact * ($UseRefinedIBM1ProbEstimate? $best_possible_cost_refined : $best_possible_cost);
    # also use the traditional IBM1 prob estimate (in case of --noUseRefinedIBM1ProbEstimate, this appears twice now)
    #    push @lexical_costs, $ruleimpact * $cost;

  }				# end: for $k


  # and finally, compute the source-conditioned lexical costs (for the --LexicalWeightFileReversed weights)
  # almost like above, but the role of $i and $j is reversed
  for (my $k = ($AlsoComputeWrongDirections? 0 : scalar(@LexicalWeightFile)); $k < $lexical_weight_entry_count; $k++) {

    # first feature: IBM1 estimate
    # traditional estimate:
    my $cost = 0;
    # refined IBM1 estimate - publication forthcoming:
    my $cost_refined = 0;

    # second feature: neg-log best-possible joint lexical prob f|e
    my $best_possible_cost = 0;
    my $best_possible_cost_refined = 0;

    if (!$abstractrule) {
      # iterate through target words
      my $j_start = $j_count==1?
	0 :			# no target words, so use NULL instead
	    1;			# start with 1 since we are not using NULL target words if we have actual words in the target phrase
      for (my $j = $j_start; $j < $j_count; $j++) {

	my $sum = 0;
	my $sum_refined = 0;
	my $bestprob = 0;
	my $bestprob_refined = 0;

	my $alignment_partner_count = 0;
	# iterate through source words - the potential alignment partners, including NULL
	for (my $i = 0; $i < $i_count; $i++) {
	  if ($i_j_w_matrix[$i]->[$j]->[$k] > 0) {
	    $alignment_partner_count++;
	    $sum += $i_j_w_matrix[$i]->[$j]->[$k];
	    $bestprob = max($bestprob, $i_j_w_matrix[$i]->[$j]->[$k]);

	    # for refined estimate, need to compute discounted weight mass:
	    my $discount_mass = 0;
	    for (my $j2 = $j_start; $j2 < $j_count; $j2++) {
	      $discount_mass += $i_j_w_matrix[$i]->[$j2]->[$k] if $j2 != $j;
	    }
	    my $margin_after_discount = 1 - $discount_mass;
	    # don't let the margin get lower than the lexical prob e_j|f_i (this can happen due to two identical source words - no, not anymore, words are now unique)
	    if ($margin_after_discount < $i_j_w_matrix[$i]->[$j]->[$k] - 0.0001) {
	      print STDERR "Warning: margin_after_discount= $margin_after_discount < $i_j_w_matrix[$i]->[$j]->[$k] = i_j_w_matrix[$i]->[$j]->[$k] for phrase pair " . join(' ',@$refsourcewords) . "#" . join(' ',@$reftargetwords) . "\n";
	      $sum_refined += 1;
	      $bestprob_refined = 1;
	    } else {
	      $sum_refined +=  
		$i_j_w_matrix[$i]->[$j]->[$k] / $margin_after_discount;
	      $bestprob_refined = max($bestprob_refined, $i_j_w_matrix[$i]->[$j]->[$k] / $margin_after_discount);
	    }
	  }
	}
	if ($alignment_partner_count > 0) {
	  $cost += -log($sum/$alignment_partner_count);
	  $cost_refined += -log($sum_refined/$alignment_partner_count);
	  $best_possible_cost += -log($bestprob);
	  $best_possible_cost_refined += -log($bestprob_refined);
	} else {
	  # wasn't aligned to anything
	  $cost += -log($lowest_lexical_weight_encountered / 10);
	  $cost_refined += -log($lowest_lexical_weight_encountered / 10);
	  $best_possible_cost += -log($lowest_lexical_weight_encountered / 10);
	  $best_possible_cost_refined += -log($lowest_lexical_weight_encountered / 10);
	}
      }
    }
    push @lexical_costs, $ruleimpact * ($UseRefinedIBM1ProbEstimate? $cost_refined : $cost);
    push @lexical_costs, $ruleimpact * ($UseRefinedIBM1ProbEstimate? $best_possible_cost_refined : $best_possible_cost);
    # also use the traditional IBM1 prob estimate (in case of --noUseRefinedIBM1ProbEstimate, this appears twice now)
    #    push @lexical_costs, $ruleimpact * $cost;

  }				# end: for $k


  return @lexical_costs;
}







# output rule to database or standard out
sub outputrule {
  my($newsourcestr, $rulebody) = @_;
  if ($RestrictToFeatures && $newsourcestr ne '@_COUNT') {
    my @tmp = split('#', $rulebody);
    my @costvector = split(' ', $tmp[2]);
    my @newcostvector = ();
    die "Assertion error: wrong length of costvector " . $tmp[2] . "(should have length $originalfeaturecount+$additionalfeaturecount) in rule $newsourcestr\#$rulebody" if scalar(@costvector) != $originalfeaturecount + $additionalfeaturecount;
    for (my $i=0; $i < scalar(@costvector); $i++) {
      if ($featureRestrictionVector[$i] != 0) {
	push @newcostvector, $costvector[$i];
      }
    }
    die "All values of feature restriction vector $RestrictToFeatures set to false!" if scalar(@newcostvector) == 0;
    $tmp[2] = join(' ', @newcostvector);
    $rulebody = join('#', @tmp);
  }
  #quick check for inf.. this is not a smart way to do it...
  if ($rulebody =~ m/\binf\b/) {
    my @tmp = split('#', $rulebody);
    my @costvector = split(' ', $tmp[2]);
    my $costline = join(' ', @costvector);
    if ($costline =~ m/\binf\b/) {
      print STDERR "inf in rule $newsourcestr\#$rulebody -- ignoring that rule.\n";
      return;
    }
  } 

  if ($outputdbfile) {
    $outputdb->db_put($newsourcestr . "\0", $rulebody . "\0");
  } else {
    print "$newsourcestr\#$rulebody\n";
  }

  $writtenoutrulenumber++;
}

# concatenates number $num $times times, each $num occurrence preceeded by ' '.
sub concatenatenumber {
  my ($num, $times) = @_;
  my $result = '';
  while (--$times >= 0) {
    $result .= ' ' . $num;
  }
  return $result;
}
