#!/usr/bin/env perl

# STDIN: words to look up, separated by \n
# --LexicalWeightFile, --LexicalWeightFileReversed (specification see below): multiple lexica to check for the word
# STDOUT: phrases of the form sourceword#targetword
# For each source word in standard in, we go through the specified lexica (first the --LexicalWeightFile files in the order specified, then the --LexicalWeightFileReversed files in the order specified) and stop at the first lexicon that matches. If that lexicon gives several translations for the source word, we choose the one with the highest probability.

use strict;
use warnings;
use Getopt::Long;

# These options can each occur multiple times and lets you specify files with word-based lexical probabilities.
# Each file passed with option --LexicalWeightFile has to consist 
# of entries of the form:
# frenchword englishword probability
#
# Each file passed with option --LexicalWeightFileReversed has to consist
# of entries of the form:
# englishword frenchword probability
#
my @LexicalWeightFile = ();
my @LexicalWeightFileReversed = ();


#my $truecount = 1;
die if !GetOptions('LexicalWeightFile=s' => \@LexicalWeightFile, 'LexicalWeightFileReversed=s' => \@LexicalWeightFileReversed);

# hash in the words to look up
my %vocCheck = ();
while (<STDIN>){
  chomp($_);
  $vocCheck{$_}++;
}


# load in Lexical weight files
my $lowest_lexical_weight_encountered = 2**30;
my $lexical_weight_entry_count = 0;
my %lexical_weight_table = ();
  foreach my $filename ((@LexicalWeightFile, @LexicalWeightFileReversed)) {
    # n.o. lexical weight files so far
    $lexical_weight_entry_count++;

    # if we're beyond the LexicalWeightFile entries, set reversemode to true to switch french and english
    my $reversemode = ($lexical_weight_entry_count > scalar(@LexicalWeightFile));
    print STDERR "Reading in lexical weight file $filename...\n";
    open(SOURCE, "< $filename") or die "Can't open lexical weight file $filename: $!";
    while (<SOURCE>) {
      chomp($_);
      my ($french, $english, $weight);
      if (!$reversemode) {
	($french, $english, $weight) = split(' ',$_);
      } else {
	($english, $french, $weight) = split(' ',$_);
      }
      if (!$vocCheck{$french}) {
	# source side doesn't occur in the restriction dev/test set -- ignore this entry
#      } elsif ($weight <= 0) {
#	print STDERR "Warning: weight $weight <= 0 occurred in entry '$_'\n";
      } else {
	$lowest_lexical_weight_encountered = $weight if $weight < $lowest_lexical_weight_encountered;
	my $value =  $lexical_weight_table{$french};
	if (defined($value)) {
	  my ($oldenglish, $oldweight,$lexicon_number) = split(' ', $value);
	  if ($lexicon_number != $lexical_weight_entry_count) {
	    # this entry is from a previously read-in, i.e., better, lexicon - leave it alone
	  } else {
	    if ($weight > $oldweight) {
	      $lexical_weight_table{$french} = $english . ' ' . $weight . ' ' . $lexical_weight_entry_count;
	    }
	  }
	} else {
	  # not entry exists yet - create one
	  $lexical_weight_table{$french} = $english . ' ' . $weight . ' ' . $lexical_weight_entry_count;
	}
      }
    }
    close(SOURCE);
  }


# now comes the actual look-up
print STDERR "Looking up the entries...\n";
my $entry_count = 0;
my $entries_found = 0;
my ($key, $value);
while (($key,$value) = each %vocCheck) { 
  $entry_count++;
  my $value = $lexical_weight_table{$key};
  if (defined($value)) {
    $entries_found++;
    my ($english, $weight, $lexicon_number) = split(' ',$value);
    print "$key\#$english\n";
  }
}

print STDERR "N.o. entries: $entry_count; n.o. entries found: $entries_found\n";
