#!/usr/bin/env perl
#Select the nbest paired alignments from a set of m alignments in one direction and n in the other
#also does alignment refinement in this one step
use strict;
use Getopt::Long;
use IO::File;
use IO::Zlib; #use this to output a zipped file

use List::Util qw[min max];
my $ef_align = "";
my $fe_align = "";
my $nbest = 10;
my $postFix = "nbest";
die if !GetOptions("E2F=s" => \$ef_align, "F2E=s" => \$fe_align, "N=s", => \$nbest, "postFix=s" => \$postFix);

$postFix = "$postFix"."$nbest";
my $ef_h = new IO::File("$ef_align") or die "Unable to open $ef_align for reading alignment\n";
my $fe_h = new IO::File("$fe_align") or die "Unable to open $fe_align for reading alignment\n";

my $ef_out = IO::Zlib->new("$ef_align.$postFix.gz", "wb9") or die "Unable to open $ef_align.$postFix.gz for reading alignment\n";
my $fe_out = IO::Zlib->new("$fe_align.$postFix.gz", "wb9") or die "Unable to open $fe_align.$postFix.gz for reading alignment\n";


my $lastSentence="";
my $hasMore = 1;

my @ef_lines = ();  my @ef_next = ();
my @fe_lines = (); my @fe_next = ();
my $snum = 1;
do{
    
    my $moreEF = readTillNext($ef_h, $snum, \@ef_lines, \@ef_next);
    my $moreFE = readTillNext($fe_h, $snum, \@fe_lines, \@fe_next);

    #process the lines, to fill the best variable with the top N pairs
    my @ef_bestN = ();
    my @fe_bestN = ();
    selectNFromN2(\@ef_lines, \@fe_lines, $nbest, \@ef_bestN, \@fe_bestN);
    
    replaceSeparator(\@ef_bestN);
    replaceSeparator(\@fe_bestN);

    $ef_out->print(join("\n",@ef_bestN)."\n");
    $fe_out->print(join("\n",@fe_bestN)."\n");
    #print "EF\n ",join(' ',@ef_lines), "\n";
    #print "FE\n ",join(' ',@fe_lines), "\n";
    if ($snum%100==0){
	print STDERR "Processing snum $snum\n";
    }

    #initialize the lines, with the previous set
    @ef_lines = @ef_next;
    @fe_lines = @fe_next;
    @ef_next = ();
    @fe_next = ();

    $snum++;
    #is there more to read?
    if ($moreEF==1 && $moreFE==1){
	$hasMore=1;
    }else{
	$hasMore=0;
    }

}while($hasMore);

sub replaceSeparator{
    my ($aref) = @_;
    for (my $i=0; $i<scalar(@$aref); $i++){
	$aref->[$i] =~ s/ \{\#\} /\n/g;
    }
}
sub selectNFromN2{
    my ($ef_aref, $fe_aref, $nselect, $ef_final_aref, $fe_final_aref) = @_;
    my %sortedProbs = ();
    for (my $ef_i = 0; $ef_i<min(scalar(@$ef_aref), $nselect); $ef_i++){

	my $efline = $ef_aref->[$ef_i];
	my $efprob = probFromLine($efline);
	for (my $fe_i = 0; $fe_i<min(scalar(@$fe_aref), $nselect); $fe_i++){
	    
	    my $feline = $fe_aref->[$fe_i];
	    my $feprob = probFromLine($feline);
	    
	    #print "Pair EF $ef_i = $efprob , FE $fe_i $feprob => ",$efprob*$feprob,"\n";
	    $sortedProbs{"$ef_i $fe_i"} = $efprob*$feprob;
	}
    }
    my @sortedKeys = reverse sort { $sortedProbs{$a} <=> $sortedProbs{$b} } keys %sortedProbs;
    my $count =0;
    foreach my $key (@sortedKeys){
	my @parts = split(/\s+/, $key);
	#print "Key $key => ",$sortedProbs{$key}, "\n",$ef_aref->[$parts[0]], "\n", $fe_aref->[$parts[1]], "\n";
	push (@$ef_final_aref, $ef_aref->[$parts[0]]);
	push (@$fe_final_aref, $fe_aref->[$parts[1]]);
	$count++;
	if ($count>=$nselect){
	    #print STDERR "Terminating: $count\n";
	    last;
	}

    }
    #print "SZ-EF: ",scalar(@$ef_final_aref), "\n";
    #print "SZ-FE: ",scalar(@$fe_final_aref), "\n";
}

sub probFromLine{
    my ($line) = @_;
    #print "Line $line \n";
    if ($line =~ m/alignment score : ([0-9\.\-e]+)/){
	my $prob = $1;
	if ($prob>1.0){
	    print STDERR "Error, prob>1 $1 : Line $line\n";
	    exit(-1);
	}
	return $prob;
    }else{
	print STDERR "Error, no alignment prob $line\n";
	exit(-1);
    }
}



sub readTillNext{
    my ($fp, $currentnum, $current_aref, $next_aref) = @_;
    #my $l1 = <$fp>;
    #my $l2 = <$fp>;
    #print "$l1 => $l2\n";
    my $line = read3($fp);
    #print "Read: $line\n";
    my $snum="";
    if ($line){
	$snum = getSentenceNumber($line);
    }else{
	$snum = -1;
    }
    while ($snum==$currentnum){
	#add a line to the current
	push(@$current_aref, $line);
	#read another line (which could come from the next sentence
	$line = read3($fp);
	if (!$line){
	    return 0;
	    last;
	}
	$snum = getSentenceNumber($line);
    }
    #if you are out here, you are from the next sentence, as long as you are a good line
    if ($line){
	#print "Adding to NEXT: $line\n";
	push (@$next_aref, $line);
    }else{
	return 0;
    }
    return 1;
}


sub getSentenceNumber{
    my ($line) = @_;
    #print "GS: $line\n";
    if ($line=~ m/Sentence pair \(([0-9]+)\).*/){
	return $1;
    }else{
	print STDERR "Line format: $line\n";
	exit(-1);
    }
}

sub read3{
    my ($fp) = @_;
    my $l1 = <$fp>;
    if (!$l1){ return $l1;}
    my $l2 = <$fp>;
    if (!$l2){ return $l2;}
    my $l3 = <$fp>;
    if (!$l3){ return $l2;}

    chomp($l1);
    chomp($l2);
    chomp($l3);
    my $line = $l1." \{\#\} ".$l2." \{\#\} ".$l3;
    return $line;
}

