#!/usr/bin/env perl
#Takes phrases exatracted over multiple alignments (produced by hiero refinement and extraction)
#and creates a posterior distribution over the multiple alternatives. The script does the following
#1: Reads all the data for a single alignment
#2: Consolidates the multiple alignments to make data for each sentence
#3: For each phrase pair line (which also indicates the alignment ranges) hash which alt it came from
#4: For each alignment, renormalize the probabilities to sum to 1
#Output: 
# --A file which is in the original format that extract rules take with additional phrase annotation
# --A file of the form sentenceID: prob1 prob2...probN (one for each alignment)
#Decisions: The alignment for a particular phrase is taken from the best scoring alignment
use strict;
use Getopt::Long;

my $NormalizeScores = 0; #by default, dont normalize! Only normalize when extract rules is called and the n is picked!
die if !GetOptions("NormalizeScores!" => \$NormalizeScores);
my $snum = 0;
my @sentenceData = (); #each line is the raw data for a whole block
while (<STDIN>){
    #indicates begin of sentence block
    if (m/LOG: NUM: .* Sentence pair \(([0-9]+)\) .* Sentence pair \(([0-9]+)\) .*/){
	my $header = $_;
	my $snum1 = $1; 
	my $snum2 = $2;
	#print "snum1 $snum1 snum2 $snum2\n";
	if ($snum1 != $snum2){
	    print STDERR "ERROR: snums done match $header\n";
	    exit;
	}
	if ($snum1!=$snum && $snum!=0){
	    processSentence(\@sentenceData);
	    #clean up the data, get ready for the next sentence
	    @sentenceData = ();
	    $snum=$snum1; #move onto the next sentence
	}else{
	    $snum=$snum1; #move onto the next sentence
	}
	#concat this whole alignment block
	my $oneAlignmentData = $header;
	my $line;
	while ( $line = <STDIN>){
	    if ($line =~ m/PHRASES_END/){
		#print "ENDING========= $line\n";
		last;
	    }else{
		$oneAlignmentData = $oneAlignmentData.$line;
	    }
	}
	#print "ONE: $oneAlignmentData\n";
	push(@sentenceData, $oneAlignmentData);
    }else{
	print STDERR "Unknown line type $_\n";
	exit;
    }    
}
#process the last sentence!
processSentence(\@sentenceData);

sub processSentence{
    my ($sent_aref) = @_;
    my $firstAltData = $sent_aref->[0];

    #build up a list of phrases that each sentence comes from
    my %phraseHash = ();
    #track the sentence number
    my $count= 0;
    my $probSum = 0;
    my @probArr = ();
    #go thru the data for each alignment, ie all its phrases and grab its probabilities
    foreach my $alt (@$sent_aref){
	#print "ALT: $alt\n";
	my ($p1, $p2) = extractPhrasesAndProbs($alt, $count, \%phraseHash);

	my $mult = $p1*$p2;
	push(@probArr, $mult);
	$probSum += $mult;

	$count++;
    }
    #renormalize the probabilities for each alignment
    if ($NormalizeScores){
	for (my $i=0; $i<scalar(@probArr); $i++){
	    $probArr[$i] = $probArr[$i]/$probSum;
	}
	#verify
	my $vsum = 0;
	for (my $i=0; $i<scalar(@probArr); $i++){
	    $vsum += $probArr[$i];
	}
    }
 
    #print "======== PROBS: ", join(' ', @probArr), " => $vsum \n";

    #The header data (and by consequence the alignment decisions) come from the 
    #first sentence. In extractrules, we have to be careful about how we calculate
    #lexical probabilities for phrases that have no alignment points according to
    #the alignment of the first sentence
    printLogHeader($firstAltData, \@probArr);
    #print "======== PHRASES: ===========\n";
    foreach my $key (keys %phraseHash){
	print "LOG: $key => ", $phraseHash{$key}, " => ", probFromOccs($phraseHash{$key}, \@probArr),"\n";
    }    
    print "LOG: PHRASES_END\n";
}

sub printLogHeader{
    my ($line, $prob_aref) = @_;
    my @parts = split(/\n/, $line);
    foreach my $p (@parts){
	#header with probs for individual alignments
	if ($p =~ m/LOG: NUM:/){
	    my $initial = $p;
	    chomp($initial);
	    print "$initial \# ALTPROBS ", join(' ', @$prob_aref), "\n";
	}
	if ($p =~ m/LOG: (SRC|TGT|ALT|PHRASES_BEGIN):/){
	    #to get it back in the format we are used to in extractrules.pl
	    #$p =~ s/PHRASES_BEGIN/PHRASES/g;
	    print $p, "\n";
	}	
    }
}

sub probFromOccs{
    my ($occLine, $prob_aref) = @_;
    my @occs = split(/\s+/, $occLine);
    my $sum = 0;
    foreach my $o (@occs){
	$sum += $prob_aref->[$o];
    }
    return $sum;
}

sub extractPhrasesAndProbs{
    my ($line, $altnum, $hash_ref) = @_;
    my @parts = split(/\n/, $line);
    my $p1; my $p2;
    foreach my $p (@parts){
	#prob data
	if ($p =~ m/LOG: NUM: .* alignment score : ([0-9\-\.e]+) .* alignment score : ([0-9\-\.e]+)/){
	    $p1 = $1;
	    $p2 = $2;
	    #print "probs $p1 $p2\n";
	}
	#phrase data, hash it all
	if ($p =~ m/LOG: (\([0-9]+,[0-9]+\) .*)/){
	    my $pdata = $1;
	    #print "PHRASE: $pdata\n";
	    if (exists $hash_ref->{$pdata}){
		$hash_ref->{$pdata} = $hash_ref->{$pdata}." ".$altnum;
	    }else{
		$hash_ref->{$pdata} = $altnum;
	    }
	}
    }
    return ($p1,$p2);
    
}
