#!/usr/bin/env perl

use strict;
use warnings;
use Getopt::Long;

#==============================================
# command line parameters

# If set to 1, we will move end-of-sentence marks ('.','!','?') from
# the end to the beginning of each sentence. This also works for lines
# with multiple sentences
my $ShiftSentenceMarks = 0;

# when there is no end of sentence mark at the end of an input line,
# then add a dot
my $AddDotForMissingSentenceMarks = 1;

# by default, we smart-case (keeping
# the casing as it is except for the first word of each sentence
# (where a line can contain multiple sentences), which we case
# according to the majority of instances of that word as non-first
# word
my $Lowercasing = 0; # if 1, then lowercase everything
my $Keepcasing = 0; # if 1, then don't modify upper-/lowercasing
my $TruecasingData = '';

# all words not stored in this map will be modified to lowercase
# whenever they occur as first-word of a sentence
my %preserve_case_first_words = ();

die if !GetOptions('Lowercasing!' => \$Lowercasing, 'Keepcasing!' => \$Keepcasing, 'TruecasingData=s' => \$TruecasingData, 'ShiftSentenceMarks!' => \$ShiftSentenceMarks, 'AddDotForMissingSentenceMarks!' => \$AddDotForMissingSentenceMarks);

die "Too many parameters" if scalar(@ARGV) != 0;
die "No text file with truecasing data specified (option --TruecasingData)" if (!$Lowercasing && !$Keepcasing && !$TruecasingData);


#============================
# program start


if ($TruecasingData) {
  # create occurrence statistics of words
  my %vocab = ();
  open(TRUECASINGDECISIONS,"> $TruecasingData.keeptruecase");
  open(TRUECASINGDATA,"< $TruecasingData");
  while (<TRUECASINGDATA>) {
    chomp;
    my @words = split(' ',$_);
    # ignore first word of sentence since that word is always
    # uppercased. Still add it to the vocab though.  
    my $ignore_next_word = 1;
    foreach my $w (@words) {
      #      print STDERR $w . "\n"; 
      #      print STDERR "Adding '$w', context '$_'\n" if $w eq 'Could';
      if ($w =~ /^(\.|\!|\?)+$/) {
	# end-of-sentence mark: skip on to the next word, which we'll
	# ignore as a count (but still add to the vocab)
	$ignore_next_word = 1;
      } else {
	# add word count (zero or one depending on $ignore_next_word)
	# to the vocab using magic += (which adds to zero in case the
	# word doesn't exist yet)
	$vocab{$w}+= $ignore_next_word ? 0 : 1;
	$ignore_next_word = 0;
      }
    }
  }
  close(TRUECASINGDATA);


  # uppercase statistics
  {
    my @uppercasewords = ();
    my ($key, $value);
    while (($key,$value) = each %vocab) { 
      # we're only interested in upper-case words:
      if ($key ne lcfirst($key)) {
	# store word, uppercase occurrence count, and lowercase occurrence count
	my $value2 = $vocab{lcfirst($key)};
	$value2 = 0 if !defined($value2);
	# push word and frequency counts of uppercase and lowercase
	# variant into statistics (just for info purposes)
	push @uppercasewords, [$key, $value, $value2] unless $value==0 && $value2 > 0;
	# if the upper-case variant is at least as popular as the
	# lower-case variant, mark it for preservation
	$preserve_case_first_words{$key} = 1 if $value >= $value2;
	print TRUECASINGDECISIONS "$key\n" if $value >= $value2;
      }
    }
    my @sortedwords = sort {$b->[1] <=> $a->[1]} @uppercasewords;
    print STDERR "Most frequent uppercase words / frequencies / frequencies of corresponding lowercase versions\n";
    for (my $i=0; $i < 30 && $i < scalar(@sortedwords); $i++) {
      print STDERR ($sortedwords[$i]->[0] . ' & ' . $sortedwords[$i]->[1] . ' & ' . $sortedwords[$i]->[2] . " \\\\\n");
    }
  }
  close(TRUECASINGDECISIONS);
}

while (<STDIN>) {
  chomp; # get rid of the newline symbol (will otherwise disturb us when removing multiple whitespace below)

  
  s/[\s\x00-\x20]+/ /g; # normalize whitespace ([ \t\n\r\f]) and control
               # characters to space character, shrink multiple
               # whitespace to one
  s/^\s+|\s+$//g; # remove leading or trailing whitespace
  
  if ($Lowercasing) {
    $_ = lc($_);
  }

  # now we really start

  my @tokens = split(' ',$_);
  my $sentence_start = 0;
  for (my $i = 0; $i < scalar(@tokens); $i++) {
    if ($i == $sentence_start) {
      if ($TruecasingData && !$Lowercasing && !$Keepcasing) {
	# lowercase first word of sentence if that word usually occurs in lowercase
	$tokens[$sentence_start] = lcfirst($tokens[$sentence_start]) unless $preserve_case_first_words{$tokens[$sentence_start]};
      }
    }
    if ($tokens[$i] =~ /^(\.|\!|\?)$/) {
      # the current token is an end-of-sentence mark
      if ($ShiftSentenceMarks) {
	# move the end-of-sentence mark to the beginning of the current sentence
	@tokens[$sentence_start..$i] = ($tokens[$i], @tokens[$sentence_start..$i-1]);
      }
      # update sentence_start marker
      $sentence_start = $i+1;
    }
  }
  if ($AddDotForMissingSentenceMarks && $sentence_start < scalar(@tokens) && $tokens[-1] ne '"') {
    # some unfinished business: there was no '.', '!' or '?' at the end of this sentence...
    # just assume it should have been a '.'
    if ($ShiftSentenceMarks) {
      @tokens[$sentence_start..$#tokens+1] = ('.', @tokens[$sentence_start..$#tokens]);
    } else {
      push @tokens, '.';
    }
  }
  
  $_ = join(' ',@tokens);
  print "$_\n"; # print out the modified line, followed by a newline symbol
}
