#!/usr/bin/env perl
#Train a SAMT grammar and runs MER starting with the raw giza files
#Inputs: refiner script, extract rule script, filter script
#Script must be able to start at any given stage of the process, using a specified
#output file form the previous stage. All outputs of this script will be 
#found in the outdir specified in the parameter settings. Naming of the output
#files of each stage come completely from the scripts provided for each stage.
#All stages AFTER the first stage will take as input, the parts from the previous
#stage, since we dont want to do all the replitting each time

use IO::File;
use strict;
use Getopt::Long;

my $EF;
my $FE;
my $partsize = 10000;
my $outdir = "."; #directory that all scripts are given to write their output into

#you start operating this script at $first_step, taking as input $input
my $first_step = 0;
my $input; #only entered as input into $first_step
my $inputPattern; #pattern that collects all the parst from the previous part

#scripts
my $refiner_script;
my $parse_script;
my $phrases_script;
die if !GetOptions("partsize=s" => \$partsize, "outdir=s" => \$outdir, "first_step=s" =>  \$first_step, "input=s" => \$input, "refiner=s" => \$refiner_script, "parse=s" => \$parse_script, "phrases=s" => \$phrases_script);

#Input: GIZA in each direction =================================================
#join the EF,FE into a single stream and calculate the refined alignment
if ($first_step <= 1){
    #Output: Refined giza file  =================================================
    $inputPattern = refinerStep();
}else{
    print STDERR "Skipping step 1 (refinement): first_step = $first_step\n";
}

#run parsing on the target side of the aligned data - note that the parsing script
#needs to be able to take the giza file as input. Save the output in the parsed_parts
#since it needs to be passed in separately
my $parsed_parts;
if ($first_step <= 2){
    $parsed_parts = runScriptFromInput($parse_script, 3);
}

#run phrase extraction
if ($first_step <= 3){
    #get info for the phrase extraction
    runScriptFromInput($phrases_script, 3);
}else{
    print STDERR "Skipping step 2 (phraseextraction): first_step = $first_step\n";
}

#the refiner stage must take a single input file
sub refinerStep{
    my ($refiner_expname, $refiner_outpostfix) = getExpAndPattern($refiner_script);
    print "SCRIPTINFO: refiner : EXPNAME $refiner_expname OUT $refiner_outpostfix\n";
    
    #create the input for the refinement stage
    my @inputBoth = split(","$input);
    die "The refiner script requires 2 inputs, EF and FE files separated by comma" if (scalar(@inputBoth)!=2);
    my $inputEF = $inputBoth[0];
    my $inputFE = $inputBoth[1];
    
    #the map step
    runRefiner($inputEF, $inputFE, $refiner_expname, $refiner_outpostfix);
    #the reduce step
    print "REFINEROUTPUT: $refinerOutput : setting as input to stage 2\n";
    return ("$outdir/$expname/*.$outpostfix");
}

sub runRefiner{
    my ($inputEF, $inputFE, $expname, $outpostfix) = @_;
    my $refinerlines = $partsize*3;
#make the requested directory
    if (-e "$outdir/$expname"){
	print STDERR "REFINER: Directory already exists $outdir/$expname\n";
    }else{
	print `mkdir $outdir/$expname`;
    }
    my $run_refine = "joinstream.pl --file $inputEF --file $inputFE --sep \"###\" | train-in-parts.sh $outdir/$expname $refinerlines $refiner_script";
    print "RUNNING: $run_refine\n";
    #funtion will block until all parts are done
    print `$run_refine`;
}


#takes in a script and an input multiplier - does the following tasks
#a) extracts the expname and file postfix from the script and creates the $outdir/$expname dir
#b) picks via input pattern or input file and sends it down to train-in-parts
sub runScriptFromInput{
    my ($script, $inputMult) = @_;
    my ($expname, $outpostfix) = getExpAndPattern($script);

    if (-e "$outdir/$expname"){
	print STDERR "PHRASES: Directory already exists $outdir/$expname\n";
    }else{
	print `mkdir $outdir/$expname`;
    }
    my $lines = $partsize*$inputMult;
    my $inputBegin = (defined($inputPattern)) ? "ls $inputPattern " : " cat $input ";
    my $remaining " | train-in-parts.sh $outdir/$expname $lines $script";
    my $inputEnd = (defined($inputPattern)) ? " 0 " : " 1 ";
    my $runme =  "$inputBegin $remaining $inputEnd";
    print "RUN: $runme\n";
    print `$runme`;
    return ("$outdir/$expname/*.$outpostfix");
}


#utility script to extract expname and outputpostfix from scripts
sub getExpAndPattern{
    my ($script) = @_;
    #the script on the command line could contain multiple arguments
    my @scriptandargs = split(/\s+/, $script); 
    #we just want to get the scriptname, which could even be a file in the path
    $script = `which $scriptandargs[0]`;
    chomp($script);

    my $expname = `fgrep EXPNAME $script`;
    if ($expname =~ m/EXPNAME =\s*\"(.*)\"/){
	$expname = $1;
    }else{
	print STDERR "Unable to extract EXPNAME from $script\n";
    }
    my $outpostfix = `fgrep OUTPOSTFIX $script`;
    if ($outpostfix =~ m/OUTPOSTFIX =\s*\"(.*)\"/){	
	$outpostfix = $1;
    }else{
	print STDERR "Unable to extract OUTPOSTFIX from $script\n";
    }
    return ($expname, $outpostfix);
}
