#!/usr/bin/env perl

use warnings;
use strict;
#use Encode::Arabic;

#use Encode qw/encode decode/;
#use Storable;

use List::Util qw(first min max);
use Fcntl;
use File::Temp qw/ tempfile tempdir /;
#use DB_File; # uncomment this line if you don't have Berkeley DB installed on your unix system
use BerkeleyDB;			# uncomment this line if you don't have Berkeley DB installed on your unix system
use Getopt::Long;


#binmode STDIN, ":utf8";
#binmode STDERR, ":utf8";
#binmode STDOUT, ":utf8";

# settings

my $BERKELEY_DB = 1;		# set to zero if Berkeley DB is not available

# Command line parameters


my $tmpdir;
my $outputdbfile;
my $VocabularyFile;
my $verbose = 0;
my $vocabRestrictionFile;
my $cachesize = 512;		# Berkeley DB cache size in MB
my $MaxMarkovOrder=1000000;	# max. Markov order allowed
my $MinHistoryFrequency=1;
my $MinFrequency=1;


die if !GetOptions('tmpdir=s' => \$tmpdir, 'cachesize=s' => \$cachesize, 'verbose=s' => \$verbose, 'outputdbfile=s' => \$outputdbfile, 'r=s' => \$vocabRestrictionFile, 'Vocabulary=s' => \$VocabularyFile, 'MaxMarkovOrder=s' => \$MaxMarkovOrder, 'MinHistoryFrequency=s' => \$MinHistoryFrequency, 'MinFrequency=s' => \$MinFrequency);

my $TrainingCorpusFile = shift(@ARGV);
die "Too many arguments (expected: 1). Superfluous arguments: " . join(' ',@ARGV) if scalar(@ARGV) != 0;

# $tmpdir = File::Spec->tmpdir() if (!$tmpdir);
$tmpdir = "." if (!$tmpdir);



################################## Global variables

# the vocabulary hash mapping a string to its integer vocab ID
my %vocabhash = ();
# the vocabulary list mapping element i to the string corresponding to vocab ID i
my @vocablist = ();

# create my own temp dir as subdir of $tmpdir, it and its contents will 
# automatically removed at program end
#my $mytmpdir = tempdir( "tmp_extractrules_XXXXXXXXX", DIR => $tmpdir, CLEANUP => 1);

my %lm = ();			# the LM: a hashtable mapping n-grams to their occurrence counts
my $sentencenumber = 0;		# counts the sentences processed
my $ruleinstancecount = 0;	# counts the number of instances of ngrams encountered (i.e. multiple occurrences count multiple times)
my $writtenoutrulenumber = 0;	# number of rules written out to STDOUT / database
my $outputdb;			# file handle for Berkeley DB

# if parameter -r was passed, restrict n-grams trained to ngrams consisting of words in %vocCheck
my %vocCheck = ();


{				# main block
  print STDERR 'trainlm.pl - started @ '. `date`;
  print STDERR ($outputdbfile ?
		"Outputting trained LM into newly created Berkeley DB file $outputdbfile\n" :
		"Outputting LM in text-form to STDOUT\n");

  die "If you're using --OutputDBFile, you have to specify a --VocabularyFile to store the vocabulary in since the n-grams in the Berkeley DB are stored as sequences of word IDs, not actual word strings." if $outputdbfile && !$VocabularyFile;

  if ($vocabRestrictionFile) {
    print STDERR "Restricting the n-grams to the words occurring in file $vocabRestrictionFile, merging all words from the training data that don't occur in that file to a special reserved word.\n";
    open(SOURCE, "< $vocabRestrictionFile") or die "Can't open $vocabRestrictionFile: $!";
    while (<SOURCE>) {
      chomp($_);
      my @sentence = ('<s>', split(' ',$_), '</s>');
      for (my $i=0; $i < scalar(@sentence); $i++) {
	$vocCheck{$sentence[$i]} = 1;
      }
    }
    close(SOURCE);
  }


  if ($outputdbfile) {
    if ($BERKELEY_DB) {
      no strict; no warnings;	# disable warnings for this block b/c BerkeleyDB module might not exist
      #      $db1 = tie %outputdbfile,
      $outputdb = new BerkeleyDB::Hash       
	  -Cachesize => 2**20 * $cachesize, 
	    -Filename => $outputdbfile, 
	      -Flags => DB_CREATE|DB_TRUNCATE,
		|| die "Can't create output rule Berkeley DB database $outputdbfile! Bye.";
    } else {
      die "Variable \$BERKELEY_DB is set to 0, therefore the rules cannot be output as database file but just as plain text! Bye.\n";
    }
  }

  # our vocabulary starts with index 1, not zero, so push a dummy into
  # the vocab list
  push @vocablist, '';
  if ($VocabularyFile && (-e $VocabularyFile)) {
    print STDERR "Vocabulary file $VocabularyFile exists -- reading it in.\n";
    # read in given vocabulary
    open(VOCIN, "< $VocabularyFile");
    my $wordsAddedCount = 0;
    while (<VOCIN>) {
      chomp($_);
      # add the word to our vocab
      GetWordIndex($_);
      $wordsAddedCount++;
    }
    close(VOCIN);
    if (scalar(@vocablist) != $wordsAddedCount+1) {
      die "Error: at least one word in vocab file $VocabularyFile occurred twice!";
    }
  }

  # iterate through the Markov-orders plus 1 
  # (i.e, through the n in n-gram, so i=0 is a zero-gram estimate)

  # the following variable will be set to true if in one iteration, we ended up skipping all n-grams
  # b/c their histories were too infrequent
  my $skippedAll = 0;

  # the following hash is for efficiency only: every line number stored in this hash causes
  # that line in $TrainingCorpusFile to be skipped b/c it was useless in the last iteration
  # and thus must be useless in the next iteration too
  my %skipLines = ();

  for (my $n=0; !$skippedAll && $n <= $MaxMarkovOrder+1; $n++) {
    print STDERR "Passing through training data to analyze $n-grams.\n";
    # start the loop with $skippedAll=true until we actually encounter something useful
    $skippedAll = 1;

    # pass through the whole training corpus
    open(TRAIN, "< $TrainingCorpusFile");
    my $lineNumber = 0;
    while (<TRAIN>) {
      $lineNumber++;
      print STDERR "At line number $lineNumber\n" if $lineNumber % 100000 == 0;

      # speed-up: if this line was useless in last iteration, it will also be useless in this iteration
      next if $skipLines{$lineNumber};

      chomp($_);
      # split the line on ' ' and add BOS and EOS markers
      my @sentence_str = ('<s>', split(' ', $_), '</s>');
    
      # convert sentence into index form, mapping words not occuring in the vocab restriction file to 0
      my @sentence = ();
      foreach my $w (@sentence_str) {
	if ($vocabRestrictionFile && !(exists($vocCheck{$w}))) {
	  # not in the restriction set - just convert it to 0
	  push @sentence, 0;
	} else {
	  # get the word index from the vocab, possibly adding the word to the vocab if it didn't exist yet
	  push @sentence, GetWordIndex($w);
	}
      }

      # now iterate through each $n-gram
      my $skippedWholeLine = 1; # will be set to false if any $n-gram of this line was used
      for (my $ngramstart=0; $ngramstart + $n <= scalar(@sentence); $ngramstart++) {
	my @ngram = @sentence[$ngramstart..$ngramstart+$n-1];
	# consider @ngram for statistics only if none of the word indices in @ngram is zero (i.e., didn't occur in $vocabRestrictionFile)
	unless (grep {$_==0} @ngram) {
	  my $ngram_packed = pack('l*', @ngram);
	  if ($n==0) {
	    # zero-gram: just keep increasing the zero-gram counter
	    ++$lm{$ngram_packed};
	    $skippedWholeLine = 0;
	  } else {
	    my $left_sub_ngram_frequency = $lm{pack('l*', @ngram[0..$n-2])};
	    $left_sub_ngram_frequency = 0 if !defined($left_sub_ngram_frequency);
	    my $right_sub_ngram_frequency = $lm{pack('l*', @ngram[1..$n-1])};
	    $right_sub_ngram_frequency = 0 if !defined($right_sub_ngram_frequency);
	    # check that the length ($n-1) sub-grams have the min. required frequencies b/c if they don't, there's no way
	    # that this $n-gram gets that frequency
	    if (($left_sub_ngram_frequency >= $MinFrequency && $right_sub_ngram_frequency >= $MinFrequency) &&
		# also check that at least one of the forward- and backward histories qualifies for $MinHistoryFrequency b/c
		# otherwise we will never be interested in this n-gram's frequency
		($left_sub_ngram_frequency >= $MinHistoryFrequency || $right_sub_ngram_frequency >= $MinHistoryFrequency)
	       ) {
	      # increase this n-gram's count
	      ++$lm{$ngram_packed};
	      $skippedWholeLine = 0;
	    }
	  }
	}
      }
      # if we didn't make use of any n-grams from this line, we also won't make use of any ($n+1)-grams from this line, 
      # so remember this
      if ($skippedWholeLine) {
	$skipLines{$lineNumber} = 1;
      } else {
	# there was at least one line we made use of, namely this one
	$skippedAll = 0;
      }
    }
    close(TRAIN);
  }

  {
    # write out the Markov order info to the DB
    # key is the reserved one-tuple <0>
    if ($MaxMarkovOrder != 1000000) {
      outputngram(pack('l*', 0), $MaxMarkovOrder);
    }

    # write out the n-grams in %lm
    print STDERR "\nWriting out the n-grams...\n";
    #    my $totalrulefreqcount_sanitytest = 0;
    my $rulenumber = 0;

    my ($key, $value);
    while (($key,$value) = each %lm) { 
      $rulenumber++;
      print STDERR "Processing n-gram number $rulenumber...\n" if $rulenumber % 1000000 == 0;
      if ($value >= $MinFrequency) {
	outputngram($key, $value);
      }
    }

#    print STDERR "!!!!!!!!!!!! Something's wrong with the underlying LM database! (Maybe an network file system issue?) \$ruleinstancecount ($ruleinstancecount) != \$totalrulefreqcount_sanitytest ($totalrulefreqcount_sanitytest)!!! \n" if abs($ruleinstancecount - $totalrulefreqcount_sanitytest) > 0.001;
  }

  print STDERR "$writtenoutrulenumber ngrams written out. ";

  if ($VocabularyFile) {
    # still have to write out the vocab file
    print STDERR "Writing out the vocabulary to $VocabularyFile.\n";
    open(VOCOUT, "> $VocabularyFile");
    # note that we start from 1, not 0, since we don't write out the dummy word
    for (my $i=1; $i < scalar(@vocablist); $i++) {
      print VOCOUT ($vocablist[$i] . "\n");
    }
    close(VOCOUT);
  }

  print STDERR 'trainlm.pl - finished @ '. `date`;

  if ($BERKELEY_DB) {
    undef $outputdb if $outputdbfile;
  }

}				# main






#################### subs


# output n-gram to database or standard out
sub outputngram {
  my($key, $frequency) = @_;

  if ($outputdbfile) {
    # $key is already in C++ int32_t format, but we still have to pack $frequency as a C++ double
    $outputdb->db_put($key, pack('d*',$frequency));
  } else {
    print join(' ',(map {$vocablist[$_]} unpack('l*', $key))) . "\#$frequency\n";
  }
  
  $writtenoutrulenumber++;
}


# get vocab ID of the word, adding it to the vocabulary if it wasn't in there yet
sub GetWordIndex {
  my ($wordstr) = @_;

  my $wordid = $vocabhash{$wordstr};
  if (!defined($wordid)) {
    $wordid = scalar(@vocablist);
    die "Oops.... too many unique words. Vocabulary's word IDs got too close too running out of bounds." if $wordid >= 2**31-10000;
    $vocabhash{$wordstr} = $wordid;
    push @vocablist, $wordstr;
  }
  return $wordid;
}
