#ifndef __BleuSearch__
#define __BleuSearch__


#include <MetricSearch.hh>
#include <Vocabulary.hh>
#include <iostream>
#include <fstream>
//#include <istream>
#include <sstream>
#include <pstream.h>

using namespace std;

typedef sgi::hash_map<std::string, double, __StringHash > GainsHash;

struct NgramDetails{
  int numOccurrences;
  int gramSize;
  int indexInSentence;
};

typedef sgi::hash_map<std::string, NgramDetails, __StringHash > SentenceHash;

//specifies a configuration on indices from which to work from
typedef int* Configuration;
typedef std::vector<std::string> NGram;
typedef std::vector<NGram> NGramList;



//Generates statistics to be used in BLEU score calculations
//Handles multiple reference translations, and takes in
//a bank of n-candidate translations
class BleuSearch : public MetricSearch{


protected:

  Score m_maxScore; //!< highest value
  Score m_minEpsilon; //!< value that we consider and actual difference


public:
  
  // loads the raw references from referenceListFile
  // there are numReferences refs per source sentence
  // pbuf is used to normalize the raw reference
  // the pVoc is passed so that any new words cause my normalization can be added to the vocab
  BleuSearch(const char* referenceListFile, redi::pstream& pbuf, bool normalizeReference, const char* gainFile, std::string scoringType, const int numReferences, const int maxGram, CVocabulary* pVoc);

  //default constructor
  BleuSearch();

  //default constructor
  BleuSearch(std::string scoringType);

  ~BleuSearch();
  
  /**
   * Scoring virtual function implementations
   */
  

  
  //================================================================
  // Virtual functions as required by MetricSearch
  //================================================================
  //efficient version that allows you to save the gram data structures
  void ScorePairwise( CCompactTSequence<int>* hypNGrams,  CCompactTSequence<int>* refNGrams,  int refLength,RawCounts& raw);
  

  //virtual void ScorePairwise(CTranslationItem& hyp, CTranslationItem& ref, RawCounts& raw);
  //void ProcessNBestList(CNBest* oneSourceList, int sentenceNumber);
  void ScoreSingleHyp(const int sentenceNumber, string candidate, RawCounts& countObject);

  Score getMetricScore(RawCounts& raw);
  AbstractScore getMetricScoreDetailed(RawCounts& raw); //with the length penalty


  //the highest score bleu can have
  Score maxScore(){
    return m_maxScore;
  }

  Score minEpsilon(){
    return m_minEpsilon;
  }
  //an overloaded version of build reference that fills the hashed datastructure that is required to do matching
  void BuildReferences(const char* fileName, redi::pstream& pbuf, bool normalizeReference, const int numReferences, CVocabulary* pVoc);
  
  // generates the count content for ths oneSource list for this sentence number
  //void SearchNBest(CNBest* oneSourceList, int sentenceNumber);
  
  // works the current sentence number to find the 
  void GetOneCandidateLine(Configuration config, int sentenceNumber, ifstream& infile, ofstream& ofile);
  
  //virtual void ProcessCounts(const int sentenceNumber, CNBestItem* nBestItem, const int numReferences);
  
  double getBleuScore(RawCounts& raw);
  double getNISTScore(RawCounts& raw);

  //score, lpen
  pair<double, double> getBleuScoreDetailed(RawCounts& raw);
  pair<double, double> getNISTScoreDetailed(RawCounts& raw);

  //RawCounts buildRawCounts(vector<CNBest*>& nBestLists, Configuration config);

  bool noDifference(Counter correctDelta, Counter suggestedDelta, double closestRefLengthDelta);
  double CDFNormal(double z);
  

  //these functions moved back into MetricSearch cause they apply for all metrics
  //std::pair<double, RawCounts> optimizeForSentence(RawCounts raw, Configuration config, int sentenceNumber);
  // the primary method that is used to select a best BLEU configuration
  //std::pair<double, RawCounts> optimizeBleuScore();
  void PrintConfiguration(Configuration config);
  void PrintConfigurationToFile(Configuration config, const char* fileName, const char* matchFile);
  // prints all the other candidates that have the same translation string as the candidate
  void PrintSharedElements(int sentenceNumber, int candidateNumber, ostream& st);
  Configuration GetFinalConfiguration();

  //double ScoreConfiguration(Configuration config);
  //score a translation run after at least one iteration has been run thru the system
  //double ScoreTranslationRun(vector<string>& translationSet, vector<CNBest*>& nBestLists);

  
  //sets a base value of 1 for all ngrams levels to do local alternative scoring
  void SetScoringBackups(bool set){ m_scoringBackup = set;}

  string GetScoringType(){
    return m_scoringType;
  }

private:
  

  //set if using local alternative scoring
  bool m_scoringBackup;

  //IBMBLEU, NISTBLEU, NIST
  std::string m_scoringType;

  std::pair<double, RawCounts> CalculateBleuNew(RawCounts raw, Counter correctDelta, Counter suggestDelta, double closestLenDelta);
  
  void PrintHashElements(int sentenceNumber);
  void PrintHashElements(SentenceHash& oneHash);
  //void BuildReferences(const char* fileName, redi::pstream& pbuf, bool normalizeReference, const int numReferences, CVocabulary* pVoc);
  void BuildInfoGains(const char* gainFile);

  SentenceHash ProcessReferences(std::vector<std::string> referenceSet);
  SentenceHash ProcessOneReference(NGram oneSentence);
  // returns a vector of strings
  std::pair<NGram, std::vector< pair<int, int> > > CollectNGrams(NGram source, int maxNSize);

  std::vector<int> m_referenceLengths;
  std::vector<SentenceHash> m_referenceList;

  //used to calculate the NIST data
  //info gains hash
  GainsHash* mp_gainsHash;
  //store the number of grams of size N are seen in the references
  Counter m_referenceNGramTotals;


  // the common indices bank for the whole training set
  //std::vector<OneSentenceCommonIndices> m_AllCommonIndices;


  vector< vector<Counter*> > m_numberCorrect;
  vector< vector<Counter*> > m_numberSuggested;
  vector< vector<double*>   > m_closestRefLength;
  
  // the number of candidates available for each source sentence
  std::vector<int> m_numCands;

  // number of reference we are scoring against
  //int m_numReferences; // moved into the metric search
  
  //number of n-grams we are considering
  //ARV
  //replaced by m_maxNGramConsidered in MetricSearch
  //int m_maxNGramSize;

  //initial configuration model
  Configuration m_config;
  //initial raw counts model
  //RawCounts m_raw;

  // set to true if we are only doing a scoring step
  // rather than a full optimization
  bool s_ScoringOnly;
  bool s_TraceOn;

  //Beta for NIST CALC
  double BETA;
};

#endif












