/*
 *  EvaluatorTER.h
 *  STTK_Anthony2
 *
 *  Created by Anton D'Auria on 4/9/06.
 *  Original perl implementation by Matthew Snover (snover@cs.umd.edu) 
 *
 */


#ifndef __EvaluatorTER__
#define __EvaluatorTER__

//#include <Evaluator.hh>
#include <Vocabulary.hh>
#include <set>
#include <deque>
#include <iostream>
#include <vector>

using namespace std;

//-- Phrase, a vector of CWordIndexS
typedef std::vector< CWordIndex > Phrase;

//-- integer set with elements in ascending order
typedef std::set< int, std::less<int> > IntSet;


//-- Hash function for Phrase so it can be used as a key in a hash_map

namespace __gnu_cxx {
  template<> struct hash< Phrase > {
    size_t operator()(const Phrase& p) const {
      size_t signature = 0;
      int primes[] = {8933, 8941, 8951, 8963, 8969, 8971, 8999, 9001, 9007, 9011};
      for( int i = p.size()-1; i >= 0; --i ){
        signature +=  p[i] * primes[i];
      }
      return signature;
    }
  };

}

/*
namespace __gnu_cxx {
  template<> struct hash< Phrase > {
    size_t operator()(const Phrase& p) const 
    {
      size_t keySum = 0;

      for(int i=1; i < p.size(); i++) { // should start with 0??
        keySum += (size_t) (int) pow(float(*(p[i])), float((j % 6)+1));
      }
      
      return (keySum % 1000000);
      
    }
  };
}
*/

//-- Equality function for Phrase so it can be used as a key in a hash_map

struct equalPhrase
{
  bool operator() (const Phrase p1, const Phrase p2) const 
  {
    if( p1.size() != p2.size() )
      return false;
    
    for(int i=1; i < p1.size(); i++)
    {
      if( p1[i] != p2[i] )
        return false;
    }
    
    return true;

  }
};

//-- hash_map, with key: Phrase, value: set<int> ordered asc
typedef __gnu_cxx::hash_map< Phrase, IntSet, __gnu_cxx::hash< Phrase >, equalPhrase > PhraseToSet_hash_map;


//====================================================================
//        CEvaluatorTERcost
//

class CEvaluatorTERcost
{
public:
  
  CEvaluatorTERcost();
  
  float Match() const;
  float Substitution() const;
  float Insertion() const;
  float Deletion() const;
  float Shift() const;
  
  void SetMatchCost(float cost);
  void SetSubstitutionsCost(float cost) ;
  void SetInsertionCost(float cost);
  void SetDeletionCost(float cost);
  void SetShiftCost(float cost);
  
  
private:
  float m_matchCost;
  float m_substitutionCost;
  float m_insertionCost;
  float m_deletionCost;
  float m_shiftCost;
};


//====================================================================
//        CEvaluatorTERshift::
//

class CEvaluatorTERshift
{
public:
  CEvaluatorTERshift();
  
  CEvaluatorTERshift(int start, int end, int moveto, int newloc);
  
  ~CEvaluatorTERshift();
  
  int Distance() const;
  
  bool LeftShift() const;
  
  int Size() const;
  
  size_t GetStart() const;
  size_t GetEnd() const;
  size_t GetMoveto() const;
  size_t GetNewloc() const;
  Phrase GetShiftedPhrase() const;
  float GetCost() const;
  void Print() const;
  
  void SetStart(size_t start);
  void SetEnd(size_t end);
  void SetMoveto(size_t moveto);
  void SetNewloc(size_t newloc);
  void SetShiftedPhrase(Phrase& shiftedPhrase);
  void SetCost(float cost);
  
private:
  size_t m_start;
  size_t m_end;
  size_t m_moveto;
  size_t m_newloc;
  Phrase m_shiftedPhrase;
  
  float m_cost;
};


//====================================================================
//        CEvaluatorTERalignment::
//

class CEvaluatorTERalignment
{
public:
  CEvaluatorTERalignment();
  CEvaluatorTERalignment(Phrase &hyp, Phrase &ref,
                         std::deque< unsigned short > medPath, 
                         float numEdits,
                         size_t numRefWords);
  ~CEvaluatorTERalignment();
  
  float                              ComputeScore() const;
  
  Phrase                              GetRef() const;
  Phrase                              GetHyp() const;
  Phrase                              GetHypPrime() const;
  std::vector< CEvaluatorTERshift* >* GetAllShifts() const; 
  float                               GetNumEdits() const;
  size_t                              GetNumRefWords() const;
  size_t                              GetNumHypWords() const;
  std::deque< unsigned short >        GetMedPath() const;
  
  void Print() const;
  
  void SetRef(Phrase r);
  void SetHyp(Phrase h);
  void SetHypPrime(Phrase s);
  void SetAllShifts(std::vector< CEvaluatorTERshift* >* shifts);
  void SetNumEdits(float numEdits);
  void SetNumRefWords(size_t numRefWords);
  void SetMedPath(std::deque< unsigned short > medPath);
  
private:
  Phrase m_ref;
  Phrase m_hyp;
  Phrase m_hypPrime;              // in UMD code: aftershift
  
  std::vector< CEvaluatorTERshift* >* m_p_allShifts; 
  
  float m_numEdits;
  size_t m_numRefWords;

  std::deque< unsigned short > m_medPath;       // const time to add at begging & random access
  
};


//====================================================================
//        CEvaluatorTER
//


class CEvaluatorTER //: public CEvaluator
{
public: 
  
  //CEvaluatorTER( CVocabulary *pVocabulary, CEvaluatorSharedData *pSharedData );
  CEvaluatorTER();
  
  ~CEvaluatorTER();
  
  // overload methods to calculate the score or scores for the translation
  // scores to be returned: cumulative NIST scores 1-maxNGram, Length Penalty, 
  // # of hypothesis senteces, sum of avg # of words in reference translations
  // bool Evaluate( const char *pFileWithTranslations, CNumericalArray< float > & Scores );
  // bool Evaluate( CArray< CString > &Translations, CNumericalArray< float > & Scores );
  // bool Evaluate( CText *pTranslations, CNumericalArray< float > & Scores );
  // bool Evaluate( const vector< CNumericalArray< CWordIndex > > *pTanslations, 
  //      CNumericalArray< float > & Scores );
  bool Evaluate( int srcSentIndex, vector<CWordIndex>& hypSentence, vector< vector<CWordIndex> >& references, vector< float >& Scores);

  
  // calculate the oracle score for a nbest list; Format: n m word sequence
  // oracle scores means: take those translations in nbest lists which optimize the scores
  // no matter if read from file, or getting the translations in arrays
  /*
  bool GetOracleScore( const char *pFileWithNBestTranslations,
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( vector< CString > *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( CText *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( vector< CNumericalArray< CWordIndex > > *pNBestTranslations, 
                       CNumericalArray< float > & Scores );
  bool GetOracleScore( vector< vector< CSentence* > >* p_NBestTranslations,
                       CNumericalArray< float >& Scores );
  */
  //std::pair<float, float> GetConfidenceInterval();
    
  //void ReportScores( CNumericalArray< float > & Scores );

  //   Min edit distance step and Scores[] keys
  size_t m_SCORE_KEY;
  size_t m_INSERTION_KEY;
  size_t m_DELETION_KEY;
  size_t m_SUBSTITUTION_KEY;
  size_t m_SHIFT_KEY;
  size_t m_MATCH_KEY;
  
  size_t m_HYP_WORD_COUNT_KEY;
  size_t m_REF_WORD_COUNT_KEY;
  size_t m_EDIT_COUNT_KEY;
  size_t m_HYP_SENT_COUNT_KEY;


protected:
  
        
  //------------------------------------------------------------------
  //      CEvaluatorTER:: BuildWordMatches
  //

  PhraseToSet_hash_map* BuildWordMatches(vector<CWordIndex>& hyp, vector<CWordIndex>& ref);
  
  
  
  //------------------------------------------------------------------
  //      CEvaluatorTER:: BuildPhrase
  //
  //        Creates a Phrase (vector<CWordIndex*>) from CSentence*
  //          creates subphrase beginIndex inclusive, endIndex exclusive

  //Phrase BuildPhrase(vector<CWordIndex>& sentence, int beginIndex, int endIndex);
  Phrase BuildPhrase(vector<CWordIndex>& sentence);
  Phrase BuildPhrase(Phrase& phrase, int beginIndex, int endIndex); //useful since it returns a shorter phrase
  
  
  
  //------------------------------------------------------------------
  //      CEvaluatorTER:: BuildValidShifts
  //
  
  std::vector< std::vector< CEvaluatorTERshift* > >* 
    BuildValidShifts(Phrase& hypPrime, Phrase& ref, 
                     PhraseToSet_hash_map* p_phraseStartHash,
                     vector< bool >& herr, vector< bool >& rerr,
                     vector< int >& ralign, 
                     CEvaluatorTERcost& costFunc);


  
  //------------------------------------------------------------------
  //      CEvaluatorTER:: ComputeMinEditDistance
  //
  
  CEvaluatorTERalignment* ComputeMinEditDistance(Phrase &hyp, Phrase &ref,
                                                 CEvaluatorTERcost &constFunc);
  
  //------------------------------------------------------------------
  //      CEvaluatorTER:: PerformShift
  //
  
  Phrase PerformShift(Phrase& origPhrase, CEvaluatorTERshift* p_shift);
  
  Phrase PerformShift(Phrase& origPhrase, int beginIndex, int endIndex,
                      int toIndex);
  
  //------------------------------------------------------------------
  //      CEvaluatorTER:: PrintPhrase
  //

  void PrintPhrase(Phrase &p) const;
  
  
  //------------------------------------------------------------------
  //      CEvaluatorTER:: FindBestShift
  //

  std::pair< CEvaluatorTERshift*, CEvaluatorTERalignment* >*
    FindBestShift(Phrase& hypPrime, Phrase& hyp, Phrase& ref,
                  PhraseToSet_hash_map* p_phraseStartHash,
                  CEvaluatorTERalignment* p_curHypRefAlignment,
                  CEvaluatorTERcost& constFunc);
  
  
  //------------------------------------------------------------------
  //      CEvaluatorTER:: FindAlignError
  //
  
  void FindAlignError(CEvaluatorTERalignment* p_alignment,
                      vector< bool >& herr, vector< bool >& rerr,
                      vector< int >& ralign );
 
  
  //-- Members--------------------------------------------------------
  
  //   Shift searching constants
  int m_MAX_SHIFT_SIZE;
  int m_MAX_SHIFT_DIST;
  int m_MAX_BEAM_WIDTH;
  
  size_t m_NUM_SHIFTS_CONSIDERED;
  size_t m_NUM_BEAM_SEARCH_CALLS;
  

};



#endif
