#ifndef __OPTIMIZEMER__
#define __OPTIMIZEMER__

#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <cassert>
#include <functional>
#include <numeric>
#include <cfloat>
#include <climits>


//my files
#include <Optimize.hh>

using namespace std;

//!A value with a corresponding index
typedef pair<double, int> IndexedValue;

//! A data instance with multiple possible features and multiple scores for each feature
struct Instance{
  pair< vector<Features*>::const_iterator, vector<Features*>::const_iterator>  m_dataRange;
  pair< vector<RawCounts*>::const_iterator, vector<RawCounts*>::const_iterator>  m_scoreRange;

  //for each dimension maintain a list of slopes in increasing order
  vector< vector<IndexedValue> > m_decreasingSlopes;

  //for each candidate, maintain a currently scaled set of dimensions and a total sum
  //this was calculation of a 'intercepts' simply involves deleting out vector_i from the sum
  //and when a value of scaling factor is changed, we need to delete the old value from the sum
  // overwrite the old fields and add the new value back to the sum
  vector< pair< vector<double>, double> > m_scaledFeaturesAndSum;

  //!number of candidates represented
  short m_numCands;
};

//!Pairs error counts - or diff counts at a location
typedef pair<double, RawCounts> ErrorAt;

//ostream& operator << (ostream& os, const IndexedValue& s){
//os<<s.first<<" "<<s.second;
//}


//!Object to sort ErrorAt
class DifferentLambda : public binary_function<const ErrorAt&, const ErrorAt&, bool>{
public: 
  bool operator()(const ErrorAt& i1, const ErrorAt& i2) const{
    return i1.first!=i2.first;
  }
};
// binary operator as non-member function for ErrorAt
ErrorAt operator+(const ErrorAt& left, const ErrorAt& right);


//!Object to sort ErrorAt by its first  field
class SortErrorAt{
public: 
  bool operator()(const ErrorAt& i1, const ErrorAt& i2) const{
    return i1.first<i2.first;
  }
};

//!Object to sort IndexedValue
class SortIndexedValue{
public: 
  bool operator()(const IndexedValue& i1, const IndexedValue& i2) const{
    return i1.first<i2.first;
  }
};

//!Object to sort IndexedValue
class SortIndexedValueDecreasing{
public: 
  bool operator()(const IndexedValue& i1, const IndexedValue& i2) const{
    return SortIndexedValue()(i2,i1);
  }
};

inline double calc_innerProduct(const Features* f, const Lambdas& lambdas){
  return inner_product(f->begin(), f->end(), lambdas.begin(), 0.0, plus<double>(), multiplies<double>());
}


class COptimizeMER : public COptimize{
  
public:

  //!Default constructor only set the print tag (which prefixes any output from this component
  COptimizeMER(){ m_printTag= "MER-Opti:";}

  //!initialize against a metric, and sets the print tag
  COptimizeMER(MetricSearch* pMetric) : COptimize(pMetric){ m_printTag= "MER-Opti:";}


  //!Sets the left and right range limits for all dimensions
  void SetRangeLimits(const Lambdas& leftRange, const Lambdas& rightRange){
    m_leftRange = leftRange;
    m_rightRange = rightRange;
  }


  /*!Initialize optimizer with this method if you dont have a param file
   *nunPermutes specifies number of times we randomly change numModify dimensions AFTER convergence
   *numStarts specifies the number of full random starts that are done
   */
  void Init(Lambdas& leftRange, Lambdas& rightRange, int numPermutes, int numModify, int numStarts, bool twoStage, bool doNorm, double permutationEpsilon, double epsilonScale=1.0);

  void SetStart(int start); //sets the m_startBlank
  
  /*! Finds lambdas that when applied to the costs (lambdas*costs) produces the lowest possible scores against the initialized pMetric
   *  firstXsentences indicate that we also want to report the metric score on the firstXsentences as well so we know what kind of gain we can expect
   */
  virtual Lambdas minModelCostMaxMetricScore(const vector<Features*>& costs, const vector<RawCounts*>&  scores, const vector<int>& cands, const Lambdas& initialValues, int firstXsentences);

protected:

  //!All output generated by this component will be preceded by this tag
  string m_printTag;

  //!Left range for the lambda values
  Lambdas m_leftRange;

  //!Right range for the lambda values
  Lambdas m_rightRange;

  //! Once you get convergence (no reduction in error), jump around a little bit and see if you can find improvement
  int m_numPermutes;

  //! Run the second stage of optimization (where certain fields are blanked out)
  bool m_twoStage;

  //!Number of parameters to modify after each iteration
  int m_numModify;

  //! Starting point for blanking for this MER training run
  static int m_startBlank;

  //!Number of times random values are selected to re start the search process
  int m_numStarts;

  //!Renormalize values after each iterationn
  bool m_doNorm;
  
  //!Value at which convergence occurs
  double m_permutationEpsilon;

  //!Scale up the epsilon used to determine terminal for two stage mode
  double m_epsilonScale;

  //! A large static double in several calculations
  const static double LargeCost = FLT_MAX;

  template<typename T>
  pair<int, T> minValueAndIndex(vector<T>& v, T notMin){
    T minValue = notMin;
    int minIndex=0;
    int index=0;
    for(typename vector<T>::iterator iter=v.begin(); iter!=v.end(); ++iter){
      if ((*iter)<minValue){
	minValue = (*iter);
	minIndex = index;
      }
      index++;
    }
    return pair<int, T>(minIndex, minValue);
  }

  //now for MER utility functions

  //verifiy the current rankings and return the error at this ranking
  double VerifyCurrentRanks(vector<Instance>& allSentences, const Lambdas& lambdas);

  //!Calculate model best counts when selected by lambdas - you can indicate that you only want the score on the firstXsentences 
  RawCounts ScoreWithLambdas(const vector<Instance>& allSentences, int firstXsentences, const Lambdas& lambdas);

  //!L2 norm of lambdas
  double L2(const Lambdas& lambdas);

  //get the lowest cost candidate for one sentence given current scaling factors
  pair<int, RawCounts> lowestCostCandidate(const Instance& oneSentence, const Lambdas& lambdas);

  //!Picks random values in range for the lambdas specified in the vector selectRandom as true
  void RandomPermutation(Lambdas& returnLambdas, const vector<bool> selectRandom);

  //!Normalize the lambdas in toModify against the length specified in reference
  void NormalizeAgainstInitial(Lambdas& toModify, const Lambdas& reference);

  //!Randomly select numToSelect dimensions to have true values, and all the others are false;
  vector<bool> SelectForSelection(const int totalNumElements, const int numToSelect);


  //see not in src
  pair<Lambdas, double> optimizeByGainTwoStage(vector<Instance>& allSentences, const Lambdas& initialValues, double initialError);

  pair<Lambdas, double> optimizeByGainWithPermutationsWithRandomStarts(vector<Instance>& allSentences, const Lambdas& initialValues, double perc, int numPermutes, int numStarts);
  //by gain with permutations across iterations
  pair<Lambdas, double> optimizeByGainWithPermutations(vector<Instance>& allSentences, const Lambdas& initialValues, const Lambdas& constantInitialValues);

  pair<Lambdas, double> optimizeByGain(vector<Instance>& allSentences, const Lambdas& initialValues);
  
  Lambdas optimizeInOrder(vector<Instance>& allSentences, Lambdas& initialValues);

  //get back a lambda and an its corresponding low error value
  pair<double, double>  optimizeDimension(int dim, const Lambdas& initialValues, vector<Instance>& allSentences);

  void StressTest(int dim, double currentError, vector<Instance>& allSentences, const Lambdas& initialValues);

  //!use the raw counts and scan thru to find the lowest error
  pair<double, double> minPointError(vector<ErrorAt>& uniqueErrorDeltas, double lowestLambda, vector< pair<double, double> >& errorSurface, bool fillErrorSurface);

  //!merge the error deltas from multiple sentences
  void mergeErrorDeltas(vector<ErrorAt>& totalErrorDelta, vector<ErrorAt>& uniqueErrorDeltas );
  vector<ErrorAt> singleSentencePoints(const int dim, Lambdas initialValueCopy, Instance& oneSentence);
  pair<double, int> findIntersection(int dim, int minLeftCandidate, double slope_minLeftCandidate, vector<double>& a, vector< IndexedValue >::iterator start, vector< IndexedValue >::iterator end);
  //pair<int, double> getBoundsCandidates(const Instance& oneSentence, vector<double>& a, vector<double>& slopes, double scalingValue, int dim);
  pair<int, double> getBoundsCandidates(const Instance& oneSentence, vector<double>& a, double scalingValue, int dim);

  void setupSentenceRanges(vector<Features*>::const_iterator iter_cost, vector<RawCounts*>::const_iterator iter_score, vector<Features*>::const_iterator end_cost, vector<RawCounts*>::const_iterator end_score, const vector<int>& startEnd, vector<Instance>& allSentences, int numDim);

  //accessor functions to get a particular score/feature or slope
  RawCounts* GetScore(const Instance& oneSentence, int candidate);
  double GetSlope(const Instance& oneSentence, int candidate, int dim);

  //functions use to take advantage of precomputation
  void InitializeDotProducts(Instance& oneSentence, const Lambdas& initial); //call before checking each dimension
  void FillIntercepts(Instance& oneSentence, int dim, vector<double>& a); //call to zero out dimension dim and return interecept values
  void RecomputeScaledSums(Instance& oneSentence, int dim, double oldScalingFactor, double newScalingFactor); //call when a change has been accepted

  //! greater than or equal to the min value, less than the max value
  vector<IndexedValue>::iterator findGeqLt(vector<IndexedValue>::iterator begin, vector<IndexedValue>::iterator end, double minValue);
  //pair<vector<IndexedValue>::iterator, vector<IndexedValue>::iterator> findGeqLt(vector<IndexedValue>::iterator begin, vector<IndexedValue>::iterator end, float minValue, float maxValue, int& numRelevant);

};

#endif
