#ifndef __CACHEDSRILM__
#define __CACHEDSRILM__

#include <Vocabulary.hh>

#include <Ngram.h>
#include <LM.h>
#include <Vocab.h>
#include <map>
#include <vector>
#include <cfloat>
#include <climits>
#include <ParserInfo.hh>
#include <LanguageModelSAMT.hh>


// undefined log prob returned when word sequence couldn't be found in Cache - 
// any positive value will do as logprobs are always negative
#define NOT_IN_CACHE FLT_MAX

//stores the lm cost for  existing sequences
typedef SGI_HASH_NAMESPACE::hash_map< std::vector<CWordIndex>, CLogProb, LMCWordSequenceHash, LMCWordSequenceEquals> LMCache;




class CachedSriLM : public CLanguageModel {

public:

  CachedSriLM(const string& paramFile, CLMType lmType);
  ~CachedSriLM();

  virtual CNegLogProb GetSingleWordLMCost(const CWordSequence& lmString, int wordPos);

  VocabIndex GetVocabularyIndex(const std::string& word){ //converts from string to LMWordIndex, note, not CWordIndex 
    VocabIndex i = vocab->getIndex(word.c_str());
    if (i==Vocab_None){
      //cout<<"unk word:"<<word<<" - "<<vocab->unkIndex<<endl;
      return vocab->getIndex(Vocab_Unknown);
    }
    return i; 
  }

//   //VocabIndex GetVocabularyIndex(CString word); //uses the CString form


  CLogProb LogProb(vector<CWordIndex> wordMostNext); //! CWordIndex form in order indicated
  CLogProb StoreInCache(CLogProb value, const vector<CWordIndex>& wordMostNext);
  CLogProb GetFromCache(const vector<CWordIndex>& wordMostNext);
  bool InCache(const vector<CWordIndex>& wordMostNext);
  //! might want to call this after each sentence to let the cache not grow too big
  virtual void ClearCache();
  
private:
  
  Vocab *vocab; //SRI vocab; gets loaded when the Ngram is loaded
  Ngram *ngramLM;   //the actual SRI model
  //uses the constant defined in the Types file, this variable avoid needing to new/delete this vector each time
  VocabIndex* m_context;
  //the caching of the lookups
  LMCache m_cache;

};



inline CLogProb CachedSriLM::StoreInCache(CLogProb value, const vector<CWordIndex>& wordMostNext){
  //ensure that its not already there?
  assert(m_cache.count(wordMostNext)==0);
  m_cache.insert(pair<vector<CWordIndex>, CLogProb>( wordMostNext, value));
  return value;
}
inline CLogProb CachedSriLM::GetFromCache(const vector<CWordIndex>& wordMostNext){
  //make sure its there
  //  if (m_cache.count(wordMostNext)!=0);
  LMCache::iterator iter = m_cache.find(wordMostNext);
  return iter==m_cache.end()? NOT_IN_CACHE : (*iter).second;
}
inline bool CachedSriLM::InCache(const vector<CWordIndex>& wordMostNext){
  if (m_cache.count(wordMostNext)!=0){
    return true;
  }else{
    return false;
  }
}

inline void CachedSriLM::ClearCache(){
  m_cache.clear();
}


// note: this function handles BOS and EOS implicitely, assuming that the respective words are in 
// SRI style, i.e., <s> and </s>, respectively
inline CLogProb CachedSriLM::LogProb(vector<CWordIndex> wordMostNext){

  // if wordMostNext is in cache, return correct value immediately
  // if (InCache(wordMostNext)){
  CLogProb value = GetFromCache(wordMostNext);
  CChartParserInfo::m_lmCall++;
    //    if (s_debugLM){cout<<"In cache: "<<"p ("<<m_pVocabulary->GetWord(wordMostNext[0])<<" | ...) = "<<value1<<endl;}
  if (value != NOT_IN_CACHE){
    CChartParserInfo::m_lmCacheHit++;
    return value;
  }

  int wi = GetVocabularyIndex(CChartParserInfo::m_pVocabulary->GetWord(wordMostNext[0]));
  // az: SRI returns inf for unknown words, so we have to do our own handling
  if (wi == vocab->getIndex(Vocab_Unknown) || wordMostNext[0]==CChartParserInfo::m_unknownWI){
      //    cout<<"UnknownToLM: "<<wi<<" word: "<<m_pVocabulary->GetWord(wordMostNext[0])<<endl;
      return StoreInCache(log10(1.0/vocab->numWords())-4, wordMostNext); //az 06/02/09
  }

#ifndef NDEBUG
  if (s_debugLM){
    cout<<"p ("<<CChartParserInfo::m_pVocabulary->GetWord(wordMostNext[0])<<" | ";
  }
#endif

  int i=0;
  //the MIN is used to allow LogProb to back off if you ask for the shorter LogProb
  //like a bi-gram context where you only have a length 2 wordMostNext
  for (i=0; i<MIN(wordMostNext.size()-1, CChartParserInfo::C_SRI_HistoryLength); i++){
    m_context[i] = 
      wordMostNext[i+1] == CChartParserInfo::m_unknownWI ? 
            vocab->getIndex(Vocab_Unknown) : // here unknown-handling does matter; we want a smart back-off strategy, let SRI do that
      GetVocabularyIndex(CChartParserInfo::m_pVocabulary->GetWord(wordMostNext[i+1])); //the first word is the word we are looking up

#ifndef NDEBUG
    //debug... spell it out
    if (s_debugLM){
      if (wordMostNext[i+1]==CChartParserInfo::m_unknownWI){
	cout<<"unkIndex ";
      }else{
	cout<<CChartParserInfo::m_pVocabulary->GetWord(wordMostNext[i+1])<<" ";
      }
    }
#endif

  }

  //put Vocab_None in all spots after the context that we are looking for including the last m_context[C_SRI_HistoryLength] marker
  for (;i<=CChartParserInfo::C_SRI_HistoryLength; i++){
    m_context[i] = Vocab_None;
  }
  value = ngramLM->wordProb(wi, m_context);

#ifndef NDEBUG
  if (s_debugLM){
    cout<<") = "<<value<<endl;
  }
#endif

  return StoreInCache(value, wordMostNext);
  
}



#endif
