#ifndef __LanguageModelSAMT__
#define __LanguageModelSAMT__

#include <Vocabulary.hh>

//make sure you include the db_cxx for the c++ calls
#include <db_cxx.h>
#include <Ngram.h>
#include <LM.h>
#include <Vocab.h>
#include <map>
#include <vector>
#include <cfloat>
#include <climits>
#include <ParserInfo.hh>

using namespace std;

//! whether our probs are of type float or double
typedef float CProb;
//! whether our log probs are of type float or double
typedef float CLogProb;
//! whether our neg-log probs are of type float or double
typedef float CNegLogProb;
//! type of an ngram frequency: make sure this is consistent with the type trainlm.pl uses
typedef double CFrequency;

//! type of LM
typedef int CLMType;
#define FORWARDLM 0 //!< left-to-right
#define BACKWARDLM 1 //!< right-to-left
//! context is considered at left and right of the current word
#define CENTERLM 2 



//! used by hash-map based LM caches
struct LMCWordSequenceHash{
  size_t operator()(const vector<CWordIndex>& c)const{
    size_t hashCode = 1;
    for (int i=0; i<c.size(); i++){
      hashCode = 31*hashCode + c[i];
    }
    return hashCode;
  }
};

//! used by hash-map based LM caches
struct LMCWordSequenceEquals{
  bool operator()(const vector<CWordIndex>& c1, const vector<CWordIndex>& c2 )const{
    return (c1==c2);
  }
};

//!only used by the SAMT LM: stores the LM costs for existing sequences
typedef SGI_HASH_NAMESPACE::hash_map< CWordSequence, pair<CNegLogProb,int>, LMCWordSequenceHash, LMCWordSequenceEquals> CCacheSAMTLM;

// the following constants are for the second (int) component of the pair returned by CLanguageModelSAMT's cache
//! the n-gram requested was not found in the cache
#define SAMTLM_NOT_IN_CACHE 0
//! the n-gram requested was found in the cache and its history is not expandable
#define SAMTLM_FINAL_NGRAM 1
//! the n-gram requested was found in the cache and its history is still expandable (this happens when LM-querying for word sequences with insufficient context)
#define SAMTLM_NONFINAL_NGRAM 2






class CLanguageModel{

public:

  // stores all the LMs we are using - possibly multiple models per file
  static vector<CLanguageModel*> m_languageModels;

  //turn this on or off from OUTSIDE the language model to see what context+scores its operating with
  static bool s_debugLM;

  // type of this LM
  CLMType m_lmType;

  CLanguageModel(CLMType lmType) : m_lmType(lmType)
  {}

  //! Gets the LM prob of the word lmString[wordPos], where the context is given by the other words in lmString. 
  /*!
    If context is longer than we need, use only the X words before (FORWARDLM) / after (BACKWARDLM) wordPos.
    In case context only has less elements than we need, back off.
    Note: the context may contain m_lmBoundaryWI ("<*>"). Only the words in lmString between the right-most <*> left of wordPos
    and the left-most <*> right of wordPos are valid context words.
  */
  virtual CNegLogProb GetSingleWordLMCost(const CWordSequence& lmString, int wordPos) = 0;

  //! might want to call this after each sentence to let the cache not grow too big
  virtual void ClearCache() 
  {}

};

class CDatabaseSAMTLM {

public:

  //! stores all the SAMT DBs statically
  static vector<CDatabaseSAMTLM*> m_samtlmDatabases;

  //!database object, needs to be a pointer cause there is no default constructor
  Db* m_db;
  
  //! construct from database file
  CDatabaseSAMTLM(const string& file, bool storedAsBTree=false);

  //! Close and delete underlying database. 
  ~CDatabaseSAMTLM();

};


class CLanguageModelSAMT : public CLanguageModel {
public:
  //!database object, needs to be a pointer cause there is no default constructor
  Db* m_db;
  //  bool m_dbWasCopied;
  int m_maxMarkovOrder;
  CFrequency m_minHistoryFrequency;
  double m_minBackoffWeight;
  double m_historyFrequencyConvergenceSpeed;
  double m_inverseBreakEvenHistoryFrequency;
  double m_uniqueEventCountImpact;

  // the following variables are set in construction time as a speed-up to not recalculate them for each probability query
  // they are all integers but can be bigger than 2 billion, so we use doubles to represent them
  CFrequency m_totalEventCount;
  CFrequency m_unassignedUnigramFrequencyMass;
  CFrequency m_underlyingVocabSize;
  double m_initialProb;

  //stores the lm cost for  existing sequences
  CCacheSAMTLM m_cache;


  //! construct from database
  CLanguageModelSAMT(Db* db, CLMType lmType, CFrequency minHistoryFrequency, double minBackoffWeight, double historyFrequencyConvergenceSpeed, CFrequency breakEvenHistoryFrequency, double uniqueEventCountImpact) :
    CLanguageModel(lmType),
    m_db(db), 
    m_minHistoryFrequency(minHistoryFrequency),
    m_minBackoffWeight(minBackoffWeight),
    m_historyFrequencyConvergenceSpeed(historyFrequencyConvergenceSpeed),
    m_inverseBreakEvenHistoryFrequency(1.0 / breakEvenHistoryFrequency),
    m_uniqueEventCountImpact(uniqueEventCountImpact)
  {
    cout<<"  Creating "<< (lmType==BACKWARDLM? "BACKWARD ":"") << "SAMT language model with SAMTLM_MinHistoryFrequency=" << m_minHistoryFrequency
	<< ", SAMTLM_MinBackoffWeight=" << m_minBackoffWeight
	<< ", SAMTLM_HistoryFrequencyConvergenceSpeed=" << m_historyFrequencyConvergenceSpeed
	<< ", SAMTLM_BreakEvenHistoryFrequency=" << breakEvenHistoryFrequency
	<< ", SAMTLM_UniqueEventCountImpact=" << uniqueEventCountImpact
      ;

    if (m_minHistoryFrequency <= 0) {
      cerr << "Error: SAMTLM_MinHistoryFrequency has to be greater than zero!" << endl;
      exit(1);
    }

    CFrequency dummyFreq; // just some dummy variables we don't care about

    // initialize m_totalEventCount with zero-gram's frequency count (i.e., the total number of word instances in the training corpus)
    CWordSequence zerogram;
    bool okay = GetTupleStats(zerogram, m_totalEventCount, m_unassignedUnigramFrequencyMass, m_underlyingVocabSize);
    if (!okay) {
      cerr << "Error: zero-tuple () is not in the language model" << endl;
      exit(1);
    }
    // and set the base prob estimate (influences mainly the probability assigned to unknown words, 
    // which is of no relevance for machine translation as every hypothesis will have the same number of unknown-words
    m_initialProb = 1.0 / m_totalEventCount;

    // check that our unknown word is not used by the LM
    if (GetTupleStats(CWordSequence(1,CChartParserInfo::m_unknownWI), dummyFreq, dummyFreq, dummyFreq)) {
      cerr << "Error: UNKNOWN-Symbol is used by the language model! Change the UNKNOWN-Symbol of the grammar to something else and specify it via --UnknownWord.\n";
      exit(2);
    }
    
    // also retrieve the max. Markov order, stored as artificial one-tuple <0> 
    // (remember that word indices start with 1; thus this will cause no collision)
    cout << ", MaxMarkovOrder=";
    CWordSequence keyForMarkovOrder(1,0);
    CFrequency maxMarkovOrder;
    if (GetTupleStats(keyForMarkovOrder, maxMarkovOrder, dummyFreq, dummyFreq)) {
      m_maxMarkovOrder = (int) maxMarkovOrder;
      cout << m_maxMarkovOrder;
    } else {
      m_maxMarkovOrder = INT_MAX/10;
      cout << "unlimited";
    }
    cout << ", m_totalEventCount=" << m_totalEventCount << endl;
  }

  virtual CNegLogProb GetSingleWordLMCost(const CWordSequence& lmString, int wordPos);
  
  bool GetTupleStats(const CWordSequence& tuple, CFrequency& tupleFreq, CFrequency& unassignedFrequencyMass, CFrequency& uniqueSuccessorCount);

  //!Utilty method to see exactly what got put into the db
  void PrintDB();

  void StoreInCache(pair<CNegLogProb,int> value, const CWordSequence& wordMostNext);
  pair<CNegLogProb,int> GetFromCache(const CWordSequence& wordMostNext);
  //! might want to call this after each sentence to let the cache not grow too big
  virtual void ClearCache() {
    m_cache.clear();
  }
  
};



inline void CLanguageModelSAMT::StoreInCache(pair<CNegLogProb,int> value, const CWordSequence& wordMostNext){
  //ensure that its not already there
  assert(m_cache.count(wordMostNext)==0);
  m_cache.insert(pair<vector<CWordIndex>, pair<CNegLogProb,int> >( wordMostNext, value));
}
inline pair<CNegLogProb,int> CLanguageModelSAMT::GetFromCache(const CWordSequence& wordMostNext){
  //make sure its there
  //  if (m_cache.count(wordMostNext)!=0);
  CCacheSAMTLM::iterator iter = m_cache.find(wordMostNext);
  return iter==m_cache.end()? pair<CNegLogProb,int>(0,SAMTLM_NOT_IN_CACHE) : (*iter).second;
}


inline bool CLanguageModelSAMT::GetTupleStats(const CWordSequence& tuple, CFrequency& tupleFreq, CFrequency& unassignedFrequencyMass, CFrequency& uniqueSuccessorCount) {
  
  //  cout<<"Dumping DB"<<endl;
  //  PrintDB();
  //  cout<<endl;
  //  cout<<"End Dumping DB"<<endl;

  //  CWordSequence tuple(tuple1);
  //  CWordIndex buf[tuple.size()];
  //  for (int i=0; i<tuple.size(); i++) {
  //    buf[i] = tuple[i];
  //  }


  //Dbt data(search_data, strlen(search_data) + 1);
  Dbt data;

  bool dbError;

  if (CChartParserInfo::SAMTLM_UseBERFormat) {
    vector<uint8_t> byteSeq;
    for (int i=0;i<tuple.size();i++) {
      CChartParserInfo::BERPack(tuple[i], byteSeq);
    }
    Dbt key((uint8_t*) &byteSeq.front(), byteSeq.size());
    dbError = (m_db->get(NULL, &key, &data, 0) != 0);
  } else {
    // Set up our DBTs - note the typcast from const CWordSequence* to CWordSequence* that Berkeley DB requires even though it doesn't modify
    // the key object
    Dbt key((CWordSequence*) &tuple.front(), sizeof(CWordIndex) * tuple.size());
    //  Dbt key(buf,sizeof(CWordIndex)*tuple.size());
    dbError = (m_db->get(NULL, &key, &data, 0) != 0);
  }

  //  cout << "hellobefore, db pointer:" << m_db << endl;
  //  CChartParserInfo::PrintVocabSequence(tuple);
  //  cout << " key size: " << key.get_size() << "\n";
  //  cout << " key data: \n";
  //  for (int i=0; i < key.get_size()/4; i++) {
  //        cout << ((int32_t*) key.get_data())[i];
  //  }
  //  cout << endl;
  //  cerr << "hellobefore\n";

  if (dbError) {
    // tuple was not found in the DB - return false
    //    cout << "not found\n";
    //    cerr << "not found\n";
    tupleFreq = 0;
    // don't care about unassignedFrequencyMass and uniqueSuccessorCount
    return false;
  } else {
    // get a Byte-size pointer to the data
    uint8_t* dataPointer = (uint8_t*) data.get_data();
    if (CChartParserInfo::SAMTLM_UseBERFormat) {
      tupleFreq = CChartParserInfo::BERUnpack<CFrequency>(dataPointer);
      unassignedFrequencyMass = CChartParserInfo::BERUnpack<CFrequency>(dataPointer);
      uniqueSuccessorCount = CChartParserInfo::BERUnpack<CFrequency>(dataPointer);
      //      CFrequency dummyFreq = CChartParserInfo::BERUnpack<CFrequency>(dataPointer);
      //    cout << "CFrequency uniqueSuccessorCount: " << dummyFreq << endl;
      assert( (uint8_t*) data.get_data() + data.get_size() == dataPointer);
    } else {
      assert(data.get_size()==sizeof(CFrequency) * 3);
      //    cout << "successful after db->get\n";
      //    cerr << "successful after db->get\n";
      
      // first sizeof(CFrequency) bytes are for tupleFreq
      tupleFreq = *( (CFrequency*) (dataPointer));
      // the next sizeof(CFrequency) bytes are for unassignedFrequencyMass
      unassignedFrequencyMass = *( (CFrequency*) (dataPointer + sizeof(CFrequency)) );
      // the last sizeof(CFrequency) bytes are for uniqueSuccessorCount
      uniqueSuccessorCount = *( (CFrequency*) (dataPointer + sizeof(CFrequency) + sizeof(CFrequency)) );
    }
    return true;
  }
}

#endif
