#include <BTree.hh>

CVocabulary* CBTree::pVoc; //!static voc set at top level

CBTree::CBTree(std::string dbFileName){

  m_db = new Db(NULL,0);
  m_db->set_cachesize(1,0,1); // 1 GB Cache (that's the first parameter) + 0 bytes (2nd parameter); using 1 cache (3rd param)
  //open the DB
  //u_int32_t oFlags = DB_CREATE | DB_EXCL; // Open flags;
  u_int32_t oFlags = DB_RDONLY; // Open flags;
  //try {
  // Redirect debugging information to std::cerr
  m_db->set_error_stream(&std::cerr);
  cout<<"Opening db==="<<dbFileName.c_str()<<"==="<<endl;
  // Open the database
  if (m_db->open(NULL, // Transaction pointer
	    dbFileName.c_str(), // Database file name
	    NULL, // Optional logical database name
	    DB_BTREE, // Database access method
	    oFlags, // Open flags
		 0) > 0){ // File mode (using defaults)
    cerr<<"Database"<<dbFileName<<" could not be opened!!!"<<endl;
    cout<<"Database"<<dbFileName<<" could not be opened!!!"<<endl;
    m_db->close(0);
    exit(-1);
  }

  if (CChartParserInfo::m_useSpanFeatures) {
    // now retrieve the COUNT statistics for all nonterminals occurring as result (i.e. left-hand-side) NTs

    CBTreeResult bTreeQueryResult = QueryBTree(CWordSequence(1,pVoc->GetIndex("@_COUNT")));
    // go through the exact matches and the possible prefix match
    if ((bTreeQueryResult.first & C_ExactMatch) != 0) { // at least one exact match

      //get the production result from the rule - and register the incomplete hyps (for the backpointers) and the generating rule
      //go thru each rule that was matched
      for ( vector<CRule>::iterator iter = bTreeQueryResult.second.begin(); iter!=bTreeQueryResult.second.end(); ++iter){
	CRule& oneItem = (*iter); // the rule found

	if (oneItem.m_negLogProbs.size() != 4) {
	  cerr << "Format error in Rule-DB when reading in the _COUNT entries: expected 4 numbers for this entry. Check STDOUT for the entry." << endl;
	  cout << "Format error in Rule-DB when reading in the _COUNT entries: expected 4 numbers for this entry. Entry is:" << endl;
	  oneItem.Display();
	  exit(1);
	}

	CChartParserInfo::m_nonterminalOccurrenceStatsMap[oneItem.m_productionResult] = CNTOccurrenceStatistics(oneItem.m_negLogProbs[0],oneItem.m_negLogProbs[1],oneItem.m_negLogProbs[2],oneItem.m_negLogProbs[3]);
      }
    } else {
      cerr << "Error: Option --UseSpanFeatures was supplied but we don't have any _COUNT entries in the rule database!" << endl;
      exit(1);
    }
  }
}

//!Utilty method to see exactly what got put into the db
void CBTree::PrintDB(){
  Dbc *cursorp;
  m_db->cursor(NULL, &cursorp, 0);
  
  Dbt key, data;
  int ret;
  
  // Iterate over the database, retrieving each record in turn.
  while ((ret = cursorp->get(&key, &data, DB_NEXT)) == 0) {
    // Do interesting things with the Dbts here.
    cout<<"key=="<<(char *)key.get_data()<<"=="<<endl;
  }
  // Cursors must be closed
  if (cursorp != NULL) 
    cursorp->close(); 
}


CBTree::~CBTree() {
  //try {
  // Close the database
  m_db->close(0);
  delete m_db;

}

/////////////// CBTree

//query the B-tree for prefix and return result as described above in definition of CBTreeResult
CBTreeResult CBTree::QueryBTree(const vector<CWordIndex>& prefix) {
  CBTreeResult bTreeResult(0,vector<CRule>()); // initialize pair with 0 and empty vector
  
  bool s_debugQuery = false;
  //create query string
  string strprefix;
  for (int i=0;i<prefix.size();i++) {
    //string wd = (const char*) (pVoc->GetWord(prefix[i]));
    string wd = pVoc->GetWord(prefix[i]);
    strprefix += (i==0 ? "" : " ") + wd;
  }

  if (s_debugQuery){
    cout<<"Looking up in btree: "<<strprefix<<endl;

  }

  Dbc *cursorp;
  // Get a cursor
  m_db->cursor(NULL, &cursorp, 0);

  // Search criteria
  const char *search_key_const =strprefix.c_str();
  char* search_key = (char*) search_key_const;
  //char *search_data = "hi";
  //cout<<"search_key:"<<search_key<<" strprefix="<<strprefix<<endl;

  // Set up our DBTs
  Dbt key(search_key, strlen(search_key) + 1);
  //Dbt data(search_data, strlen(search_data) + 1);
  Dbt data;

  bool stopit = false;

  // Position the cursor to the first record in the database whose
  // key begins with the prefix

  // cursorp return 0 if something was found
  int ret = cursorp->get(&key, &data, DB_SET_RANGE);
  if (s_debugQuery){
    cout<<"ret="<<ret<<endl;
  }

  //as we loop thru the db results, we keep track of what is going on in the bTreeResult.first
  //if there is an exact match, the least significant bit of the short gets set to 1, and the 
  //array gets filled. If there are multiple matches it keeps getting filled.
  //When there is no exact match, we record if the next record was a prefix match and quit looping
  //since we just need to record that it happened.
  while (ret==0 && !stopit) {
      // do we have an exact match?
      string strkey((char *)key.get_data());
      if (strprefix == strkey) {
	//yes
	if (s_debugQuery){
	  cout<<"exact match: "<<strkey<<endl;
	}
	bTreeResult.first |= C_ExactMatch;

	string strdata((char *)data.get_data());
    
	//process strdata, creating a new CRule
	assert(pVoc!=NULL);
	CRule rule(strkey, strdata);

	//push back rule into vector of rules
	bTreeResult.second.push_back(rule);

	ret = cursorp->get(&key, &data, DB_NEXT); // try to get next DB record
      } else {
	// all exact matches have been processed
	// check if strprefix is prefix of strkey
	if (s_debugQuery){
	  cout<<"checking if:"<<strkey<<" has strprefix="<<strprefix << " followed by a space" <<endl;
	}
	// if the source side prefix we're looking for is e.g. "population", a prefix match occurs if the next entry
	// is e.g. "population de", but "populations" 
	// should not be considered a prefix match. Since space is before any proper character in the lexicographical 
	// alphabet, to exclude these inproper matches (and thus speed up the parsing
	// by not storing unnecessary incomplete hyps), we merely need to add a space to strprefix
	strprefix += " ";
	if (strkey.find( strprefix, 0 ) == 0) {
	  // that's all we need to know, we don't care about the data yet
	  bTreeResult.first |= C_StrictPrefixMatch;	  
	}
	stopit = true; // don't look for further records
      }
  }
      
  // Close the cursor
  if (cursorp != NULL) cursorp->close();

  return bTreeResult;
}
