#ifndef __ChartParserHyp__
#define __ChartParserHyp__


#include <map>
#include <set>
#include <queue>
#include <ParsingHeader.hh>
#include <ParserInfo.hh>
#include <CachedSriLM.hh>
#include <LanguageModelSAMT.hh>
#include <UniqueCompleteHyps.hh>
#include <cfloat>
#include <climits>

using namespace std;

//! some static utility functions
class CUtilities {
public:
  //! Computes the LM context of this lmString - does the equivalent of the q function in Chiang(Comp.Ling.2007)
  static CWordSequence ComputeLMContext(const CWordSequence& lmString);

  //! Computes the LM cost(s) of this lmString - does the equivalent of the p function in Chiang(Comp.Ling.2007)
  /*! computeInsideCost: compute inside LM cost (ignoring first C_HistoryLength words if estimateBoundaryCost==false)
    estimateBoundaryCost==true: add back-off costs for the first C_HistoryLength words (but of course 
    not the costs of the C_HistoryLength words following immediately after m_lmBoundaryWI-symbols 
    as these are only representing history context, not actual words to be LM-processed)
  */
  static float ComputeLMCost(CLanguageModel& lm, const CWordSequence& lmString, bool estimateBoundaryCost, bool computeInsideCost);
};


//! The unique data for a hyp - need this because we can't hash the hyp pointer since it might
//  get deleted.
struct HypData{
  HypData(const CWordSequence& lmContext, const CWordIndex& productionResult){
    m_lmContext = lmContext;
    m_productionResult = productionResult;
  }
  CWordSequence m_lmContext;
  CWordIndex m_productionResult;
};

struct HypDataHash
{
  size_t operator()(const HypData& h ) const{
      size_t hashCode = h.m_productionResult;
      if (CChartParserInfo::m_compareTargetWords==false){
	// done
      }else{
	for (CWordSequence::const_iterator iter=h.m_lmContext.begin(); iter!=h.m_lmContext.end(); ++iter){
	  hashCode = 31*hashCode + (*iter);
	}
      }
      return hashCode;
  }
  
};

struct HypDataEquals
{
  bool operator()(const HypData& h1, const HypData& h2 ) const{
    // first compare production result b/c that's the most efficient comparison
    if (h1.m_productionResult != h2.m_productionResult){
      return false;
    }
    // now, if compareTargetWords, compare LM context
    if ( CChartParserInfo::m_compareTargetWords ) {
      if (h1.m_lmContext != h2.m_lmContext ) {
	return false;
      }
    }
    // hyps are equivalent
    return true;
  }
};


typedef SGI_HASH_NAMESPACE::hash_map<HypData, CChartParserNode*, HypDataHash, HypDataEquals> CChartParserNodeContainer;


//Compare coarse grained eq classes based on their m_pLowestCostHyp
struct CompareProductionNode{
  bool operator()(const CProductionNode* p1, const CProductionNode* p2)const;
};

//!Utility functor - Compare the best total cost of the chart parser nodes 
//Used in RankFineGrainedClasses when Finalizing the node to prepare it to be used in further combinations
struct CompareChartParserNode{
  bool operator()(const CChartParserNode* p1, const CChartParserNode* p2)const;
};

//!Utility functor - Compare the best total cost of the chart parser nodes
struct CompareCompleteHyps{
  bool operator()(const CChartParserHypComplete* p1, const CChartParserHypComplete* p2)const;
};








//!Coarse grained eq class used during rules matching
/*!
 *Contains bank of fine grained eq-classes
*/
class CProductionNode{

public:

  CProductionNode(int i, int j, CWordIndex productionResult);

  //! delete all the contained fine grained classes
  ~CProductionNode();

  //! inserts into this production category, handles the creation of fine-grained equivalance classes under this node and manages the best-rule app count - called by FinalizeProductionNode
  bool FileHyp(CChartParserHypComplete* pHyp);

  //!Inserts a hyp into a flat vector of hyps IFF it fits into the probabilistic beam for this node, returning true if this hyp was kept; if hyp wasn't inserted, it will be deleted here
  bool CheckBeamInsertHyp(CChartParserHypComplete* pHyp, bool addAll=false);

  //elements to do beam based insertion
  float m_cutoffLogProb; //lowest prob allowed by the beam 
  CChartParserHypComplete* m_pLowestCostHyp; //and the current lowest cost hyp

  vector<CChartParserHypComplete*> m_unfiledHyps;
  pair<double, double> m_pruningOptions; //!< filled when this hyp is first constructed
  

  //!Does a final sorting of the flat vector of hyps and adjusts it to meet the beam criteria
  /*!
   * Fills the m_lowestToHighestCostFineGrained vector with real fine grained equivalence classes by calling RankFineGrainedClasses
   * Set the m_bestTotalRuleApplicationCount
   * Sets m_lexical
   */
  void FinalizeProductionNode(bool addAll=false);

  //!Rank the fine grained eq classes from lowest to highest cost, filling the m_lowestToHighestCostFineGrained array
  /*
   * Fills the m_lowestToHighestCostFineGrained so that combo pruning has an order to go from
   * Return total number of hyps across fine grained classes
   */
  int RankFineGrainedClasses();

  //!Apply the beam criteria on the m_unfiledHyps - the hyps that are thrown out are also deleted
  void PruneUnfiledHyps(bool addAll=false);

  //! Display data for this node
  virtual void Display(CVocabulary* pVoc = NULL);

  //! Accessor for the production result
  CWordIndex GetResult() const {
    return m_productionResult;
  }

  //! Accessor for the number of words spanned
  int I() const {
    return m_i;
  }

  //! Accessor for the position of this hypergraph node
  int J() const {
    return m_j;
  }

  int NumHyps(){
    return m_numHyps;
  }

  //!Hash function for fine grained eq classes
  //  static size_t FineHash(const CWordSequence& lmContext,  CWordIndex productionResult );
  

  //!Prune the fine grained eq classes based on these two criteria
  //Returns the number of hyps kept vs possibly kept after pruning - the more important thing to track is number of 
  //fine grained eq classes which can be checked by looking at m_fineGrainedNodes.size()
  // -- replaced by FinalizeProductionNode
  // -- this function is no longer supported since it implies a wasteful process, ie creating a bunch of filled eq classes and then discarding them
  //pair<int, int> EqClassBeamPruning();


  //!The resulting nonterminal (unless considering the base sentence words, then: a terminal symbol)
  CWordIndex m_productionResult;
  
   //!number of source words spanned and position of first source word spanned by this hyp
  int m_i, m_j;

  //========================================================================================================================
  // All member variables below are only filled AFTER FinalizeProductionNode() has been called
  //  --- eventually we can just protect these variables and allow access via function to enforce this constraint
  //========================================================================================================================
  //!Keep track of the hyp with the lowest rule application count
  float m_bestTotalRuleApplicationCount;

  //!If this node only contains purely lexical entries
  bool m_lexical;

  //!Fine grained eq classes within this node
  CChartParserNodeContainer m_fineGrainedNodes;
  
  //!Fine grained eq classes sorted by lowest cost to highest cost - only valid after calling RankFineGrainedClasses();
  vector<CChartParserNode*> m_lowestToHighestCostFineGrained;

  //! total number of hyps in the class, ONLY valid once the class has been finalized
  int m_numHyps;
  
  
};

//!Generic sorting object (by cost) for CChartParserHypComplete pointers, typically by total cost, this template is used in several places
typedef multimap<float, CChartParserHypComplete*> CSortedHypContainer;

//keep incomplete hyps sorted by rule app count
typedef multimap<float, CChartParserHypIncomplete*> CSortedIncompleteHypContainer;

//! Base class for hypothesis in the search space
/*!
  Each hyp is represented by a span and production result as well as back pointers to coarse- or fine-grained ECs,
  one backpointer for each nonterminal instance in the hyp's translation output
*/
class CChartParserHyp {

public:

  //!number of source words spanned and position of first source word spanned by this hyp
  int m_i, m_j;
  
  //! Back pointers to the coarse grained classes
  /*! backtrace information: since the rules used at this hyp can have both terminals and non-terminals
    we need a list of back pointers ie S->the NP hit my NP (needs 2 back pointers). 
    Each of the terminals are simply completed edges  backpointer info, one pointer to another hypergraph node for each source-side nonterminal of the rule corresponding to this hyp
  */
  //Removed this from the general hyp. Complete hyps point back to fine-grained
  //equivalence classes, while incompletes point back to course-grained classes
  //vector<CChartParserNode*> m_backPointers;

  //!Sum of the rule application count feature at position C_RuleApplicationCountLocation over all backpointer
  //best-hyps plus rule app count feature of this hyp (if complete)
  float m_totalRuleApplicationCount;
    
  //!Default constructor - needed by inheriting classes
  //  CChartParserHyp() {}

  //!Constructor initializing the values - used by inheriting classes
  CChartParserHyp(int i, int j, float RAC) :
    m_i(i),
    m_j(j),
    m_totalRuleApplicationCount(RAC) {}

//   //!Copy constructor - used by inheriting classes
//   CChartParserHyp(CChartParserHyp* pHyp) :
//     m_i(pHyp->m_i),
//     m_j(pHyp->m_j),
//     m_totalRuleApplicationCount(pHyp->m_totalRuleApplicationCount) {
//     //m_backPointers = pHyp->m_backPointers;
//   }


  //! Accessor for the number of words spanned
  CWordIndex I() const{
    return m_i;
  }

  //! Accessor for the position of this hyp
  CWordIndex J() const{
    return m_j;
  }


};



class CChartParserHypComplete; //define here so that CChartParserHypComplete can be an element in the heap
class CChartParserHypCompleteEqClass; //define here so that CChartParserHypCompleteNBest can reference CChartParserHypCompleteEq
class CChartParserHypCompleteNBest; //define here so that CChartParserHypCompleteNBest can reference CChartParserHypCompleteEq

//define this here so that we can maintain a heap pointer IN the node
class KBestCandidateHeap;

//========= To keep derivations in the n-best list separate ============================
struct KBestCandidateHash{
  size_t operator()(const CChartParserHypComplete* p) const;
};
struct KBestCandidateEq{
  bool operator()(const CChartParserHypComplete* p1, const CChartParserHypComplete* p2) const;
};

//========= To compare derivations in the n-best list ============================
struct KBestCandidateMinCompare{
  bool operator()(const CChartParserHypComplete* p1, const CChartParserHypComplete* p2) const;
};

//we need to define this here since CChartParserNode needs to maintain the UniqueHypsKBest
typedef SGI_HASH_NAMESPACE::hash_set<CChartParserHypComplete*, KBestCandidateHash, KBestCandidateEq> UniqueHypsKBest; //the one used specifically for derivations
typedef std::priority_queue<CChartParserHypComplete*, vector<CChartParserHypComplete*>, KBestCandidateMinCompare> HeapKBest;




//! This is an equivalence class of hypotheses
/*!
 [X,i,j,LMFeatures] a node in the hypergraph (X: Production result, i: span length, 
 j: span start index within source sentence, LMFeatures: possibly first two and last two target words spanned by all the hyps in this node)

  CChartParserNode has full memory-responsibility for all hyps inserted and will set all its hyps free upon
  destruction. After inserting a hyp pointer, this pointer is not guarranteed to be valid anymore as only
  the m_maxHypCount best hyps are kept in the node at any time
*/
class CChartParserNode {
public:

  //! iterates through the backpointer-targetword-sequence, replacing each substitution site with the m_lmContext of the respective backpointed node in m_backPointers; this is basically the substitution operation alpha[e_1/X_1,...,e_n/X_n] in Chiang(Comp.Ling.2007)
  static CWordSequence HarvestLMString(const CWordSequence& bpTargetSequence, const vector<CChartParserNode*>& m_backPointers);

//   //! true iff this node represents hyps that have already had their BOS calculated. 
//   bool m_BOSDone;

//   //! true iff this node represents hyps that have already had their EOS calculated. 
//   bool m_EOSDone;

  //! true iff this node is lexical (i.e., represents a source word and thus has a terminal symbol as production result)
  bool m_lexical;

  //! The complete hyps stored in this node
  CSortedHypContainer m_hypContainer;

  //! a hash of already seen LM costs; (not done yet)

  //!The maximum number of hyps that will be stored in this node; only the m_maxHypCount best hyps are stored
  int m_maxHypCount;

  //ARV - this stuff moved to the production parent
  //!The resulting nonterminal (unless considering the base sentence words, then: a terminal symbol)
  //CWordIndex m_productionResult;
  //!Number of source words spanned and position of span start in source sentence (starting with 1)
  //int m_i, m_j;

  //! (m_totalCost) of the best hypothesis in this equivalence class
  /*!
    Updated each time a new hyp is filed in here. Use with m_scaledFutureCostEstimate to know the real best cost inside estimate of this node and with m_scaledBoundaryCost to get an estimate that can compare nodes of different fine-grained ECs
  */
  float m_bestTotalCost;

  //! LM m_scaledFutureCostEstimate cost of the best performing hyp in this eq class
  float m_scaledFutureCostEstimate;

  //!Boundary word cost for this fine-grained class - if m_compareTargetWords==0, this is the best hyp's m_scaledFutureCostEstimate, otherwise all hyps have the same m_scaledFutureCostEstimate
  /*
   * Value set on intialization of node
   */
  float m_scaledBoundaryCost;

  //! includes boundary costs if m_pruningMapUseBoundary is turned on - the m_scaledBoundaryCost is a constant across all hyps in this
  float BestHypAllComponentCost() const {
    return m_bestTotalCost + m_scaledFutureCostEstimate + m_scaledBoundaryCost;
  }

  //!just the m_totalCost + m_scaledFutureCostEstimate - use this when you explicitly want to make sure that you have not included boundary costs
  float BestHypScaledTMLMCost() const {
    return m_bestTotalCost + m_scaledFutureCostEstimate;
  }

  //! rule app count of the best hyp - NOT the best rule app count
  float m_totalRuleApplicationCountBestHyp;

  /*! the vectors representing back and forward context, i.e. first and last C_HistoryLength target words
    now represented a la Chiang(Comp.Ling. 2007) as [firstwords m_lmBoundaryWI lastwords]
    If the spanned target words are less than or equal to C_HistoryLength words,
    m_lmContext contains these words, without an m_lmBoundaryWI.
    Note also that these words might be based on underlying backpointer nodes. In that case, they
    refer to the best hyps in these nodes (for fine grained mode cmp1 this is not relevant
    as the lmContext of all hyps in an underlying node must be the same).
  */
  CWordSequence m_lmContext;


  //!The `owner' production node
  CProductionNode* m_productionNodeParent;

  //! Primary constructor, specifying span and production result as well as a maxHypCount to limit memory usage
  /*
    //dont use this direct constructor, instead assign a parent production node
  CChartParserNode(int i, int j, CWordIndex productionResult, int maxHypCount) {
    m_i = i;
    m_j = j;
    m_maxHypCount = maxHypCount;
    m_productionResult = productionResult;
    // no need to initialize m_lexical, m_scaledFutureCostEstimate,
    // m_firstWords, m_lastWords, m_bestTotalRuleApplicationCount,
    // m_bestTotalCost and m_bestNegLogProbs; this will be done as soon as the first hyp is added
  }
  */
  
  CChartParserNode(CProductionNode* parent, int maxHypCount);

  //!Inserts a new hyp into this equivalence class and update the m_best variables
  /*!
    Limits the number of hyps stored to m_maxHypCount
    when an additional hyp is supposed to inserted such that the number of hyps stored is higher
    than m_maxHypCount, the highest-cost hyp is removed and deleted
    (that could also be the hyp supposed to be inserted)!
    Returns true iff m_bestTotalRuleApplicationCount was updated.
   */
  bool InsertHyp(CChartParserHypComplete* pHyp);

  //! Accessor for the production result
  CWordIndex GetResult() const {
    //return m_productionResult;
    return m_productionNodeParent->GetResult();
  }

  //! Accessor for the number of words spanned
  int I() const {
    //return m_i;
    return m_productionNodeParent->I();
  }

  //! Accessor for the position of this hypergraph node
  int J() const {
    //return m_j;
    return m_productionNodeParent->J();
  }

//   //!Accessor for the first words
//   CWordSequence GetFirstWords() const{
//     return m_firstWords;
//   }

//   //!Accessor for the last words  
//   CWordSequence GetLastWords() const{
//     return m_lastWords;
//   }

  //! Display data for this node
  virtual void Display(CVocabulary* pVoc = NULL);

  //! Display this node and recursively the best hyp chain
  virtual void DisplayRecursive();

  //========================================== KBest list stuff =====================
  KBestCandidateHeap* m_pHeap;          // heap of elements that might need exploration
  CUniqueCompleteHyps m_uniqueFinalNBest;   // need to keep track of unique hyps at the top level as well
  vector<CChartParserHypComplete*> m_candidateVector; // possibly (cmp0) un-sorted list of popped elements

  //Returns a vector of CChartParserHypComplete pointers in lowest-highest cost order using Huang/Chiang(2005) K-best retrieval
  vector<CChartParserHypComplete*> RetrieveKBestHC(int k);

  //returns the Kth best translation (starting with 0) found at this node in lazy fashion, returns NULL if such an element was not available
  CChartParserHypComplete* GetKthBest(int k, bool isTopLevel=false); 

  //! destructor - frees the memory for all hyps that have been inserted
  ~CChartParserNode();

};





//! Complete hyps represent hierarchical matched rules 
/*
  CChartParserHypComplete pointers have owners that represent the hyp's course grained equivalence class
*/
class CChartParserHypComplete : public CChartParserHyp{

public:

  //trace the LM cost of this hyp back to its spawning ancestors. In cmp=0 mode, you can use this variable to see the 
  //reduction in LM costs as the hyps get spawned off. we would expect the absolute range to increase with sentence length
  vector<float> m_costTrackBack;

  void AddParentHypCost(float v){
    m_costTrackBack.push_back(v);
  }
  
  //! true iff this hyp is lexical (i.e., represents a source word and thus has a terminal symbol as production result)
  bool m_lexical;

  //! Hypergraph node owning this hyp, i.e., fine grained equivalence class of this hyp
  /*
    set in CChartParserNode.InsertHyp
  */
  CChartParserNode* m_ownerNode;

  CWordIndex m_productionResult;   //!< The resulting nonterminal (unless considering the base sentence words, then it's the source word itself)

  string m_rule; //!< the rule in form of a string that was applied to create this hyp. useful for analysis of which rules a translation in the n-best list is composed of
  CWordSequence m_translationOutput;   //!< the translation output for this hyp. Will be used e.g.by GetLMCost and CChartParser::RetrieveKBestParses. Contains backpointer references at the place of nonterminal occurrences

  CWordSequence m_parsingOutput; //!< parsing output annotated to allow printing the target parse tree - instead of -BPIndex we annotate with -NT

  CWordSequence m_taggedTargets; //!Translations for the tagged words, these are propogated and then assembled by UnwindHyp

  //!Indicates whether the hyp has been completely unwound by UnwindHyp() such that m_translationOutput contains the whole translation
  /*!
    Will be used by CChartParser::RetrieveKBestParses and CChartParser::UnwindHyp()
  */
  bool m_completelyUnwound;

  //! For completely unwound hyps, contains the unknown source words in target order, one source word for each m_unknownWI occurring in m_translationOutput
  /*! 
   The reason why we don't replace unknown-markers in m_translationOutput immediately with the corresponding source words or delete them is
   that we need the unknown-marker for correct calculation of the language model cost
  */
  vector<CWordIndex> m_targetOrderedUnknownSourceWords;

  /*! the vectors representing back and forward context, i.e. first and last C_HistoryLength target words
    now represented a la Chiang(Comp.Ling. 2007) as [firstwords m_lmBoundaryWI lastwords]
    If the spanned target words are less than or equal to C_HistoryLength words,
    m_lmContext contains these words, without an m_lmBoundaryWI.
    Note also that these words might be based on underlying backpointer nodes. In that case, they
    refer to the best hyps in these nodes (for fine grained mode cmp1 this is not relevant
    as the lmContext of all hyps in an underlying node must be the same).
  */
  CWordSequence m_lmContext;

  //  //! LM cost estimate of this hyp, unscaled and NOT including the underlying hyps' LM costs
  //  float m_unscaledInsideLMCost;

  //! scaled estimate of not-yet-determinable costs + underlying best-hyps' m_scaledFutureCostEstimate values - currently zero
  float m_scaledFutureCostEstimate;

  //ARV 07/18
  //a set of raw model costs that get updated as this hyp moves along
  vector<float> m_negLogProbs;            //!<negative log prob of the rule connected to this hypothesis (but not of underlying hyps)

  float m_totalCost; //!<total cost for this hyp, based on summed scaled m_negLogProbs and the the underlying back-pointer nodes' best hyps' m_totalCost values

  float m_scaledBoundaryCost; //!< sum of all feature-weight-scaled LM boundary costs for this hyp (used in combo-pruning and when comparing equivalence classes)

  //! Complete hyps point back to find grained eq-classes 
  /*! backtrace information: since the rules used at this hyp can have both terminals and non-terminals
    we need a list of back pointers ie S->the NP hit my NP (needs 2 back pointers). 
    Each of the terminals are simply completed edges  backpointer info, one pointer to another hypergraph node for each source-side nonterminal of the rule corresponding to this hyp
  */
  vector<CChartParserNode*> m_backPointers;

  //KBEST member variables... these need to be copied over when cloning takes place
  //Once this candidate has been removed (and placed on some permanent listing), call for its neighbors
  CChartParserHypComplete* m_pParentHyp; //the real parse hyp from where this k-best hyp was formed
  vector<int> m_derivationVector;        //references to the set of derivation choices taken below this hyp
  vector<int> m_derivationLimit;         //set only for the parent hyp, empty otherwise


  //! Constructor used by InitSearch
  /*!
    Initializes a complete chart item with a lexical entry, most importantly, it does
    not put anything into the m_translationOutput, only the m_productionResult is set
    This constructor is used by InitSearch and for untranslated words
  */
  CChartParserHypComplete(CWordIndex wi, int i, int j) :
    CChartParserHyp(i,j,0.0), // initialize inherited fields with i,j, and rule app count 0
    // mark this hyp as lexical
    m_lexical(true),
    m_productionResult(wi),
    m_ownerNode(NULL),

    //no need to initialize the cost vector - will never be used
    //     for (int m=0; m<C_NumModelsTotal; m++){
    //       m_negLogProbs.push_back(0);
    //     }

    m_completelyUnwound(false),
    m_totalCost(0),
    m_scaledFutureCostEstimate (0),
    m_scaledBoundaryCost(0)

    //    m_lmContext = CWordSequence();
    //    m_taggedTargets = CWordSequence();
    //    m_targetOrderedUnknownSourceWords = CWordSequence();

    //for tracking back the nbest list in cmp=0
    //    m_costTrackBack = vector<float>();
  {}

  //! Constructor completing an incomplete hyp that completely matches a rule
  /*!
    Take an incomplete hyp (the tail, containing the backpointers and accumulated costs) and a rule (the head, containing 
    the production result, translation output, and the costs of applying the rule) and form a complete hyp out of these two parts
    This constructor is used in the Fundamental and in the Self rule - now during combo pruning in CCombinationTuple::Complete()
  */
  CChartParserHypComplete( CChartParserHypIncomplete* pIncompleteHyp, const vector<CChartParserNode*>& backPointers, CRule* pItem);// :CChartParserHyp((CChartParserHyp*) pIncompleteHyp);
  

  //!Copy constructor; used in CChartParser::UnwindHyp and to create reference copies for the key of the CChartParserNodeContainer multimap
  //  CChartParserHypComplete(CChartParserHypComplete* pHyp):CChartParserHyp((CChartParserHyp*) pHyp) {
//     //copy over the HypComplete-specific stuff
    
//     //during n-best list, this elements gets incremented
//     m_costTrackBack = pHyp->m_costTrackBack;

//     m_lexical = pHyp->m_lexical;
//     m_productionResult = pHyp->m_productionResult;
//     //m_pTransducerItem = pHyp->m_pTransducerItem;
//     m_translationOutput = pHyp->m_translationOutput;
//     m_rule = pHyp->m_rule;
//     m_parsingOutput = pHyp->m_parsingOutput; //for kbest retrieval
//     //arv the m_totalRuleApplicationCount is not used in UnwindHyp but we should still maintain it correctly
//     //az: no, this is done by the inherited constructor
//     //    m_totalRuleApplicationCount = pHyp->m_totalRuleApplicationCount;
  
//     m_ownerNode = pHyp->m_ownerNode;
//     m_lmContext = pHyp->m_lmContext;
//     m_completelyUnwound = pHyp->m_completelyUnwound;

//     //propagate the unknown words and the num tagged words
//     m_taggedTargets = pHyp->m_taggedTargets;
//     m_targetOrderedUnknownSourceWords = pHyp->m_targetOrderedUnknownSourceWords;
//     m_negLogProbs = pHyp->m_negLogProbs;
//     m_totalCost = pHyp->m_totalCost;
//     m_scaledFutureCostEstimate = pHyp->m_scaledFutureCostEstimate;  
//     //dont really need this since this constructor is only used in n-best extraction
//     //where we should only be dealing with ScaledTMLMCost()
//     m_scaledBoundaryCost = pHyp->m_scaledBoundaryCost;
  
//     m_backPointers = pHyp->m_backPointers;

//     //m_numTargetWordsGenerated = pHyp->m_numTargetWordsGenerated;
//     //    m_BOSDone = pHyp->m_BOSDone;
//     //    m_EOSDone = pHyp->m_EOSDone;

//     //for K-Best
//     m_derivationVector = pHyp->m_derivationVector;
//     m_derivationLimit = pHyp->m_derivationLimit;
//     m_pParentHyp = pHyp->m_pParentHyp;
//   }

static bool CompareAllComponentCosts(CChartParserHypComplete* p1, CChartParserHypComplete* p2) {
  return (p1->AllComponentCost() < p2->AllComponentCost());
}

    //! Adds the result category, indices and source words to the parsing string provided
  static void AddParseOutputSubTreeStart(CWordSequence& parsingOutput, CChartParserHypComplete* pBackHyp);
  //! Adds the marker designated to end parse subtrees
  static void AddParseOutputSubTreeEnd(CWordSequence& parsingOutput);

  //! Once the translation is recovered, handle untranslated words and tags and remove translationOutput's <s> and </s> tags
  void HandleUnknownAndTags();

  //! Accessor for the production result
  CWordIndex GetResult() const{
    return m_productionResult;
  }

  //!Accessor for the total cost in the m_negLogProbs array of this completed hyp - doesnt include the LM cost which is in m_scaledFutureCostEstimate
  //  float TotalCost() const{
  //    return m_totalCost;
  //  }
  
  //! includes boundary costs if m_pruningMapUseBoundary is turned on, so consider this a cost estimate
  // this function will return the same value as ScaledTMLMCost when m_pruningMapUseBoundary is false
  float AllComponentCost() const {
    //    if (CChartParserInfo::m_pruningMapUseBoundary){
    //    //if you are cmp=1, then you and your fine node share the same first words
    //    if (CChartParserInfo::m_compareTargetWords==1 && m_ownerNode!=NULL){
    //      assert(fabs(m_ownerNode->m_scaledBoundaryCost-m_scaledBoundaryCost)<0.01);
    //    }
    //  }
    return m_totalCost + m_scaledFutureCostEstimate + m_scaledBoundaryCost;
  }

  //!just the m_totalCost + m_scaledFutureCostEstimate - use this when you explicitly want to make sure that you have not included boundary costs
  float ScaledTMLMCost() const {
    return m_totalCost + m_scaledFutureCostEstimate;
  }


  //! Updates the cost of this hyp after construction
  /*!
    Computes the total cost based on the scaling factors and the existing model scores
    as well as the bestTotalCosts of the underlying (backpointer) hyps
  */
  //  void UpdateTotalCost();

  //!Updates the m_totalRuleApplicationCount for this hyp by looking at the real rule app counts used in the back pointers
  //  void UpdateTotalRuleAppCount();


  //! Accessor for the LM cost (raw) for use in other computations
  //az: disabled to avoid confusion on whether this is the futurecostestimate, the raw, or the scaled LM cost
  //  float GetLMCost() const{
  //    return m_negLogProbs[C_LMLocation];
  //  }

  //  //ARV 01/05 return the Glue cost (raw) for use in other computations
  //float GetGlueCost() const{
  //  return m_negLogProbs[C_GlueLocation];
  //}

  //   //!Accessor for the LM context
  //   CWordSequence GetLMCon() const{
//     return m_firstWords;
//   }

//   //!Accessor for the last words  
//   CWordSequence GetLastWords() const{
//     return m_lastWords;
//   }

//  //! Add the course grained CChartParserNode pointer as a back pointer for this hyp
//  void AddBackPointerToCompleteClass(CChartParserNode* pClass){
//    m_backPointers.push_back(pClass);
//  }

  // Returns the vector of back pointers
  vector<CChartParserNode*>  GetBackPointers(){
    return m_backPointers;
  }

  //! Apply BOS and EOS if applicable
  //  void ProcessMonotoneHyp(const CWordSequence& firstWords, const CWordSequence& mostRecentWords, double& pHypLMCost);

   //! Updates the pHyp's m_firstWords, m_lastWords, and LM cost by traversing the backp pointers in target order
  //  static double ComputeLMData(const CWordSequence& m_translationOutput, const vector<CChartParserNode*>& m_backPointers, CWordSequence& firstWords, CWordSequence& mostRecentWords);


  //! Destructor
  /* 
     Everything else in the CChartParserHyp is statically allocated, the back pointer containers
     are deleted from the CChartParser level
  */
  ~CChartParserHypComplete() { }
  

  //! display this hyp in its current form, main use is the translation
  void Display(CVocabulary* pVoc = NULL);

  //! display this completed hyp, along with its trace back down the path
  void DisplayRecursive(CVocabulary* pVoc = NULL, int tabs = 0);

  
  //! Print the translation output with resolved backpointers
  void PrintTranslationOutput();


  //called on a clone of the very first set of hyps added to the beam, sets the m_derivationVector to 0's
  //return false if the hyp was not creatable ie A DUPLICATE. This should be the only reason false comes up
  //the backWardStarSize is the number of backpointers that we have to initialize for
  bool InitializeHeapData(CChartParserHypComplete* pParentHyp, int backWardStarSize);

  //display simple information for k-best processing
  void DisplaySimpleKBest();

  //! Computes the costs and model features for this hyp based on the choices in its derivation vector
  /* 
   *  This functions is overloaded in dervied classes to perform appropriate cost calculations
   *  For example: In regular HC05 with cmp1, the derivation refers to each back pointer in the m_translationOutput
   *               While in HC05 on a Left-to-right agenda, the derivation refers to 1 less unwound state
   */
  virtual bool ComputeCostsLazyKBest();

  //!All versions return pointers to CChartParserHypComplete* , but they can be instantiated differently
  virtual vector<CChartParserHypComplete*> ExpandNeighbors(const UniqueHypsKBest& m_uniqueTuples);
  
};

//! Extend the complete hyp class to participate in the Left-to-right heuristic search
/*!
 *  Elements of this type are stored on the the beam's agenda and then files into equivalence classes
 *  The parent hyp points back to the previous equivalence class
 */
class CChartParserHypCompleteNBest : public CChartParserHypComplete{
  
public:

  //! If this is hyp a, and hyp b is filed into the same equivalence class, then hyp a goes !m_active
  // being !m_active prevents you from ever being expanded
  bool m_active;
  int m_posNT; //!< Location of the first non-terminal that needs to be expanded

  CChartParserHypCompleteEqClass* m_pBackEqClass; //!< the parent hyp that was used before unrolling created this hyp
  //! NOTE: we need the m_pBackHyp since edges on the agenda are formed from other EDGEs and we need to subtract the model costs
  CChartParserHypCompleteNBest* m_pBackHyp; //! the actual back pointer that generated this hyp from the m_pBackEqClass

  //! Calculates costs based on alternative translations in the back edge
  virtual bool ComputeCostsLazyKBest();

  //!scan forward until you find the next non-terminal. If none are found, then m_posNT will be equal to m_translationOutput.size()
  void UpdatePosNT();
  
  //copy the hyp
  CChartParserHypCompleteNBest(const CChartParserHypComplete& rHyp):
    CChartParserHypComplete(rHyp),
    m_active(true),
    m_pBackEqClass(NULL),
    m_pBackHyp(NULL)
  {
    m_posNT=0;
    UpdatePosNT();
  }

  //! copy constructor
  CChartParserHypCompleteNBest(const CChartParserHypCompleteNBest& rHyp):CChartParserHypComplete(rHyp){
      //during a* search you should never be cloning a non-active hyp, but during the following HC05 its fine to do
    m_active = true;
    m_pBackEqClass = rHyp.m_pBackEqClass;
    m_pBackHyp = rHyp.m_pBackHyp;
    m_posNT = rHyp.m_posNT;
  }
  
  virtual vector<CChartParserHypComplete*> ExpandNeighbors(const UniqueHypsKBest& m_uniqueTuples);
  
  ~CChartParserHypCompleteNBest(){}

  //! Give this hyp acess to its backpointer eq class and actual hyp
  void SetBackInfo(CChartParserHypCompleteEqClass* pBackEqClass, CChartParserHypCompleteNBest* pBackHyp );

  void Deactivate(){ m_active=false; }
  
}; 


//! Equivalence class of CChartParserHypCompleteNBest generated during Left-to-right search
/*!
 *  Also handles n-best list operations for HC05 on the resulting hypergraph
 */
class CChartParserHypCompleteEqClass : public CChartParserHypCompleteNBest{


public:

  //==================================================================================
  // member variables
  //==================================================================================

  //! This vector is not-sorted until we go all the way to the end
  vector<CChartParserHypCompleteNBest*> m_recombinedHyps; //!< hyps that have been recombined into THIS hyp
  CChartParserHypCompleteNBest* m_pCurrentBest; //!< keep track of the current best hyp in this eq class


  //========================================== KBest list stuff =====================
  KBestCandidateHeap* m_pHeap;          // heap of elements that might need exploration
  CUniqueCompleteHyps m_uniqueFinalNBest;   // need to keep track of unique hyps at the top level as well
  vector<CChartParserHypComplete*> m_candidateVector; // un-sorted list of popped elements - serves as the n-best list

  //==================================================================================
  // functions used during forward Left-to-right heuristic search
  //==================================================================================
  //! make sure we dont clone the m_recombinedHyps vector
  CChartParserHypCompleteEqClass(CChartParserHypCompleteNBest* pHyp):CChartParserHypCompleteNBest(*pHyp){
    m_pCurrentBest = NULL;
    m_pHeap = NULL;
  }
  
  //destructor - this does NOT clear the underlying recombined hyps, you need to call ClearRecombined to do that
  //but when you finally delete this class, you also delete the K-Best candidate heap
  ~CChartParserHypCompleteEqClass();
  
  //This hyp is in charge of the memory of the recombined hyps
  void ClearRecombined();
  
  //! Just add the recombined hyps during parsing, returns true IF this hyp is better than the current prototype value
  bool AddRecombinedHyp(CChartParserHypCompleteNBest* pHyp, bool& olddeactivated);

  //! Return only those hyps from the m_recombinedHyps that have different parents. the higher cost duplicates are discarded
  vector<CChartParserHypCompleteNBest*> IdentifyUniqueHyps();
  
  //==================================================================================
  // functions used HC05 on this node - once this equivalence class is filled after a* search
  //==================================================================================
  CChartParserHypCompleteNBest* GetKthBest(int k, bool isTopLevel=false);
  
  //! print out the m_candidateVector
  void PrintNBest();

}; 
  


//! Incomplete hyps track rules that might sucessfully span a part of the sentence
//In contrast to a complete hypothesis, an incomplete hyp tells us that there is a rule whose prefix
// matched with the requirements of the current cell (m_i,m_j). 
class CChartParserHypIncomplete : public CChartParserHyp {
  
public:
  
  //! The source symbols (sequence of terminals and/or nonterminals) that correspond to the cell this hyp is filed in and which matched as strict prefix in the B-tree
  vector<CWordIndex> m_prefix;

  //! Incomplete hyps point back to coarse grained eq-classes 
  vector<CProductionNode*> m_productionBackPointers;

  //!Sum of the rule application count feature at position C_RuleApplicationCountLocation over all backpointer
  //hyps
  //az:now in inherited CChartParserHyp
  //  float m_totalRuleApplicationCount;

  // creates an empty incomplete hyp at source word position j
  CChartParserHypIncomplete(int j) : 
    CChartParserHyp(
		    0, // span length is zero
		    j, // source word position
		    0) // start with 0 cumulative rule app count
  {}

//   // copy constructor
//   CChartParserHypIncomplete(CChartParserHypIncomplete* pHyp):CChartParserHyp((CChartParserHyp*) pHyp){
//     //copy over the prefix info as well as the production backpointer
//     m_prefix = pHyp->m_prefix;
//     m_productionBackPointers = pHyp->m_productionBackPointers;
//     //done by inherited constructor://    m_totalRuleApplicationCount = pHyp->m_totalRuleApplicationCount;
//   }

  //! Add the tail information from the completed coarse grained class
  /*
    expand the hyp with pNode, meaning: adjusting hyp span length m_i, extending prefix and backpointer info, updating cost variables and LM stuff
  */
  void ExpandTail(CProductionNode* pExpansionClass);

  //!ReCalculateRuleAppCount - used if we know that the m_bestTotalRuleApplicationCount for the backpointers might have changed
  //and we want to make sure that when we call AddIncompleteHyp its uses the best m_totalRuleApplicationCount
  void ReCalculateRuleAppCount();

  //!Add a production back pointer to this hyp's backpointer vector
  //  void AddProductionBP(CProductionNode* pClass);

  //!display this hyp in its current form
  void Display(CVocabulary* pVoc = NULL);
  
};

//=======================================================================================
// For K-Best extraction in Huang,Chiang(2005) style, since we need to maintain lists in CChartParserNode
//=======================================================================================

/**
 * Maintains the search heap within each node in the k-best search
 */
class KBestCandidateHeap{

public: 
  
  KBestCandidateHeap();
  ~KBestCandidateHeap();
  
  //!Add this tuple to the beam - it may or may not be added or might kick out another tuple in the beam
  //returns true if this element actually made it onto the beam
  bool InsertBaseCombination(CChartParserHypComplete* toAdd);

  //!Returns the next tuple on the list
  //CChartParserHypComplete* NextTuple();

  CChartParserHypComplete* PopHeap();
  void AddNeighbors(CChartParserHypComplete* pTopElement);

  //!Check if the target side is unique (NULL if its not)
  CChartParserHypComplete* CheckUniqueTarget(CChartParserHypComplete* pHyp){
    return m_uniqueChecker.CheckUnique(pHyp);
  }

  void Display();

  //clear it when its done
  //every cloned item is in the m_uniqueTuples or it as already deleted
  void Clear();

  HeapKBest m_candidateSet; //!< maintains candidate set

protected:
  UniqueHypsKBest m_uniqueTuples; //! keep track of ptrs to unique tuples
  CUniqueCompleteHyps m_uniqueChecker; //! keeps track of unique target sequences

  //Utility function that gives you a UniqueCombinations structure for easy lookup from a SortedCombinations
  //structure which gives you easy ordering
  //static void ConvertSortedToUnique(SortedCombinations* pFrom, UniqueCombinations* pTo);

};


// actually not used in the moment, but we keep it in for potential future use
inline bool CompareProductionNode::operator()(const CProductionNode* p1, const CProductionNode* p2)const{
  assert(p1->m_pLowestCostHyp != NULL && p2->m_pLowestCostHyp != NULL); 
  double p1total = p1->m_pLowestCostHyp->AllComponentCost();
  double p2total = p2->m_pLowestCostHyp->AllComponentCost();
  return ( p1total < p2total);
}


inline bool CompareChartParserNode::operator()(const CChartParserNode* p1, const CChartParserNode* p2)const{
  return (p1->BestHypAllComponentCost()<p2->BestHypAllComponentCost()); //possibly_crashes_with_G++3.2.2

  //  //the BestHypAllComponentCost includes the boundary costs if m_pruningMapUseBoundary==true
  //  double p1total = p1->BestHypAllComponentCost();
  //  double p2total = p2->BestHypAllComponentCost();
  //  if ( p1total < p2total){
  //    return true;
  //  }else{
  //    return false;
  //  }
}


inline bool CompareCompleteHyps::operator()(const CChartParserHypComplete* p1, const CChartParserHypComplete* p2)const{
  //  float c1 = p1->AllComponentCost();
  //  float c2 = p2->AllComponentCost();
  //return (c1<c2); // this is the safe line
  return (p1->AllComponentCost()<p2->AllComponentCost()); //crashes_with_G++3.2.2
}


#endif
