#ifndef LatticeChart_hh
#define LatticeChart_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <fortify.h>

#include <cfloat>
#include <climits>
#include <map>
#include <vector>
#include <iostream>
#include <sstream>
#include <set>
#include <iterator>

#include <BTree.hh>
//includes all the hyp classes
#include <ParserInfo.hh>
#include <ChartParserHyp.hh>
#include <ChartParserHypBeam.hh>
#include <CombinationBeam.hh>

using namespace std;

//! a secondary index that also holds the pointers to hypergraph nodes of a cell, but this time sorted by their m_bestTotalRuleApplicationCount
//typedef multimap<float, CChartParserNode*> CRuleAppCountSortedNodeIndex;
//its multimap because there could be MANY production nodes at a given CRuleAppCountSortedNodeIndex
typedef multimap<float, CProductionNode*> CRuleAppCountSortedNodeIndex;
//coarse grained eq
typedef SGI_HASH_NAMESPACE::hash_map<CWordIndex, CProductionNode* > CCoarseGrainedEq;

//! This class is the contents of one cell in the Complete Hypotheses Chart
/*
 * Maintains a set of production nodes (i.e., coarse-grained ECs) in two indexing schemes
 */
class CCompleteHypChartCell {
public:
  //CChartParserNodeContainer nodes;
  CCoarseGrainedEq m_productionNodes; //!< CProductionNode(s) separated by production result
  CRuleAppCountSortedNodeIndex ruleAppCountSortedNodeIndex; //!< CProductionNode(s) arranged by rule app ct
  //  set<CWordIndex> productionResults; //!The set of production results at this cell
  int m_cellType; //!< whether it's a BOS (1) or EOS (2) or regular cell (0)
  float m_bestGluedRuleAppCount; //!< best RAC that we can get by gluing subcells together to get a translation spanning this cell

  CCompleteHypChartCell() {
    m_cellType = 0; // set default cell type to normal
    m_bestGluedRuleAppCount = FLT_MAX/10;
  }
};


//! Class that drives the search procedure based on smart CKY parsing
/*
  Contains several methods that are used while combining and processing hyps, like LM calculation
*/
class CChartParser
{
public:

  //!Sets static variables for CChartParser
  CChartParser(CVocabulary *pVocabulary,CWordIndex unknownWI, bool noTranslationOutputOfUnknownWords, float maxRuleAppCountDifference, string ruleDBFile, int maxHypCount, int MaxCombinationCount, double aFutureCostImpact, int aBeamSizeFactor, bool extractUnique);

  //! accessor to cells of the complete hyp chart
  CCompleteHypChartCell& GetCompleteCell(int i, int j) {
    return m_completeHypsChart[i][j];
  }

  //! accessor to cells of the incomplete hyp chart
  CSortedIncompleteHypContainer& GetIncompleteCell(int i, int j) {
    return m_incompleteHypsChart[i][j];
  }

  //scaling factors for ranking in decoding
  void SetScalingFactors(vector<double> scalingFactors){
    if (scalingFactors.size()!= CRule::C_NumModelsTotal){
      cerr<<"SetScalingFactors: scalingFactors.size(): "<<scalingFactors.size()<<" BUT C_NumModelsTotal: "<<CRule::C_NumModelsTotal<<" they must match"<<endl;
      exit(2);
    }
    m_scalingFactors = scalingFactors;
    CChartParserInfo::m_scalingFactors = scalingFactors; // az: also set for the static variable in CChartParserHyp
  }

  //display a rule with backpointers as given by the pIncompleteHyp
  void DisplayInstantiatedRule(CRule* pRule, CChartParserHypIncomplete* pIncompleteHyp);


  //!go thru the complete and incomplete hyps and give a full row wise dump of the chart at this point
  void Display();

  //!display the completed cells, in equivalence class batches
  void DisplayCompletedHypsInCell(int i, int j);

  //! Display the incomplete hyps of the cell i,j
  void DisplayIncompletedHypsInCell(int i, int j);

  //! display the complete and incomplete hyps of a single cell in the chart
  void DisplayCell(int i, int j);

  static void SetVocabulary(  CVocabulary *pVocabulary ){
    CChartParserInfo::m_pVocabulary = pVocabulary;
  }


  static void SetMaxHypCount(int max){
    CChartParserInfo::MAX_HYP_COUNT = max;
  }
  

  // initialize the search process with the following steps
  // apply all the source words as completed edges in the m_completeHypsChart
  // we don't need any special handling for fully lexicalized rules since terminals have completed
  //void InitSearch(CSentence* pSentence);
  void InitSearch(CWordSequence& pSentence);

  //! Reserve space for the chart
  void SetupChartCells();

  // true if we want to keep hyps separated by first and last target words
  void SetCompareTargetWords(bool b){
    CChartParserInfo::m_compareTargetWords = b;
  }

  // true if we want to keep hyps separated by first and last target words
  void SetUseLMInCoarse(bool b){
    CChartParserInfo::m_lmInCoarse = b;
  }

  // true if we want to keep hyps separated by first and last target words
  bool GetCompareTargetWords() const{
    return CChartParserInfo::m_compareTargetWords;
  }

  // let the parser know what it takes to be considered a completed sentence
  void SetFullSentenceNonTerminal(CWordIndex w){
    CChartParserInfo::m_FullSentenceNT = w;
  }
  // not needed anymore
  void SetGlueNonTerminal(CWordIndex w){
    CChartParserInfo::m_GlueNT = w;
  }
  
  //!call ClearSearch() after each sentence to clear out the memory used for the previous sentence
  void ClearSearch();
  //! clear cell in the complete hyps chart
  void ClearCHCCell(int i, int j);
  //! clear each cell in the incomplete hyps chart
  void ClearIHCCell(int i, int j);

  //! prune away coarse grained classes - only prunes within newlyAddedProductionNodes b/c other nodes could already by backpointed to by SelfRule-induced hyps
  void  PruneCell(const int i, const int j);

  //!Call finalize method on all nodes in productionNodes
  int FinalizeAllNodes(const CCoarseGrainedEq& productionNodes,  bool addAll=false);

  //! Called after all lexical or fundamental rules have been applied to create new nodes in cell i,j;
  /*!
    Should be renamed b/c it does much more than indexing
    This function is the finalizer for a cell, we call this function when we know that no more hyps
    will ever be added into this cell.
   adds special lexical _UNKNOWN hyp for unknown words
   the strategy for unk words is as follows. when the word is initialized we will know if it has any possible 
  //translations by looking at the complete edges at the cell (1,w)
  //if there are no translations, use the word info initialized here to create an edge that will participate with
  //other edges in normal search either by glue rule or by special _UNKNOWN rules in the rule file.
  //this unknown-edge is a complete lexical (i.e., initial) hyp with  production result unknownWI
  //implicit assumption here: if no word-rule (i.e., rule whose source side is the source word itself) for a source word can be found,
  //then also no phrase-rule (i.e., longer rule containing the source word amongst other word and/or NTs) will be found
  //this assumption is satisfied by our rule extraction method


  */
  void IndexNodesAccordingToRuleAppCount(int i, int j);

  //returns the number 1 parse thru the chart by applying the CYK+ algorithm using the 
  //rule tables accessed in pTransducer
  void ParseChart();

  // simply file the incomplete hyp into the IH chart
  void AddIncompleteHyp(CChartParserHypIncomplete* pIncompleteHyp);


  // compute the LM data, call FileHypInCompletedAndRecurse
  bool AddCompletedHyp(CChartParserHypComplete* pHyp, bool addAll=false);

  /*!
    Returns true iff skipping of hypergraph nodes or of incomplete
    hyps due to too high rule application counts occurred
  */
  std::pair<int,int> ApplyFundamentalRule(int sourceCellRow, int sourceCellTarget, int targetCellRow, int targetCellColumn, float bestResultingRuleRuleAppCountEstimate, RulesByProductionResult& hypRuleCombination,  vector<CChartParserHypIncomplete*>& incompleteHypsToDelete);  

  //!Check where the production results generated in the hypRuleCombinations are valid when considered for self rule application on node pNode
  /*
   * Production results are invalid if they are generating hyps to be filed into already finalized CProductionNode (s)
   */
  //bool SelfRuleCompliance(CProductionNode* pNode, const RulesByProductionResult& hypRuleCombinations);

  //!expand the equivalence class node and adds more complete and incomplete hyps to chart
  /*!
    Calls the btree lookup on the production result of this node
    Calls ProcessBTreeResult to handle incomplete hyps (file or delete)
    Adds new pNewIncompleteHypPrototype to the scheduleForAdding list so that they can be filed with AddIncompleteHyp after ReCalculateRuleAppCount is called on them
    at the end of IndexNodesAccordingToRuleAppCount
  */
  void p_applySelfRuleRecursive(CProductionNode* pHyp, vector<CChartParserHypIncomplete*>& scheduleForAdding);


  //helper functions for ProcessBTreeResult
  //they take the incomplete hyp (that points to coarse-grained eq classes) and extracts all possible CChartParserNode* sequences
  //and then spawns off a complete hyp for each of these
  //void GenerateCompleteFromIncomplete(CChartParserHypIncomplete* pNewIncompleteHypPrototype, CRule &oneItem, bool selfRuleMode);


  //!takes a current state of the bp vector where bp's have been partially mapped over, and returns an expanded list
  //void GenerateNextBP(vector < vector<CChartParserNode*> >& allBackPointerSeqs, vector<CProductionNode*>::iterator iter );

  //tester function that using 2 different beam values and checks that everything in the stricter beam exists in the looser beam
  //void TestKBest(RulesByProductionResult::iterator begin, RulesByProductionResult::iterator end);

  //initialize the beam with the RulesByProductionResult elements before calling FillNBestReturnPtr on the beam
  void InitializeComboBeam(CCombinationTupleBeam* pBeam, RulesByProductionResult::iterator begin, RulesByProductionResult::iterator end, int& allPossibleExpansions, bool addAll);


  //check beam compliance - its a separate function to make it easy to change pruning schemes
  //  bool IsTupleFuzzCompliant(double hypCost, double firstFailedCost);

  //do spawning for a particular production result only - verify that all combos do produce the same production result
  //returns a pair of int's representing the number expanded, and the number that could have been expanded
  pair<int, int> SpawnForProductionResult(RulesByProductionResult::iterator begin, RulesByProductionResult::iterator end, bool addAll);
  //take the list of rules that match this one source side span + the back pointers that each rule would point to
  //and instantiate the top K real CChartParserNode based complete hyps before app
  //void SpawnKBestCompleteHyps(vector<CRule>& matchedRules, CChartParserHypIncomplete* pNewIncompleteHypPrototype, int K, bool selfRuleMode);
  //uses the variable CChartParserInfo::m_comboPruningBeamSize CChartParserInfo::m_comboPruningBeamPerc
  //returns a pair of int's representing the number expanded, and the number that could have been expanded
  pair<int, int> SpawnKBestCompleteHyps(RulesByProductionResult& hypRuleCombinations, bool addAll);


  //!called by ApplyFundamentalRule and p_applySelfRuleRecursive; look in implementation for details
  //CChartParserHypIncomplete* ProcessBTreeResult(CBTreeResult& bTreeQueryResult, CChartParserHypIncomplete* pNewIncompleteHypPrototype, bool selfRuleMode);
  //return true if pNewIncompleteHypPrototype is a prefix match - if its false you should delete the pNewIncompleteHypPrototype
  //bool ProcessBTreeResult(CBTreeResult& bTreeQueryResult, CChartParserHypIncomplete* pNewIncompleteHypPrototype);

  //filles the hypRuleCombinations if neccesary (ie for non-lexical items with backpointers)
  //bool ProcessBTreeResult( vector< pair<CChartParserHypIncomplete*, vector<CRule> > >& hypRuleCombinations, CBTreeResult& bTreeQueryResult, CChartParserHypIncomplete* pNewIncompleteHypPrototype);
  bool ProcessBTreeResult(RulesByProductionResult& hypRuleCombinations, CBTreeResult& bTreeQueryResult, CChartParserHypIncomplete* pNewIncompleteHypPrototype);

//! returns a pointer to a hyp beam of the K best completely unwound translation hypotheses. 
/*! 
  An A* style lazy unwinding algorithm a la Huang&Chiang 2005 is used.
  Start from the top node as specified here
Don't forget to delete the returned beam and its contents! the beam contains only freshly created hyps that all have to be deleted
*/
  //CHypBeam* RetrieveKBestParses(int K, CProductionNode* pTopNode);
  //CHypBeam* RetrieveKBestParses(int K, CChartParserNode* pTopFineNode);
  //CHypBeam* RetrieveKBestParsesSpan(int K, CProductionNode* pTopNode);

  //!Top level n-best list functions
  /*!
   * Will perform n-best extraction either using HC05 or Heuristic Search
   */
  vector<CChartParserHypComplete*> RetrieveKBestParses(int K);
  vector<CChartParserHypComplete*> RetrieveKBestParsesCell(int K, int i, int j);
  vector<CChartParserHypComplete*> RetrieveKBestParsesSpan(int K, CProductionNode* pTopNode);

  //! Uses the A* search method
  void InitializeBeam(CHypBeam& beam, CSortedHypContainer* sortedHyps);
  CHypBeam* RetrieveByAgenda(int K, CChartParserNode* pTopFineNode); //do it by expanding all items on the agenda
  CHypBeam* RetrieveKBestParses(int K, CChartParserNode* pTopFineNode);
  CChartParserHypCompleteNBest* m_pFullSentenceHyp;

  //! Call Huang,Chiang(2005) K-best on a beam from the A* method
  vector<CChartParserHypComplete*> HC05onBeam(int K, CHypBeam& beam);

  //! Rescore the n-best list with the LM
  void RescoreLM(vector<CChartParserHypComplete*>& nBestList);

  CProductionNode* CompleteTranslationNode(){
    if (GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes.count(CChartParserInfo::m_FullSentenceNT)==0) {
      return NULL;
    } else {
      return GetCompleteCell(CChartParserInfo::m_sourceLength,1).m_productionNodes[CChartParserInfo::m_FullSentenceNT];
    }
  }


  //void PrintParsingOutput(const CWordSequence& v, ostream& os=cout);
  //! sub of PrintParsingOutput
  //void PrintParsingOutput_SafePrint(const int i, ostream& os);

  //! Output the unwound hypotheses referenced by the iterators
  /*!
   * Standardizes the way in which n-best output is written
  */
  //void OutputBeam(CHypBeam* pBeam, int translationCount = -1);
  void OutputNBest(vector<CChartParserHypComplete*>::iterator first, vector<CChartParserHypComplete*>::iterator last, int translationCount);


  //ARV 09/24/06 this function used to make appropriate annotations in the pNewHyp to handle UNKs and tagged words that the pBackHyp generated
  //             without modifying the m_translationOutput of the pNewHyp so that the LM operated over these tags
  void UnwindHyp_HandleUnknownAndTags(CWordIndex word, CChartParserHypComplete* pNewHyp, CChartParserHypComplete* pBackHyp);

  //!sub used by RetrieveKBestParses
  /*! unwinds pHyp
    
    pHyp has to be on the beam and will be removed from the beam before unwinding unless pHyp turns out to be completely unwound
    
    auto-delete hyps worse than k best when inserting into CHypContainer
    if you do identify the first sentence spanning hyp eq class, then return it 
  */
  CChartParserHypCompleteNBest* UnwindHyp(CChartParserHypCompleteNBest* pHyp, CHypBeam& beam, float lowestCostHyp);
  //CChartParserHypCompleteNBest* UnwindHyp(CSortedHypNBestContainer::iterator& iterPHyp, CHypBeam& beam, float lowestCostHyp);
  //ARV 09/21 moved to ParseInfo.hh/cc
  //! sub used by UnwindHyp
  //  float UnwindHyp_LMProcessWord(CWordSequence& processedWords, CWordIndex w, bool allowBOS=true);
  //! sub used by UnwindHyp to calculate the EOS cost given a workingContext which either a) the whole translation output b) last context words
  //float UnwindHyp_EOS(const CWordSequence& workingContext);
  
  //! total n.o. combination trials between nodes and incomplete hyps performed; 
  /*! once this number exceeds MaxCombinationCount, only _G rules are allowed to combine
   */
  int m_combinationCount;




  /*! maximum allowed difference in ApplyFundamentalRule between m_totalRuleApplicationCount (in case of
    incomplete hyps) respectively m_bestTotalRuleApplicationCount (in
    case of hypergraph nodes) and the respective lowest values of the
    incomplete hyps resp. hypergraph nodes considered. Nodes
    resp. incompl. hyps with higher differences are ignored.
  */
  float m_maxRuleAppCountDifference;

  
  vector< vector<CSortedIncompleteHypContainer> > m_incompleteHypsChart; //!< Incomplete hyps corresponding to the type-2 list in the CYK+ paper

  vector< vector<CCompleteHypChartCell> > m_completeHypsChart;  //!< Complete hyps in course equivalence class buckets

 private:

  //az06/11/01: now stored in GetCompleteCell(sourcelength,1), see also accessor CompleteTranslationNode()
  //  CProductionNode* m_pCompleteTranslationsNode;   //!< Node containing the complete translations

  vector<CChartParserHypCompleteNBest*> m_poppedHyps; //! Hyps that have been popped in n-best retrieval (clear after sentence)

  //! the nodes that have been newly created within the current hyp-adding wave for the current CH chart cell
  //vector<CProductionNode*> m_newlyCreatedNodes;
  //its a hash map now so that we can easily kill elements from within it
  CCoarseGrainedEq m_newlyCreatedNodes;

  CBTree m_bTree; //! the BTRee interface communicating with the rule database

  //scaling factors for the underlying models
  vector<double> m_scalingFactors;

  //arv: in the future we should change the timing operation, not the place of the deletion...
  CHypBeam* m_pBeamToDelete; //beam that needs to be deleted after the sentences is handled after retrievekbest

};


#endif


