#ifndef Vocabulary_hh
#define Vocabulary_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fortify.h>
#include <assert.h>
// CVocabulary
// Base class for vocabulary handling
//
#include <iostream>
using std::istream;
using std::ostream;

#include <Types.hh>

#ifdef HAVE_EXT_HASH_MAP
#include <ext/hash_map>
#include <ext/hash_set>
// Hey, what's that? seems like overkill to me. --db
namespace sgi {
  using __gnu_cxx::hash;
  using __gnu_cxx::hash_map;
  using __gnu_cxx::hash_set;
}
#define SGI_HASH_NAMESPACE __gnu_cxx
#else
#include <hash_map>
#include <hash_set>

namespace sgi {
     using std::hash;
     using std::hash_map;
     using std::hash_set;
}
#define SGI_HASH_NAMESPACE std
#endif

struct eqstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) == 0;
  }
};

struct __StringHash{
  size_t operator()(const std::string& s1) const{
    return sgi::hash<const char*>()(s1.c_str());
  }
};

//! Note: it's important that CWordIndex be a 32-bit int to ensure consistency when loading word indices from e.g. Berkeley DBs. Also note that we need signed ints as we use negative values or zero to denote backpointer sites
typedef int32_t CWordIndex;

using std::string;
using std::vector;


class CVocabulary
{
public:
  CVocabulary();
  ~CVocabulary();

  //! Initialze some constants so that we can use a blank vocab
  bool InitBlank();

  // allocate memory for MaxNumOfVocItems vocabulary items
  // does not affect the size of the vocabulary, but allows faster loading
  bool Allocate( int MaxNumOfVocItems );

  // remove all entries in the vocabulary
  void Clear();

  // load vocabulary with/without comment lines/<UNK>/</s> 
  bool Load( const char *FileName );
  bool Load( std::istream &InStream ); 

  // store vocabulary with/without <UNK>/</s>
  // no comment line is written
  bool Store( const char *FileName ) const;
  bool Store( std::ostream &OutStream) const;

  // number of stored vocabulary items
  int Size() const;

  //const CVocabularyItem & GetItem( CWordIndex WordIndex ) const;
  //const CString &         GetWord( CWordIndex WordIndex ) const;
  //const CString *         GetWordPtr( CWordIndex WordIndex ) const;
  const string &         GetWord( CWordIndex WordIndex ) const;
  const string *         GetWordPtr( CWordIndex WordIndex ) const;
  
  const bool         IsInRange( CWordIndex WordIndex ) const;

  //inline bool IsContentWord( CWordIndex WordIndex ) const;
  //inline bool IsMultiWord( CWordIndex WordIndex ) const;

  // for multi-word items, get index of first and last word
  //inline CWordIndex FirstWord( CWordIndex WordIndex ) const;
  //inline CWordIndex LastWord( CWordIndex WordIndex ) const;

  // return index of Word if in vocabulary
  // if Word is not in the vocab, add it on-the-fly 
  // Behavior is thus identical to AddWord
  //CWordIndex GetIndex( const CString & Word ) const;
  CWordIndex GetIndex( const string& Word );

  // search for the nearest vocabulary item <= Word
  //CWordIndex SearchWord( const char * Word ) const;

  // add a word to the vocabulary if not existing yet
  // return the word's word index
  CWordIndex AddWord(const string& word);
  //CWordIndex AddWord(const CString & word);

  
  // convert string into CWordIndices
  //BOOL ConvertToIndices( char * const str, CArray<CWordIndex> *pSent, int *lenn);

  // display methods
  void Display() const;
  //void DisplayItem( CWordIndex WordIndex ) const;
  //void DisplayWord( CWordIndex WordIndex ) const;
  //void DisplaySentence( const CWordIndex *WordIndices, uint Length ) const;

  //void WriteWord( CWordIndex WordIndex, std::ostream &OutStream ) const;
  //void WriteSentence( const CWordIndex *WordIndices, uint Length, 
  //std::ostream &OutStream ) const;
  
  // special vocabulary entries
  // EOS = VocSize + 1, EOT = -1, UNK = 0

  // EOT = -1 and UNK = 0 for all vocabularies
  //  static CWordIndex EOT, UNK;


  //  static CWordIndex BOS, EOS; //beginning of sentence / end of sentence markers

    //! trims an std::string
  static void string_trim(string& str)
  {
    string::size_type pos = str.find_last_not_of(" \n\r");
    if(pos != string::npos) {
      str.erase(pos + 1);
      pos = str.find_first_not_of(" \n\r");
      if(pos != string::npos) str.erase(0, pos);
    }
    else str.erase(str.begin(), str.end());
  }




protected:
  //void UpdateMultiWords();
  
  //CArray< CVocabularyItem * > m_VocItems;
  vector< string > m_VocItems;
  //sgi::hash_map< const char *, CWordIndex, sgi::hash<const char*>, eqstr > m_VocIndices;
  sgi::hash_map< string, CWordIndex, __StringHash > m_VocIndices;
};

#endif
