/**
   @file ChainedAllocator.hh

   Provides an allocator that simply delegates the job to another allocator.

   CVS id: $Id: ChainedAllocator.hh,v 1.6 2005/02/11 01:37:06 dietmar Exp $
 */
#ifndef ChainedAllocator_hh
#define ChainedAllocator_hh

// autoconf's symbols
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ObjectAllocator.hh>
#include <DefaultAllocator.hh>

/**
   @brief An allocator that delegates the job of allocation to another allocator.

   Derive more elaborate allocators from this class in order to provide the
   user a way to also handle classes that do not provide a default constructor:
   Derive a simple allocator from CObjectAllocator that allocates such an object
   using your favourite constructor.

   You can even construct whole allocation chains, e.g. a buffered allocator to
   smooth frequent new/delete calls, followed by a segment allocator, allocating
   huge numbers of objects at once, followed by a special allocator for a class that
   does not provide a standard constructor etc.

   If constructed using the default constructor, objects will simply be allocated
   by a CDefaultAllocator. Pass a pointer to another allocator to the constructor in
   order to link / chain two allocators together.

   Multiple chained allocators can share the same allocator for delegation.

   This class provides no methods to detect or prevent loop (which would eventually
   result in lifelocks), so be sure to avoid them.
 */
template<class T>
class CChainedAllocator : public CObjectAllocator<T> {
protected:
  //! The allocator object encapsulated in this object.
  CObjectAllocator<T>* m_pAllocator;
  /**
     Indicator whether this object is responsible to free the encapsulated allocator on destruction.
     This is only the case if the standard constructor had to create a new CDefaultAllocator object.
  */
  bool m_blDelAllocator;
public:
  /**
     @brief Standard constructor.

     Use a new default allocator object for chaining, which will be freed automatically on destruction.
  */
  inline CChainedAllocator()
    : m_pAllocator(new CDefaultAllocator<T>), m_blDelAllocator(true) {}
  /**
     Constructor.
     Link allocator with another one. The allocator will NOT free the linked allocator on
     destruction. The calling application is responsible for that.

     If pAllocator is NULL, however, this constructor just acts like the default constructor
     and creates a new CDefaultAllocator object for delegation that will be automatically
     freed on destruction.

     @param pAllocator The allocator to which this allocator will be chained to.
   */
  inline CChainedAllocator(CObjectAllocator<T>* pAllocator)
    : m_pAllocator(pAllocator), m_blDelAllocator(false)
  { if (m_pAllocator == NULL) { m_blDelAllocator = true; m_pAllocator = new CDefaultAllocator<T>; } }
  //! Destructor. If this allocator was chained to another one, the other allocator will NOT be freed.
  inline ~CChainedAllocator()     { if (m_blDelAllocator) delete m_pAllocator; }
  //! Allocate object by delegating the job to other allocator.
  virtual T* Create()             { ++m_cObjects; ++m_cCreateCalls; return m_pAllocator->Create(); }
  //! Copy object by delegating the job to another allocator.
  virtual T* Copy( const T& obj ) { ++m_cObjects; ++m_cCopyCalls;   return m_pAllocator->Copy(obj); }
  //! Delete object by delegating the job to another allocator.
  virtual void Del(T* deletable)  { assert(m_cObjects > 0); --m_cObjects; ++m_cDelCalls; m_pAllocator->Del(deletable); }
  //! Get a pointer to the allocator object linked to this allocator.
  inline CObjectAllocator<T>* GetChainedAllocator() const { return m_pAllocator; }
}; // CChainedAllocator

#endif // header already included?
