/**
   @file ObjectAllocator.hh

   Provides base class for STTK's object allocation scheme.

   CVS id: $Id: ObjectAllocator.hh,v 1.11 2005/02/16 07:53:36 dietmar Exp $
 */
#ifndef ObjectAllocator_hh
#define ObjectAllocator_hh

// autoconf's symbols
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <functional>
using std::unary_function;

/**
   This is the base class for STTK's object allocation scheme.

   There are various possible types for allocators: buffered allocators, allocators that
   allocate huge amounts of objects at once etc. They all share the interface of
   this class.

   class T is the type of the objects to allocated. There are no restrictions on the
   class interface so far. Read the documentation of the corresponding subclasses.

   Subclasses must implement the factory methods to create, copy and delete an object
   of class T. The subclass is responsible to keep track of the object counter,
   m_cObjects. This counter is usually incremented by 1 by the copy and create methods
   and decremented by 1 by the deletion method Del().

   The subclass must also provide statistics about the number of physically allocated objects
   of class T via the method getNumberOfAllocatedObjects() and the number of active
   objects via getNumberOfObjects(). Both default to the object counter, but as
   an allocator can allocate more objects than necessary, these number may eventually
   differ. Hence, it's up to the subclass to interpret the object counter correctly.

   The subclass must also maintain the trace counters for Create(), Copy() and Del().

   Objects allocated by this allocator (using Create() or Copy()) must eventually be
   freed using Del(). Otherwise the behaviour of your application is undefined or
   memory leaks can occur.

   As you should not call delete() directly for objects allocated with an allocator,
   this allocator also provides an unary functor for use with STTK container classes.

\code
// alloc is an allocator of base class type CObjectAllocator<CMyClass>

// create object
CMyClass *foo = alloc.Create();
// copy object
CMyClass *bar = alloc.Copy(*foo);
// delete objects
alloc.Del(foo);
// ERROR!
delete bar;

// usage with STL container
std::list<CMyClass*> c;
c.push_back(alloc.Create());
...
// delete all objects in container c
std::foreach(c.begin(),c.end(),alloc.GetDeleter())
\endcode
 */
template<class T>
class CObjectAllocator {
protected:
  //! Object counter for active / allocated objects created by this allocator.
  long int m_cObjects;
  //! Number of calls to Create().
  long int m_cCreateCalls;
  //! Number of calls to Copy().
  long int m_cCopyCalls;
  //! Number of calls to Del().
  long int m_cDelCalls;
public:
  //! Default constructor.
  inline CObjectAllocator() : m_cObjects(0), m_cCreateCalls(0), m_cCopyCalls(0), m_cDelCalls(0) {}
  //! Factory method: Allocate object.
  virtual T* Create() = 0;
  //! Factory method: Copy object.
  virtual T* Copy( const T& obj )  = 0;
  //! Factory method: Delete object.
  virtual void Del( T* deletable ) = 0;
  //! Get number of active objects (allocated and in use by application).
  virtual long int GetNumberOfObjects()          const { return m_cObjects; }
  //! Get number if allocated objects (can be greater than number of active objects).
  virtual long int GetNumberOfAllocatedObjects() const { return m_cObjects; }
  //! Get number of Create() calls.
  inline long int GetNumberOfCreateCalls() const { return m_cCreateCalls; }
  //! Get number of Copy() calls.
  inline long int GetNumberOfCopyCalls()   const { return m_cCopyCalls; }
  //! Get number of Del() calls.
  inline long int GetNumberOfDelCalls()    const { return m_cDelCalls; }
  /**
     @brief A deletion functor for use with STL containers.

     Example: Use
     \code
     foreach(c.begin(),c.end(),allocator.GetDeleter())
     \endcode
     to "delete" (using this allocator) all objects in a STL container c of
     value_type T* (c contains pointers to objects of class T).
  */
  class CDeleter : public unary_function<T*, void> {
  private:
    //! The associated allocator that handles deletion actually.
    CObjectAllocator* m_pAllocator;
  public:
    //! Constructor: Link functor to allocator.
    inline CDeleter(CObjectAllocator* pAlloc)
      : m_pAllocator(pAlloc) { assert(m_pAllocator); }
    //! The functor's operator called by STL container: do the deletion.
    inline void operator() (T * deletable)
    { assert(m_pAllocator); m_pAllocator->Del(deletable); }
  }; // class CMedHypDeleter
  //! Get a deletion functor associated with this allocator.
  CDeleter virtual GetDeleter() { return CDeleter(this); }
}; // CObjectAllocator

#endif // header already included?
