/**
   @file

   Provides a class that make frequent allocation / deletions more effecient.

   CVS file ID: $Id: StaticBufferedAllocator.hh,v 1.12 2005/02/11 01:53:09 dietmar Exp $
 */
#ifndef StaticBufferedAllocator_hh
#define StaticBufferedAllocator_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ChainedAllocator.hh> // for base class

/**
   @brief An object factory class to make frequent allocation / deletions more effecient.

   Instead of freeing objects immediately, the allocator buffers the object for a potential
   future allocation up to a limited, fixed size. The buffer size stays constant throughout
   the lifetime of the allocator object (this is why it's called "static").

   class T is the type of the objects to be buffered. T must either provide standard
   and copy constructors (the prerequesites for use with a CDefaultAllocator) or
   the allocator must explicitely be chained to another allocator
   (using the pAllocator parameter in the constructor) to delegate the allocation job to.
   In the first case you might also want to provide
   an "Init" function to initialize the object appropriately (outside this allocator, then).

   In any case, an assignment operator "=" must be defined, as well as an accessible
   function "Reset()" which is called if an object is reused instead of allocated newly.
 */
template<class T>
class CStaticBufferedAllocator : public CChainedAllocator<T> {
protected:
  //! Current number of buffered objects.
  unsigned long int m_cBuffered;
  //! Object buffer, a fixed size array.
  T** m_ObjBuffer;
  //! Size of the object buffer.
  unsigned long m_ObjBufSize;
public:
  /**
     @brief Constructor.
     @param bufSize Size of the object buffer.
     @param pAlloc The allocator object for delegation if buffering fails. See @c CChainedAllocator.
  */
  inline CStaticBufferedAllocator(unsigned long bufSize, CObjectAllocator<T>* pAlloc = NULL)
    : CChainedAllocator<T>(pAlloc),
      m_cBuffered(0), m_ObjBufSize(bufSize), m_ObjBuffer(NULL)
  { assert(m_pAllocator); if (bufSize > 0) m_ObjBuffer = new (T*)[bufSize]; }
  //! Destructor.
  ~CStaticBufferedAllocator();
  //! Allocate / reuse object.
  virtual T* Create();
  //! Factory method to copy an object.
  virtual T* Copy( const T& obj );
  //! Delete / buffer object.
  virtual void Del(T* deletable);
  //! Clear buffer.
  virtual void Clear();
  //! Get (static) buffer size.
  virtual inline unsigned long GetBufferSize()  const { return m_ObjBufSize; }
  //! Get number of currently buffered objects.
  virtual inline unsigned long GetBufferUsage() const { return m_cBuffered; }
  //! In this class, the number of active objects and the number of allocated objects may differ.
  virtual inline long int      GetNumberOfObjects() const
  { assert(m_cObjects >= m_cBuffered); return m_cObjects-m_cBuffered; }
}; // CStaticBufferedAllocator

// template class, must include definitions
#include <StaticBufferedAllocator_inl.hh>

#endif // header already included?
