/**
   @file

   @brief Inline definitions for template class CStaticBufferedAllocator.

   Never include this file directly, use StaticBufferedAllocator.hh instead.

   CVS file ID: $Id: StaticBufferedAllocator_inl.hh,v 1.11 2005/02/11 21:00:53 dietmar Exp $
 */
#ifndef StaticBufferedAllocator_inl_hh
#define StaticBufferedAllocator_inl_hh

#include <assert.h>

// The using module should already have included Trace.hh, so either
// TRACE_ENABLED or TRACE_DISABLED and the trace macros should be defined.
#ifndef Trace // check for trace macro

// No trace macros! This is an error, however instead of complaining, default to "no tracing".
#ifndef NO_TRACE    // avoid double definition
#define NO_TRACE 1
#endif
#include <Trace.hh> // If already included, this does nothing

#endif // Trace defined?

// check if this header was included correctly.
#ifndef StaticBufferedAllocator_hh
#error ERROR: Header was not included by StaticBufferedAllocator.hh. This is an internal header. Do not attempt to include it directly. Use StaticBufferedAllocator.hh instead.
#endif

//#include <StaticBufferedAllocator.hh>

/**
   @brief Destructor.

   Frees objects that are still in buffer.

   Also prints out some statistics for debugging, if tracing is enabled.
   As this is a template class, it depends on the instanciating class whether
   tracing is enabled or not. As tracing only occurs in the destructor which
   is rarely called (for allocators usually once per program execution), this
   should be OK.

   The statistics are interpreted as follows:

<ul>
<li><b>Objects created</b>

Number of calls to Create().</li>
<li><b>Objects copied</b>

Number of calls to Copy().</li>
<li><b>Buffered copies/creates</b>

Number of Create() / Copy() calls where an object could be reused from the object buffer.
This number is calculated from the total number of Create() and Copy() calls minus the number
of Create() and Copy() calls to the remote allocator.
Hence, if other allocators share the same remote ("chained") allocator, this
number may be lower than expected or even negative.</li>
<li><b>Delegated copies/creates</b>

Number of Create() / Copy() calls where the allocator had to delegate the request to the chained allocator
because the buffer was empty.
However, if other allocators share the same remote ("chained") allocator, this
number may be higher than expected. Otherwise (esp. if the default remote allocator is used),
the sum of buffered and delegated copies/creates equals the sum of "Objects created" and "Objects copied".</li>
<li><b>Objects deleted</b>

Number of calls to Del().
The sum of Create() calls ("Objects created") and Copy() calls ("Objects copied") should equal this number
on allocator destruction.</li>
<li><b>Objects freed</b>

Number of Del() calls where the allocator had to free the object because the buffer was already full.
However, if other allocators share the same remote ("chained") allocator, this
number may be higher than expected. Otherwise (esp. if the default remote allocator is used),
the (positive) difference between "Objects deleted" and "Objects freed" should equal the number of
objects still in buffer ("Objects in buffer" below).

<li><b>Objects in buffer</b>

Number of objects that still remained in buffer when destructor was called. They were deleted, but
not freed yet. However, they will be freed on destruction of the buffered allocator (that is, now).</li>
<li><b>Objects still active</b>

Number of objects that are still active, that is, objects that have been allocated using this allocator
but not been deleted using Del().
Should be 0 on allocator destruction and equals the difference between the sum of "Objects created"
plus "Objects copied" and the number of "Objects deleted".
</ul>
 */
template< class T >
CStaticBufferedAllocator<T>::~CStaticBufferedAllocator() {
  Trace(1,"CStaticBufferedAllocator Destructor.");

  // trace messages
#ifndef NDEBUG
  CObjectAllocator<T>* pAlloc = GetChainedAllocator();
  assert(pAlloc);
  // we assume we're the only ones using the nested default allocator
  assert(pAlloc->GetNumberOfCreateCalls() <= GetNumberOfCreateCalls());
  assert(pAlloc->GetNumberOfCopyCalls()   <= GetNumberOfCopyCalls());
  assert(pAlloc->GetNumberOfDelCalls()    <= GetNumberOfDelCalls());
#ifdef TRACE_ENABLED
  if (TraceLevel >= 2) {
    clog << "CStaticBufferedAllocator Statistics:" << endl
	 << "  Objects created             : " << GetNumberOfCreateCalls() << endl
	 << "  Objects copied              : " << GetNumberOfCopyCalls() << endl;
    clog << "  Buffered copies/creates     : " << ( GetNumberOfCreateCalls() + GetNumberOfCopyCalls() -
						( pAlloc->GetNumberOfCreateCalls() + pAlloc->GetNumberOfCopyCalls() ) ) << endl
	 << "  Delegated copies/creates    : " << pAlloc->GetNumberOfCreateCalls() + pAlloc->GetNumberOfCopyCalls() << endl
	 << "  Objects deleted             : " << GetNumberOfDelCalls() << endl
	 << "  Objects freed               : " << pAlloc->GetNumberOfDelCalls() << endl
	 << "  Objects in buffer           : " << GetBufferUsage() << endl
	 << "  Objects still active        : " << GetNumberOfObjects() << endl;
  } // endif: TraceLevel >= 2
#endif // TRACE_ENABLED
  //  assert(GetNumberOfAllocatedObjects() == pAlloc->GetNumberOfAllocatedObjects());
#endif

  // clean up
  Clear();
  assert(m_cBuffered == 0);
  if (m_ObjBuffer != NULL)
    delete[] m_ObjBuffer;
} // CStaticBufferedAllocator destructor

/**
   @brief Allocate / reuse object.

   This is the object factory method to create an object.

   The object is either taken from the buffer and reseted (object
   must support Reset() call) or newly allocated (via the superclass,
   that is, using the chained allocator).

   @return A "new" object.
 */
template< class T >
T* CStaticBufferedAllocator<T>::Create() {
  assert(m_cBuffered <= m_ObjBufSize);
  assert(m_ObjBuffer != NULL || m_ObjBufSize == 0);

  // Check if there are objects left from a recent delete
  if (m_cBuffered > 0) {
    // pop
    --m_cBuffered;
    // buffer size is now index in array of just popped hyp
    assert(m_ObjBuffer[m_cBuffered]);
    // reset object before returning
    m_ObjBuffer[m_cBuffered]->Reset();

    // adjust statistics
    ++m_cCreateCalls;
    return m_ObjBuffer[m_cBuffered];
  } else {
    // buffer empty. Parent will update m_cCreateCalls
    return CChainedAllocator<T>::Create();
  } // endif: buffered objects available?
} // CStaticBufferedAllocator<T>::Create()

/**
   @brief Factory method to copy an object.

   If there are buffered objects, return one of them and copy data,
   using the assignment operator "=".

   Otherwise allocate a new hypothesis via the superclass, that is,
   use the chained allocator.

   @param obj Object to be copied.
   @return A copy of obj.
 */
template<class T>
T* CStaticBufferedAllocator<T>::Copy(const T& obj) {
  assert(m_cBuffered <= m_ObjBufSize);
  assert(m_ObjBuffer != NULL || m_ObjBufSize == 0);

  // Let's see if there are objects left from a recent delete
  if (m_cBuffered > 0) {
    // pop
    --m_cBuffered;
    // buffer size is now index in array of just popped hyp
    assert(m_ObjBuffer[m_cBuffered]);
    // copy hypothesis before returning
    *m_ObjBuffer[m_cBuffered] = obj;

    // update statistics
    ++m_cCopyCalls;
    return m_ObjBuffer[m_cBuffered];
  } else {
    // buffer empty: allocate new. Parent will update m_cCopyCalls
    return CChainedAllocator<T>::Copy(obj);
  } // endif: cached objects available?
} // CStaticBufferedAllocator<T>::Copy(const T& obj)

/**
   @brief Delete / buffer object.

   Put object in deletion buffer.

   If the buffer is full, delete the object via superclass, using
   the chained allocator, otherwise buffer it for later re-allocation.

   @param deletable (Not NULL) The object to be deleted.
                    Consider the object gone after function call..
 */
template< class T >
void CStaticBufferedAllocator<T>::Del( T* deletable )
{
  assert(m_ObjBuffer != NULL || m_ObjBufSize == 0);
  assert(deletable);

  // check if bufferable
  if (m_cBuffered < m_ObjBufSize) {
    // object can be buffered
    m_ObjBuffer[m_cBuffered++] = deletable;
    // update statistics
    ++m_cDelCalls;
  } else {
    assert(m_cBuffered == m_ObjBufSize);
    // buffer full. Parent updates m_cDelCalls.
    CChainedAllocator<T>::Del(deletable);
  } // endif: buffer full?
} // CStaticBufferedAllocator<T>::Del(...)

/**
   @brief Clear buffer.

   Frees all buffered objects.
 */
template< class T >
void CStaticBufferedAllocator<T>::Clear() {
  assert(m_cBuffered <= m_ObjBufSize);
  assert(m_ObjBuffer != NULL || m_ObjBufSize == 0);

  // delete all objects
  while (m_cBuffered > 0)
    // parent updates m_cDelCalls. That's OK.
    CChainedAllocator<T>::Del(m_ObjBuffer[--m_cBuffered]);
  assert(m_cBuffered == 0);
} // CStaticBufferedAllocator<T>::Clear()

#endif // already included?
